unit DrawObjs;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Windows, Graphics, SysUtils, Controls, Classes, ClipLine,
  FileSave, GrDraw, DefFont, SeqList, PtShared;

type
  TDrawMouseEvent = (deDownNew, deMoveNew, deUpNew,
                     deDownSelectDrag, deDownSelectResize,
                     deMoveBeginDrag, deMoveContinueDrag, deUpEndDrag,
                     deMoveBeginResize, deMoveContinueResize, deUpEndResize);

  TLabelResizeSide = (rsDragging, rsOriginPoint, rsDrawToPoint,
                      rsTopLeft, rsTopRight, rsBottomLeft, rsBottomRight,
                      rsTopSide, rsBottomSide, rsLeftSide, rsRightSide);

  TDrawLabelModes = (dlmErase, dlmNormalCopy, dlmNotXor);

type
  TDrawBaseClass = class(TSequenceListObj) {this is an abstract class}
  private
  protected
    CurrDragOriginPoint : TPoint;
    CurrMouseDownPoint : TPoint;
    FSelectMarksVisible : boolean;
  public
    property SelectMarksVisible : boolean read FSelectMarksVisible;
    constructor Create;
    procedure ErrorSetToNormalMode;
    function GetLabelNameString : string; dynamic; abstract;
    function GetDragResizeCursor : TCursor; dynamic;
    procedure SetAttributes(TheTBAttribClass : TToolBarAttribClass); dynamic; abstract;
    procedure ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); dynamic; abstract;
    procedure Draw(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); virtual; abstract;
    procedure RedrawMoveResizeMode(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); virtual; abstract;
    procedure RedrawSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
    procedure MouseEvent(XPix, YPix : integer;
                         DrawMouseEvent : TDrawMouseEvent;
                         TheBitmap : TBitmap;
                         const TheBmpAttribRec : TBitmapAttribRec); virtual; abstract;
    function DistanceFromPix(XPix, YPix : integer) : integer; virtual; abstract;
    function IsLabelOnBmp(const TheBmpAttribRec : TBitmapAttribRec) : boolean; dynamic; abstract;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile); dynamic; abstract;
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile;
                           TheBitmap : TBitmap;
                           const TheBmpAttribRec : TBitmapAttribRec); dynamic; abstract;
  end; {TDrawBaseClass}

{------------------------------------}

  TDrawFloodFill = class(TDrawBaseClass)
  private
    FLabelBrushStyle : TBrushStyle;
    procedure SetLabelBrushStyle(TheStyle : TBrushStyle);
  protected
    procedure DrawSelectBox(XPix, YPix : integer; TheBitmap : TBitmap;
                            const TheBmpAttribRec : TBitmapAttribRec);
    function CreateNewSeqObj : TObject; override;
    procedure AssignSeqObj(Source : TObject); override;
  public
    DrawPoint : TPoint;
    LabelBrushColor : TColor;
    property LabelBrushStyle : TBrushStyle read FLabelBrushStyle
                                           write SetLabelBrushStyle;
    constructor Create;
    function GetLabelNameString : string; override;
    procedure SetAttributes(TheTBAttribClass : TToolBarAttribClass); override;
    procedure ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure Draw(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure RedrawMoveResizeMode(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure ToggleLocationMark(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
    procedure MouseEvent(XPix, YPix : integer;
                         DrawMouseEvent : TDrawMouseEvent;
                         TheBitmap : TBitmap;
                         const TheBmpAttribRec : TBitmapAttribRec); override;
    function DistanceFromPix(XPix, YPix : integer) : integer; override;
    function IsLabelOnBmp(const TheBmpAttribRec : TBitmapAttribRec) : boolean; override;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile); override;
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile;
                           TheBitmap : TBitmap;
                           const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure Assign(Source : TDrawFloodFill);
  end; {TDrawFloodFill}

{------------------------------------}

  TDrawLabel = class(TDrawBaseClass) {this is an abstract class}
  protected
    CurrDragDrawToPoint : TPoint;
    procedure DrawSelectBox(XPix, YPix : integer; TheBitmap : TBitmap;
                            const TheBmpAttribRec : TBitmapAttribRec);
    procedure DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                            TheMode : TDrawLabelModes;
                            TheBitmap : TBitmap;
                            const TheBmpAttribRec : TBitmapAttribRec); virtual; abstract;
    procedure AssignDrawLabel(Source : TDrawLabel);
  public
    OriginPoint,
    DrawToPoint : TPoint;
    constructor Create;
    procedure Draw(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure RedrawMoveResizeMode(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure MouseEvent(XPix, YPix : integer;
                         DrawMouseEvent : TDrawMouseEvent;
                         TheBitmap : TBitmap;
                         const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); override;
    function DistanceFromPix(XPix, YPix : integer) : integer; override;
    function AreCordinatesOnBmp(ThePoint1, ThePoint2 : TPoint;
                                const TheBmpAttribRec : TBitmapAttribRec) : boolean; dynamic;
    function IsLabelOnBmp(const TheBmpAttribRec : TBitmapAttribRec) : boolean; override;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile); override;
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile;
                           TheBitmap : TBitmap;
                           const TheBmpAttribRec : TBitmapAttribRec); override;
  end; {TDrawLabel}

{------------------------------------}

  TDrawShape = class(TDrawLabel) {this is an abstract class}
  protected
    CurrLabelResizeSide : TLabelResizeSide;
    FLabelPenStyle : TPenStyle;
    FLabelBrushStyle : TBrushStyle;
    procedure SetLabelPenStyle(ThePenStyle : TPenStyle); virtual;
    procedure SetLabelBrushStyle(TheBrushStyle : TBrushStyle); virtual;
    procedure AssignDrawShape(Source : TDrawShape);
  public
    {save var's to disk}
    LabelPenColor : TColor;
    LabelPenWidth : integer;
    LabelBrushColor : TColor;
    property LabelPenStyle : TPenStyle read FLabelPenStyle
                                       write SetLabelPenStyle;
    property LabelBrushStyle : TBrushStyle read FLabelBrushStyle
                                           write SetLabelBrushStyle;
    constructor Create;
    function GetDragResizeCursor : TCursor; override;
    procedure SetAttributes(TheTBAttribClass : TToolBarAttribClass); override;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile); override;
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile;
                           TheBitmap : TBitmap;
                           const TheBmpAttribRec : TBitmapAttribRec); override;
  end; {TDrawShape}

{------------------------------------}

  TDrawInclosedShape = class(TDrawShape)
  protected
    procedure DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                            TheMode : TDrawLabelModes;
                            TheBitmap : TBitmap;
                            const TheBmpAttribRec : TBitmapAttribRec); override;
    function CreateNewSeqObj : TObject; override;
    procedure AssignSeqObj(Source : TObject); override;
  public
    DrawingTool : TDrawingTool;
    constructor Create(TheDrawingTool : TDrawingTool);
    function GetLabelNameString : string; override;
    procedure MouseEvent(XPix, YPix : integer;
                         DrawMouseEvent : TDrawMouseEvent;
                         TheBitmap : TBitmap;
                         const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile); override;
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile;
                           TheBitmap : TBitmap;
                           const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure Assign(Source : TDrawInclosedShape);
  end; {TDrawInclosedShape}

{------------------------------------}

const
  FreeHandArrayCount = 32;
  MaxFreeHandPoints = 500 * FreeHandArrayCount;

type
  TFreeHandBoundsRect = packed record
    TopleftDeltaX, TopleftDeltaY,
    BottomRightDeltaX, BottomRightDeltaY : integer;
  end; {TFreeHandBoundsRect}

  TFreeHandPointRec = packed record
    DeltaX, DeltaY : smallint; {use 16bit integers to save memory}
  end; {TFreeHandPointRec}

  TFreeHandPointArray = packed array[1..FreeHandArrayCount] of TFreeHandPointRec;

  PFreeHandPointsNode = ^TFreeHandPointsNode;
  TFreeHandPointsNode = packed record
    NextNode : PFreeHandPointsNode;
    NodeBoundsRect : TFreeHandBoundsRect;
    PointArray : TFreeHandPointArray;
  end; {TFreeHandPointsNode}

  TDrawFreeHand = class(TDrawShape)
  protected
    ShapeBoundsRect : TFreeHandBoundsRect;
    PointCount : integer;
    NodeCount : integer;
    RecCountInEndNode : integer;
    BeginNode, EndNode : PFreeHandPointsNode;
    LinePointArray : TLinePointArray;
    procedure SetLabelPenStyle(ThePenStyle : TPenStyle); override;
    procedure SetLabelBrushStyle(TheBrushStyle : TBrushStyle); override;
    function AddNewNode : boolean;
    procedure DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                            TheMode : TDrawLabelModes;
                            TheBitmap : TBitmap;
                            const TheBmpAttribRec : TBitmapAttribRec); override;
    function CreateNewSeqObj : TObject; override;
    procedure AssignSeqObj(Source : TObject); override;
  public
    constructor Create;
    destructor Destroy; override;
    function GetLabelNameString : string; override;
    procedure ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure MouseEvent(XPix, YPix : integer;
                         DrawMouseEvent : TDrawMouseEvent;
                         TheBitmap : TBitmap;
                         const TheBmpAttribRec : TBitmapAttribRec); override;
    function DistanceFromPix(XPix, YPix : integer) : integer; override;
    function AreCordinatesOnBmp(ThePoint1, ThePoint2 : TPoint;
                                const TheBmpAttribRec : TBitmapAttribRec) : boolean; override;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile); override;
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile;
                           TheBitmap : TBitmap;
                           const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure Assign(Source : TDrawFreeHand);
  end; {TDrawFreeHand}

{------------------------------------}

  TDrawLine = class(TDrawShape)
  protected
    procedure SetLabelPenStyle(ThePenStyle : TPenStyle); override;
    procedure SetLabelBrushStyle(TheBrushStyle : TBrushStyle); override;
    procedure DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                            TheMode : TDrawLabelModes;
                            TheBitmap : TBitmap;
                            const TheBmpAttribRec : TBitmapAttribRec); override;
    function CreateNewSeqObj : TObject; override;
    procedure AssignSeqObj(Source : TObject); override;
  public
    function GetLabelNameString : string; override;
    procedure ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure MouseEvent(XPix, YPix : integer;
                         DrawMouseEvent : TDrawMouseEvent;
                         TheBitmap : TBitmap;
                         const TheBmpAttribRec : TBitmapAttribRec); override;
    function AreCordinatesOnBmp(ThePoint1, ThePoint2 : TPoint;
                                const TheBmpAttribRec : TBitmapAttribRec) : boolean; override;
    function DistanceFromPix(XPix, YPix : integer) : integer; override;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile); override;
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile;
                           TheBitmap : TBitmap;
                           const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure Assign(Source : TDrawLine);
  end; {TDrawLine}

{------------------------------------}

  TDrawText = class(TDrawLabel)
  protected
    FLabelTextFont : TFont;
    procedure DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                            TheMode : TDrawLabelModes;
                            TheBitmap : TBitmap;
                            const TheBmpAttribRec : TBitmapAttribRec); override;
    function CreateNewSeqObj : TObject; override;
    procedure AssignSeqObj(Source : TObject); override;
  public
    LabelTextStr : string;
    property LabelTextFont : TFont read FLabelTextFont;
    constructor Create;
    destructor Destroy; override;
    function GetLabelNameString : string; override;
    function ForceTextToStayOnBmp(const TheBmpAttribRec : TBitmapAttribRec) : boolean;
    procedure SetFontAttributes(TheTopLeftPoint : TPoint;
                                const TheLabelTextStr : string;
                                TheLabelTextFont : TFont;
                                TheBitmap : TBitmap;
                                const TheBmpAttribRec : TBitmapAttribRec);
    procedure MouseEvent(XPix, YPix : integer;
                         DrawMouseEvent : TDrawMouseEvent;
                         TheBitmap : TBitmap;
                         const TheBmpAttribRec : TBitmapAttribRec); override;
    function DistanceFromPix(XPix, YPix : integer) : integer; override;
    function CalcTextBottomRight(TheTopLeftPoint : TPoint;
           const TheTextStr : string; TheTextFont : TFont;
           TheBitmap : TBitmap;
           const TheBmpAttribRec : TBitmapAttribRec) : TPoint;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile); override;
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile;
                           TheBitmap : TBitmap;
                           const TheBmpAttribRec : TBitmapAttribRec); override;
    procedure Assign(Source : TDrawText);
  end; {TDrawText}

{------------------------------------}

implementation

const
  MinSelectEnd = 8;
  MinSelectMid = 12;
  LabelOutOfRangeDist = 10000;

{======================================================================}

function CalcPointsDistance(P1, P2 : TPoint) : integer;
var
  FX, FY : Extended;
begin {function CalcPointsDistance}
  try
    FX := P1.X - P2.X;
    FY := P1.Y - P2.Y;
    Result := Trunc(sqrt(abs(FX*FX)+abs(FY*FY)));
  except
    on EMathError do
      Result := LabelOutOfRangeDist;
    on EIntError do
      Result := LabelOutOfRangeDist;
  end; {try..except}
end; {function CalcPointsDistance}

{======================================================================}

constructor TDrawBaseClass.Create;
begin {constructor TDrawBaseClass.Create}
  inherited Create;
  with CurrDragOriginPoint do
    begin
      X := 0;
      Y := 0;
    end; {with}
  with CurrMouseDownPoint do
    begin
      X := 0;
      Y := 0;
    end; {with}
  FSelectMarksVisible := false;
end; {constructor TDrawBaseClass.Create}

{----------------------------------------------------------------------}

procedure TDrawBaseClass.ErrorSetToNormalMode;
begin {procedure TDrawBaseClass.ErrorSetToNormalMode;}
  DidFieldsChange := true;
  FSelectMarksVisible := false;
  {The GraphWindow will redraw all labels after this error event.}
end; {procedure TDrawBaseClass.ErrorSetToNormalMode;}

{----------------------------------------------------------------------}

function TDrawBaseClass.GetDragResizeCursor : TCursor;
begin {function TDrawBaseClass.GetDragResizeCursor}
  Result := crDrag;
end; {function TDrawBaseClass.GetDragResizeCursor}

{----------------------------------------------------------------------}

procedure TDrawBaseClass.RedrawSelectMarks(TheBitmap: TBitmap;
  const TheBmpAttribRec: TBitmapAttribRec);
begin {procedure TDrawBaseClass.RedrawSelectMarks}
  FSelectMarksVisible := false;
  ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
end; {procedure TDrawBaseClass.RedrawSelectMarks}

{======================================================================}

procedure TDrawFloodFill.SetLabelBrushStyle(TheStyle : TBrushStyle);
begin {procedure TDrawFloodFill.SetLabelBrushStyle}
  DidFieldsChange := true;
  if TheStyle = bsClear then
    FLabelBrushStyle := bsSolid
  else
    FLabelBrushStyle := TheStyle;
end; {procedure TDrawFloodFill.SetLabelBrushStyle}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.DrawSelectBox(XPix, YPix : integer; TheBitmap : TBitmap;
                                       const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawFloodFill.DrawSelectBox}
  with TheBitmap.Canvas do
    begin
      Brush.Color := clWhite;
      Brush.Style := bsClear;
      Pen.Mode := pmNotXor;
      Pen.Style := psSolid;
      Pen.Color := clBlack;
      Pen.Width := 2;
      MoveTo(XPix-6, YPix);
      LineTo(XPix+7, YPix);
      MoveTo(XPix, YPix-6);
      LineTo(XPix, YPix+7);
    end; {with}
end; {procedure TDrawFloodFill.DrawSelectBox}

{----------------------------------------------------------------------}

constructor TDrawFloodFill.Create;
begin {constructor TDrawFloodFill.Create}
  inherited;
  DrawPoint.X := 0;
  DrawPoint.Y := 0;
  LabelBrushColor := clBlack;
  FLabelBrushStyle := bsSolid;
end; {constructor TDrawFloodFill.Create}

{----------------------------------------------------------------------}

function TDrawFloodFill.GetLabelNameString : string;
begin {function TDrawFloodFill.GetLabelNameString}
  with DrawPoint do
    Result := 'Flood Fill: Location('+IntToStr(X)+', '+IntToStr(Y)+')';
end; {function TDrawFloodFill.GetLabelNameString}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.SetAttributes(TheTBAttribClass : TToolBarAttribClass);
begin {procedure TDrawFloodFill.SetAttributes}
  DidFieldsChange := true;
  if Assigned(TheTBAttribClass) then
   begin
     with TheTBAttribClass do
      begin
        LabelBrushColor := BrushColor;
        LabelBrushStyle := BrushStyle;
      end; {with}
   end
  else
   begin
     LabelBrushColor := Dflt_ToolBar_BrushColor;
     LabelBrushStyle := Dflt_ToolBar_BrushStyle;
   end; {else}
end; {procedure TDrawFloodFill.SetAttributes}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawFloodFill.ToggleSelectMarks;}
  FSelectMarksVisible := not FSelectMarksVisible;
  DrawSelectBox(DrawPoint.X, DrawPoint.Y, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawFloodFill.ToggleSelectMarks;}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.Draw(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawFloodFill.Draw;}
  FSelectMarksVisible := false;
  if IsLabelOnBmp(TheBmpAttribRec) then
    with TheBitmap.Canvas do
      begin
        Pen.Mode := pmCopy;
        Pen.Color := LabelBrushColor;
        Brush.Color := LabelBrushColor;
        if LabelBrushStyle = bsClear then
          Brush.Style := bsSolid
        else
          Brush.Style := LabelBrushStyle;
        //Don't need anymore. JunkHandle := Handle; //fix for VCL bug
        FloodFill(DrawPoint.X, DrawPoint.Y,
                  Pixels[DrawPoint.X,DrawPoint.Y], fsSurface);
      end; {with}
end; {procedure TDrawFloodFill.Draw;}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.RedrawMoveResizeMode(TheBitmap: TBitmap;
  const TheBmpAttribRec: TBitmapAttribRec);
begin {procedure TDrawFloodFill.RedrawMoveResizeMode}
  DrawSelectBox(CurrDragOriginPoint.X, CurrDragOriginPoint.Y, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawFloodFill.RedrawMoveResizeMode}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.ToggleLocationMark(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawFloodFill.ToggleLocationMark;}
  if not FSelectMarksVisible then
    begin
      with TheBitmap.Canvas do
        begin
          Brush.Color := clWhite;
          Brush.Style := bsClear;
          Pen.Mode := pmNotXor;
          Pen.Style := psSolid;
          Pen.Color := clBlack;
          Pen.Width := 2;
          MoveTo(DrawPoint.X-5, DrawPoint.Y-5);
          LineTo(DrawPoint.X+6, DrawPoint.Y+6);
          MoveTo(DrawPoint.X+5, DrawPoint.Y-5);
          LineTo(DrawPoint.X-6, DrawPoint.Y+6);
        end; {with}
    end; {if}
end; {procedure TDrawFloodFill.ToggleLocationMark;}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.MouseEvent(XPix, YPix : integer;
                                    DrawMouseEvent : TDrawMouseEvent;
                                    TheBitmap : TBitmap;
                                    const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawFloodFill.MouseEvent}
  DidFieldsChange := true;
  case DrawMouseEvent of
    deDownNew:
      begin
        DrawPoint.X := XPix;
        DrawPoint.Y := YPix;
        Draw(TheBitmap, TheBmpAttribRec);
      end;
    {deMoveNew: deUpNew: Ignore these events}
    deDownSelectDrag, deDownSelectResize:
      begin
        CurrMouseDownPoint.X := XPix;
        CurrMouseDownPoint.Y := YPix;
        if not FSelectMarksVisible then
          ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
      end;
    deMoveBeginDrag, deMoveBeginResize:
      begin
        {begin dragging label}
        if FSelectMarksVisible then
          ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
        {draw in pmNotXor mode at new cordinates}
        CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + DrawPoint.X;
        CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawPoint.Y;
        DrawSelectBox(CurrDragOriginPoint.X, CurrDragOriginPoint.Y, TheBitmap, TheBmpAttribRec);
      end;
    deMoveContinueDrag, deMoveContinueResize:
      begin
        {continue dragging label}
        {erase dragging image at old cordinates}
        DrawSelectBox(CurrDragOriginPoint.X, CurrDragOriginPoint.Y, TheBitmap, TheBmpAttribRec);
        {draw in pmNotXor mode at new cordinates}
        CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + DrawPoint.X;
        CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawPoint.Y;
        DrawSelectBox(CurrDragOriginPoint.X, CurrDragOriginPoint.Y, TheBitmap, TheBmpAttribRec);
      end;
    deUpEndDrag, deUpEndResize:
      begin
        CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + DrawPoint.X;
        CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawPoint.Y;
        {if new cordinates are on the Bmp then save them}
        {  else keep original cordinates.}
         if IsPointOnCanvas(CurrDragOriginPoint.X, CurrDragOriginPoint.Y,
                            TheBmpAttribRec.BmpWidth, TheBmpAttribRec.BmpHeight) then
           DrawPoint := CurrDragOriginPoint;
        {The GraphWindow will redraw all labels after this event.}
      end;
  end; {case}
end; {procedure TDrawFloodFill.MouseEvent}

{----------------------------------------------------------------------}

function TDrawFloodFill.DistanceFromPix(XPix, YPix : integer) : integer;
begin {function TDrawFloodFill.DistanceFromPix}
  Result := CalcPointsDistance(Point(XPix, YPix), DrawPoint);
end; {function TDrawFloodFill.DistanceFromPix}

{----------------------------------------------------------------------}

function TDrawFloodFill.IsLabelOnBmp(const TheBmpAttribRec : TBitmapAttribRec) : boolean;
begin {function TDrawFloodFill.IsLabelOnBmp}
  Result := IsPointOnCanvas(DrawPoint.X, DrawPoint.Y, TheBmpAttribRec.BmpWidth, TheBmpAttribRec.BmpHeight);
end; {function TDrawFloodFill.IsLabelOnBmp}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.SaveToDisk(WriteGraphFile : TWriteGraphFile);
begin {procedure TDrawFloodFill.SaveToDisk;}
  with WriteGraphFile do
    begin
      WriteLongint(dl_FileType_FloodFill);
      WriteRecord(FLabelBrushStyle, sizeof(FLabelBrushStyle));
      WritePoint(DrawPoint);
      WriteColor(LabelBrushColor);
    end; {with}
end; {procedure TDrawFloodFill.SaveToDisk;}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.ReadFromDisk(ReadGraphFile : TReadGraphFile;
                                      TheBitmap : TBitmap;
                                      const TheBmpAttribRec : TBitmapAttribRec);

{----------------------------------}

procedure ReadVer1Fields;
type
  TDiskV1_FloodFillRec = packed record
    X, Y : smallint;
    FillColor : TColor;
    FBrushStyle : TBrushStyle;
  end; {TDiskV1_FloodFillRec = packed record}
var
  DiskRec : TDiskV1_FloodFillRec;
begin {procedure ReadVer1Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  DrawPoint.X := DiskRec.X;
  DrawPoint.Y := DiskRec.Y;
  LabelBrushColor := DiskRec.FillColor;
  FLabelBrushStyle := DiskRec.FBrushStyle;
end; {procedure ReadVer1Fields;}

{----------------------------------}

procedure ReadVer2_3Fields;
begin {procedure ReadVer2_3Fields;}
  with ReadGraphFile do
    begin
      ReadRecord(FLabelBrushStyle, sizeof(FLabelBrushStyle));
      DrawPoint := Read32BitPoint;
      LabelBrushColor := ReadColor;
    end; {with}
end; {procedure ReadVer2_3Fields;}

{----------------------------------}

begin {procedure TDrawFloodFill.ReadFromDisk}
  DidFieldsChange := true;
  if ReadGraphFile.FileVersion = 1 then
      ReadVer1Fields
    else {if (ReadGraphFile.FileVersion = 2) or (ReadGraphFile.FileVersion = 3) then}
      ReadVer2_3Fields;
end; {procedure TDrawFloodFill.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.Assign(Source: TDrawFloodFill);
begin {procedure TDrawFloodFill.Assign}
  FLabelBrushStyle := Source.FLabelBrushStyle;
  DrawPoint := Source.DrawPoint;
  LabelBrushColor := Source.LabelBrushColor;
end; {procedure TDrawFloodFill.Assign}

{----------------------------------------------------------------------}

procedure TDrawFloodFill.AssignSeqObj(Source: TObject);
begin {procedure TDrawFloodFill.AssignSeqObj}
  Assign(Source as TDrawFloodFill);
end; {procedure TDrawFloodFill.AssignSeqObj}

{----------------------------------------------------------------------}

function TDrawFloodFill.CreateNewSeqObj: TObject;
begin {function TDrawFloodFill.CreateNewSeqObj}
  Result := TDrawFloodFill.Create;
end; {function TDrawFloodFill.CreateNewSeqObj}

{======================================================================}

procedure TDrawLabel.DrawSelectBox(XPix, YPix : integer; TheBitmap : TBitmap;
                                   const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawLabel.DrawSelectBox}
  {check to make sure it's near the bmp}
  if (XPix >= -10) and (XPix <= (TheBmpAttribRec.BmpWidth+10)) and
     (YPix >= -10) and (YPix <= (TheBmpAttribRec.BmpHeight+10)) then
    with TheBitmap.Canvas do
      begin
        Brush.Color := clWhite;
        Brush.Style := bsClear;
        Pen.Mode := pmNotXor;
        Pen.Style := psSolid;
        Pen.Color := clBlack;
        Pen.Width := 5;
        Rectangle(Xpix-3, YPix-3, Xpix+3, YPix+3);
      end; {with}
end; {procedure TDrawLabel.DrawSelectBox}

{----------------------------------------------------------------------}

constructor TDrawLabel.Create;
begin {constructor TDrawLabel.Create;}
  inherited;
  with CurrDragDrawToPoint do
    begin
      X := 0;
      Y := 0;
    end; {with}
  with OriginPoint do
    begin
      X := 0;
      Y := 0;
    end; {with}
  with DrawToPoint do
    begin
      X := 0;
      Y := 0;
    end; {with}
end; {constructor TDrawLabel.Create;}

{----------------------------------------------------------------------}

procedure TDrawLabel.Draw(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawLabel.Draw;}
  FSelectMarksVisible := false;
  DrawWithAMode(OriginPoint, DrawToPoint, dlmNormalCopy, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawLabel.Draw;}

{----------------------------------------------------------------------}

procedure TDrawLabel.RedrawMoveResizeMode(TheBitmap: TBitmap;
  const TheBmpAttribRec: TBitmapAttribRec);
begin {procedure TDrawLabel.RedrawMoveResizeMode}
  DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawLabel.RedrawMoveResizeMode}

{----------------------------------------------------------------------}

procedure TDrawLabel.MouseEvent(XPix, YPix : integer;
                                DrawMouseEvent : TDrawMouseEvent;
                                TheBitmap : TBitmap;
                                const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawLabel.MouseEvent}
  DidFieldsChange := true;
  case DrawMouseEvent of
    deDownNew:
      begin
        OriginPoint.X := XPix;
        OriginPoint.Y := YPix;
        DrawToPoint := OriginPoint;
        DrawWithAMode(OriginPoint, DrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
      end;
    deMoveNew:
      begin
        DrawWithAMode(OriginPoint, DrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
        DrawToPoint.X := XPix;
        DrawToPoint.Y := YPix;
        DrawWithAMode(OriginPoint, DrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
      end;
    deUpNew:
      begin
        DrawWithAMode(OriginPoint, DrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
        DrawToPoint.X := XPix;
        DrawToPoint.Y := YPix;
        DrawWithAMode(OriginPoint, DrawToPoint, dlmNormalCopy, TheBitmap, TheBmpAttribRec);
        {The GraphWindow will redraw all labels after this event.}
      end;
    deDownSelectDrag, deDownSelectResize:
      begin
        CurrMouseDownPoint.X := XPix;
        CurrMouseDownPoint.Y := YPix;
        if not FSelectMarksVisible then
          ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
      end;
    deMoveBeginDrag, deMoveBeginResize:
      begin
        {begin dragging label}
        if FSelectMarksVisible then
          ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
        {erase existing label}
        DrawWithAMode(OriginPoint, DrawToPoint, dlmErase, TheBitmap, TheBmpAttribRec);
        {draw in pmNotXor mode at new cordinates}
        CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + OriginPoint.X;
        CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + OriginPoint.Y;
        CurrDragDrawToPoint.X := (XPix - CurrMouseDownPoint.X) + DrawToPoint.X;
        CurrDragDrawToPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawToPoint.Y;
        DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
      end;
    deMoveContinueDrag, deMoveContinueResize:
      begin
        {continue dragging label}
        {erase dragging image at old cordinates}
        DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
        {draw in pmNotXor mode at new cordinates}
        CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + OriginPoint.X;
        CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + OriginPoint.Y;
        CurrDragDrawToPoint.X := (XPix - CurrMouseDownPoint.X) + DrawToPoint.X;
        CurrDragDrawToPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawToPoint.Y;
        DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
      end;
    deUpEndDrag, deUpEndResize:
      begin
        CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + OriginPoint.X;
        CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + OriginPoint.Y;
        CurrDragDrawToPoint.X := (XPix - CurrMouseDownPoint.X) + DrawToPoint.X;
        CurrDragDrawToPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawToPoint.Y;
        {if new cordinates are on the Bmp then save them}
        {  else keep original cordinates.}
        if AreCordinatesOnBmp(CurrDragOriginPoint, CurrDragDrawToPoint, TheBmpAttribRec) then
          begin
            OriginPoint := CurrDragOriginPoint;
            DrawToPoint := CurrDragDrawToPoint;
          end; {if}
        {The GraphWindow will redraw all labels after this event.}
      end;
  end; {case}
end; {procedure TDrawLabel.MouseEvent}

{----------------------------------------------------------------------}

procedure TDrawLabel.ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
{This Procedure Assumes that TopLeft, BottomRight represent a rectangle}
var
  TmpTopLeft, TmpBottomRight : TPoint;
  Draw4Th : integer;
begin {procedure TDrawLabel.ToggleSelectMarks}
  FSelectMarksVisible := not FSelectMarksVisible;
  TmpTopLeft := OriginPoint;
  TmpBottomRight := DrawToPoint;
  FindTopLeftPoint(TmpTopLeft, TmpBottomRight);
  DrawSelectBox(TmpTopLeft.X, TmpTopLeft.Y, TheBitmap, TheBmpAttribRec);
  Draw4Th := 0;
  if (abs(TmpTopLeft.X - TmpBottomRight.X) > MinSelectEnd) then
    begin
      DrawSelectBox(TmpBottomRight.X, TmpTopLeft.Y, TheBitmap, TheBmpAttribRec);
      inc(Draw4Th);
    end; {if}
  if (abs(TmpTopLeft.Y - TmpBottomRight.Y) > MinSelectEnd) then
    begin
      DrawSelectBox(TmpTopLeft.X, TmpBottomRight.Y, TheBitmap, TheBmpAttribRec);
      inc(Draw4Th);
    end; {if}
  if Draw4Th = 2 then
    DrawSelectBox(TmpBottomRight.X, TmpBottomRight.Y, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawLabel.ToggleSelectMarks}

{----------------------------------------------------------------------}

function TDrawLabel.DistanceFromPix(XPix, YPix : integer) : integer;
{This Procedure Assumes that TopLeft, BottomRight represent a rectangle}
var
  TmpTopLeft, TmpBottomRight : TPoint;
  D1, D2, D3, D4 : integer;
  FoundVert, FoundHorz : boolean;
begin {function TDrawLabel.DistanceFromPix}
  try
    TmpTopLeft := OriginPoint;
    TmpBottomRight := DrawToPoint;
    FindTopLeftPoint(TmpTopLeft, TmpBottomRight);
    D1 := LabelOutOfRangeDist; {set to a large value}
    D2 := LabelOutOfRangeDist; {set to a large value}
    D3 := LabelOutOfRangeDist; {set to a large value}
    D4 := LabelOutOfRangeDist; {set to a large value}
    FoundVert := false;
    if (XPix >= TmpTopLeft.X) and (XPix <= TmpBottomRight.X) then
      begin
        D1 := abs(YPix - TmpTopLeft.Y);
        D2 := abs(YPix - TmpBottomRight.Y);
        FoundVert := true;
      end; {if}
    FoundHorz := false;
    if (YPix >= TmpTopLeft.Y) and (YPix <= TmpBottomRight.Y) then
      begin
        D3 := abs(XPix - TmpTopLeft.X);
        D4 := abs(XPix - TmpBottomRight.X);
        FoundHorz := true;
      end; {if}
    if (not FoundHorz) and (not FoundVert) then
      begin
        if (XPix <= TmpTopLeft.X) then
          begin
            if (YPix <= TmpTopLeft.Y) then
              Result := CalcPointsDistance(TmpTopLeft, Point(XPix, YPix))
            else
              Result := CalcPointsDistance(Point(TmpTopLeft.X, TmpBottomRight.Y),
                                           Point(XPix, YPix));
          end
        else
          begin
            if (YPix <= TmpTopLeft.Y) then
              Result := CalcPointsDistance(Point(TmpBottomRight.X, TmpTopLeft.Y),
                                           Point(XPix, YPix))
            else
              Result := CalcPointsDistance(TmpBottomRight, Point(XPix, YPix));
          end; {else}
      end
    else
      if (not FoundHorz) and FoundVert then
      begin
        if D1 <= D2 then
          Result := D1
        else
          Result := D2;
      end
    else
      if FoundHorz and (not FoundVert) then
      begin
        if D3 <= D4 then
          Result := D3
        else
          Result := D4;
      end
    else
      begin {if FoundHorz and FoundVert then}
        if D1 <= D2 then
          Result := D1
        else
          Result := D2;
        if D3 <= Result then
          Result := D3;
        if D4 <= Result then
          Result := D4;
      end; {else}
  except
    on EMathError do
      Result := LabelOutOfRangeDist;
    on EIntError do
      Result := LabelOutOfRangeDist;
  end; {try..except}
end; {function TDrawLabel.DistanceFromPix}

{----------------------------------------------------------------------}

function TDrawLabel.AreCordinatesOnBmp(ThePoint1, ThePoint2 : TPoint;
                                       const TheBmpAttribRec : TBitmapAttribRec) : boolean;
{This Procedure Assumes that ThePoint1, ThePoint2 represent a rectangle}
var
  TheBmpWidthM1, TheBmpHeightM1 : integer;
begin {function TDrawLabel.AreCordinatesOnBmp}
  {test all 4 points to see if any of them are on the Bmp.}
  Result := IsPointOnCanvas(ThePoint1.X, ThePoint1.Y, TheBmpAttribRec.BmpWidth, TheBmpAttribRec.BmpHeight) or
            IsPointOnCanvas(ThePoint2.X, ThePoint2.Y, TheBmpAttribRec.BmpWidth, TheBmpAttribRec.BmpHeight) or
            IsPointOnCanvas(ThePoint1.X, ThePoint2.Y, TheBmpAttribRec.BmpWidth, TheBmpAttribRec.BmpHeight) or
            IsPointOnCanvas(ThePoint2.X, ThePoint1.Y, TheBmpAttribRec.BmpWidth, TheBmpAttribRec.BmpHeight);
  if not Result then
    begin
      TheBmpWidthM1 := TheBmpAttribRec.BmpWidth - 1;
      TheBmpHeightM1 := TheBmpAttribRec.BmpHeight - 1;
      {test all 4 lines to see if they cross the Bmp.}
      FindTopLeftPoint(ThePoint1, ThePoint2);
      {Test horizontal lines.}
      if (ThePoint1.Y < 0) and (ThePoint2.Y > TheBmpHeightM1) then
        Result := ((ThePoint1.X >= 0) and (ThePoint1.X <= TheBmpWidthM1))or
              ((ThePoint2.X >= 0) and (ThePoint2.X <= TheBmpWidthM1));
      {test vertical lines if horizontal lines did not cross the bmp.}
      if not Result then
        begin
          if (ThePoint1.X < 0) and (ThePoint2.X > TheBmpWidthM1) then
            Result := ((ThePoint1.Y >= 0)and(ThePoint1.Y <= TheBmpHeightM1))or
                  ((ThePoint2.Y >= 0)and(ThePoint2.Y <= TheBmpHeightM1));
        end; {if}
    end; {if}
end; {function TDrawLabel.AreCordinatesOnBmp}

{----------------------------------------------------------------------}

function TDrawLabel.IsLabelOnBmp(const TheBmpAttribRec : TBitmapAttribRec) : boolean;
begin {function TDrawLabel.IsLabelOnBmp}
  Result := AreCordinatesOnBmp(OriginPoint, DrawToPoint, TheBmpAttribRec);
end; {function TDrawLabel.IsLabelOnBmp}

{----------------------------------------------------------------------}

procedure TDrawLabel.SaveToDisk(WriteGraphFile : TWriteGraphFile);
begin {procedure TDrawLabel.SaveToDisk}
  with WriteGraphFile do
    begin
      WritePoint(OriginPoint);
      WritePoint(DrawToPoint);
    end; {with}
end; {procedure TDrawLabel.SaveToDisk}

{----------------------------------------------------------------------}

procedure TDrawLabel.ReadFromDisk(ReadGraphFile : TReadGraphFile;
                                  TheBitmap : TBitmap;
                                  const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawLabel.ReadFromDisk}
  DidFieldsChange := true;
  if ReadGraphFile.FileVersion >= 2 then
    begin
      with ReadGraphFile do
        begin
          OriginPoint := Read32BitPoint;
          DrawToPoint := Read32BitPoint;
        end; {with}
    end; {if}
end; {procedure TDrawLabel.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TDrawLabel.AssignDrawLabel(Source: TDrawLabel);
begin {procedure TDrawLabel.AssignDrawLabel}
  OriginPoint := Source.OriginPoint;
  DrawToPoint := Source.DrawToPoint;
end; {procedure TDrawLabel.AssignDrawLabel}

{======================================================================}

procedure TDrawShape.SetLabelPenStyle(ThePenStyle : TPenStyle);
begin {procedure TDrawShape.SetLabelPenStyle}
  DidFieldsChange := true;
  FLabelPenStyle := ThePenStyle;
end; {procedure TDrawShape.SetLabelPenStyle}

{----------------------------------------------------------------------}

procedure TDrawShape.SetLabelBrushStyle(TheBrushStyle : TBrushStyle);
begin {procedure TDrawShape.SetLabelBrushStyle}
  DidFieldsChange := true;
  FLabelBrushStyle := TheBrushStyle;
end; {procedure TDrawShape.SetLabelBrushStyle}

{----------------------------------------------------------------------}

constructor TDrawShape.Create;
begin {constructor TDrawShape.Create}
  inherited;
  FLabelPenStyle := psSolid;
  FLabelBrushStyle := bsClear;
  LabelPenColor := clBlack;
  LabelPenWidth := 1;
  LabelBrushColor := clWhite;
  CurrLabelResizeSide := rsDragging;
end; {constructor TDrawShape.Create}

{----------------------------------------------------------------------}

function TDrawShape.GetDragResizeCursor : TCursor;
begin {function TDrawShape.GetDragResizeCursor}
  case CurrLabelResizeSide of
    rsDragging:
      Result := crDrag;
    rsOriginPoint, rsDrawToPoint:
      Result := crSize;
    rsTopLeft, rsBottomRight:
      Result := crSizeNWSE;
    rsTopRight, rsBottomLeft:
      Result := crSizeNESW;
    rsTopSide, rsBottomSide:
      Result := crSizeNS;
    rsLeftSide, rsRightSide:
      Result := crSizeWE;
    else
      Result := crDefault;
  end; {case}
end; {function TDrawShape.GetDragResizeCursor}

{----------------------------------------------------------------------}

procedure TDrawShape.SetAttributes(TheTBAttribClass : TToolBarAttribClass);
begin {procedure TDrawShape.SetAttributes}
  DidFieldsChange := true;
  if Assigned(TheTBAttribClass) then
    begin
      with TheTBAttribClass do
        begin
          LabelPenStyle := PenStyle;
          LabelPenColor := PenColor;
          LabelPenWidth := PenWidth;
          LabelBrushColor := BrushColor;
          LabelBrushStyle := BrushStyle;
        end; {with}
     end
  else
    begin
      LabelPenStyle := Dflt_ToolBar_PenStyle;
      LabelPenColor := Dflt_ToolBar_PenColor;
      LabelPenWidth := Dflt_ToolBar_PenWidth;
      LabelBrushColor := Dflt_ToolBar_BrushColor;
      LabelBrushStyle := Dflt_ToolBar_BrushStyle;
    end; {else}
end; {procedure TDrawShape.SetAttributes}

{----------------------------------------------------------------------}

procedure TDrawShape.SaveToDisk(WriteGraphFile : TWriteGraphFile);
begin {procedure TDrawShape.SaveToDisk;}
  with WriteGraphFile do
    begin
      WriteRecord(FLabelPenStyle, sizeof(FLabelPenStyle));
      WriteColor(LabelPenColor);
      WriteLongint(LabelPenWidth);
      WriteColor(LabelBrushColor);
      WriteRecord(FLabelBrushStyle, sizeof(FLabelBrushStyle));
    end; {with}
  inherited SaveToDisk(WriteGraphFile);
end; {procedure TDrawShape.SaveToDisk;}

{----------------------------------------------------------------------}

procedure TDrawShape.ReadFromDisk(ReadGraphFile : TReadGraphFile;
                                  TheBitmap : TBitmap;
                                  const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawShape.ReadFromDisk}
  DidFieldsChange := true;
  if ReadGraphFile.FileVersion >= 2 then
    begin
      with ReadGraphFile do
        begin
          ReadRecord(FLabelPenStyle, sizeof(FLabelPenStyle));
          LabelPenColor := ReadColor;
          LabelPenWidth := ReadLongint;
          LabelBrushColor := ReadColor;
          ReadRecord(FLabelBrushStyle, sizeof(FLabelBrushStyle));
        end; {with}
      inherited ReadFromDisk(ReadGraphFile, TheBitmap, TheBmpAttribRec);
    end; {if}
end; {procedure TDrawShape.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TDrawShape.AssignDrawShape(Source: TDrawShape);
begin {procedure TDrawShape.AssignDrawShape}
  AssignDrawLabel(Source);
  FLabelPenStyle := Source.FLabelPenStyle;
  FLabelBrushStyle := Source.FLabelBrushStyle;
  LabelPenColor := Source.LabelPenColor;
  LabelPenWidth := Source.LabelPenWidth;
  LabelBrushColor := Source.LabelBrushColor;
end; {procedure TDrawShape.AssignDrawShape}

{======================================================================}

procedure TDrawInclosedShape.DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                                           TheMode : TDrawLabelModes;
                                           TheBitmap : TBitmap;
                                           const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawInclosedShape.DrawWithAMode}
  with TheBitmap.Canvas do
    begin
      case TheMode of
        dlmErase:
          begin
            Pen.Mode := pmCopy;
            Pen.Color := TheBmpAttribRec.BackgroundColor;
            Brush.Color := TheBmpAttribRec.BackgroundColor;
          end;
        dlmNormalCopy:
          begin
            Pen.Mode := pmCopy;
            Pen.Color := LabelPenColor;
            Brush.Color := LabelBrushColor;
          end;
        dlmNotXor:
          begin
            Pen.Mode := pmNotXor;
            Pen.Color := LabelPenColor;
            Brush.Color := LabelBrushColor;
          end;
      end; {case}
      Brush.Style := LabelBrushStyle;
      if (LabelBrushStyle = bsClear) and (LabelPenStyle = psClear) then
        Pen.Style := psSolid
      else
        Pen.Style := LabelPenStyle;
      Pen.Width := LabelPenWidth;
      FindTopLeftPoint(ThePoint1, ThePoint2);
      case DrawingTool of
        dtRectangle:
          Rectangle(ThePoint1.X, ThePoint1.Y, ThePoint2.X, ThePoint2.Y);
        dtEllipse:
          Ellipse(ThePoint1.X, ThePoint1.Y, ThePoint2.X, ThePoint2.Y);
        dtRoundRect:
          RoundRect(ThePoint1.X, ThePoint1.Y, ThePoint2.X, ThePoint2.Y,
                    (ThePoint1.X - ThePoint2.X) div 2,
                    (ThePoint1.Y - ThePoint2.Y) div 2);
      end; {case}
    end; {with}
end; {procedure TDrawInclosedShape.DrawWithAMode}

{----------------------------------------------------------------------}

constructor TDrawInclosedShape.Create(TheDrawingTool : TDrawingTool);
begin {constructor TDrawInclosedShape.Create}
  inherited Create;
  DrawingTool := TheDrawingTool;
end; {constructor TDrawInclosedShape.Create}

{----------------------------------------------------------------------}

function TDrawInclosedShape.GetLabelNameString : string;
begin {function TDrawInclosedShape.GetLabelNameString}
  case DrawingTool of
    dtRectangle:
      Result := 'Rectangle';
    dtEllipse:
      Result := 'Ellipse';
    dtRoundRect:
      Result := 'Round Rectangle';
    else
      Result := '';
  end; {case}
  Result := Result+': Top Left('+IntToStr(OriginPoint.X)+', '+IntToStr(OriginPoint.Y)+
            ') Bottom Right('+IntToStr(DrawToPoint.X)+', '+IntToStr(DrawToPoint.Y)+')';
end; {function TDrawInclosedShape.GetLabelNameString}

{----------------------------------------------------------------------}

procedure TDrawInclosedShape.MouseEvent(XPix, YPix : integer;
                                        DrawMouseEvent : TDrawMouseEvent;
                                        TheBitmap : TBitmap;
                                        const TheBmpAttribRec : TBitmapAttribRec);

{-------------------}

procedure FindSideToResizeFrom;
var
  LoCornerDist, LoSideDist,
  TopSideDist, BottomSideDist, LeftSideDist, RightSideDist,
  TopLeftDist, TopRightDist, BottomLeftDist, BottomRightDist : integer;
begin {procedure FindSideToResizeFrom;}
  CurrLabelResizeSide := rsDragging;
  {begin calculate distances to corners}
  TopLeftDist := CalcPointsDistance(Point(XPix, YPix), CurrDragOriginPoint);
  TopRightDist := CalcPointsDistance(Point(XPix, YPix),
                    Point(CurrDragDrawToPoint.X, CurrDragOriginPoint.Y));
  BottomLeftDist := CalcPointsDistance(Point(XPix, YPix),
                    Point(CurrDragOriginPoint.X, CurrDragDrawToPoint.Y));
  BottomRightDist := CalcPointsDistance(Point(XPix, YPix), CurrDragDrawToPoint);
  {end calculate distances to corners}
  {begin calculate LoCornerDist}
  if TopLeftDist <= TopRightDist then
    LoCornerDist := TopLeftDist
  else
    LoCornerDist := TopRightDist;
  if BottomLeftDist < LoCornerDist then
    LoCornerDist := BottomLeftDist;
  if BottomRightDist < LoCornerDist then
    LoCornerDist := BottomRightDist;
  {end calculate LoCornerDist}
  if LoCornerDist <= MaxSelectDistance then
    begin
      {select the corner to resize with}
      if TopLeftDist <= LoCornerDist then
        CurrLabelResizeSide := rsTopLeft
      else
      if BottomRightDist <= LoCornerDist then
        CurrLabelResizeSide := rsBottomRight
      else
      if TopRightDist <= LoCornerDist then
        CurrLabelResizeSide := rsTopRight
      else
      if BottomLeftDist <= LoCornerDist then
        CurrLabelResizeSide := rsBottomLeft;
    end
  else
    begin
      {begin calculate distances to sides and LoSideDist}
      if (XPix >= CurrDragOriginPoint.X)and(XPix <= CurrDragDrawToPoint.X) then
        begin
          TopSideDist := abs(YPix - CurrDragOriginPoint.Y);
          BottomSideDist := abs(YPix - CurrDragDrawToPoint.Y);
          if TopSideDist <= BottomSideDist then
            LoSideDist := TopSideDist
          else
            LoSideDist := BottomSideDist;
        end
      else
        begin
          TopSideDist := LabelOutOfRangeDist;
          BottomSideDist := LabelOutOfRangeDist;
          LoSideDist := LabelOutOfRangeDist;
        end; {else}
      if (YPix >= CurrDragOriginPoint.Y)and(YPix <= CurrDragDrawToPoint.Y) then
        begin
          LeftSideDist := abs(XPix - CurrDragOriginPoint.X);
          RightSideDist := abs(XPix - CurrDragDrawToPoint.X);
          if LeftSideDist < LoSideDist then
            LoSideDist := LeftSideDist;
          if RightSideDist < LoSideDist then
            LoSideDist := RightSideDist;
        end
      else
        begin
          LeftSideDist := LabelOutOfRangeDist;
          RightSideDist := LabelOutOfRangeDist;
        end; {else}
      {end calculate distances to sides and LoSideDist}
      {select the side to resize with}
      if LoSideDist <= MaxSelectDistance then
        begin
          if BottomSideDist <= LoSideDist then
            CurrLabelResizeSide := rsBottomSide
          else
          if RightSideDist <= LoSideDist then
            CurrLabelResizeSide := rsRightSide
          else
          if TopSideDist <= LoSideDist then
            CurrLabelResizeSide := rsTopSide
          else
          if LeftSideDist <= LoSideDist then
            CurrLabelResizeSide := rsLeftSide;
        end; {if}
    end; {else}
end; {procedure FindSideToResizeFrom;}

{-------------------}

procedure CalcNewCurrResizePoints;
begin {procedure CalcNewCurrResizePoints;}
  case CurrLabelResizeSide of
    rsTopLeft:
      begin
        CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + OriginPoint.X;
        CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + OriginPoint.Y;
      end;
    rsTopRight:
      begin
        CurrDragDrawToPoint.X := (XPix - CurrMouseDownPoint.X) + DrawToPoint.X;
        CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + OriginPoint.Y;
      end;
    rsBottomLeft:
      begin
        CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + OriginPoint.X;
        CurrDragDrawToPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawToPoint.Y;
      end;
    rsBottomRight:
      begin
        CurrDragDrawToPoint.X := (XPix - CurrMouseDownPoint.X) + DrawToPoint.X;
        CurrDragDrawToPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawToPoint.Y;
      end;
    rsTopSide:
      CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + OriginPoint.Y;
    rsBottomSide:
      CurrDragDrawToPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawToPoint.Y;
    rsLeftSide:
      CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + OriginPoint.X;
    rsRightSide:
      CurrDragDrawToPoint.X := (XPix - CurrMouseDownPoint.X) + DrawToPoint.X;
  end; {case}
end; {procedure CalcNewCurrResizePoints;}

{-------------------}

begin {procedure TDrawInclosedShape.MouseEvent}
  DidFieldsChange := true;
  case DrawMouseEvent of
    deDownSelectResize:
      begin
        CurrMouseDownPoint.X := XPix;
        CurrMouseDownPoint.Y := YPix;
        FindTopLeftPoint(OriginPoint, DrawToPoint);
        CurrDragOriginPoint := OriginPoint;
        CurrDragDrawToPoint := DrawToPoint;
        if not FSelectMarksVisible then
          ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
        {determine which point or side to resize from}
        FindSideToResizeFrom;
      end;
    deMoveBeginResize:
      begin
        if CurrLabelResizeSide = rsDragging then
          inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec)
        else
          begin
            {begin resizing label}
            if FSelectMarksVisible then
              ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
            {erase existing label}
            DrawWithAMode(OriginPoint, DrawToPoint, dlmErase, TheBitmap, TheBmpAttribRec);
            {draw in pmNotXor mode at new cordinates}
            CalcNewCurrResizePoints;
            DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
          end; {else}
      end;
    deMoveContinueResize:
      begin
        if CurrLabelResizeSide = rsDragging then
          inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec)
        else
          begin
            {continue resizing label}
            {erase resizing image at old cordinates}
            DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
            {draw in pmNotXor mode at new cordinates}
            CalcNewCurrResizePoints;
            DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
          end; {else}
      end;
    deUpEndResize:
      begin
        if CurrLabelResizeSide = rsDragging then
          inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec)
        else
          begin
            CalcNewCurrResizePoints;
            {if new cordinates are on the Bmp then save them}
            {  else keep original cordinates.}
            if AreCordinatesOnBmp(CurrDragOriginPoint, CurrDragDrawToPoint, TheBmpAttribRec) then
              begin
                OriginPoint := CurrDragOriginPoint;
                DrawToPoint := CurrDragDrawToPoint;
              end; {if}
            {The GraphWindow will redraw all labels after this event.}
          end; {else}
      end;
    else
      begin
        CurrLabelResizeSide := rsDragging;
        inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec);
      end; {else}
  end; {case}
end; {procedure TDrawInclosedShape.MouseEvent}

{----------------------------------------------------------------------}

procedure TDrawInclosedShape.SaveToDisk(WriteGraphFile : TWriteGraphFile);
begin {procedure TDrawInclosedShape.SaveToDisk;}
  case DrawingTool of
    dtRectangle: WriteGraphFile.WriteLongint(dl_FileType_Rectangle);
    dtEllipse: WriteGraphFile.WriteLongint(dl_FileType_Ellipse);
    dtRoundRect: WriteGraphFile.WriteLongint(dl_FileType_RoundRect);
    else
      WriteGraphFile.WriteLongint(dl_FileType_Rectangle);
  end; {case}
  inherited SaveToDisk(WriteGraphFile);
end; {procedure TDrawInclosedShape.SaveToDisk;}

{----------------------------------------------------------------------}

procedure TDrawInclosedShape.ReadFromDisk(ReadGraphFile : TReadGraphFile;
                                          TheBitmap : TBitmap;
                                          const TheBmpAttribRec : TBitmapAttribRec);

{----------------------------------}

procedure ReadVer1Fields;
begin {procedure ReadVer1Fields;}
  with ReadGraphFile do
    begin
      ReadRecord(FLabelPenStyle, sizeof(FLabelPenStyle));
      SkipBytes(2);
      LabelPenColor := ReadColor;
      LabelPenWidth := ReadSmallint;
      OriginPoint.X := ReadSmallint;
      OriginPoint.Y := ReadSmallint;
      DrawToPoint.X := ReadSmallint;
      DrawToPoint.Y := ReadSmallint;
    end; {with}
end; {procedure ReadVer1Fields;}

{----------------------------------}

begin {procedure TDrawInclosedShape.ReadFromDisk}
  DidFieldsChange := true;
  if ReadGraphFile.FileVersion = 1 then
    ReadVer1Fields
  else {if FileVersion >= 2 then}
    inherited ReadFromDisk(ReadGraphFile, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawInclosedShape.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TDrawInclosedShape.Assign(Source: TDrawInclosedShape);
begin {procedure TDrawInclosedShape.Assign}
  AssignDrawShape(Source);
  DrawingTool := Source.DrawingTool;
end; {procedure TDrawInclosedShape.Assign}

{----------------------------------------------------------------------}

procedure TDrawInclosedShape.AssignSeqObj(Source: TObject);
begin {procedure TDrawInclosedShape.AssignSeqObj}
  Assign(Source as TDrawInclosedShape);
end; {procedure TDrawInclosedShape.AssignSeqObj}

{----------------------------------------------------------------------}

function TDrawInclosedShape.CreateNewSeqObj: TObject;
begin {function TDrawInclosedShape.CreateNewSeqObj}
  Result := TDrawInclosedShape.Create(DrawingTool);
end; {function TDrawInclosedShape.CreateNewSeqObj}

{======================================================================}

procedure TDrawFreeHand.SetLabelPenStyle(ThePenStyle : TPenStyle);
begin {procedure TDrawFreeHand.SetLabelPenStyle}
  DidFieldsChange := true;
  if ThePenStyle = psClear then
    FLabelPenStyle := psSolid
  else
    FLabelPenStyle := ThePenStyle;
end; {procedure TDrawFreeHand.SetLabelPenStyle}

{----------------------------------------------------------------------}

procedure TDrawFreeHand.SetLabelBrushStyle(TheBrushStyle : TBrushStyle);
begin {procedure TDrawFreeHand.SetLabelBrushStyle}
  DidFieldsChange := true;
  if TheBrushStyle in [bsClear, bsSolid] then
    FLabelBrushStyle := TheBrushStyle
  else
    FLabelBrushStyle := bsClear;
end; {procedure TDrawFreeHand.SetLabelBrushStyle}

{----------------------------------------------------------------------}

function TDrawFreeHand.AddNewNode : boolean;
var
  TmpNode : PFreeHandPointsNode;
begin {function TDrawFreeHand.AddNewNode}
  Result := true;
  TmpNode := nil;
  try
    TmpNode := new(PFreeHandPointsNode);
  except
    on EOutOfMemory do
      Result := false;
  end; {try..except}
  if Result then
    begin
      inc(NodeCount);
      RecCountInEndNode := 0;
      with TmpNode^.NodeBoundsRect do
        begin
          TopLeftDeltaX := 0;
          TopLeftDeltaY := 0;
          BottomRightDeltaX := 0;
          BottomRightDeltaY := 0;
        end; {with}
      TmpNode^.NextNode := nil;
      if BeginNode = nil then
        BeginNode := TmpNode;
      if EndNode <> nil then
        EndNode^.NextNode := TmpNode;
      EndNode := TmpNode;
    end; {if}
end; {function TDrawFreeHand.AddNewNode}

{----------------------------------------------------------------------}

procedure TDrawFreeHand.DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                                      TheMode : TDrawLabelModes;
                                      TheBitmap : TBitmap;
                                      const TheBmpAttribRec : TBitmapAttribRec);
var
  I, LinePointArrayCount, NodeArrayI : integer;
  CurrNodePtr : PFreeHandPointsNode;
begin {procedure TDrawFreeHand.DrawWithAMode}
  if PointCount >= 1 then
  with TheBitmap.Canvas do
    begin
      case TheMode of
        dlmErase:
          begin
            Pen.Mode := pmCopy;
            Pen.Color := TheBmpAttribRec.BackgroundColor;
            Brush.Color := TheBmpAttribRec.BackgroundColor;
          end;
        dlmNormalCopy:
          begin
            Pen.Mode := pmCopy;
            Pen.Color := LabelPenColor;
            Brush.Color := LabelBrushColor;
          end;
        dlmNotXor:
          begin
            Pen.Mode := pmNotXor;
            Pen.Color := LabelPenColor;
            Brush.Color := LabelBrushColor;
          end;
      end; {case}
      if LabelBrushStyle in [bsClear, bsSolid] then
        Brush.Style := LabelBrushStyle
      else
        Brush.Style := bsClear;
      if LabelPenStyle = psClear then
        Pen.Style := psSolid
      else
        Pen.Style := LabelPenStyle;
      Pen.Width := LabelPenWidth;
      {draw it}
      if PointCount = 1 then
        begin
          MoveTo(ThePoint1.X, ThePoint1.Y);
          LineTo(ThePoint2.X, ThePoint2.Y);
          Exit;
        end; {if}
      LinePointArrayCount := 0;
      NodeArrayI := 0;
      CurrNodePtr := BeginNode;
      for I := 1 to PointCount do
       begin
         {Add a point to LinePointArray}
         inc(LinePointArrayCount);
         inc(NodeArrayI);
         if NodeArrayI > FreeHandArrayCount then
           begin
             NodeArrayI := 1;
             CurrNodePtr := CurrNodePtr^.NextNode;
           end; {if}
         with CurrNodePtr^.PointArray[NodeArrayI] do
           begin
             LinePointArray[LinePointArrayCount].X := ThePoint1.X + DeltaX;
             LinePointArray[LinePointArrayCount].Y := ThePoint1.Y + DeltaY;
           end; {with}
         {draw a PolyLine if LinePointArray is full}
         if LinePointArrayCount >= LinePointArraySize then
           begin
             PolyLine(LinePointArray);
             LinePointArray[1] := LinePointArray[LinePointArraySize];
             LinePointArrayCount := 1;
           end; {if}
       end; {for}
      {draw last points in LinePointArray}
      if LinePointArrayCount > 1 then
        begin
          if LinePointArrayCount >= LinePointArraySize then
            PolyLine(LinePointArray)
          else
            PolyLine(Slice(LinePointArray, LinePointArrayCount));
        end; {if}
    end; {with}
end; {procedure TDrawFreeHand.DrawWithAMode}

{----------------------------------------------------------------------}

constructor TDrawFreeHand.Create;
begin {constructor TDrawFreeHand.Create}
  inherited;
  with ShapeBoundsRect do
    begin
      TopLeftDeltaX := 0;
      TopLeftDeltaY := 0;
      BottomRightDeltaX := 0;
      BottomRightDeltaY := 0;
    end; {with}
  PointCount := 0;
  NodeCount := 1;
  RecCountInEndNode := 0;
  BeginNode := new(PFreeHandPointsNode);
  EndNode := BeginNode;
  BeginNode^.NextNode := nil;
  with BeginNode^.NodeBoundsRect do
    begin
      TopLeftDeltaX := 0;
      TopLeftDeltaY := 0;
      BottomRightDeltaX := 0;
      BottomRightDeltaY := 0;
    end; {with}
end; {constructor TDrawFreeHand.Create}

{----------------------------------------------------------------------}

destructor TDrawFreeHand.Destroy;
var
  CurrNode, DeleteNode : PFreeHandPointsNode;
begin {destructor TDrawFreeHand.Destroy;}
  CurrNode := BeginNode;
  while CurrNode <> nil do
    begin
      DeleteNode := CurrNode;
      CurrNode := CurrNode^.NextNode;
      Dispose(DeleteNode);
    end; {while}
  inherited Destroy;
end; {destructor TDrawFreeHand.Destroy;}

{----------------------------------------------------------------------}

function TDrawFreeHand.GetLabelNameString : string;
begin {function TDrawFreeHand.GetLabelNameString}
  with OriginPoint do
    Result := 'Free Hand: Origin Point('+IntToStr(X)+', '+IntToStr(Y)+')';
end; {function TDrawFreeHand.GetLabelNameString}

{----------------------------------------------------------------------}

procedure TDrawFreeHand.ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
var
  I, J, MiddleMarksCount : integer;
  CanDrawCurrent : boolean;
  DistanceBetweenMarks : integer;
  CurrTogglePoint, CurrSearchPoint : TPoint;
  TogglePCount, SearchPCount,
  TogglePNodeNdx, SearchPNodeNdx : integer;
  TogglePNode, SearchPNode : PFreeHandPointsNode;

{--------------}

procedure ResetTogglePointVars;
begin {procedure ResetTogglePointVars;}
  TogglePCount := 1;
  TogglePNodeNdx := 1;
  TogglePNode := BeginNode;
  with CurrTogglePoint do
    begin
      X := OriginPoint.X;
      Y := OriginPoint.Y;
    end; {with}
end; {procedure ResetTogglePointVars;}

{--------------}

procedure ResetSearchPointVars;
begin {procedure ResetSearchPointVars;}
  SearchPCount := 1;
  SearchPNodeNdx := 1;
  SearchPNode := BeginNode;
  with CurrSearchPoint do
    begin
      X := OriginPoint.X;
      Y := OriginPoint.Y;
    end; {with}
end; {procedure ResetSearchPointVars;}

{--------------}

procedure GetNextTogglePoint;
var
  SeekNodeNdx : integer;
  TmpDeltaRec : TFreeHandPointRec;
begin {procedure GetNextTogglePoint;}
  TogglePCount := TogglePCount + DistanceBetweenMarks;
  if TogglePCount >= PointCount then
    begin
      with CurrTogglePoint do
        begin
          X := DrawToPoint.X;
          Y := DrawToPoint.Y;
        end; {with}
    end
  else
    begin
      SeekNodeNdx := (TogglePCount div FreeHandArrayCount) + 1;
      while TogglePNodeNdx < SeekNodeNdx do
        begin
          inc(TogglePNodeNdx);
          TogglePNode := TogglePNode^.NextNode;
        end; {while}
      TmpDeltaRec:= TogglePNode^.PointArray[(TogglePCount mod FreeHandArrayCount)+1];
      with CurrTogglePoint do
        begin
          X := OriginPoint.X + TmpDeltaRec.DeltaX;
          Y := OriginPoint.Y + TmpDeltaRec.DeltaY;
        end; {with}
    end; {else}
end; {procedure GetNextTogglePoint;}

{--------------}

procedure GetNextSearchPoint;
var
  SeekNodeNdx : integer;
  TmpDeltaRec : TFreeHandPointRec;
begin {procedure GetNextSearchPoint;}
  SearchPCount := SearchPCount + DistanceBetweenMarks;
  if SearchPCount >= PointCount then
    begin
      with CurrSearchPoint do
        begin
          X := DrawToPoint.X;
          Y := DrawToPoint.Y;
        end; {with}
    end
  else
    begin
      SeekNodeNdx := (SearchPCount div FreeHandArrayCount) + 1;
      while SearchPNodeNdx < SeekNodeNdx do
        begin
          inc(SearchPNodeNdx);
          SearchPNode := SearchPNode^.NextNode;
        end; {while}
      TmpDeltaRec:= SearchPNode^.PointArray[(SearchPCount mod FreeHandArrayCount)+1];
      with CurrSearchPoint do
        begin
          X := OriginPoint.X + TmpDeltaRec.DeltaX;
          Y := OriginPoint.Y + TmpDeltaRec.DeltaY;
        end; {with}
    end; {else}
end; {procedure GetNextSearchPoint;}

{--------------}

function ArePointsFarEnoughAppart(P1, P2 : TPoint) : boolean;
begin {function ArePointsFarEnoughAppart}
  Result := (abs(P1.X - P2.X) > MinSelectEnd)or
            (abs(P1.Y - P2.Y) > MinSelectEnd);
end; {function ArePointsFarEnoughAppart}

{--------------}

begin {procedure TDrawFreeHand.ToggleSelectMarks;}
  FSelectMarksVisible := not FSelectMarksVisible;
  {draw end point select marks}
  DrawSelectBox(OriginPoint.X, OriginPoint.Y, TheBitmap, TheBmpAttribRec);
  if ArePointsFarEnoughAppart(OriginPoint, DrawToPoint) then
    DrawSelectBox(DrawToPoint.X, DrawToPoint.Y, TheBitmap, TheBmpAttribRec);
  {draw middle select marks}
  if PointCount >= 8 then
    begin
      MiddleMarksCount := PointCount div 32;
      if MiddleMarksCount > 25 then
        MiddleMarksCount := 25;
      if MiddleMarksCount < 1 then
        MiddleMarksCount := 1;
      DistanceBetweenMarks := PointCount div (MiddleMarksCount + 1);
      ResetTogglePointVars;
      for I := 1 to MiddleMarksCount do
        begin
          GetNextTogglePoint;
          if ArePointsFarEnoughAppart(OriginPoint, CurrTogglePoint) and
             ArePointsFarEnoughAppart(DrawToPoint, CurrTogglePoint) then
            begin
              CanDrawCurrent := true;
              ResetSearchPointVars;
              for J := 1 to (I-1) do
                begin
                  GetNextSearchPoint;
                  if not ArePointsFarEnoughAppart(CurrSearchPoint,
                                                  CurrTogglePoint) then
                    begin
                      CanDrawCurrent := false;
                      Break;
                    end; {if}
                end; {for}
              if CanDrawCurrent then
                DrawSelectBox(CurrTogglePoint.X, CurrTogglePoint.Y, TheBitmap, TheBmpAttribRec);
            end; {if}
        end; {for}
    end; {if}
end; {procedure TDrawFreeHand.ToggleSelectMarks;}

{----------------------------------------------------------------------}

procedure TDrawFreeHand.MouseEvent(XPix, YPix : integer;
                                   DrawMouseEvent : TDrawMouseEvent;
                                   TheBitmap : TBitmap;
                                   const TheBmpAttribRec : TBitmapAttribRec);
var
  NewDeltaX, NewDeltaY : integer;

{--------------------------------}

procedure SetPenAndBrushForLineTo;
begin {procedure SetPenAndBrushForLineTo;}
  with TheBitmap.Canvas do
    begin
      Brush.Color := LabelBrushColor;
      if LabelBrushStyle in [bsClear, bsSolid] then
        Brush.Style := LabelBrushStyle
      else
        Brush.Style := bsClear;
      Pen.Mode := pmCopy;
      Pen.Color := LabelPenColor;
      Pen.Style := LabelPenStyle;
      Pen.Width := LabelPenWidth;
    end; {with}
end; {procedure SetPenAndBrushForLineTo;}

{--------------------------------}

begin {procedure TDrawFreeHand.MouseEvent}
  DidFieldsChange := true;
  case DrawMouseEvent of
    deDownNew:
      begin
        PointCount := 1;
        RecCountInEndNode := 1;
        OriginPoint.X := XPix;
        OriginPoint.Y := YPix;
        DrawToPoint := OriginPoint;
        with BeginNode^.PointArray[1] do
          begin
            DeltaX := 0;
            DeltaY := 0;
          end;
        SetPenAndBrushForLineTo;
        with TheBitmap.Canvas do
          begin
            MoveTo(XPix, YPix);
            LineTo(XPix, YPix);
          end; {with}
      end;
    deMoveNew, deUpNew:
      begin
        if ((DrawToPoint.X <> XPix) or (DrawToPoint.Y <> YPix))and
           (PointCount <= MaxFreeHandPoints) then
          begin
            if RecCountInEndNode < FreeHandArrayCount then
              Inc(RecCountInEndNode)
            else
              begin
                if AddNewNode then
                  RecCountInEndNode := 1
                else
                  Exit; {an out of memory error has occured}
              end; {else}
            NewDeltaX := XPix - OriginPoint.X;
            NewDeltaY := YPix - OriginPoint.Y;
            with EndNode^.NodeBoundsRect do
              begin
                if NewDeltaX < TopLeftDeltaX then
                  TopLeftDeltaX := NewDeltaX;
                if NewDeltaY < TopLeftDeltaY then
                  TopLeftDeltaY := NewDeltaY;
                if NewDeltaX > BottomRightDeltaX then
                  BottomRightDeltaX := NewDeltaX;
                if NewDeltaY > BottomRightDeltaY then
                  BottomRightDeltaY := NewDeltaY;
              end; {with}
            with ShapeBoundsRect do
              begin
                if NewDeltaX < TopLeftDeltaX then
                  TopLeftDeltaX := NewDeltaX;
                if NewDeltaY < TopLeftDeltaY then
                  TopLeftDeltaY := NewDeltaY;
                if NewDeltaX > BottomRightDeltaX then
                  BottomRightDeltaX := NewDeltaX;
                if NewDeltaY > BottomRightDeltaY then
                  BottomRightDeltaY := NewDeltaY;
              end; {with}
            with EndNode^.PointArray[RecCountInEndNode] do
              begin
                DeltaX := NewDeltaX;
                DeltaY := NewDeltaY;
              end; {with}
            DrawToPoint.X := XPix;
            DrawToPoint.Y := YPix;
            Inc(PointCount);
            SetPenAndBrushForLineTo;
            TheBitmap.Canvas.LineTo(XPix, YPix);
          end; {if}
      end;
    else
      inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec);
  end; {case}
end; {procedure TDrawFreeHand.MouseEvent}

{----------------------------------------------------------------------}

function TDrawFreeHand.DistanceFromPix(XPix, YPix : integer) : integer;
var
  I, J, PointsInCurrArray, CurrDistance : integer;
  CurrNode : PFreeHandPointsNode;

{--------------}

function IsPixNearBoundsRect(const TheBoundsRect :
                                   TFreeHandBoundsRect) : boolean;
begin {function IsPixNearBoundsRect}
  with TheBoundsRect do
  Result := not ((XPix < (TopleftDeltaX + OriginPoint.X - MaxSelectDistance))or
    (XPix > (BottomRightDeltaX + OriginPoint.X + MaxSelectDistance))or
    (YPix < (TopleftDeltaY + OriginPoint.Y - MaxSelectDistance))or
    (YPix > (BottomRightDeltaY + OriginPoint.Y + MaxSelectDistance)));
end; {function IsPixNearBoundsRect}

{--------------}

function NestedDistanceFromLine(P1X, P1Y, P2X, P2Y : integer) : integer;
var
  LoX, HiX, LoY, HiY : integer;
  Slope : Extended;
  Tmp : integer;
begin {function NestedDistanceFromLine}
  if (abs(P1X - P2X) <= 1) and (abs(P1Y - P2Y) <= 1) then
    begin
      {line is to short to test}
      Result := LabelOutOfRangeDist;
      Exit;
    end; {if}
  if P1X >= P2X then
    begin
      HiX := P1X;
      LoX := P2X;
    end
  else
    begin
      LoX := P1X;
      HiX := P2X;
    end; {else}
  if P1Y >= P2Y then
    begin
      HiY := P1Y;
      LoY := P2Y;
    end
  else
    begin
      LoY := P1Y;
      HiY := P2Y;
    end; {else}
  if P1Y = P2Y then {Horizontal Line}
    begin
      if (XPix < LoX)or(XPix > HiX) then
        Result := LabelOutOfRangeDist
      else
        Result := abs(YPix - P1Y);
    end
  else
  if P1X = P2X then {Vertical Line}
    begin
      if (YPix < LoY)or(YPix > HiY) then
        Result := LabelOutOfRangeDist
      else
        Result := abs(XPix - P1X);
    end
  else
    begin
      if (XPix < LoX)or(XPix > HiX)or(YPix < LoY)or(YPix > HiY) then
        Result := LabelOutOfRangeDist
      else
        begin
          { Slope = (Y2 - Y1)/(X2 - X1) }
          Slope:= ((P1Y - P2Y)/
                   (P1X - P2X));
          if abs(Slope) <= 1 then
            begin
                { Y = M(X - X1) + Y1 }
              Tmp := Round(Slope * (XPix - P2X));
              Result := abs(YPix - (Tmp + P2Y));
            end
          else
            begin
                { X = (Y - Y1)/Slope + X1 }
              Tmp := Round((YPix - P2Y) / Slope);
              Result := abs(XPix - (Tmp + P2X));
            end; {else}
          if Result = (MaxSelectDistance + 1) then
            Result := MaxSelectDistance; {account for Rounding}
        end; {else}
    end; {else}
end; {function NestedDistanceFromLine}

{--------------}

begin {function TDrawFreeHand.DistanceFromPix}
  try
    Result := LabelOutOfRangeDist;
    if (PointCount > 0) and IsPixNearBoundsRect(ShapeBoundsRect) then
      begin
        {check OriginPoint}
        CurrDistance := CalcPointsDistance(Point(XPix, YPix), OriginPoint);
        if CurrDistance < Result then
          Result := CurrDistance;
        {check DrawToPoint}
        CurrDistance := CalcPointsDistance(Point(XPix, YPix), DrawToPoint);
        if CurrDistance < Result then
          Result := CurrDistance;
        if Result = 0 then
          Exit; {can't get any closer so quit searching}
        {check middle points}
        CurrNode := BeginNode;
        for I := 1 to NodeCount do
          begin
            if IsPixNearBoundsRect(CurrNode^.NodeBoundsRect) then
              begin
                if I = NodeCount then
                  PointsInCurrArray := RecCountInEndNode
                else
                  PointsInCurrArray := FreeHandArrayCount;
                {begin test distance from points}
                for J := 1 to PointsInCurrArray do
                  begin
                    with CurrNode^.PointArray[J] do
                      CurrDistance := CalcPointsDistance(Point(XPix, YPix),
                        Point((DeltaX+OriginPoint.X), (DeltaY+OriginPoint.Y)));
                    if CurrDistance < Result then
                      Result := CurrDistance;
                  end; {while}
                if Result = 0 then
                  Exit; {can't get any closer so quit searching}
                {end test distance from points}
                {begin test distance from line segments}
                for J := 2 to PointsInCurrArray do
                  begin
                    CurrDistance := NestedDistanceFromLine(
                      (CurrNode^.PointArray[J-1].DeltaX + OriginPoint.X),
                      (CurrNode^.PointArray[J-1].DeltaY + OriginPoint.Y),
                      (CurrNode^.PointArray[J].DeltaX + OriginPoint.X),
                      (CurrNode^.PointArray[J].DeltaY + OriginPoint.Y));
                    if CurrDistance < Result then
                      Result := CurrDistance;
                  end; {while}
                if Result = 0 then
                  Exit; {can't get any closer so quit searching}
                {end test distance from line segments}
              end; {if}
            CurrNode := CurrNode^.NextNode;
          end; {for}
      end; {if}
  except
    on EMathError do
      Result := LabelOutOfRangeDist;
    on EIntError do
      Result := LabelOutOfRangeDist;
  end; {try..except}
end; {function TDrawFreeHand.DistanceFromPix}

{----------------------------------------------------------------------}

function TDrawFreeHand.AreCordinatesOnBmp(ThePoint1, ThePoint2 : TPoint;
                                          const TheBmpAttribRec : TBitmapAttribRec) : boolean;
var
  I, J, PointsInCurrArray : integer;
  CurrNode : PFreeHandPointsNode;
begin {function TDrawFreeHand.AreCordinatesOnBmp}
  with TheBmpAttribRec do
    Result := IsPointOnCanvas(ThePoint1.X, ThePoint1.Y, BmpWidth, BmpHeight) or
              IsPointOnCanvas(ThePoint2.X, ThePoint2.Y, BmpWidth, BmpHeight);
  if not Result then
    begin
      CurrNode := BeginNode;
      for I := 1 to NodeCount do
        begin
          if I = NodeCount then
            PointsInCurrArray := RecCountInEndNode
          else
            PointsInCurrArray := FreeHandArrayCount;
          for J := 1 to PointsInCurrArray do
            begin
              with TheBmpAttribRec do
              with CurrNode^.PointArray[J] do
                Result := IsPointOnCanvas((ThePoint1.X + DeltaX),
                                          (ThePoint1.Y + DeltaY), BmpWidth, BmpHeight);
              if Result then
                Exit; {found a point in the bmp}
            end; {for}
          CurrNode := CurrNode^.NextNode;
        end; {for}
    end; {if}
end; {function TDrawFreeHand.AreCordinatesOnBmp}

{----------------------------------------------------------------------}

procedure TDrawFreeHand.SaveToDisk(WriteGraphFile : TWriteGraphFile);
var
  I : integer;
  CurrNode : PFreeHandPointsNode;
begin {procedure TDrawFreeHand.SaveToDisk;}
  with WriteGraphFile do
    begin
      WriteLongint(dl_FileType_FreeHand);
      WriteRecord(ShapeBoundsRect, sizeof(ShapeBoundsRect));
      WriteLongint(PointCount);
      WriteLongint(NodeCount);
      WriteLongint(RecCountInEndNode);
      CurrNode := BeginNode;
      for I := 1 to NodeCount do
        begin
          with CurrNode^ do
            begin
              WriteRecord(NodeBoundsRect, sizeof(NodeBoundsRect));
              WriteRecord(PointArray, sizeof(PointArray));
            end; {with}
          CurrNode := CurrNode^.NextNode;
        end; {for}
    end; {with}
  inherited SaveToDisk(WriteGraphFile);
end; {procedure TDrawFreeHand.SaveToDisk;}

{----------------------------------------------------------------------}

procedure TDrawFreeHand.ReadFromDisk(ReadGraphFile : TReadGraphFile;
                                     TheBitmap : TBitmap;
                                     const TheBmpAttribRec : TBitmapAttribRec);
var
  I : integer;
  TmpPointCount, TmpNodeCount, TmpRecCountInEndNode : integer;
  CurrNode, NewNode : PFreeHandPointsNode;
begin {procedure TDrawFreeHand.ReadFromDisk}
  DidFieldsChange := true;
  {make sure this is a new class}
  Assert((PointCount = 0)and(NodeCount = 1)and(RecCountInEndNode = 0));
  with ReadGraphFile do
    begin
      ReadRecord(ShapeBoundsRect, sizeof(ShapeBoundsRect));
      TmpPointCount := ReadLongint;
      TmpNodeCount := ReadLongint;
      TmpRecCountInEndNode := ReadLongint;
      if TmpNodeCount >= 1 then
        begin
          {read the first node}
          with BeginNode^ do
            begin
              ReadRecord(NodeBoundsRect, sizeof(NodeBoundsRect));
              ReadRecord(PointArray, sizeof(PointArray));
            end; {with}
          {read the rest of the nodes}
          CurrNode := BeginNode;
          for I := 2 to TmpNodeCount do
            begin
              NewNode := new(PFreeHandPointsNode);
              NewNode^.NextNode := nil;
              {read the new node}
              try
                with NewNode^ do
                  begin
                    ReadRecord(NodeBoundsRect, sizeof(NodeBoundsRect));
                    ReadRecord(PointArray, sizeof(PointArray));
                  end; {with}
              except
                Dispose(NewNode);
                raise;
              end; {try..except}
              CurrNode^.NextNode := NewNode;
              EndNode := NewNode;
              CurrNode := NewNode;
              inc(NodeCount);
            end; {for}
        end; {if}
      PointCount := TmpPointCount;
      NodeCount := TmpNodeCount;
      RecCountInEndNode := TmpRecCountInEndNode;
    end; {with}
  inherited ReadFromDisk(ReadGraphFile, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawFreeHand.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TDrawFreeHand.Assign(Source: TDrawFreeHand);
var
  TmpNewNode, TmpSourceNode : PFreeHandPointsNode;

{--------------}

procedure DoDisposeNodeList;
var
  CurrNode, DeleteNode : PFreeHandPointsNode;
begin {procedure DoDisposeNodeList;}
  CurrNode := BeginNode;
  while CurrNode <> nil do
    begin
      DeleteNode := CurrNode;
      CurrNode := CurrNode^.NextNode;
      Dispose(DeleteNode);
    end; {while}
  PointCount := 0;
  NodeCount := 0;
  RecCountInEndNode := 0;
  BeginNode := nil;
  EndNode := nil;
end; {procedure DoDisposeNodeList;}

{--------------}

begin {procedure TDrawFreeHand.Assign}
  DoDisposeNodeList;
  AssignDrawShape(Source);
  ShapeBoundsRect := Source.ShapeBoundsRect;
  PointCount := Source.PointCount;
  NodeCount := Source.NodeCount;
  RecCountInEndNode := Source.RecCountInEndNode;
  try
    TmpSourceNode := Source.BeginNode;
    while TmpSourceNode <> nil do
      begin
        TmpNewNode := new(PFreeHandPointsNode);
        TmpNewNode^ := TmpSourceNode^;
        TmpNewNode^.NextNode := nil;
        if BeginNode = nil then
          BeginNode := TmpNewNode;
        if EndNode <> nil then
          EndNode^.NextNode := TmpNewNode;
        EndNode := TmpNewNode;
        TmpSourceNode := TmpSourceNode^.NextNode;
      end; {while}
  except
    DoDisposeNodeList;
  end; {try..except}
end; {procedure TDrawFreeHand.Assign}

{----------------------------------------------------------------------}

procedure TDrawFreeHand.AssignSeqObj(Source: TObject);
begin
  Assign(Source as TDrawFreeHand);
end;

{----------------------------------------------------------------------}

function TDrawFreeHand.CreateNewSeqObj: TObject;
begin
  Result := TDrawFreeHand.Create;
end;

{======================================================================}

procedure TDrawLine.SetLabelPenStyle(ThePenStyle : TPenStyle);
begin {procedure TDrawLine.SetLabelPenStyle}
  DidFieldsChange := true;
  if ThePenStyle = psClear then
    FLabelPenStyle := psSolid
  else
    FLabelPenStyle := ThePenStyle;
end; {procedure TDrawLine.SetLabelPenStyle}

{----------------------------------------------------------------------}

procedure TDrawLine.SetLabelBrushStyle(TheBrushStyle : TBrushStyle);
begin {procedure TDrawLine.SetLabelBrushStyle}
  DidFieldsChange := true;
  if TheBrushStyle in [bsClear, bsSolid] then
    FLabelBrushStyle := TheBrushStyle
  else
    FLabelBrushStyle := bsClear;
end; {procedure TDrawLine.SetLabelBrushStyle}

{----------------------------------------------------------------------}

procedure TDrawLine.DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                                  TheMode : TDrawLabelModes;
                                  TheBitmap : TBitmap;
                                  const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawLine.DrawWithAMode}
  with TheBitmap.Canvas do
    begin
      case TheMode of
        dlmErase:
          begin
            Pen.Mode := pmCopy;
            Pen.Color := TheBmpAttribRec.BackgroundColor;
            Brush.Color := TheBmpAttribRec.BackgroundColor;
          end;
        dlmNormalCopy:
          begin
            Pen.Mode := pmCopy;
            Pen.Color := LabelPenColor;
            Brush.Color := LabelBrushColor;
          end;
        dlmNotXor:
          begin
            Pen.Mode := pmNotXor;
            Pen.Color := LabelPenColor;
            Brush.Color := LabelBrushColor;
          end;
      end; {case}
      if LabelBrushStyle in [bsClear, bsSolid] then
        Brush.Style := LabelBrushStyle
      else
        Brush.Style := bsClear;
      if LabelPenStyle = psClear then
        Pen.Style := psSolid
      else
        Pen.Style := LabelPenStyle;
      Pen.Width := LabelPenWidth;
      //Don't need anymore. JunkHandle := Handle; //fix up for VCL bug in TCanvas.LineTo method
      MoveTo(ThePoint1.X, ThePoint1.Y);
      LineTo(ThePoint2.X, ThePoint2.Y);
    end; {with}
end; {procedure TDrawLine.DrawWithAMode}

{----------------------------------------------------------------------}

function TDrawLine.GetLabelNameString : string;
begin {function TDrawLine.GetLabelNameString}
  Result := 'Line: Origin Point('+
            IntToStr(OriginPoint.X)+', '+IntToStr(OriginPoint.Y)+
            ') End Point('+
            IntToStr(DrawToPoint.X)+', '+IntToStr(DrawToPoint.Y)+')';
end; {function TDrawLine.GetLabelNameString}

{----------------------------------------------------------------------}

procedure TDrawLine.ToggleSelectMarks(TheBitmap : TBitmap; const TheBmpAttribRec : TBitmapAttribRec);
var
  MidX, MidY, HalfX, HalfY : integer;
begin {procedure TDrawLine.ToggleSelectMarks}
  FSelectMarksVisible := not FSelectMarksVisible;
  DrawSelectBox(OriginPoint.X, OriginPoint.Y, TheBitmap, TheBmpAttribRec);
  if (abs(OriginPoint.X - DrawToPoint.X) > MinSelectEnd)or
     (abs(OriginPoint.Y - DrawToPoint.Y) > MinSelectEnd)then
     begin
       DrawSelectBox(DrawToPoint.X, DrawToPoint.Y, TheBitmap, TheBmpAttribRec);
       {Calculate Middle X}
       if OriginPoint.X >= DrawToPoint.X then
         begin
           HalfX := (OriginPoint.X - DrawToPoint.X) div 2;
           MidX := HalfX + DrawToPoint.X;
         end
       else
         begin
           HalfX := (DrawToPoint.X - OriginPoint.X) div 2;
           MidX := HalfX + OriginPoint.X;
         end; {else}
       {Calculate Middle Y}
       if OriginPoint.Y >= DrawToPoint.Y then
         begin
           HalfY := (OriginPoint.Y - DrawToPoint.Y) div 2;
           MidY := HalfY + DrawToPoint.Y;
         end
       else
         begin
           HalfY := (DrawToPoint.Y - OriginPoint.Y) div 2;
           MidY := HalfY + OriginPoint.Y;
         end; {else}
       if (HalfX > MinSelectMid)or(HalfY > MinSelectMid)then
         DrawSelectBox(MidX, MidY, TheBitmap, TheBmpAttribRec);
     end; {if}
end; {procedure TDrawLine.ToggleSelectMarks}

{----------------------------------------------------------------------}

procedure TDrawLine.MouseEvent(XPix, YPix : integer;
                               DrawMouseEvent : TDrawMouseEvent;
                               TheBitmap : TBitmap;
                               const TheBmpAttribRec : TBitmapAttribRec);
var
  OriginDist, DrawToDist : integer;

{-------------------}

procedure CalcNewCurrResizePoints;
begin {procedure CalcNewCurrResizePoints;}
  if CurrLabelResizeSide = rsOriginPoint then
    begin
      CurrDragOriginPoint.X := (XPix - CurrMouseDownPoint.X) + OriginPoint.X;
      CurrDragOriginPoint.Y := (YPix - CurrMouseDownPoint.Y) + OriginPoint.Y;
    end
  else {if CurrLabelResizeSide = rsDrawToPoint then}
    begin
     CurrDragDrawToPoint.X := (XPix - CurrMouseDownPoint.X) + DrawToPoint.X;
     CurrDragDrawToPoint.Y := (YPix - CurrMouseDownPoint.Y) + DrawToPoint.Y;
    end; {if}
end; {procedure CalcNewCurrResizePoints;}

{-------------------}

begin {procedure TDrawLine.MouseEvent}
  DidFieldsChange := true;
  case DrawMouseEvent of
    deDownSelectResize:
      begin
        CurrMouseDownPoint.X := XPix;
        CurrMouseDownPoint.Y := YPix;
        CurrDragOriginPoint := OriginPoint;
        CurrDragDrawToPoint := DrawToPoint;
        if not FSelectMarksVisible then
          ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
        {determine which point to resize from}
        CurrLabelResizeSide := rsDragging;
        OriginDist := CalcPointsDistance(Point(XPix, YPix), OriginPoint);
        DrawToDist := CalcPointsDistance(Point(XPix, YPix), DrawToPoint);
        if DrawToDist <= OriginDist then
          begin
            if DrawToDist <= MaxSelectDistance then
              CurrLabelResizeSide := rsDrawToPoint;
          end
        else
          begin
            if OriginDist <= MaxSelectDistance then
              CurrLabelResizeSide := rsOriginPoint;
          end; {else}
      end;
    deMoveBeginResize:
      begin
        if CurrLabelResizeSide = rsDragging then
          inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec)
        else
          begin
            {begin resizing label}
            if FSelectMarksVisible then
              ToggleSelectMarks(TheBitmap, TheBmpAttribRec);
            {erase existing label}
            DrawWithAMode(OriginPoint, DrawToPoint, dlmErase, TheBitmap, TheBmpAttribRec);
            {draw in pmNotXor mode at new cordinates}
            CalcNewCurrResizePoints;
            DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
          end; {else}
      end;
    deMoveContinueResize:
      begin
        if CurrLabelResizeSide = rsDragging then
          inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec)
        else
          begin
            {continue resizing label}
            {erase resizing image at old cordinates}
            DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
            {draw in pmNotXor mode at new cordinates}
            CalcNewCurrResizePoints;
            DrawWithAMode(CurrDragOriginPoint, CurrDragDrawToPoint, dlmNotXor, TheBitmap, TheBmpAttribRec);
          end; {else}
      end;
    deUpEndResize:
      begin
        if CurrLabelResizeSide = rsDragging then
          inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec)
        else
          begin
            CalcNewCurrResizePoints;
            {if new cordinates are on the Bmp then save them}
            {  else keep original cordinates.}
            if AreCordinatesOnBmp(CurrDragOriginPoint, CurrDragDrawToPoint, TheBmpAttribRec) then
              begin
                OriginPoint := CurrDragOriginPoint;
                DrawToPoint := CurrDragDrawToPoint;
              end; {if}
            {The GraphWindow will redraw all labels after this event.}
          end; {else}
      end;
    else
      begin
        CurrLabelResizeSide := rsDragging;
        inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec);
      end; {else}
  end; {case}
end; {procedure TDrawLine.MouseEvent}

{----------------------------------------------------------------------}

function TDrawLine.AreCordinatesOnBmp(ThePoint1, ThePoint2 : TPoint;
                                      const TheBmpAttribRec : TBitmapAttribRec) : boolean;
begin {function TDrawLine.AreCordinatesOnBmp}
  with TheBmpAttribRec do
    Result := ClipLineToBmp(ThePoint1.X, ThePoint2.X,
                            ThePoint1.Y, ThePoint2.Y,
                            (BmpHeight-1), (BmpWidth-1));
end; {function TDrawLine.AreCordinatesOnBmp}

{----------------------------------------------------------------------}

function TDrawLine.DistanceFromPix(XPix, YPix : integer) : integer;
var
  LoX, HiX, LoY, HiY : integer;
  Slope : Extended;
  Tmp : integer;

{-----------------------------}

procedure CalcDistToCorners;
var
  DrawToDistance, OriginDistance : integer;
begin
  DrawToDistance := CalcPointsDistance(Point(XPix, YPix), DrawToPoint);
  OriginDistance := CalcPointsDistance(Point(XPix, YPix), OriginPoint);
  if DrawToDistance <= OriginDistance then
    Result := DrawToDistance
  else
    Result := OriginDistance;
end;

{-----------------------------}

begin {function TDrawLine.DistanceFromPix}
  try
    if OriginPoint.X >= DrawToPoint.X then
      begin
        HiX := OriginPoint.X;
        LoX := DrawToPoint.X;
      end
    else
      begin
        LoX := OriginPoint.X;
        HiX := DrawToPoint.X;
      end; {else}
    if OriginPoint.Y >= DrawToPoint.Y then
      begin
        HiY := OriginPoint.Y;
        LoY := DrawToPoint.Y;
      end
    else
      begin
        LoY := OriginPoint.Y;
        HiY := DrawToPoint.Y;
      end; {else}
    if OriginPoint.Y = DrawToPoint.Y then {Horizontal Line}
      begin
        if (XPix < LoX)or(XPix > HiX) then
          CalcDistToCorners
        else
          Result := abs(YPix - OriginPoint.Y);
      end
    else
    if OriginPoint.X = DrawToPoint.X then {Vertical Line}
      begin
        if (YPix < LoY)or(YPix > HiY) then
          CalcDistToCorners
        else
          Result := abs(XPix - OriginPoint.X);
      end
    else
      begin
        if (XPix < LoX)or(XPix > HiX)or(YPix < LoY)or(YPix > HiY) then
          CalcDistToCorners
        else
          begin
            { Slope = (Y2 - Y1)/(X2 - X1) }
            Slope:= ((OriginPoint.Y - DrawToPoint.Y)/
                     (OriginPoint.X - DrawToPoint.X));
            if abs(Slope) <= 1 then
              begin
                  { Y = M(X - X1) + Y1 }
                Tmp := Round(Slope * (XPix - DrawToPoint.X));
                Result := abs(YPix - (Tmp + DrawToPoint.Y));
              end
            else
              begin
                  { X = (Y - Y1)/Slope + X1 }
                Tmp := Round((YPix - DrawToPoint.Y) / Slope);
                Result := abs(XPix - (Tmp + DrawToPoint.X));
              end; {else}
            if Result = (MaxSelectDistance + 1) then
              Result := MaxSelectDistance; {account for Rounding}
          end; {else}
      end; {else}
  except
    on EMathError do
      Result := LabelOutOfRangeDist;
    on EIntError do
      Result := LabelOutOfRangeDist;
  end; {try..except}
end; {function TDrawLine.DistanceFromPix}

{----------------------------------------------------------------------}

procedure TDrawLine.SaveToDisk(WriteGraphFile : TWriteGraphFile);
begin {procedure TDrawLine.SaveToDisk}
  WriteGraphFile.WriteLongint(dl_FileType_Line);
  inherited SaveToDisk(WriteGraphFile);
end; {procedure TDrawLine.SaveToDisk}

{----------------------------------------------------------------------}

procedure TDrawLine.ReadFromDisk(ReadGraphFile : TReadGraphFile;
                                 TheBitmap : TBitmap;
                                 const TheBmpAttribRec : TBitmapAttribRec);

{----------------------------------}

procedure ReadVer1Fields;
begin {procedure ReadVer1Fields;}
  with ReadGraphFile do
    begin
      ReadRecord(FLabelPenStyle, sizeof(FLabelPenStyle));
      SkipBytes(2);
      LabelPenColor := ReadColor;
      LabelPenWidth := ReadSmallint;
      OriginPoint.X := ReadSmallint;
      OriginPoint.Y := ReadSmallint;
      DrawToPoint.X := ReadSmallint;
      DrawToPoint.Y := ReadSmallint;
    end; {with}
end; {procedure ReadVer1Fields;}

{----------------------------------}

begin {procedure TDrawLine.ReadFromDisk}
  DidFieldsChange := true;
  if ReadGraphFile.FileVersion = 1 then
    ReadVer1Fields
  else {if FileVersion >= 2 then}
    inherited ReadFromDisk(ReadGraphFile, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawLine.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TDrawLine.Assign(Source: TDrawLine);
begin {procedure TDrawLine.Assign}
  AssignDrawShape(Source);
end; {procedure TDrawLine.Assign}

{----------------------------------------------------------------------}

procedure TDrawLine.AssignSeqObj(Source: TObject);
begin {procedure TDrawLine.AssignSeqObj}
  Assign(Source as TDrawLine);
end; {procedure TDrawLine.AssignSeqObj}

{----------------------------------------------------------------------}

function TDrawLine.CreateNewSeqObj: TObject;
begin {function TDrawLine.CreateNewSeqObj}
  Result := TDrawLine.Create;
end; {function TDrawLine.CreateNewSeqObj}

{======================================================================}

procedure TDrawText.DrawWithAMode(ThePoint1, ThePoint2 : TPoint;
                                  TheMode : TDrawLabelModes;
                                  TheBitmap : TBitmap;
                                  const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawText.DrawWithAMode}
  if TheMode <> dlmErase then
  with TheBitmap.Canvas do
    begin
      Brush.Color := clwhite;
      Brush.Style := bsClear;
      case TheMode of
        dlmNormalCopy:
          begin
            Pen.Mode := pmCopy;
            Font.Assign(FLabelTextFont);
            if LabelTextStr <> '' then
              TextOut(ThePoint1.X, ThePoint1.Y, LabelTextStr);
          end;
        dlmNotXor:
          begin
            Pen.Mode := pmNotXor;
            Pen.Style := psDot;
            Pen.Color := clBlack;
            Pen.Width := 1;
            Rectangle(ThePoint1.X, ThePoint1.Y, ThePoint2.X, ThePoint2.Y);
          end;
      end; {case}
    end; {with}
end; {procedure TDrawText.DrawWithAMode}

{----------------------------------------------------------------------}

constructor TDrawText.Create;
begin {constructor TDrawText.Create;}
  inherited;
  FLabelTextFont := CreateANewTFontClass;
end; {constructor TDrawText.Create;}

{----------------------------------------------------------------------}

Destructor TDrawText.Destroy;
begin {Destructor TDrawText.Destroy;}
  FLabelTextFont.free;
  inherited Destroy;
end; {Destructor TDrawText.Destroy;}

{----------------------------------------------------------------------}

function TDrawText.GetLabelNameString : string;
begin {function TDrawText.GetLabelNameString}
  with OriginPoint do
    Result := 'Text: Origin Point('+IntToStr(X)+', '+IntToStr(Y)+
              ') "'+LabelTextStr+'"';
end; {function TDrawText.GetLabelNameString}

{----------------------------------------------------------------------}

function TDrawText.ForceTextToStayOnBmp(const TheBmpAttribRec : TBitmapAttribRec) : boolean;
{Returns true if label was moved}
var
  BmpHeightSub1, BmpWidthSub1 : integer;
{----------------}
procedure MoveX(DeltaX : integer);
begin {procedure MoveX}
  Result := true;
  OriginPoint.X := OriginPoint.X + DeltaX;
  DrawToPoint.X := DrawToPoint.X + DeltaX;
end; {procedure MoveX}
{----------------}
procedure MoveY(DeltaY : integer);
begin {procedure MoveY}
  Result := true;
  OriginPoint.Y := OriginPoint.Y + DeltaY;
  DrawToPoint.Y := DrawToPoint.Y + DeltaY;
end; {procedure MoveY}
{----------------}
begin {procedure TDrawText.ForceTextToStayOnBmp}
  Result := false;
  BmpHeightSub1 := TheBmpAttribRec.BmpHeight - 1;
  BmpWidthSub1 := TheBmpAttribRec.BmpWidth - 1;
  if (OriginPoint.X <= DrawToPoint.X) then
    begin
      if (OriginPoint.X > BmpWidthSub1) then
        MoveX(BmpWidthSub1 - OriginPoint.X)
      else
        if (DrawToPoint.X < 0) then
          MoveX(abs(DrawToPoint.X));
    end
  else
    begin
      if (DrawToPoint.X > BmpWidthSub1) then
        MoveX(BmpWidthSub1 - DrawToPoint.X)
      else
        if (OriginPoint.X < 0) then
          MoveX(abs(OriginPoint.X));
    end; {else}

  if (OriginPoint.Y <= DrawToPoint.Y) then
    begin
      if (OriginPoint.Y > BmpHeightSub1) then
        MoveY(BmpHeightSub1 - OriginPoint.Y)
      else
        if (DrawToPoint.Y < 0) then
          MoveY(abs(DrawToPoint.Y));
    end
  else
    begin
      if (DrawToPoint.Y > BmpHeightSub1) then
        MoveY(BmpHeightSub1 - DrawToPoint.Y)
      else
        if (OriginPoint.Y < 0) then
          MoveY(abs(OriginPoint.Y));
    end; {else}
  if Result then
    DidFieldsChange := true;
end; {procedure TDrawText.ForceTextToStayOnBmp}

{----------------------------------------------------------------------}

procedure TDrawText.SetFontAttributes(TheTopLeftPoint : TPoint;
                                      const TheLabelTextStr : string;
                                      TheLabelTextFont : TFont;
                                      TheBitmap : TBitmap;
                                      const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawText.SetFontAttributes}
  DidFieldsChange := true;
  OriginPoint := TheTopLeftPoint;
  DrawToPoint := CalcTextBottomRight(TheTopLeftPoint,
                     TheLabelTextStr, TheLabelTextFont, TheBitmap, TheBmpAttribRec);
  LabelTextStr := TheLabelTextStr;
  if Assigned(TheLabelTextFont) then
    FLabelTextFont.Assign(TheLabelTextFont);
end; {procedure TDrawText.SetFontAttributes}

{----------------------------------------------------------------------}

procedure TDrawText.MouseEvent(XPix, YPix : integer;
                               DrawMouseEvent : TDrawMouseEvent;
                               TheBitmap : TBitmap;
                               const TheBmpAttribRec : TBitmapAttribRec);
begin {procedure TDrawText.MouseEvent}
  DidFieldsChange := true;
  if not (DrawMouseEvent in [deDownNew, deMoveNew, deUpNew]) then
    inherited MouseEvent(XPix, YPix, DrawMouseEvent, TheBitmap, TheBmpAttribRec);
end; {procedure TDrawText.MouseEvent}

{----------------------------------------------------------------------}

function TDrawText.DistanceFromPix(XPix, YPix : integer) : integer;
begin {function TDrawText.DistanceFromPix}
  if (XPix >= OriginPoint.X)and(XPix <= DrawToPoint.X)and
     (YPix >= OriginPoint.Y)and(YPix <= DrawToPoint.Y) then
    Result := 0
  else
    Result := inherited DistanceFromPix(XPix, YPix);
end; {function TDrawText.DistanceFromPix}

{----------------------------------------------------------------------}

function TDrawText.CalcTextBottomRight(TheTopLeftPoint : TPoint;
           const TheTextStr : string; TheTextFont : TFont;
           TheBitmap : TBitmap;
           const TheBmpAttribRec : TBitmapAttribRec) : TPoint;
begin {function TDrawText.CalcTextBottomRight}
  if TheTextStr = '' then
    Result := TheTopLeftPoint
  else
    begin
      with TheBitmap.Canvas do
        begin
          font.assign(TheTextFont);
          Result.X := TextWidth(TheTextStr) + TheTopLeftPoint.X;
          Result.Y := TextHeight(TheTextStr) + TheTopLeftPoint.Y;
        end; {with}
    end; {else}
end; {function TDrawText.CalcTextBottomRight}

{----------------------------------------------------------------------}

procedure TDrawText.SaveToDisk(WriteGraphFile : TWriteGraphFile);
begin {procedure TDrawText.SaveToDisk;}
  with WriteGraphFile do
    begin
      WriteLongint(dl_FileType_Text);
      WriteString(LabelTextStr);
      WriteFont(FLabelTextFont);
    end; {with}
  inherited SaveToDisk(WriteGraphFile);
end; {procedure TDrawText.SaveToDisk;}

{----------------------------------------------------------------------}

procedure TDrawText.ReadFromDisk(ReadGraphFile : TReadGraphFile;
                                 TheBitmap : TBitmap;
                                 const TheBmpAttribRec : TBitmapAttribRec);

{----------------------------------}

procedure ReadVer1Fields;
begin {procedure ReadVer1Fields;}
  with ReadGraphFile do
    begin
      OriginPoint.X := ReadSmallint;
      OriginPoint.Y := ReadSmallint;
      DrawToPoint.X := ReadSmallint;
      DrawToPoint.Y := ReadSmallint;
      LabelTextStr := Read16BitString(255);
      Read16BitFont(FLabelTextFont);
    end; {with}
end; {procedure ReadVer1Fields;}

{----------------------------------}

procedure ReadVer2_3Fields;
begin {procedure ReadVer2_3Fields;}
  with ReadGraphFile do
    begin
      LabelTextStr := Read32BitString;
      Read32BitFont(FLabelTextFont);
    end; {with}
  inherited ReadFromDisk(ReadGraphFile, TheBitmap, TheBmpAttribRec);
end; {procedure ReadVer2_3Fields;}

{----------------------------------}

begin {procedure TDrawText.ReadFromDisk}
  DidFieldsChange := true;
  if ReadGraphFile.FileVersion = 1 then
    ReadVer1Fields
  else {if FileVersion >= 2 then}
    ReadVer2_3Fields;
  {force text on bmp incase font is different}
  DrawToPoint := CalcTextBottomRight(OriginPoint,LabelTextStr,FLabelTextFont,TheBitmap,TheBmpAttribRec);
  ForceTextToStayOnBmp(TheBmpAttribRec);
end; {procedure TDrawText.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TDrawText.Assign(Source: TDrawText);
begin {procedure TDrawText.Assign}
  AssignDrawLabel(Source);
  LabelTextStr := Source.LabelTextStr;
  UniqueString(LabelTextStr);
  FLabelTextFont.Assign(Source.FLabelTextFont);
end; {procedure TDrawText.Assign}

{----------------------------------------------------------------------}

procedure TDrawText.AssignSeqObj(Source: TObject);
begin {procedure TDrawText.AssignSeqObj}
  Assign(Source as TDrawText);
end; {procedure TDrawText.AssignSeqObj}

{----------------------------------------------------------------------}

function TDrawText.CreateNewSeqObj: TObject;
begin {function TDrawText.CreateNewSeqObj}
  Result := TDrawText.Create;
end; {function TDrawText.CreateNewSeqObj}

{======================================================================}

end.
