unit Fg3dline;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  SysUtils, graphics, W32_asm, PtShared;

const
  MaxColRowCount = 100;

type
  TFixedPoint = longint;

  TTrigAngles = packed record
    CosRadX, CosRadY, CosRadZ,
    SinRadX, SinRadY, SinRadZ : TFixedPoint;
  end; {TTrigAngles}

type
  TLineListError = (ll_NoError, ll_ListFull, ll_NoMem);

const
  FixedMul = 65536;
  FixedShift = 16;
  Max3DDomain = +700;
  Min3DDomain = -700;
  Max3DDomainStr = '+700';
  Min3DDomainStr = '-700';

const
  Max3DFuncCount = 252; {May not be Greater than 252}
  Low3DFuncNum = 0;
  High3DFuncNum = 0;
  XAxisFuncNum = 253;
  YAxisFuncNum = 254;
  ZAxisFuncNum = 255;
  LowFuncColorNum = 0;
  HighFuncColorNum = 255;

type
  TFuncColorArray = packed array[LowFuncColorNum..HighFuncColorNum] of TColor;

type
  PPlotRec = ^TPlotRec;
  TPlotRec = packed record
    XPlot : longint;
    YPlot : longint;
    Z : TFixedPoint;
  end;

  PPointRec = ^TPointRec;
  TPointRec = packed record
    X,
    Y,
    Z : TFixedPoint;
  end;

  PPixRec = ^TPixRec;
  TPixRec = packed record
    Static : TPointRec;
    Rotate : TPlotRec;
  end;

const
  TPixArraySize = 1024;
  TPixPtrArraySize = 64;
  MaxPixElements = 65536;

type
  PPixArray = ^TPixArray;
  {Arrays start at 0 so they will be easier to access with DIV and MOD}
  TPixArray = packed array[0..TPixArraySize-1] of TPixRec;
  TPixPtrArray = packed array[0..TPixPtrArraySize-1] of PPixArray;

  PLineRec = ^TLineRec;
  TLineRec = packed record
    P1,
    P2 : PPixRec;
    FuncNum : byte;
  end;

const
  TLineArraySize = 1024;
  TLinePtrArraySize = 64;
  MaxLineElements = 65535; //Reserve 1 value for the NilLineNumber
  NilLineNumber = 65535;

type
  PLineArray = ^TLineArray;
  {Arrays start at 0 so they will be easier to access with DIV and MOD}
  TLineArray = packed array[0..TLineArraySize-1] of TLineRec;
  TLinePtrArray = packed array[0..TLinePtrArraySize-1] of PLineArray;

type
  TLineObj = class
    private
    { Private declarations }
      ObjFieldsChanged : boolean;
      FLineCount : integer;
      FPixCount : integer;
      FPixPtrArrayCount : integer;
      FLinePtrArrayCount : integer;
      FPixPtrArray : TPixPtrArray;
      FLinePtrArray : TLinePtrArray;
      MarkPixCount : integer;
      FLowestX,
      FHighestX,
      FLowestY,
      FHighestY,
      FLowestZ,
      FHighestZ : TFixedPoint;
      FFuncColorArray : TFuncColorArray;
    protected
    { Protected declarations }
    public
    { Public declarations }
      property FuncColorArray : TFuncColorArray read FFuncColorArray;
      property LineCount : integer read FLineCount;
      property PixCount : integer read FPixCount;
      property PixPtrArrayCount : integer read FPixPtrArrayCount;
      property LinePtrArrayCount : integer read FLinePtrArrayCount;
      property PixPtrArray : TPixPtrArray read FPixPtrArray;
      property LinePtrArray : TLinePtrArray read FLinePtrArray;
      property LowestX : TFixedPoint read FLowestX;
      property HighestX : TFixedPoint read FHighestX;
      property LowestY : TFixedPoint read FLowestY;
      property HighestY : TFixedPoint read FHighestY;
      property LowestZ : TFixedPoint read FLowestZ;
      property HighestZ : TFixedPoint read FHighestZ;
      procedure RotateSort(const InTrigAngles : TTrigAngles;
                       ViewZoom, HalfBmpHeight, HalfBmpWidth : integer;
                       OnIsTerminatedHandler : TIsTerminatedEvent);
      constructor Create;
      destructor Destroy; override;
      procedure ReInit;
      procedure MarkBeginFunc;
      procedure GetLineRec(TheLineNumber : integer;var TheLineRec : TLineRec);
      function AddItem_Points(TheFuncNum : byte;
                   const StaticPoint1, StaticPoint2 : TPointRec;
                   var P1Pointer, P2Pointer : pointer) : TLineListError;
      procedure RegisterColor(TheFuncNumber : integer; TheColor : TColor);
  end; {TLineObj = class}

function CalcHighestDetail(TheLineVal : integer) : integer;
procedure RotateAPix(const InTrigAngles : TTrigAngles;
                     ViewZoom, HalfBmpHeight, HalfBmpWidth : integer;
                     const InPointRec : TPointRec;
                     var OutPlotRec : TPlotRec);
function LineListErrorToString(TheLineListError: TLineListError) : string;

implementation

{------------------------------------------------------------------------}

function CalcHighestDetail(TheLineVal : integer) : integer;
begin {function CalcHighestDetail}
  if TheLineVal <= 0 then
    Result := MaxColRowCount
  else
    if TheLineVal = 1 then
      Result := (MaxColRowCount - 2)
    else
      begin
        {Calc highest detail where ColCount <= MaxColRowCount}
        {  ColCount := ((Circles-1)*LineDetail)+Circles;}
        Result:= abs(trunc((MaxColRowCount - TheLineVal)/(TheLineVal - 1)));
        while(Result > 0)and
             ((((TheLineVal-1)*Result)+TheLineVal) > MaxColRowCount)do
          dec(Result); {account for rounding errors}
      end; {else}
end; {function CalcHighestDetail}

{------------------------------------------------------------------------}

procedure RotateAPix(const InTrigAngles : TTrigAngles;
                     ViewZoom, HalfBmpHeight, HalfBmpWidth : integer;
                     const InPointRec : TPointRec;
                     var OutPlotRec : TPlotRec);
var
  TmpStatic, TmpRotate : TPointRec;
  HoldXCord : TFixedPoint;
  XProj, YProj : longint;

{--------------------------------------}

  function LongMul(ThePoint, TheAngle : TFixedPoint) : TFixedPoint;
  var
    TmpPoint, TmpAngle, TmpAns : Extended;
  begin {function LongMul}
    TmpPoint := ThePoint;
    TmpAngle := TheAngle;
    TmpAns := (TmpPoint * TmpAngle) / 65536; {2^16}
    Result := trunc(TmpAns);
  end; {function LongMul}

{--------------------------------------}

  function SafeMulByZoom(TheFixed : TFixedPoint) : longint;
  var
    TmpAxisCord : Extended;
  begin {function SafeMulByZoom}
    TmpAxisCord := (TheFixed / FixedMul) * ViewZoom;
    Result := trunc(TmpAxisCord);
  end; {function SafeMulByZoom}

{--------------------------------------}

begin {procedure RotateAPix}
 TmpStatic := InPointRec;
 with InTrigAngles do
 with TmpRotate do
   Begin
     {Rotate Parall to X Axis}
     {X := TmpStatic.X;}
     Y := LongMul(TmpStatic.Y,CosRadX)-
           LongMul(TmpStatic.Z,SinRadX);
     Z := LongMul(TmpStatic.Y,SinRadX)+
           LongMul(TmpStatic.Z,CosRadX);
     {Rotate Parall to Y Axis}
     X := LongMul(TmpStatic.X,CosRadY) + LongMul(Z,SinRadY);
     Z := longMul(Z,CosRadY) - LongMul(TmpStatic.X,SinRadY);
     {Rotate Parall to Z Axis}
     HoldXCord := X;
     X := LongMul(X,CosRadZ) - LongMul(Y,SinRadZ);
     Y := LongMul(HoldXCord,SinRadZ) + LongMul(Y,CosRadZ);
     {do ViewZoom and HalfBmpHeight, HalfBmpWidth}
     OutPlotRec.Z := Z;
     XProj := SafeMulByZoom(X);
     YProj := SafeMulByZoom(Y);
     OutPlotRec.XPlot := XProj + HalfBmpWidth;
     OutPlotRec.YPlot := (-1)*YProj+HalfBmpHeight;
   end; {with}
end; {procedure RotateAPix}

{------------------------------------------------------------------------}

procedure TLineObj.RotateSort(const InTrigAngles : TTrigAngles;
                          ViewZoom, HalfBmpHeight, HalfBmpWidth :integer;
                          OnIsTerminatedHandler : TIsTerminatedEvent);

{---------------------------------------------------------}

function LocalDoIsTerminated: boolean;
begin {function LocalDoIsTerminated}
  Result := false;
  if Assigned(OnIsTerminatedHandler) then
    OnIsTerminatedHandler(Result);
end; {function LocalDoIsTerminated}

{---------------------------------------------------------}

(*
procedure Pascal_DoRotations;
var
  I,PixPtrArrayI : integer;
  NumLeftToRotate : integer;
  ITimes : integer;
  CurrPixArray : PPixArray;
  TmpStatic, TmpRotate : TPointRec;
  TrAngles : TTrigAngles;
  HoldXCord : TFixedPoint;
  XProj, YProj : longint;

{--------------------------------------}

  function LongMul(ThePoint, TheAngle : TFixedPoint) : TFixedPoint;
  var
    TmpPoint, TmpAngle, TmpAns : Extended;
  begin {function LongMul}
    TmpPoint := ThePoint;
    TmpAngle := TheAngle;
    TmpAns := (TmpPoint * TmpAngle) / 65536; {2^16}
    Result := trunc(TmpAns);
  end; {function LongMul}

{--------------------------------------}

  function SafeMulByZoom(TheFixed : TFixedPoint) : longint;
  var
    TmpAxisCord : Extended;
  begin {function SafeMulByZoom}
    TmpAxisCord := (TheFixed / FixedMul) * ViewZoom;
    Result := trunc(TmpAxisCord);
  end; {function SafeMulByZoom}

{--------------------------------------}

begin {procedure Pascal_DoRatations}
  TrAngles := InTrigAngles;
  NumLeftToRotate := FPixCount;
  for PixPtrArrayI := 0 to FPixPtrArrayCount-1 do
    begin
      CurrPixArray := FPixPtrArray[PixPtrArrayI];
      if NumLeftToRotate >= TPixArraySize then
        begin
          ITimes := TPixArraySize;
          NumLeftToRotate := NumLeftToRotate - TPixArraySize;
        end {if}
      else
        ITimes := NumLeftToRotate;
      for I := 0 to (ITimes - 1) do
        begin
          TmpStatic := CurrPixArray^[I].Static;
          with TrAngles do
          with TmpRotate do
            Begin
              {Rotate Parall to X Axis}
              {X := TmpStatic.X;}
              Y := LongMul(TmpStatic.Y,CosRadX)-
                    LongMul(TmpStatic.Z,SinRadX);
              Z := LongMul(TmpStatic.Y,SinRadX)+
                    LongMul(TmpStatic.Z,CosRadX);
              {Rotate Parall to Y Axis}
              X := LongMul(TmpStatic.X,CosRadY) + LongMul(Z,SinRadY);
              Z := longMul(Z,CosRadY) - LongMul(TmpStatic.X,SinRadY);
              {Rotate Parall to Z Axis}
              HoldXCord := X;
              X := LongMul(X,CosRadZ) - LongMul(Y,SinRadZ);
              Y := LongMul(HoldXCord,SinRadZ) + LongMul(Y,CosRadZ);
              {do ViewZoom and HalfBmpHeight, HalfBmpWidth}
              CurrPixArray^[I].Rotate.Z := Z;
              XProj := SafeMulByZoom(X);
              YProj := SafeMulByZoom(Y);
              CurrPixArray^[I].Rotate.XPlot := XProj + HalfBmpWidth;
              CurrPixArray^[I].Rotate.YPlot := (-1)*YProj+HalfBmpHeight;
            end; {with}
        end; {for}
    end; {for}
end; {procedure Pascal_DoRotations}
*)

{----------------------------------------------------------}

procedure CallAsm_DoRotations;
var
  PixPtrArrayI : integer;
  NumLeftToRotate : integer;
  ITimes : integer;
  CurrPixArray : PPixArray;
begin {procedure CallAsm_DoRotations;}
  if LocalDoIsTerminated then Exit;
  NumLeftToRotate := FPixCount;
  for PixPtrArrayI := 0 to FPixPtrArrayCount-1 do
    begin
      if LocalDoIsTerminated then Exit;
      CurrPixArray := FPixPtrArray[PixPtrArrayI];
      if NumLeftToRotate >= TPixArraySize then
        begin
          ITimes := TPixArraySize;
          NumLeftToRotate := NumLeftToRotate - TPixArraySize;
        end {if}
      else
        ITimes := NumLeftToRotate;
      if LocalDoIsTerminated then Exit;
      if ITimes >= 1 then
        with InTrigAngles do
          AsmRotateOneArray(CurrPixArray, ITimes,
                            ViewZoom, HalfBmpHeight, HalfBmpWidth,
                            CosRadX, CosRadY, CosRadZ,
                            SinRadX, SinRadY, SinRadZ);
    end; {for}
end; {procedure CallAsm_DoRotations;}

{-------------------------------------------------------------}

begin {procedure TLineObj.RotateSort}
  if LocalDoIsTerminated then Exit;
  if ObjFieldsChanged then
    begin
      with FLinePtrArray[0]^[0] do
        begin
          P1 := addr(FPixPtrArray[0]^[0]);
          P2 := addr(FPixPtrArray[0]^[1]);
          FuncNum := XAxisFuncNum;
        end; {with}
      with FLinePtrArray[0]^[1] do
        begin
          P1 := addr(FPixPtrArray[0]^[2]);
          P2 := addr(FPixPtrArray[0]^[3]);
          FuncNum := YAxisFuncNum;
        end; {with}
      with FLinePtrArray[0]^[2] do
        begin
          P1 := addr(FPixPtrArray[0]^[4]);
          P2 := addr(FPixPtrArray[0]^[5]);
          FuncNum := ZAxisFuncNum;
        end; {with}
      with FPixPtrArray[0]^[0].Static do
        begin
          X := LowestX;
          Y := 0;
          Z := 0;
        end; {with}
      with FPixPtrArray[0]^[1].Static do
        begin
          X := HighestX;
          Y := 0;
          Z := 0;
        end; {with}
      with FPixPtrArray[0]^[2].Static do
        begin
          X := 0;
          Y := LowestY;
          Z := 0;
        end; {with}
      with FPixPtrArray[0]^[3].Static do
        begin
          X := 0;
          Y := HighestY;
          Z := 0;
        end; {with}
      with FPixPtrArray[0]^[4].Static do
        begin
          X := 0;
          Y := 0;
          Z := LowestZ;
        end; {with}
      with FPixPtrArray[0]^[5].Static do
        begin
          X := 0;
          Y := 0;
          Z := HighestZ;
        end; {with}
    end; {if}
  if LocalDoIsTerminated then Exit;
  if (FPixCount > 6) and (FLineCount > 3) then
    begin
      {begin Rotate}
      CallAsm_DoRotations;
      {Pascal_DoRotations;}
    end; {if}
end; {procedure TLineObj.RotateSort}

{------------------------------------------------------------------------}

constructor TLineObj.Create;
var
  I : integer;
begin
  inherited Create;
  ObjFieldsChanged := true;
  FLineCount := 3;
  FPixCount := 6;
  FPixPtrArrayCount := 1;
  FLinePtrArrayCount := 1;
  MarkPixCount := 7;
  FPixPtrArray[0] := New(PPixArray);
  FLinePtrArray[0] := New(PLineArray);
  for I := 1 to TPixPtrArraySize-1 do
    FPixPtrArray[I] := nil;
  for I := 1 to TLinePtrArraySize-1 do
    FLinePtrArray[I] := nil;
  FLowestX := 0;
  FHighestX := 0;
  FLowestY := 0;
  FHighestY := 0;
  FLowestZ := 0;
  FHighestZ := 0;
  for I := LowFuncColorNum to HighFuncColorNum do {Init array to valid values}
    FFuncColorArray[I] := clblack;
end; {constructor TLineObj.Create;}

{------------------------------------------------------------------------}

destructor TLineObj.Destroy;
var
  I : integer;
begin
  for I := 0 to FPixPtrArrayCount-1 do
    dispose(FPixPtrArray[I]);
  for I := 0 to FLinePtrArrayCount-1 do
    dispose(FLinePtrArray[I]);
  inherited Destroy;
end; {destructor TLineObj.Destroy;}

{------------------------------------------------------------------------}

procedure TLineObj.ReInit;
var
  I : integer;
begin
  for I := 1 to FPixPtrArrayCount-1 do
    dispose(FPixPtrArray[I]);
  for I := 1 to FLinePtrArrayCount-1 do
    dispose(FLinePtrArray[I]);
  ObjFieldsChanged := true;
  FLineCount := 3;
  FPixCount := 6;
  FPixPtrArrayCount := 1;
  FLinePtrArrayCount := 1;
  MarkPixCount := 7;
  for I := 1 to TPixPtrArraySize-1 do
    FPixPtrArray[I] := nil;
  for I := 1 to TLinePtrArraySize-1 do
    FLinePtrArray[I] := nil;
  FLowestX := 0;
  FHighestX := 0;
  FLowestY := 0;
  FHighestY := 0;
  FLowestZ := 0;
  FHighestZ := 0;
  for I := LowFuncColorNum to HighFuncColorNum do {Init array to valid values}
    FFuncColorArray[I] := clblack;
end; {procedure TLineObj.ReInit;}

{------------------------------------------------------------------------}

procedure TLineObj.MarkBeginFunc;
begin {procedure TLineObj.MarkBeginFunc;}
  if FPixCount = 0 then
    MarkPixCount := 1
  else
    MarkPixCount := FPixCount + 1;
end; {procedure TLineObj.MarkBeginFunc;}

{------------------------------------------------------------------------}

procedure TLineObj.GetLineRec(TheLineNumber : integer;
                              var TheLineRec : TLineRec);
begin {procedure TLineObj.GetLineRec}
  TheLineRec := FLinePtrArray[(TheLineNumber div TLineArraySize)]^
                             [(TheLineNumber mod TLineArraySize)];
end; {procedure TLineObj.GetLineRec}

{------------------------------------------------------------------------}

function TLineObj.AddItem_Points(TheFuncNum : byte;
                  const StaticPoint1,StaticPoint2 : TPointRec;
                  var P1Pointer, P2Pointer : pointer) : TLineListError;
{ Returns an error state }
{ Possible States are ll_NoError = 0; ll_ListFull = 1; ll_NoMem = 2}
var
  ErStatus : TLineListError;
  PixRec1, PixRec2 : PPixRec;
  ThePointRec : TPointRec;

{---------------------------------------}

procedure SetPossibleHiLoVals;
begin {procedure SetPossibleHiLoVals}
  if (ThePointRec.X < FLowestX) then
    FLowestX := ThePointRec.X
  else
    if (ThePointRec.X > FHighestX) then
      FHighestX := ThePointRec.X;

  if (ThePointRec.Y < FLowestY) then
    FLowestY := ThePointRec.Y
  else
    if (ThePointRec.Y > FHighestY) then
      FHighestY := ThePointRec.Y;

  if (ThePointRec.Z < FLowestZ) then
    FLowestZ := ThePointRec.Z
  else
    if (ThePointRec.Z > FHighestZ) then
      FHighestZ := ThePointRec.Z;
end; {procedure SetPossibleHiLoVals}

{---------------------------------------}

  function PutPix : PPixRec;
  var
    PixI : integer;
  begin {function PutPix}
    Result := nil;
    PixI := FPixCount mod TPixArraySize;
    if (PixI = 0) then
      begin
        try
          FPixPtrArray[FPixPtrArrayCount] := new(PPixArray);
        except
          on EOutOfMemory do ErStatus := ll_NoMem;
        end; {except}
        if ErStatus = ll_NoError then
          begin
            Result := addr(FPixPtrArray[FPixPtrArrayCount]^[0]);
            inc(FPixPtrArrayCount);
            inc(FPixCount);
            Result^.Static := ThePointRec;
          end; {if}
      end {if}
    else
      begin
        inc(FPixCount);
        Result := addr(FPixPtrArray[(FPixPtrArrayCount-1)]^[PixI]);
        Result^.Static := ThePointRec;
      end; {else}
  end; {function PutPix}

{---------------------------------------}

  procedure PutLine;
  var
    LineI : integer;
  begin {procedure PutLine}
    LineI := FLineCount mod TLineArraySize;
    if LineI = 0 then
      begin
        try
          FLinePtrArray[FLinePtrArrayCount] := new(PLineArray);
        except
          on EOutOfMemory do ErStatus := ll_NoMem;
        end; {except}
        if ErStatus = ll_NoError then
          begin
            with FLinePtrArray[FLinePtrArrayCount]^[0] do
              begin
                FuncNum := TheFuncNum;
                P1 := PixRec1;
                P2 := PixRec2;
              end; {with}
            inc(FLinePtrArrayCount);
            inc(FLineCount);
          end; {if}
      end {if}
    else
      begin
        inc(FLineCount);
        with FLinePtrArray[(FLinePtrArrayCount-1)]^[LineI] do
          begin
            FuncNum := TheFuncNum;
            P1 := PixRec1;
            P2 := PixRec2;
          end; {with}
      end; {else}
  end; {procedure PutLine}

{---------------------------------------}

  function SearchAndPutPix : PPixRec;
  var
    I,CompCount : integer;
    DidFind : boolean;
    CompPoint : PPixRec;
  begin {function SearchAndPutPix;}
    CompPoint := nil;
    DidFind := False;
    for CompCount := FPixCount downto MarkPixCount do
      begin
        {can optomize the div and mod here}
        I := CompCount - 1;
        CompPoint := addr(FPixPtrArray[(I div TPixArraySize)]^
                                      [(I mod TPixArraySize)]);
        with CompPoint^.Static do
          if (X = ThePointRec.X)and(Y = ThePointRec.Y)and
             (Z = ThePointRec.Z) then
            begin
              DidFind := true;
              break; {CompCount := MarkPixCount;} {leave for loop}
            end; {if}
      end; {for}
    if DidFind then
      SearchAndPutPix := CompPoint
    else
      SearchAndPutPix := PutPix;
  end; {function SearchAndPutPix;}

{---------------------------------------}

begin {function TLineObj.AddItem_Points}
  ObjFieldsChanged := true;
  ErStatus := ll_NoError;
  if (FPixCount > (MaxPixElements-2))or
        (FLineCount > (MaxLineElements-1)) then
    ErStatus := ll_ListFull
  else
    begin
      if P1Pointer = nil then
        begin
          ThePointRec := StaticPoint1;
          SetPossibleHiLoVals;
          PixRec1 := PutPix;
          P1Pointer := PixRec1;
        end
      else
        PixRec1 := P1Pointer;
      if ErStatus = ll_NoError then
        begin
          if (StaticPoint1.X = StaticPoint2.X)and
             (StaticPoint1.Y = StaticPoint2.Y)and
             (StaticPoint1.Z = StaticPoint2.Z) then
            begin
              PixRec2 := PixRec1;
              PutLine;
            end {if}
          else
            begin
              if P2Pointer = nil then
                begin
                  ThePointRec := StaticPoint2;
                  SetPossibleHiLoVals;
                  PixRec2 := SearchAndPutPix;
                  P2Pointer := PixRec2;
                end
              else
                PixRec2 := P2Pointer;
              if ErStatus = ll_NoError then
                PutLine;
            end; {else}
        end {if}
    end; {else}
  Result := ErStatus;
end; {function TLineObj.AddItem_Points}

{------------------------------------------------------------------------}

procedure TLineObj.RegisterColor(TheFuncNumber : integer; TheColor : TColor);
begin {procedure TLineObj.RegisterColor}
  ObjFieldsChanged := true;
  if (TheFuncNumber >= LowFuncColorNum) and
     (TheFuncNumber <= HighFuncColorNum) then
    begin
      FFuncColorArray[TheFuncNumber] := TheColor;
    end; {if}
end; {procedure TLineObj.RegisterColor}

{------------------------------------------------------------------------}

function LineListErrorToString(TheLineListError: TLineListError) : string;
begin {function LineListErrorToString}
  Result := '';
  case TheLineListError of
    ll_NoError: Result := '';
    ll_ListFull: Result := '3D window full, partial function will be displayed';
    ll_NoMem: Result := 'Out of memory';
    else
      Result := 'Undetermined error';
  end; {case}
end; {function LineListErrorToString}

{------------------------------------------------------------------------}

end.
