unit Filesave;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Windows, SysUtils, Controls, Dialogs, CentrDlg, Graphics, Jpeg,
  MathCalc, MathFunc, ReadEdit, classes;

{$I+}

const
  grType2D = 1;
  grType2DPolar = 2;
  grType3D = 3;

type
  TGraphFileBaseClass = class
  private
    { Private declarations }
    FileHandle : integer;
    FIsFileOpen : boolean;
  protected
    { Protected declarations }
    function DoOpenFileRead(const FString : string) : boolean;
    function DoOpenFileRewrite(const FString : string) : boolean;
    function DoGetFileSize(out TheFileSize : integer) : boolean;
    function DoFileSeek(TheFilePosition : integer) : boolean;
    function DoFileReadRecord(var TheRec; TheRecSize : integer) : integer;
    function DoFileWriteRecord(const TheRec; TheRecSize : integer) : integer;
    procedure DoCloseFile;
  public
    { Public declarations }
    property IsFileOpen : boolean read FIsFileOpen;
    constructor Create;
    destructor Destroy; override;
  end; {TGraphFileBaseClass}

const
  CheckSumBufferSize = 4096;
type
  TCheckSumBufferArray = packed array[1..CheckSumBufferSize] of byte;

type
  TReadGraphFile = class(TGraphFileBaseClass)
  private
    { Private declarations }
    CheckSumBufferArray : TCheckSumBufferArray;
    MathFunction : TMathFunction;
    FProgramVersion : string;
    FFileVersion : longint;
    FGraphType : longint;
    FCurrentDecSeparator : char;
  public
    { Public declarations }
    property ProgramVersion : string read FProgramVersion;
    property FileVersion : longint read FFileVersion;
    property GraphType : longint read FGraphType;
    property CurrentDecSeparator : char read FCurrentDecSeparator;
    constructor Create;
    destructor Destroy; override;
    function EditStrReadErrorDlg(const TheErrorString : string; var TheEditString : string): boolean;
    function OpenAndCheckFile(const FString : string; TheCurrDecSeparator : char) : boolean;
    procedure CloseDataFile;
    procedure FixMathString(var TheStr : string);
    procedure FixCalcMathConstString(var TheString : string; var TheMathValue : TMathValue);
    procedure ReadRecord(var TheRec; TheRecSize : integer);
    function ReadCreateBitmap : TBitmap;
    function ReadCreateJPEGImage : TJPEGImage;
    procedure Read16BitFont(TheFont : TFont);
    procedure Read32BitFont(TheFont : TFont);
    function Read16BitPoint : TPoint;
    function Read32BitPoint : TPoint;
    function ReadBoolean : boolean;
    function ReadByte : byte;
    function ReadWord : word;
    function ReadSmallint : Smallint;
    function ReadLongint : Longint;
    function ReadExtended : Extended;
    function ReadColor : TColor;
    function Read16BitString(TheStrCharCount : integer) : string;
    function Read32BitString : string;
    function ReadMathString : string;
    procedure ReadMathConstString(var TheString : string; var TheMathValue : TMathValue; IsEmpty : boolean);
    procedure SkipBytes(SkipCount : integer);
    function ReadMathValue : TMathValue;
    procedure RaiseExceptionLoading;
  end; {TReadGraphFile}

  TWriteGraphFile = class(TGraphFileBaseClass)
  private
    { Private declarations }
    FGraphType : longint;
    CurrCheckSum : word;
    FCurrentDecSeparator : char;
    procedure CheckSum(const TheRec; TheRecSize : integer);
  public
    { Public declarations }
    property GraphType : longint read FGraphType;
    property CurrentDecSeparator : char read FCurrentDecSeparator;
    constructor Create;
    destructor Destroy; override;
    function RewriteSaveFile(const FString : string;
                             TheGraphType : longint;
                             TheCurrDecSeparator : char) : boolean;
    function CloseValidFile : boolean;
    procedure CloseFileAfterError;
    procedure FixMathString(var TheStr : string);
    procedure WriteRecord(const TheRec; TheRecSize : integer);
    procedure WriteBitmap(TheBitmap : TBitmap);
    procedure WriteJPEGImage(TheJPEGImage : TJPEGImage);
    procedure WriteFont(TheFont : TFont);
    procedure WritePoint(ThePoint : TPoint);
    procedure WriteBoolean(TheBoolean : boolean);
    procedure WriteByte(TheByte : byte);
    procedure WriteWord(TheWord : word);
    procedure WriteSmallint(TheSmallint : Smallint);
    procedure WriteLongint(TheLongint : Longint);
    procedure WriteExtended(TheExtended : Extended);
    procedure WriteColor(TheColor : TColor);
    procedure WriteString(const TheStr : string);
    procedure WriteMathString(const TheStr : string);
    procedure WriteMathValue(const TheMathValue : TMathValue);
  end; {TWriteGraphFile}

procedure FileMsgViewCenterChanged;

implementation

const
  ApplicationNameStr = 'MathGV'; {TradeMark name used in program.}
  msgExceptLoadingFileStr = 'Error loading file';
  msgExceptSavingFileStr = 'Error saving file';
  msgErOpening = 'Error opening file';
  msgErCorruptedFile = 'Error in file; file data is corrupted';
  msgErReading = 'Error reading from file';
  msgErCreating = 'Error creating file';
  msgErWriting = 'Error writing to file';
  msgErIncorrectFileType = 'File is not a ' + ApplicationNameStr +
                           ' data file, and can not be opened';
  msgErReadingSize = 'Error reading file size';
  msgErVeifyingExisting = 'Error verifying existing file type';

const {these constants define the current file and program version numbers}
  TheFileVerNum = 4;
  ValidFileVerNums = [1..4];
  {length(ProgramVerNumberStr) must be <= ProgramVerStrSize}
  ProgramVerNumber = '4.1';
  ProgramVerStrSize = 10;

type
  TProgramVerStr = string[ProgramVerStrSize];
  TMVHeaderStr = string[37];
  TMVVerHeader = packed record
    ProgramVer : TProgramVerStr;
    FileVer : longint;
    GraphType : longint;
    DecimalSepar, Reserv2, Reserv3, Reserv4 : byte; //before ver 3 Reserved : longint;
  end; {TMVVerHeader}

var
  {length byte is saved in file}
  MVHeaderStr : TMVHeaderStr = 'Mathematical Function Graph Data File';

{---------------------------------------------------------------------}

procedure FileMsgViewCenterChanged;
begin {procedure FileMsgViewCenterChanged;}
  MessageDlgCenter('The graph view center has been changed for compatibility reasons.',
                   mtInformation, [mbOk], 0, nil);
end; {procedure FileMsgViewCenterChanged;}

{---------------------------------------------------------------------}

{ TGraphFileBaseClass }

constructor TGraphFileBaseClass.Create;
begin {constructor TGraphFileBaseClass.Create}
  inherited;
  FileHandle := -1;
  FIsFileOpen := false;
end; {constructor TGraphFileBaseClass.Create}

{---------------------------------------------------------------------}

destructor TGraphFileBaseClass.Destroy;
begin {destructor TGraphFileBaseClass.Destroy}
  DoCloseFile;
  inherited;
end; {destructor TGraphFileBaseClass.Destroy}

{---------------------------------------------------------------------}

procedure TGraphFileBaseClass.DoCloseFile;
begin {procedure TGraphFileBaseClass.DoCloseFile}
  if FIsFileOpen then
    FileClose(FileHandle);
  FileHandle := -1;
  FIsFileOpen := false;
end; {procedure TGraphFileBaseClass.DoCloseFile}

{---------------------------------------------------------------------}

function TGraphFileBaseClass.DoFileReadRecord(var TheRec;
  TheRecSize: integer): integer;
begin {function TGraphFileBaseClass.DoFileReadRecord}
  Assert(TheRecSize > 0, 'Error the TheRecSize <= 0 in function TGraphFileBaseClass.DoFileReadRecord');
  Assert(FIsFileOpen, 'File not open error in function TGraphFileBaseClass.DoFileReadRecord');
  Result := FileRead(FileHandle, TheRec, TheRecSize);
  if Result <= -1 then Result := 0;
end; {function TGraphFileBaseClass.DoFileReadRecord}

{---------------------------------------------------------------------}

function TGraphFileBaseClass.DoFileSeek(TheFilePosition: integer): boolean;
var
  NewPos : integer;
begin {function TGraphFileBaseClass.DoFileSeek}
  Assert(FIsFileOpen, 'File not open error in function TGraphFileBaseClass.DoFileSeek');
  Assert(TheFilePosition >= 0, 'Error file seek position < 0');
  NewPos := FileSeek(FileHandle, TheFilePosition, 0);
  Result := (NewPos <> -1) and (NewPos = TheFilePosition);
end; {function TGraphFileBaseClass.DoFileSeek}

{---------------------------------------------------------------------}

function TGraphFileBaseClass.DoFileWriteRecord(const TheRec;
  TheRecSize: integer): integer;
begin {function TGraphFileBaseClass.DoFileWriteRecord}
  Assert(FIsFileOpen, 'File not open error in function TGraphFileBaseClass.DoFileWriteRecord');
  Assert(TheRecSize > 0, 'Error the TheRecSize <= 0 in function TGraphFileBaseClass.DoFileWriteRecord');
  Result := FileWrite(FileHandle, TheRec, TheRecSize);
  if Result <= -1 then Result := 0;
end; {function TGraphFileBaseClass.DoFileWriteRecord}

{---------------------------------------------------------------------}

function TGraphFileBaseClass.DoGetFileSize(
  out TheFileSize: integer): boolean;
var
  FS : Cardinal;
begin {function TGraphFileBaseClass.DoGetFileSize}
  Assert(FIsFileOpen, 'File not open error in function TGraphFileBaseClass.DoGetFileSize');
  Result := false;
  FS := GetFileSize(FileHandle, nil);
  if FS <> $FFFFFFFF then //check for error
    if FS < (2147483647 - 128) then  //limit file size to integer range minus 128
      begin
        TheFileSize := FS;
        Result := true;
      end; {if}
end; {function TGraphFileBaseClass.DoGetFileSize}

{---------------------------------------------------------------------}

function TGraphFileBaseClass.DoOpenFileRead(
  const FString: string): boolean;
begin {function TGraphFileBaseClass.DoOpenFileRead}
  Assert(FString <> '', 'Empty file string in function DoOpenFileRead');
  Assert(FIsFileOpen = false, 'File already open error in function TGraphFileBaseClass.DoOpenFileRead');
  FileHandle := FileOpen(FString, (fmOpenRead or fmShareDenyWrite));
  FIsFileOpen := FileHandle >= 0;
  Result := FIsFileOpen;
end; {function TGraphFileBaseClass.DoOpenFileRead}

{---------------------------------------------------------------------}

function TGraphFileBaseClass.DoOpenFileRewrite(
  const FString: string): boolean;
begin {function TGraphFileBaseClass.DoOpenFileRewrite}
  Assert(FString <> '', 'Empty file string in function DoOpenFileRewrite');
  Assert(FIsFileOpen = false, 'File already open error in function TGraphFileBaseClass.DoOpenFileRewrite');
  FileHandle := FileCreate(FString);
  FIsFileOpen := FileHandle >= 0;
  Result := FIsFileOpen;
end; {function TGraphFileBaseClass.DoOpenFileRewrite}

{---------------------------------------------------------------------}

{ TReadGraphFile }

constructor TReadGraphFile.Create;
begin {constructor TReadGraphFile.Create;}
  inherited;
  FProgramVersion := '';
  FFileVersion := 0;
  FGraphType := 0;
  FCurrentDecSeparator := '.';
  MathFunction := TMathFunction.Create;
end; {constructor TReadGraphFile.Create;}

{---------------------------------------------------------------------}

destructor TReadGraphFile.Destroy;
begin {destructor TReadGraphFile.Destroy;}
  MathFunction.Free;
  inherited;
end; {destructor TReadGraphFile.Destroy;}

{---------------------------------------------------------------------}

function TReadGraphFile.EditStrReadErrorDlg(const TheErrorString: string;
  var TheEditString: string): boolean;
begin {function TReadGraphFile.EditStrReadErrorDlg}
  Result := ShowEditStrReadErrorDlg(TheErrorString, TheEditString);
end; {function TReadGraphFile.EditStrReadErrorDlg}

{---------------------------------------------------------------------}

function TReadGraphFile.OpenAndCheckFile(const FString: string; TheCurrDecSeparator : char): boolean;
var
  FSize : longint;
  LoadedMVHeaderStr : TMVHeaderStr;
  LoadedMVVerHeader : TMVVerHeader;
  CurrCheckSum : word;

{------------------------------------------}

function DoOpenForInput : boolean;
var
  TryAgain : boolean;
begin {function DoOpenForInput}
  repeat
    Result := DoOpenFileRead(FString);
    if (not Result) then
      TryAgain := (mrRetry = MessageDlgCenter(msgErOpening,mtError,[mbAbort,mbRetry],0,nil))
    else
      TryAgain := false;
  until (not TryAgain);
end; {function DoOpenForInput}

{------------------------------------------}

function ReadCurrFileSize : boolean;
begin {function ReadCurrFileSize}
  Result := DoGetFileSize(FSize);
  if not Result then
    MessageDlgCenter(msgErReadingSize, mtError, [mbOk], 0, nil)
  else
    begin
      if not (FSize > (sizeof(TMVHeaderStr) + sizeof(TMVVerHeader) +
                       (2*sizeof(CurrCheckSum)))) then
        begin
          Result := false;
          MessageDlgCenter(msgErIncorrectFileType, mtError, [mbOk], 0, nil);
        end; {if}
    end; {else}
end; {function ReadCurrFileSize}

{------------------------------------------}

procedure CheckSum(const TheRec; TheRecSize : integer);
type
  TByteArrayCast = packed array[1..maxint] of byte;
var
  I : integer;
begin {function CheckSum}
  {$Q-}{$R-}
  for I := 1 to TheRecSize do
    CurrCheckSum := CurrCheckSum + TByteArrayCast(TheRec)[I];
end; {function CheckSum}

{------------------------------------------}

function VerifyHeaderStr : boolean;
type
  TByteArrayCast = packed array[0..(sizeof(LoadedMVHeaderStr)-1)] of byte;
var
  I : integer;
begin {function VerifyHeaderStr}
  Result := true;
  try
    ReadRecord(LoadedMVHeaderStr, sizeof(LoadedMVHeaderStr));
  except
    Result := false;
    Exit;
  end; {try..except}
  for I := 0 to (sizeof(LoadedMVHeaderStr)-1) do
    if TByteArrayCast(LoadedMVHeaderStr)[I] <> TByteArrayCast(MVHeaderStr)[I] then
      begin
        Result := false;
        Break;
      end; {if}
  if Result then
    CheckSum(LoadedMVHeaderStr, sizeof(LoadedMVHeaderStr))
  else
    MessageDlgCenter(msgErIncorrectFileType, mtError, [mbOk], 0, nil);
end; {function VerifyHeaderStr}

{------------------------------------------}

function VerifyHeaderRec : boolean;
var
  TheHeaderCheckSum : word;
begin {function VerifyHeaderRec}
  Result := true;
  try
    ReadRecord(LoadedMVVerHeader, sizeof(LoadedMVVerHeader));
  except
    Result := false;
    Exit;
  end; {try..except}
  CheckSum(LoadedMVVerHeader, sizeof(LoadedMVVerHeader));
  try
    ReadRecord(TheHeaderCheckSum, sizeof(TheHeaderCheckSum));
  except
    Result := false;
    Exit;
  end; {try..except}
  if TheHeaderCheckSum <> CurrCheckSum then
    begin
      Result := false;
      MessageDlgCenter(msgErCorruptedFile, mtError, [mbOk], 0, nil);
      Exit;
    end; {if}
  CheckSum(TheHeaderCheckSum, sizeof(TheHeaderCheckSum));
end; {function VerifyHeaderRec}

{------------------------------------------}

function ParceHeaderRec : boolean;
begin {function ParceHeaderRec}
  { verify that File Version Number is compatable }
  Result := LoadedMVVerHeader.FileVer in ValidFileVerNums;
  if not Result then
    begin
      if (Length(LoadedMVVerHeader.ProgramVer) >
          (sizeof(LoadedMVVerHeader.ProgramVer)-1))or
         (Length(LoadedMVVerHeader.ProgramVer) = 0) then{corrupted Length Byte}
        MessageDlgCenter(msgErCorruptedFile, mtError, [mbOk], 0, nil)
      else
        MessageDlgCenter('File created by version number '+
                         LoadedMVVerHeader.ProgramVer+
                         ', and can not be read by this version.',
                         mtError, [mbOk], 0, nil);
      Exit;
    end; {if}
  FProgramVersion := LoadedMVVerHeader.ProgramVer;
  FFileVersion := LoadedMVVerHeader.FileVer;
  FGraphType := LoadedMVVerHeader.GraphType;
end; {function ParceHeaderRec}

{------------------------------------------}

function FinishFileCheckSum : boolean;
var
  BytesLeftToRead : integer;
  I : integer;
  TheFileCheck : word;
begin {function FinishFileCheckSum}
  Result := false;
  {Read Rest of the Data}
  BytesLeftToRead := FSize - (sizeof(TMVHeaderStr) +
                 sizeof(TMVVerHeader) + (2*sizeof(CurrCheckSum)));
  repeat
    if BytesLeftToRead >= CheckSumBufferSize then
      I := CheckSumBufferSize
    else
      I := BytesLeftToRead;
    BytesLeftToRead := BytesLeftToRead - CheckSumBufferSize;
    try
      if I > 0  then
        begin
          ReadRecord(CheckSumBufferArray, I);
          CheckSum(CheckSumBufferArray, I);
        end; {if}
    except
      Exit;
    end; {try..except}
  until (BytesLeftToRead <= 0);
  {Read Check Sum From End of File}
  try
    ReadRecord(TheFileCheck, sizeof(TheFileCheck));
  except
    Exit;
  end; {try..except}
  if TheFileCheck <> CurrCheckSum then
    begin
      MessageDlgCenter(msgErCorruptedFile, mtError, [mbOk], 0, nil);
      Exit;
    end; {if}
  {Put file position at end of header}
  if not DoFileSeek((sizeof(TMVHeaderStr) + sizeof(TMVVerHeader) + sizeof(CurrCheckSum))) then
    begin
      MessageDlgCenter(msgErReading, mtError, [mbOk], 0, nil);
      Exit;
    end; {if}
  Result := true;
end; {function FinishFileCheckSum}

{------------------------------------------}

begin {function TReadGraphFile.OpenAndCheckFile}
  FCurrentDecSeparator := TheCurrDecSeparator;
  CurrCheckSum := 0;
  LoadedMVHeaderStr := '';
  CloseDataFile;
  Result := false;
  if FString <> '' then
  if DoOpenForInput then
  if ReadCurrFileSize then
  if VerifyHeaderStr then
  if VerifyHeaderRec then
  if ParceHeaderRec then
  if FinishFileCheckSum then
    Result := true;
  if not Result then
    CloseDataFile;
end; {function TReadGraphFile.OpenAndCheckFile}

{---------------------------------------------------------------------}

procedure TReadGraphFile.CloseDataFile;
begin {procedure TReadGraphFile.CloseDataFile}
  DoCloseFile;
end; {procedure TReadGraphFile.CloseDataFile}

{---------------------------------------------------------------------}

procedure TReadGraphFile.FixMathString(var TheStr: string);
var
  I : integer;
begin {procedure TReadGraphFile.FixMathString}
  if CurrentDecSeparator <> '.' then
    begin
      for I := 1 to length(TheStr) do
        if TheStr[I] = '.' then
          TheStr[I] := CurrentDecSeparator;
    end; {if}
end; {procedure TReadGraphFile.FixMathString}

{---------------------------------------------------------------------}

procedure TReadGraphFile.FixCalcMathConstString(var TheString: string;
  var TheMathValue: TMathValue);
var
  DidOK, VerifySuccessful : boolean;
  TmpValueStr : string;
begin {procedure TReadGraphFile.FixCalcMathConstString}
  if TheString = '' then
    Exit; //do nothing if string is empty
  FixMathString(TheString);
  MathFunction.FuncString := TheString;
  if MathFunction.VerifyMathFunc then
    TheMathValue := MathFunction.MathResultValue
  else
    begin
      InitMathValue(TheMathValue);
      TmpValueStr := TheString;
      VerifySuccessful := false;
      repeat
        DidOK := ShowEditStrReadErrorDlg(MathFunction.GetLastErrorString, TmpValueStr);
        if DidOK then
          begin
            MathFunction.FuncString := TmpValueStr;
            VerifySuccessful := MathFunction.VerifyMathFunc;
          end; {if}
      until (not DidOK) or (VerifySuccessful);
      if VerifySuccessful then
        begin
          TheString := TmpValueStr;
          TheMathValue := MathFunction.MathResultValue;
        end
      else
        begin
          raise Exception.Create(msgExceptLoadingFileStr);
        end; {else}
    end; {else}
end; {procedure TReadGraphFile.FixCalcMathConstString}

{---------------------------------------------------------------------}

procedure TReadGraphFile.ReadRecord(var TheRec; TheRecSize: integer);
var
  TryAgain : boolean;
  CountRead : integer;
begin {procedure TReadGraphFile.ReadRecord}
  repeat
    TryAgain := false;
    CountRead := DoFileReadRecord(TheRec, TheRecSize);
    if CountRead <> TheRecSize then
      begin
        if CountRead = 0 then
          TryAgain := (mrRetry = MessageDlgCenter(msgErReading,mtError,[mbAbort,mbRetry],0,nil))
        else
          MessageDlgCenter(msgErReading,mtError,[mbOk],0,nil);
        if not TryAgain then
          raise Exception.Create(msgExceptLoadingFileStr);
      end; {if}
  until (not TryAgain);
end; {procedure TReadGraphFile.ReadRecord}

{---------------------------------------------------------------------}

procedure TReadGraphFile.Read16BitFont(TheFont: TFont);
var
  {redeclared for ver2 Name : TFontName;}
  TmpName : string;
  TmpColor : TColor;
  TmpStyle : TFontStyles;
  TmpSize : Smallint;
begin {procedure TReadGraphFile.Read16BitFont}
  TmpName := Read16BitString(31);
  TmpColor := ReadColor;
  ReadRecord(TmpStyle, sizeof(TmpStyle));
  TmpSize := ReadSmallint;
  TheFont.Name := TmpName;
  TheFont.Color := TmpColor;
  TheFont.Style := TmpStyle;
  TheFont.Size := TmpSize;
end; {procedure TReadGraphFile.Read16BitFont}

{---------------------------------------------------------------------}

procedure TReadGraphFile.Read32BitFont(TheFont: TFont);
var
  {redeclared for ver2 Name : TFontName;}
  TmpName : string;
  TmpColor : TColor;
  TmpStyle : TFontStyles;
  TmpSize : Longint;
begin {procedure TReadGraphFile.Read32BitFont}
  TmpName := Read32BitString;
  TmpColor := ReadColor;
  ReadRecord(TmpStyle, sizeof(TmpStyle));
  TmpSize := ReadLongint;
  TheFont.Name := TmpName;
  TheFont.Color := TmpColor;
  TheFont.Style := TmpStyle;
  TheFont.Size := TmpSize;
end; {procedure TReadGraphFile.Read32BitFont}

{---------------------------------------------------------------------}

function TReadGraphFile.Read16BitPoint : TPoint;
var
  X, Y : Smallint;
begin {function TReadGraphFile.Read16BitPoint}
  ReadRecord(X, sizeof(X));
  ReadRecord(Y, sizeof(Y));
  Result.X := X;
  Result.Y := Y;
end; {function TReadGraphFile.Read16BitPoint}

{---------------------------------------------------------------------}

function TReadGraphFile.Read32BitPoint : TPoint;
var
  X, Y : longint;
begin {function TReadGraphFile.Read32BitPoint}
  ReadRecord(X, sizeof(X));
  ReadRecord(Y, sizeof(Y));
  Result.X := X;
  Result.Y := Y;
end; {function TReadGraphFile.Read32BitPoint}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadBoolean: boolean;
begin {function TReadGraphFile.ReadBoolean}
  ReadRecord(Result, sizeof(Result));
end; {function TReadGraphFile.ReadBoolean}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadByte: byte;
begin {function TReadGraphFile.ReadByte}
  ReadRecord(Result, sizeof(Result));
end; {function TReadGraphFile.ReadByte}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadWord: word;
begin {function TReadGraphFile.ReadWord}
  ReadRecord(Result, sizeof(Result));
end; {function TReadGraphFile.ReadWord}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadSmallint: Smallint;
begin {function TReadGraphFile.ReadSmallint}
  ReadRecord(Result, sizeof(Result));
end; {function TReadGraphFile.ReadSmallint}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadLongint: Longint;
begin {function TReadGraphFile.ReadLongint}
  ReadRecord(Result, sizeof(Result));
end; {function TReadGraphFile.ReadLongint}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadExtended: Extended;
begin {function TReadGraphFile.ReadExtended}
  ReadRecord(Result, sizeof(Result));
end; {function TReadGraphFile.ReadExtended}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadColor: TColor;
begin {function TReadGraphFile.ReadColor}
  ReadRecord(Result, sizeof(Result));
end; {function TReadGraphFile.ReadColor}

{---------------------------------------------------------------------}

function TReadGraphFile.Read16BitString(TheStrCharCount: integer): string;
var
  TmpS : string[255];
begin {function TReadGraphFile.Read16BitString}
  Assert((TheStrCharCount >= 1) and (TheStrCharCount <= 255));
  ReadRecord(TmpS, (TheStrCharCount + 1));
  Result := TmpS;
end; {function TReadGraphFile.Read16BitString}

{---------------------------------------------------------------------}

function TReadGraphFile.Read32BitString: string;
var
  CharCount : integer;
begin {function TReadGraphFile.Read32BitString}
  Result := '';
  CharCount := ReadLongint;
  if CharCount > 0 then
    begin
      SetLength(Result, CharCount);
      ReadRecord(PChar(Result)^, CharCount);
    end; {if}
end; {function TReadGraphFile.Read32BitString}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadMathString: string;
begin {function TReadGraphFile.ReadMathString}
  Result := Read32BitString;
  FixMathString(Result);
end; {function TReadGraphFile.ReadMathString}

{---------------------------------------------------------------------}

procedure TReadGraphFile.ReadMathConstString(var TheString: string;
  var TheMathValue: TMathValue; IsEmpty : boolean);
begin {procedure TReadGraphFile.ReadMathConstString}
  TheString := Read32BitString;
  if not IsEmpty then
    FixCalcMathConstString(TheString, TheMathValue);
end; {procedure TReadGraphFile.ReadMathConstString}

{---------------------------------------------------------------------}

procedure TReadGraphFile.SkipBytes(SkipCount : integer);
var
  ReadCount : integer;
  SkipedBuffer : packed array[1..16] of byte;
begin {procedure TReadGraphFile.SkipBytes}
  Assert(SkipCount > 0);
  repeat
    if SkipCount >= 16 then
      ReadCount := 16
    else
      ReadCount := SkipCount;
    SkipCount := SkipCount - ReadCount;
    if ReadCount > 0 then
      ReadRecord(SkipedBuffer, ReadCount);
  until SkipCount <= 0;
end; {procedure TReadGraphFile.SkipBytes}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadMathValue: TMathValue;
var
  PackedMathValue : TPackedMathValue;
begin {function TReadGraphFile.ReadMathValue}
  ReadRecord(PackedMathValue, Sizeof(PackedMathValue));
  Result := UnPackMathValue(PackedMathValue);
end; {function TReadGraphFile.ReadMathValue}

{---------------------------------------------------------------------}

procedure TReadGraphFile.RaiseExceptionLoading;
begin {procedure TReadGraphFile.RaiseExceptionLoading;}
  MessageDlgCenter(msgExceptLoadingFileStr, mtError, [mbOk], 0, nil);
  raise Exception.Create(msgExceptLoadingFileStr);
end; {procedure TReadGraphFile.RaiseExceptionLoading;}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadCreateBitmap: TBitmap;
var
  MemS : TMemoryStream;
  PicSize : integer;
begin {function TReadGraphFile.ReadCreateBitmap}
  PicSize := ReadLongint;
  MemS := TMemoryStream.Create;
  try
    MemS.SetSize(PicSize);
    Result := TBitmap.Create;
    try
      Result.HandleType := bmDIB;
      ReadRecord((MemS.Memory)^, PicSize);
      MemS.Position := 0;
      Result.LoadFromStream(MemS);
    except
      Result.free;
    end; {try..except}
  finally
    MemS.free;
  end; {try..finally}
end; {function TReadGraphFile.ReadCreateBitmap}

{---------------------------------------------------------------------}

function TReadGraphFile.ReadCreateJPEGImage: TJPEGImage;
var
  MemS : TMemoryStream;
  PicSize : integer;
begin {function TReadGraphFile.ReadCreateJPEGImage}
  PicSize := ReadLongint;
  MemS := TMemoryStream.Create;
  try
    MemS.SetSize(PicSize);
    Result := TJPEGImage.Create;
    try
      ReadRecord((MemS.Memory)^, PicSize);
      MemS.Position := 0;
      Result.LoadFromStream(MemS);
    except
      Result.free;
    end; {try..except}
  finally
    MemS.free;
  end; {try..finally}
end; {function TReadGraphFile.ReadCreateJPEGImage}

{---------------------------------------------------------------------}

{ TWriteGraphFile }

procedure TWriteGraphFile.CheckSum(const TheRec; TheRecSize: integer);
type
  TByteArrayCast = packed array[1..maxint] of byte;
var
  I : integer;
begin {procedure TWriteGraphFile.CheckSum}
  {$Q-}{$R-}
  for I := 1 to TheRecSize do
    CurrCheckSum := CurrCheckSum + TByteArrayCast(TheRec)[I];
end; {procedure TWriteGraphFile.CheckSum}

{---------------------------------------------------------------------}

constructor TWriteGraphFile.Create;
begin {constructor TWriteGraphFile.Create;}
  inherited;
  FGraphType := grType2D;
  CurrCheckSum := 0;
  FCurrentDecSeparator := '.';
end; {constructor TWriteGraphFile.Create;}

{---------------------------------------------------------------------}

destructor TWriteGraphFile.Destroy;
begin {destructor TWriteGraphFile.Destroy;}
  CloseFileAfterError;
  inherited;
end; {destructor TWriteGraphFile.Destroy;}

{---------------------------------------------------------------------}

function TWriteGraphFile.RewriteSaveFile(const FString: string;
  TheGraphType: Integer; TheCurrDecSeparator : char): boolean;

{------------------------------------------}

function PromptOnSaveToExisting : boolean;
type
  TByteArrayCast = packed array[0..(sizeof(TMVHeaderStr)-1)] of byte;
var
  LoadedMVHeaderStr : TMVHeaderStr;
  TheFileSize : longint;
  GotMathFileHeaderStr : boolean;
  I : integer;
begin {function PromptOnSaveToExisting}
  LoadedMVHeaderStr := '';
  Result := true;
  if not DoOpenFileRead(FString) then
    begin
      DoCloseFile;
      Exit;
    end; {if}
  if not DoGetFileSize(TheFileSize) then
    begin
      DoCloseFile;
      Result := false;
      MessageDlgCenter(msgErVeifyingExisting, mtError, [mbOk], 0, nil);
      Exit;
    end; {if}
  GotMathFileHeaderStr := false;
  if (TheFileSize > (sizeof(TMVHeaderStr) + sizeof(TMVVerHeader)+
                    (2*sizeof(CurrCheckSum)))) then
    begin
      if DoFileReadRecord(LoadedMVHeaderStr, sizeof(LoadedMVHeaderStr))
         <> sizeof(LoadedMVHeaderStr) then
        begin
          DoCloseFile;
          Result := false;
          MessageDlgCenter(msgErVeifyingExisting, mtError, [mbOk], 0, nil);
          Exit;
        end; {if}
      DoCloseFile;
      GotMathFileHeaderStr := true;
      for I := 0 to (sizeof(LoadedMVHeaderStr)-1) do
        if TByteArrayCast(LoadedMVHeaderStr)[I] <> TByteArrayCast(MVHeaderStr)[I] then
          begin
            GotMathFileHeaderStr := false;
            Break;
          end; {if}
    end; {if}
  DoCloseFile;
  if Result and (not GotMathFileHeaderStr) then
    begin
      if mrCancel = MessageDlgCenter('Existing file is not a '+ApplicationNameStr+
                                     ' data file, do you want to overwrite?',
                                     mtWarning, [mbOk,mbCancel], 0, nil) then
        Result := false;
    end; {if}
end; {function PromptOnSaveToExisting}

{------------------------------------------}

function OpenForOutput : boolean;
var
  TryAgain : boolean;
begin {function OpenForOutput}
  repeat
    Result := DoOpenFileRewrite(FString);
    if (not Result) then
      TryAgain := (mrRetry = MessageDlgCenter(msgErCreating,mtError,[mbAbort,mbRetry],0,nil))
    else
      TryAgain := false;
  until (not TryAgain);
end; {function OpenForOutput}

{------------------------------------------}

function HeaderToSaveFile : boolean;
var
  MVVerHeader : TMVVerHeader;
  CurrSavedHeaderChS : word;
  I : integer;
begin {function HeaderToSaveFile}
  Result := true;
  with MVVerHeader do
    begin
      ProgramVer := ProgramVerNumber;
      for I := (Length(ProgramVerNumber)+1) to ProgramVerStrSize do
        ProgramVer[I] := ' ';
      FileVer := TheFileVerNum;
      GraphType := grType2D;
      DecimalSepar := byte(TheCurrDecSeparator);
      Reserv2 := 0;
      Reserv3 := 0;
      Reserv4 := 0;
    end; {with}
  try
    WriteRecord(MVHeaderStr, sizeof(MVHeaderStr));
    MVVerHeader.GraphType := TheGraphType;
    WriteRecord(MVVerHeader, sizeof(MVVerHeader));
    {Save Header CheckSum}
    CurrSavedHeaderChS := CurrCheckSum;
    WriteRecord(CurrSavedHeaderChS, sizeof(CurrSavedHeaderChS));
  except
    Result := false;
  end; {try..except}
end; {function HeaderToSaveFile}

{------------------------------------------}

begin {function RewriteSaveFile}
  Result := false;
  FCurrentDecSeparator := TheCurrDecSeparator;
  CurrCheckSum := 0;
  CloseFileAfterError;
  if FString <> '' then
  if PromptOnSaveToExisting then
  if OpenForOutput then
  if HeaderToSaveFile then
    Result := true;
  if not Result then
    CloseFileAfterError;
end; {function RewriteSaveFile}

{---------------------------------------------------------------------}

function TWriteGraphFile.CloseValidFile: boolean;
var
  TmpCheckSum : word;
begin {function TWriteGraphFile.CloseValidFile}
  Result := true;
  if IsFileOpen then
    begin
      {begin write Check Sum to end of file}
      try
        TmpCheckSum := CurrCheckSum;
        WriteRecord(TmpCheckSum, sizeof(TmpCheckSum));
      except
        Result := false;
      end; {try..except}
      {end write Check Sum to end of file}
      DoCloseFile;
    end; {if}
end; {function TWriteGraphFile.CloseValidFile}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.CloseFileAfterError;
begin {procedure TWriteGraphFile.CloseFileAfterError}
  DoCloseFile;
end; {procedure TWriteGraphFile.CloseFileAfterError}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.FixMathString(var TheStr: string);
var
  I : integer;
begin {procedure TWriteGraphFile.FixMathString}
  if CurrentDecSeparator <> '.' then
    begin
      for I := 1 to length(TheStr) do
        if TheStr[I] = CurrentDecSeparator then
          TheStr[I] := '.';
    end; {if}
end; {procedure TWriteGraphFile.FixMathString}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteRecord(const TheRec; TheRecSize: integer);
var
  TryAgain : boolean;
  CountWritten : integer;
begin {procedure TWriteGraphFile.WriteRecord}
  repeat
    TryAgain := false;
    CountWritten := DoFileWriteRecord(TheRec, TheRecSize);
    if CountWritten <> TheRecSize then
      begin
        if CountWritten = 0 then
          TryAgain := (mrRetry = MessageDlgCenter(msgErWriting,mtError,[mbAbort,mbRetry],0,nil))
        else
          MessageDlgCenter(msgErWriting,mtError,[mbOk],0, nil);
        if not TryAgain then
          raise Exception.Create(msgExceptSavingFileStr);
      end; {if}
  until (not TryAgain);
  CheckSum(TheRec, TheRecSize);
end; {procedure TWriteGraphFile.WriteRecord}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteFont(TheFont: TFont);
var
  {redeclared for ver2 Name : TFontName;}
  TmpName : string;
  TmpColor : TColor;
  TmpStyle : TFontStyles;
  TmpSize : Longint;
begin {procedure TWriteGraphFile.WriteFont}
  TmpName := TheFont.Name;
  TmpColor := TheFont.Color;
  TmpStyle := TheFont.Style;
  TmpSize := TheFont.Size;
  WriteString(TmpName);
  WriteColor(TmpColor);
  WriteRecord(TmpStyle, sizeof(TmpStyle));
  WriteLongint(TmpSize);
end; {procedure TWriteGraphFile.WriteFont}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WritePoint(ThePoint : TPoint);
begin {procedure TWriteGraphFile.WritePoint}
  with ThePoint do
    begin
      WriteLongint(X);
      WriteLongint(Y);
    end; {with}
end; {procedure TWriteGraphFile.WritePoint}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteBoolean(TheBoolean: boolean);
begin {procedure TWriteGraphFile.WriteBoolean}
  WriteRecord(TheBoolean, sizeof(TheBoolean));
end; {procedure TWriteGraphFile.WriteBoolean}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteByte(TheByte: byte);
begin {procedure TWriteGraphFile.WriteByte}
  WriteRecord(TheByte, sizeof(TheByte));
end; {procedure TWriteGraphFile.WriteByte}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteWord(TheWord: word);
begin {procedure TWriteGraphFile.WriteWord}
  WriteRecord(TheWord, sizeof(TheWord));
end; {procedure TWriteGraphFile.WriteWord}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteSmallint(TheSmallint: Smallint);
begin {procedure TWriteGraphFile.WriteSmallint}
  WriteRecord(TheSmallint, sizeof(TheSmallint));
end; {procedure TWriteGraphFile.WriteSmallint}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteLongint(TheLongint: Integer);
begin {procedure TWriteGraphFile.WriteLongint}
  WriteRecord(TheLongint, sizeof(TheLongint));
end; {procedure TWriteGraphFile.WriteLongint}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteColor(TheColor: TColor);
begin {procedure TWriteGraphFile.WriteColor}
  WriteRecord(TheColor, sizeof(TheColor));
end; {procedure TWriteGraphFile.WriteColor}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteExtended(TheExtended: Extended);
begin {procedure TWriteGraphFile.WriteExtended}
  WriteRecord(TheExtended, sizeof(TheExtended));
end; {procedure TWriteGraphFile.WriteExtended}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteString(const TheStr: string);
var
  CharCount : longint;
begin {procedure TWriteGraphFile.WriteString}
  CharCount := Length(TheStr);
  WriteLongint(CharCount);
  if CharCount > 0 then
    WriteRecord(PChar(TheStr)^, CharCount);
end; {procedure TWriteGraphFile.WriteString}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteMathString(const TheStr: string);
var
  TmpString : string;
begin {procedure TWriteGraphFile.WriteMathString}
  TmpString := TheStr;
  FixMathString(TmpString);
  WriteString(TmpString);
end; {procedure TWriteGraphFile.WriteMathString}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteMathValue(const TheMathValue: TMathValue);
var
  PackedMathValue : TPackedMathValue;
begin {procedure TWriteGraphFile.WriteMathValue}
  PackedMathValue := PackMathValue(TheMathValue);
  WriteRecord(PackedMathValue, Sizeof(PackedMathValue));
end; {procedure TWriteGraphFile.WriteMathValue}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteBitmap(TheBitmap: TBitmap);
var
  MemS : TMemoryStream;
  PicSize : integer;
begin {procedure TWriteGraphFile.WriteBitmap}
  MemS := TMemoryStream.Create;
  try
    TheBitmap.SaveToStream(MemS);
    MemS.Position := 0;
    PicSize := MemS.Size;
    WriteLongint(PicSize);
    WriteRecord(MemS.Memory^, PicSize);
  finally
    MemS.free;
  end; {try..finally}
end; {procedure TWriteGraphFile.WriteBitmap}

{---------------------------------------------------------------------}

procedure TWriteGraphFile.WriteJPEGImage(TheJPEGImage: TJPEGImage);
var
  MemS : TMemoryStream;
  PicSize : integer;
begin {procedure TWriteGraphFile.WriteJPEGImage}
  MemS := TMemoryStream.Create;
  try
    TheJPEGImage.SaveToStream(MemS);
    MemS.Position := 0;
    PicSize := MemS.Size;
    WriteLongint(PicSize);
    WriteRecord(MemS.Memory^, PicSize);
  finally
    MemS.free;
  end; {try..finally}
end; {procedure TWriteGraphFile.WriteJPEGImage}

{---------------------------------------------------------------------}

end.
