unit MathCalc;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Sysutils, dialogs;

type
  PMathFloat = ^TMathFloat;
  TMathFloat = Extended;

  TMathValueType = (vtEmpty, vtFloat, vtFraction);

  PMathValue = ^TMathValue;
  TMathValue = record
    Numerator,
    Denominator : Int64;
    FloatValue : TMathFloat;
    ValueType : TMathValueType;
  end; {TMathValue}

  PPackedMathValue = ^TPackedMathValue;
  TPackedMathValue = packed record
    Numerator,
    Denominator : Int64;
    FloatValue : TMathFloat;
    ValueType : TMathValueType;
  end; {TPackedMathValue}

procedure InitMathValue(var MathValue : TMathValue);
function ReturnInitMathValue : TMathValue;

function PackMathValue(const TheMathValue : TMathValue) : TPackedMathValue;
function UnPackMathValue(const ThePackedMathValue : TPackedMathValue) : TMathValue;

function IsMathValueReducedFraction(const TheMathValue : TMathValue) : boolean;

function FloatToMathValue(const MathFloat : TMathFloat) : TMathValue;
function LongintToMathValue(TheInteger : longint) : TMathValue;
function FractionToMathValue(const TheNumer, TheDenom : Int64) : TMathValue;

function AddMathValue(MathValue1, MathValue2 : TMathValue) : TMathValue;
function SubMathValue(MathValue1, MathValue2 : TMathValue) : TMathValue;
function MulMathValue(MathValue1, MathValue2 : TMathValue) : TMathValue;
function DivMathValue(MathValue1, MathValue2 : TMathValue;
                      var DivByZeroError : boolean) : TMathValue;
function PowerMathValue(MathValue1, MathValue2 : TMathValue;
                        var InvalidPowerError : boolean) : TMathValue;

function ExpFuncMathValue(MathValue : TMathValue) : TMathValue;
function GrintFuncMathValue(MathValue : TMathValue) : TMathValue;
function IntFuncMathValue(MathValue : TMathValue) : TMathValue;
function SqrtFuncMathValue(MathValue : TMathValue; var FuncOutDomainError : boolean) : TMathValue;
function SquareFuncMathValue(MathValue : TMathValue) : TMathValue;
function RoundFuncMathValue(MathValue : TMathValue) : TMathValue;
function FracFuncMathValue(MathValue : TMathValue) : TMathValue;
function AbsFuncMathValue(MathValue : TMathValue) : TMathValue;
function FactFuncMathValue(MathValue : TMathValue;
                           var FuncOutDomainError : boolean;
                           var FactToBigError : boolean) : TMathValue;

function StringToMathValue(RealNumString : string;
                           var ErrorPos : integer;
                           var MathValue : TMathValue) : boolean;

implementation

uses
  Primes;

const
  MaxFracValue = High(Int64);    // 9,223,372,036,854,775,807 =  2^63-1
  MinFracValue = Low(Int64) + 1; //-9,223,372,036,854,775,807 = -2^63-1
  MaxFracDecimalDigits = 18;

{----------------------------------------------------------------------}

function MoveNegativeToNumerator(var MathValue : TMathValue) : boolean;
begin {function MoveNegativeToNumerator}
  Result := true;
  if MathValue.ValueType = vtFraction then
    begin
      if (MathValue.Numerator = low(Int64)) or (MathValue.Denominator = low(Int64)) then
        begin
          Result := false;
          MathValue.ValueType := vtFloat;
        end
      else
        if MathValue.Denominator < 0 then
         begin
           MathValue.Numerator := MathValue.Numerator * (-1);
           MathValue.Denominator := abs(MathValue.Denominator);
         end; {if}
    end; {if}
end; {function MoveNegativeToNumerator}

{----------------------------------------------------------------------}

function IsMathFractionEqualToZero(const MathValue : TMathValue) : boolean;
begin {function IsMathFractionEqualToZero}
  Result := False;
  if MathValue.FloatValue = 0 then
    Result := True
  else
    if MathValue.ValueType = vtFraction then
      if MathValue.Numerator = 0 then
        Result := true;
end; {function IsMathFractionEqualToZero}

{----------------------------------------------------------------------}

function CountBits(TheInt64 : Int64) : integer;
type
  TBArray = packed array[0..7] of byte;
var
  I, SigByteIndex : integer;
  SigByte, BitMask : byte;
begin {function CountBits}
  Result := 64;
  if TheInt64 = Low(Int64) then
    Exit;
  TheInt64 := abs(TheInt64);
  SigByteIndex := -1;
  for I := 7 downto 0 do
    begin
      SigByte := TBArray(TheInt64)[I];
      if SigByte <> 0 then
        begin
          SigByteIndex := I;
          Break;
        end; {if}
    end; {for}
  if SigByteIndex < 0 then
    begin
      Result := 0;
      Exit;
    end; {if}
  BitMask := $80;
  for I := 8 downto 1 do
    begin
      if (SigByte and BitMask) <> 0 then
        begin
          Result := I + (SigByteIndex * 8);
          Exit;
        end; {if}
      BitMask := BitMask shr 1;
    end; {for}
end; {function CountBits}

{----------------------------------------------------------------------}

function IsFloatValidFracValue(const TheMathFloat : TMathFloat) : boolean;
begin {function IsValidFracValue}
  Result := ((TheMathFloat <= MaxFracValue) and (TheMathFloat >= MinFracValue));
end; {function IsValidFracValue}

{----------------------------------------------------------------------}

function SafeIntAdd(const Num1, Num2 : Int64; var Answer : Int64) : boolean;
begin {function SafeIntAdd}
  Result := false;
  try
    if (CountBits(Num1) <= 61) and (CountBits(Num2) <= 61) then
      begin
        Answer := Num1 + Num2;
        Result := true;
      end; {if}
  except
    Result := false;
  end; {try..except}
end; {function SafeIntAdd}

{----------------------------------------------------------------------}

function SafeIntSub(const Num1, Num2 : Int64; var Answer : Int64) : boolean;
begin {function SafeIntSub}
  Result := false;
  try
    if (CountBits(Num1) <= 61) and (CountBits(Num2) <= 61) then
      begin
        Answer := Num1 - Num2;
        Result := true;
      end; {if}
  except
    Result := false;
  end; {try..except}
end; {function SafeIntSub}

{----------------------------------------------------------------------}

function SafeIntMul(const Num1, Num2 : Int64; var Answer : Int64) : boolean;
begin {function SafeIntMul}
  Result := false;
  try
    if Num1 = (-1) then
      begin
        if Num2 <> Low(Int64) then
          begin
            Answer := Num2 * (-1);
            Result := true;
          end; {if}
        Exit;
      end; {if}
    if Num2 = (-1) then
      begin
        if Num1 <> Low(Int64) then
          begin
            Answer := Num1 * (-1);
            Result := true;
          end; {if}
        Exit;
      end; {if}
    if (Num1 = 0) or (Num2 = 0) then
      begin
        Answer := 0;
        Result := true;
        Exit;
      end; {if}
    if Num1 = 1 then
      begin
        Answer := Num2;
        Result := true;
        Exit;
      end; {if}
    if Num2 = 1 then
      begin
        Answer := Num1;
        Result := true;
        Exit;
      end; {if}
    if (CountBits(Num1) + CountBits(Num2)) <= 62 then
      begin
        Answer := Num1 * Num2;
        Result := true;
      end; {if}
  except
    Result := false;
  end; {try..except}
end; {function SafeIntMul}

{----------------------------------------------------------------------}

function SafeIntPower(const Base, Exponent : Int64; var Answer : Int64) : boolean;
//Exponent must be >= 0
var
  I, IntegerExp : integer;
begin {function SafeIntPower}
  Assert(Exponent >= 0, 'Error negative integer power');
  Result := false;
  try
    if Exponent = 0 then
      begin
        if Base <> 0 then
          begin
            Answer := 1;
            Result := true;
          end; {if}
        Exit;
      end; {if}
    if Base = (-1) then
      begin
        if (Exponent mod 2) = 0 then //even number
          Answer := 1
        else
          Answer := (-1);
        Result := True;
        Exit;
      end; {if}
    if Base = 0 then
      begin
        Answer := 0;
        Result := True;
        Exit;
      end; {if}
    if Base = 1 then
      begin
        Answer := 1;
        Result := True;
        Exit;
      end; {if}
    if Exponent > 63 then  //Answer is guaranteed to overflow
      Exit;
    IntegerExp := Exponent;
    Answer := Base;
    for I := 1 to (IntegerExp - 1) do
      begin
        if (CountBits(Answer) + CountBits(Base)) > 62 then
          Exit; //overflow error
        Answer := Answer * Base;
      end; {for}
    if CountBits(Answer) <= 62 then
      Result := True;
  except
    Result := false;
  end; {try..except}
end; {function SafeIntPower}

{----------------------------------------------------------------------}

function SafeIntSqrt(const Number : Int64; var Answer : Int64) : boolean;
var
  NumFloat : TMathFloat;
  NumInt : int64;
begin {function SafeIntSqrt}
  Result := false;
  try
    if Number >= 0 then
      begin
        if (Number = 0) or (Number = 1) then
          begin
            Answer := Number;
            Result := true;
          end
        else
          begin
            NumFloat := Number;
            NumFloat := sqrt(NumFloat);
            if Frac(NumFloat) = 0 then
              begin
                NumInt := Trunc(NumFloat);
                if (NumInt * NumInt) = Number then
                  begin
                    Answer := NumInt;
                    Result := true;
                  end; {if}
              end; {if}
          end; {else}
      end; {if}
  except
    Result := false;
  end; {try..except}
end; {function SafeIntSqrt}

{----------------------------------------------------------------------}

function IsKnownPrimeNumber(TheInt64Value : Int64) : boolean;
var
  LoNdx, MidNdx, HiNdx : integer;
  TheNum : longint;
begin {function IsKnownPrimeNumber}
  Result := false;
  if (TheInt64Value > High(longint)) or (TheInt64Value < Low(longint)) then
    Exit;
  TheNum := abs(TheInt64Value);
  if (TheNum >= PrimesArray[MinPrimesNdx]) and (TheNum <= PrimesArray[MaxPrimesNdx]) then
    begin
      //begin do fast check for small numbers
      if TheNum <= 29 then
        begin
          Result := (TheNum in [2,3,5,7,11,13,17,19,23,29]);
          Exit;
        end; {if}
      //end do fast check for small numbers
      if (TheNum = PrimesArray[MinPrimesNdx]) or (TheNum = PrimesArray[MaxPrimesNdx]) then
        begin
          Result := true;
          Exit;
        end; {if}
      LoNdx := MinPrimesNdx;
      HiNdx := MaxPrimesNdx;
      while (HiNdx - LoNdx) > 1 do
        begin
          MidNdx := LoNdx + ((HiNdx - LoNdx) div 2);
          if TheNum = PrimesArray[MidNdx] then
            begin
              Result := true;
              Exit;
            end; {if}
          if TheNum < PrimesArray[MidNdx] then
            HiNdx := MidNdx
          else
            LoNdx := MidNdx;
        end; {while}
    end; {if}
end; {function IsKnownPrimeNumber}

{----------------------------------------------------------------------}

procedure ReduceValidateMathFraction(var MathValue : TMathValue);
var
  CurrNumer, CurrDenom, SmallNumber, LargeNumber, CurrPrime, Remainder : Int64;
  PrimeNdx : integer;
  CanReduceAgain, DidLoopReduceFraction : boolean;
begin {procedure ReduceValidateMathFraction}
  Assert(MathValue.ValueType <> vtEmpty, 'Reduce Empty MathValue');
  if MathValue.ValueType <> vtFraction then
    Exit;
  Assert(MathValue.Denominator <> 0, 'Fraction divide by zero, in Reduce');
  try
    with MathValue do
      begin
        if Numerator = 0 then
          begin //value is zero
            Denominator := 1;
            Exit;
          end; {if}
        if (Numerator = Low(Int64)) or (Denominator = Low(Int64)) then
          begin
            if Numerator = Denominator then
              begin
                Numerator := 1;
                Denominator := 1;
              end
            else
              ValueType := vtFloat;
            Exit;
          end; {if}
        if not MoveNegativeToNumerator(MathValue) then //Make Denominator > 0
          Exit;
        if Denominator = 1 then
          Exit; //can't reduce, value is an integer
        if abs(Numerator) <= 1 then
          Exit; //can't reduce
        if abs(Numerator) = Denominator then
          begin
            if Numerator < 0 then
              begin
                Numerator := -1;
                Denominator := 1;
              end
            else
              begin
                Numerator := 1;
                Denominator := 1;
              end; {else}
            Exit;
          end; {if}
        //begin easy reduction by dividing smaller number into bigger number.
        if abs(Numerator) < Denominator then
          begin
            Remainder := Denominator mod Numerator;
            if Remainder = 0 then
              begin
                Denominator := Denominator div abs(Numerator);
                if Numerator < 0 then //Numerator := Numerator div abs(Numerator);
                  Numerator := -1
                else
                  Numerator := 1;
                if Denominator = Low(Int64) then //validate result
                  ValueType := vtFloat;
                Exit;
              end; {if}
          end
        else
          begin //if abs(Numerator) > Denominator then
            Remainder := abs(Numerator) mod Denominator;
            if Remainder = 0 then
              begin
                Numerator := Numerator div Denominator;
                Denominator := 1;
                if Numerator = Low(Int64) then //validate result
                  ValueType := vtFloat;
                Exit;
              end; {if}
          end; {else}
        //end easy reduction by dividing smaller number into bigger number.
        if IsKnownPrimeNumber(Numerator) or IsKnownPrimeNumber(Denominator) then
          Exit; //can't reduce prime number
        //begin loop to search for reduced fraction.
        DidLoopReduceFraction := false;
        CurrNumer := abs(Numerator);
        CurrDenom := Denominator;
        repeat
          CanReduceAgain := False;
          if CurrNumer <= CurrDenom then
            begin
              SmallNumber := CurrNumer;
              LargeNumber := CurrDenom;
            end
          else
            begin
              SmallNumber := CurrDenom;
              LargeNumber := CurrNumer;
            end; {else}
          PrimeNdx := 1;
          repeat
            CurrPrime := PrimesArray[PrimeNdx];
            if CurrPrime <= SmallNumber then
              begin
                Remainder := SmallNumber mod CurrPrime;
                if Remainder = 0 then
                  begin
                    Remainder := LargeNumber mod CurrPrime;
                    if Remainder = 0 then
                      begin
                        CurrNumer := CurrNumer div CurrPrime;
                        CurrDenom := CurrDenom div CurrPrime;
                        CanReduceAgain := true;
                        DidLoopReduceFraction := true;
                      end; {if}
                  end; {if}
              end
            else
              break; //can not reduce further.
            inc(PrimeNdx);
          until (CanReduceAgain) or (PrimeNdx > MaxPrimesNdx);
          //begin check for prime numbers
          if CanReduceAgain then
            if IsKnownPrimeNumber(CurrNumer) or IsKnownPrimeNumber(CurrDenom) then
              CanReduceAgain := False; //can't reduce prime number
          //end check for prime numbers
        until (not CanReduceAgain);
        if DidLoopReduceFraction then
          begin
            if Numerator < 0 then
              Numerator := CurrNumer * (-1)
            else
              Numerator := CurrNumer;
            Denominator := CurrDenom;
          end; {if}
        //end loop to search for reduced fraction.
        //validate result
        if (Numerator = Low(int64)) or (Denominator = Low(Int64)) then
          ValueType := vtFloat;
      end; {with}
  except
    MathValue.ValueType := vtFloat;
  end; {try..except}
end; {procedure ReduceValidateMathFraction}

{----------------------------------------------------------------------}

procedure InitMathValue(var MathValue : TMathValue);
begin {procedure InitMathValue}
  with MathValue do
    begin
      FloatValue := 0;
      Numerator := 0;
      Denominator := 1;
      ValueType := vtEmpty;
    end; {with}
end; {procedure InitMathValue}

{----------------------------------------------------------------------}

function ReturnInitMathValue : TMathValue;
begin {function ReturnInitMathValue}
  with Result do
    begin
      FloatValue := 0;
      Numerator := 0;
      Denominator := 1;
      ValueType := vtEmpty;
    end; {with}
end; {function ReturnInitMathValue}

{----------------------------------------------------------------------}

function PackMathValue(const TheMathValue : TMathValue) : TPackedMathValue;
begin {function PackMathValue}
  Result.Numerator := TheMathValue.Numerator;
  Result.Denominator := TheMathValue.Denominator;
  Result.FloatValue := TheMathValue.FloatValue;
  Result.ValueType := TheMathValue.ValueType;
end; {function PackMathValue}

{----------------------------------------------------------------------}

function UnPackMathValue(const ThePackedMathValue : TPackedMathValue) : TMathValue;
begin {function UnPackMathValue}
  Result.Numerator := ThePackedMathValue.Numerator;
  Result.Denominator := ThePackedMathValue.Denominator;
  Result.FloatValue := ThePackedMathValue.FloatValue;
  Result.ValueType := ThePackedMathValue.ValueType;
end; {function UnPackMathValue}

{----------------------------------------------------------------------}

function IsMathValueReducedFraction(const TheMathValue : TMathValue) : boolean;
begin {function IsMathValueReducedFraction}
  Result := false;
  if TheMathValue.ValueType = vtFraction then
  if TheMathValue.Denominator > 0 then
  if ((TheMathValue.Numerator mod 2) <> 0) or ((TheMathValue.Denominator mod 2) <> 0) then
  if (TheMathValue.Numerator <= (int64(MaxPrimesValue) * int64(MaxPrimesValue))) and
     (TheMathValue.Numerator >= ((-1) * (int64(MaxPrimesValue) * int64(MaxPrimesValue)))) and
     (TheMathValue.Denominator <= (int64(MaxPrimesValue) * int64(MaxPrimesValue))) and
     (TheMathValue.Denominator >= ((-1) * (int64(MaxPrimesValue) * int64(MaxPrimesValue)))) then
    Result := true;
end; {function IsMathValueReducedFraction}

{----------------------------------------------------------------------}

function FloatToMathValue(const MathFloat : TMathFloat) : TMathValue;
begin {function FloatToMathValue}
  with Result do
    begin
      FloatValue := MathFloat;
      Numerator := 0;
      Denominator := 1;
      ValueType := vtFloat;
    end; {with}
end; {function FloatToMathValue}

{----------------------------------------------------------------------}

function LongintToMathValue(TheInteger : longint) : TMathValue;
begin {function LongintToMathValue}
  with Result do
    begin
      FloatValue := TheInteger;
      Numerator := TheInteger;
      Denominator := 1;
      ValueType := vtFraction;
    end; {with}
end; {function LongintToMathValue}

{----------------------------------------------------------------------}

function FractionToMathValue(const TheNumer, TheDenom : Int64) : TMathValue;
var
  FloatNumer, FloatDenom : TMathFloat;
begin {function FractionToMathValue}
  with Result do
    begin
      FloatValue := 0;
      Numerator := 0;
      Denominator := 1;
      ValueType := vtEmpty;
    end; {with}
  if TheDenom = 0 then
    Exit;
  //Set Float Value
  try
    FloatNumer := TheNumer;
    FloatDenom := TheDenom;
    Result.FloatValue := FloatNumer / FloatDenom;
    Result.ValueType := vtFloat;
  except
    Result.FloatValue := 0;
    Result.ValueType := vtEmpty;
    Exit;
  end; {try..except}
  //Set Fractional Value
  try
    with Result do
      begin
        Numerator := TheNumer;
        Denominator := TheDenom;
        ValueType := vtFraction;
      end; {with}
    ReduceValidateMathFraction(Result);
  except
    Result.ValueType := vtFloat;
  end; {try..except}
end; {function FractionToMathValue}

{----------------------------------------------------------------------}

function CalcCommonDenoninators(var MathValue1, MathValue2 : TMathValue) : boolean;
var
  MV1, MV2 : TMathValue;
begin {function CalcCommonDenoninators}
  Assert(((MathValue1.ValueType = vtFraction) and (MathValue2.ValueType = vtFraction)), 'Error can not calculate common denoninator');
  Assert(((MathValue1.Denominator <> 0) and (MathValue2.Denominator <> 0)), 'Fraction divide by zero, in Common Denominator');
  Result := false;
  try
    MV1 := MathValue1;
    MV2 := MathValue2;
    //begin move negative from Denominator into Numerator.
    if not MoveNegativeToNumerator(MV1) then
      Exit;
    if not MoveNegativeToNumerator(MV2) then
      Exit;
    //end move negative from Denominator into Numerator.
    if MV1.Denominator = MV2.Denominator then
      begin
        MathValue1 := MV1;
        MathValue2 := MV2;
        Result := true;
        Exit;
      end; {if}
    if SafeIntMul(MV1.Numerator, MV2.Denominator, MathValue1.Numerator) then
      if SafeIntMul(MV1.Denominator, MV2.Denominator, MathValue1.Denominator) then
        if SafeIntMul(MV2.Numerator, MV1.Denominator, MathValue2.Numerator) then
          begin
            MathValue2.Denominator := MathValue1.Denominator;
            Result := true;
          end; {if}
  except
    Result := false;
  end; {try..except}
end; {function CalcCommonDenoninators}

{----------------------------------------------------------------------}

function AddMathValue(MathValue1, MathValue2 : TMathValue) : TMathValue;
begin {function AddMathValue}
  Assert(((MathValue1.ValueType <> vtEmpty) and (MathValue2.ValueType <> vtEmpty)), 'Add Empty MathValue');
  Assert(((MathValue1.ValueType <> vtFraction) or (MathValue1.Denominator > 0)), 'Invaid fraction #1 denominator in function AddMathValue');
  Assert(((MathValue2.ValueType <> vtFraction) or (MathValue2.Denominator > 0)), 'Invaid fraction #2 denominator in function AddMathValue');
  Result.FloatValue := MathValue1.FloatValue + MathValue2.FloatValue;
  Result.ValueType := vtFloat;
  try
    if (MathValue1.ValueType = vtFraction) and (MathValue2.ValueType = vtFraction) then
      if CalcCommonDenoninators(MathValue1, MathValue2) then
        if SafeIntAdd(MathValue1.Numerator, MathValue2.Numerator, Result.Numerator) then
          begin
            Result.Denominator := MathValue1.Denominator;
            Result.ValueType := vtFraction;
            ReduceValidateMathFraction(Result);
          end; {if}
  except
    Result.ValueType := vtFloat;
  end; {try..except}
end; {function AddMathValue}

{----------------------------------------------------------------------}

function SubMathValue(MathValue1, MathValue2 : TMathValue) : TMathValue;
begin {function SubMathValue}
  Assert(((MathValue1.ValueType <> vtEmpty) and (MathValue2.ValueType <> vtEmpty)), 'Sub Empty MathValue');
  Assert(((MathValue1.ValueType <> vtFraction) or (MathValue1.Denominator > 0)), 'Invaid fraction #1 denominator in function SubMathValue');
  Assert(((MathValue2.ValueType <> vtFraction) or (MathValue2.Denominator > 0)), 'Invaid fraction #2 denominator in function SubMathValue');
  Result.FloatValue := MathValue1.FloatValue - MathValue2.FloatValue;
  Result.ValueType := vtFloat;
  try
    if (MathValue1.ValueType = vtFraction) and (MathValue2.ValueType = vtFraction) then
      if CalcCommonDenoninators(MathValue1, MathValue2) then
        if SafeIntSub(MathValue1.Numerator, MathValue2.Numerator, Result.Numerator) then
          begin
            Result.Denominator := MathValue1.Denominator;
            Result.ValueType := vtFraction;
            ReduceValidateMathFraction(Result);
          end; {if}
  except
    Result.ValueType := vtFloat;
  end; {try..except}
end; {function SubMathValue}

{----------------------------------------------------------------------}

function MulMathValue(MathValue1, MathValue2 : TMathValue) : TMathValue;
begin {function MulMathValue}
  Assert(((MathValue1.ValueType <> vtEmpty) and (MathValue2.ValueType <> vtEmpty)), 'Multiply Empty MathValue');
  Assert(((MathValue1.ValueType <> vtFraction) or (MathValue1.Denominator > 0)), 'Invaid fraction #1 denominator in function MulMathValue');
  Assert(((MathValue2.ValueType <> vtFraction) or (MathValue2.Denominator > 0)), 'Invaid fraction #2 denominator in function MulMathValue');
  Result.FloatValue := MathValue1.FloatValue * MathValue2.FloatValue;
  Result.ValueType := vtFloat;
  //if MathValue1 = 0
  if MathValue1.ValueType = vtFraction then
    if (MathValue1.Numerator = 0) and (MathValue1.FloatValue = 0) and
       (Result.FloatValue = 0) then
      begin
        with Result do
          begin
            FloatValue := 0;
            Numerator := 0;
            Denominator := 1;
            ValueType := vtFraction;
          end; {with}
        Exit;
      end; {if}
  //if MathValue2 = 0
  if MathValue2.ValueType = vtFraction then
    if (MathValue2.Numerator = 0) and (MathValue2.FloatValue = 0) and
       (Result.FloatValue = 0) then
      begin
        with Result do
          begin
            FloatValue := 0;
            Numerator := 0;
            Denominator := 1;
            ValueType := vtFraction;
          end; {with}
        Exit;
      end; {if}
  try
    if (MathValue1.ValueType = vtFraction) and (MathValue2.ValueType = vtFraction) then
      if SafeIntMul(MathValue1.Numerator, MathValue2.Numerator, Result.Numerator) then
        if SafeIntMul(MathValue1.Denominator, MathValue2.Denominator, Result.Denominator) then
          begin
            Result.ValueType := vtFraction;
            ReduceValidateMathFraction(Result);
          end; {if}
  except
    Result.ValueType := vtFloat;
  end; {try..except}
end; {function MulMathValue}

{----------------------------------------------------------------------}

function DivMathValue(MathValue1, MathValue2 : TMathValue;
                      var DivByZeroError : boolean) : TMathValue;
begin {function DivMathValue}
  Assert(((MathValue1.ValueType <> vtEmpty) and (MathValue2.ValueType <> vtEmpty)), 'Division Empty MathValue');
  Assert(((MathValue1.ValueType <> vtFraction) or (MathValue1.Denominator > 0)), 'Invaid fraction #1 denominator in function DivMathValue');
  Assert(((MathValue2.ValueType <> vtFraction) or (MathValue2.Denominator > 0)), 'Invaid fraction #2 denominator in function DivMathValue');
  DivByZeroError := False;
  if MathValue2.FloatValue = 0 then
    begin
      DivByZeroError := True;
      Result.ValueType := vtEmpty;
    end
  else
    begin
      Result.FloatValue := MathValue1.FloatValue / MathValue2.FloatValue;
      Result.ValueType := vtFloat;
      //if MathValue1 = 0
      if MathValue1.ValueType = vtFraction then
        if (MathValue1.Numerator = 0) and (MathValue1.FloatValue = 0) and
           (Result.FloatValue = 0) then
          begin
            with Result do
              begin
                FloatValue := 0;
                Numerator := 0;
                Denominator := 1;
                ValueType := vtFraction;
              end; {with}
            Exit;
          end; {if}
      try
        if (MathValue1.ValueType = vtFraction) and (MathValue2.ValueType = vtFraction) then
         if MathValue2.Numerator <> 0 then
          if SafeIntMul(MathValue1.Numerator, MathValue2.Denominator, Result.Numerator) then
           if SafeIntMul(MathValue1.Denominator, MathValue2.Numerator, Result.Denominator) then
             begin
               Result.ValueType := vtFraction;
               ReduceValidateMathFraction(Result);
             end; {if}
      except
        Result.ValueType := vtFloat;
      end; {try..except}
    end; {else}
end; {function DivMathValue}

{----------------------------------------------------------------------}

function PowerMathValue(MathValue1, MathValue2 : TMathValue;
                        var InvalidPowerError : boolean) : TMathValue;
{------------------------------}
function IsEvenNumberInt64(const TheNum : int64) : boolean;
begin {function IsEvenNumberInt64}
  Result := ((TheNum mod 2) = 0);
end; {function IsEvenNumberInt64}
{------------------------------}
procedure AttemptFractionToInteger;
begin {procedure AttemptFractionToInteger;}
  Result.ValueType := vtFloat;
  try
    if (MathValue1.ValueType = vtFraction) and (MathValue2.ValueType = vtFraction) and
       (abs(MathValue2.Denominator) = 1) then
      begin
         if MathValue2.Numerator >= 0 then
           begin
             if not SafeIntPower(MathValue1.Numerator, MathValue2.Numerator, Result.Numerator) then
               Exit;
             if not SafeIntPower(MathValue1.Denominator, MathValue2.Numerator, Result.Denominator) then
               Exit;
           end
         else
           begin
             if not SafeIntPower(MathValue1.Numerator, abs(MathValue2.Numerator), Result.Denominator) then
               Exit;
             if not SafeIntPower(MathValue1.Denominator, abs(MathValue2.Numerator), Result.Numerator) then
               Exit;
           end; {else}
        Result.ValueType := vtFraction;
        ReduceValidateMathFraction(Result);
      end; {if}
  except
    Result.ValueType := vtFloat;
  end; {try..finally}
end; {procedure AttemptFractionToInteger;}
{------------------------------}
begin {function PowerMathValue}
  Assert(((MathValue1.ValueType <> vtEmpty) and (MathValue2.ValueType <> vtEmpty)), 'Power Empty MathValue');
  Assert(((MathValue1.ValueType <> vtFraction) or (MathValue1.Denominator > 0)), 'Invaid fraction #1 denominator in function PowerMathValue');
  Assert(((MathValue2.ValueType <> vtFraction) or (MathValue2.Denominator > 0)), 'Invaid fraction #2 denominator in function PowerMathValue');
  InvalidPowerError := False;
  with Result do
    begin
      FloatValue := 0;
      Numerator := 0;
      Denominator := 1;
      ValueType := vtEmpty;
    end; {with}
  if IsMathFractionEqualToZero(MathValue1) then
    begin
      if IsMathFractionEqualToZero(MathValue2) then //illeagle 0^0 power.
        begin
          InvalidPowerError := True;
          Result.ValueType := vtEmpty;
        end
      else //if 0^X power.
        if (MathValue1.ValueType = vtFraction) and (MathValue1.Numerator = 0) then
          Result.ValueType := vtFraction //result is Zero.
        else
          Result.ValueType := vtFloat; //result is Zero.
      Exit;
    end; {if}
  if IsMathFractionEqualToZero(MathValue2) then //X^0 = 1
    begin
      with Result do
        begin
          FloatValue := 1;
          Numerator := 1;
          Denominator := 1;
          if (MathValue2.ValueType = vtFraction) and (MathValue2.Numerator = 0) then
            ValueType := vtFraction
          else
            ValueType := vtFloat;
        end; {with}
      Exit;
    end; {if}
  if MathValue1.FloatValue >= 0 then
    begin
      Result.FloatValue := Exp(MathValue2.FloatValue * Ln(abs(MathValue1.FloatValue)));
      AttemptFractionToInteger;
      Exit;
    end; {if}
  //if MathValue1.FloatValue < 0 then
  if Frac(MathValue2.FloatValue) = 0 then {Value2 is an integer}
    begin
      if MathValue2.FloatValue = 1 then
        Result.FloatValue := MathValue1.FloatValue
      else
        if MathValue2.FloatValue = (-1) then
          Result.FloatValue := 1 / MathValue1.FloatValue
        else
          if Frac(MathValue2.FloatValue / 2) = 0 then{if MathValue2.FloatValue is even int}
            Result.FloatValue := Exp(MathValue2.FloatValue * Ln(abs(MathValue1.FloatValue)))
          else {FDataClass.MEStk[MStkI+1] is odd int}
            Result.FloatValue := -Exp(MathValue2.FloatValue * Ln(abs(MathValue1.FloatValue)));
      AttemptFractionToInteger;
      Exit;
    end; {if}
  if IsMathValueReducedFraction(MathValue2) then {Value2 is a reduced integer}
    begin
      Assert((((MathValue2.Numerator mod 2) <> 0) or ((MathValue2.Denominator mod 2) <> 0)), 'Error in function PowerMathValue; fraction not reduced, numerator and denominator are both even');
      {All Combinations: 1/1 = Neg; Even/1 = Pos; Odd/1 = Neg; 1/Even = Undef;}
      {    Odd/Even = Undef; 1/Odd = Neg; Even/Odd = Pos; Odd/Odd = Neg}
      {Even/Even; 0/X; X/0; are not possible due to previous checks}
      if MathValue2.Denominator = 1 then
        begin
          if IsEvenNumberInt64(MathValue2.Numerator) then {Even/1 = Pos}
            Result.FloatValue := Exp(MathValue2.FloatValue * Ln(abs(MathValue1.FloatValue)))
          else {1/1 = Neg; Odd/1 = Neg}
            Result.FloatValue := -Exp(MathValue2.FloatValue * Ln(abs(MathValue1.FloatValue)));
          AttemptFractionToInteger;
        end
      else
        begin
          if IsEvenNumberInt64(MathValue2.Denominator) then {1/Even = Undef; Odd/Even = Undef}
            begin
              InvalidPowerError := True; {Got A Negative Num to Fraction Power}
              Result.ValueType := vtEmpty;
            end
          else
            begin
              if (MathValue2.Numerator = 1) or (MathValue2.Numerator = (-1)) then {1/Odd = Neg; -1/Odd = Neg}
                Result.FloatValue := -Exp(MathValue2.FloatValue * Ln(abs(MathValue1.FloatValue)))
              else
                if IsEvenNumberInt64(MathValue2.Numerator) then {Even/Odd = Pos}
                  Result.FloatValue := Exp(MathValue2.FloatValue * Ln(abs(MathValue1.FloatValue)))
                else {Odd/Odd = Neg}
                  Result.FloatValue := -Exp(MathValue2.FloatValue * Ln(abs(MathValue1.FloatValue)));
              AttemptFractionToInteger;
            end; {else}
        end; {else}
      Exit;
    end; {if}
  InvalidPowerError := True; {Got A Negative Num to Fraction Power}
  Result.ValueType := vtEmpty;
end; {function PowerMathValue}

{----------------------------------------------------------------------}

function ExpFuncMathValue(MathValue : TMathValue) : TMathValue;
begin {function ExpFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  if IsMathFractionEqualToZero(MathValue) then //X^0 = 1
    begin
      with Result do
        begin
          FloatValue := 1;
          Numerator := 1;
          Denominator := 1;
          if (MathValue.ValueType = vtFraction) and (MathValue.Numerator = 0) then
            ValueType := vtFraction
          else
            ValueType := vtFloat;
        end; {with}
    end
  else
    begin
      with Result do
        begin
          ValueType := vtFloat;
          FloatValue := Exp(MathValue.FloatValue);
        end; {with}
    end; {else}
end; {function ExpFuncMathValue}

{----------------------------------------------------------------------}

procedure SetFloatToFracIfPossible(var MathValue : TMathValue);
begin {procedure SetFloatToFracIfPossible}
  MathValue.ValueType := vtFloat;
  try
    if IsFloatValidFracValue(MathValue.FloatValue) then
      begin
        MathValue.Numerator := Trunc(MathValue.FloatValue);
        if MathValue.Numerator <> Low(Int64) then
          if MathValue.Numerator = MathValue.FloatValue then
            begin
              MathValue.Denominator := 1;
              MathValue.ValueType := vtFraction;
            end; {if}
      end; {if}
  except
    MathValue.ValueType := vtFloat;
  end; {try..except}
end; {procedure SetFloatToFracIfPossible}

{----------------------------------------------------------------------}

function GrintFuncMathValue(MathValue : TMathValue) : TMathValue;
begin {function GrintFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  if MathValue.FloatValue >= 0 then
    Result.FloatValue := int(MathValue.FloatValue)
  else
    begin
      if Frac(MathValue.FloatValue) = 0 then
        Result.FloatValue := MathValue.FloatValue
      else
        Result.FloatValue := int(MathValue.FloatValue) - 1;
    end; {else}
  //test to see if result is a fraction
  SetFloatToFracIfPossible(Result);
end; {function GrintFuncMathValue}

{----------------------------------------------------------------------}

function IntFuncMathValue(MathValue : TMathValue) : TMathValue;
begin {function IntFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  Result.FloatValue := Int(MathValue.FloatValue);
  //test to see if result is a fraction
  SetFloatToFracIfPossible(Result);
end; {function IntFuncMathValue}

{----------------------------------------------------------------------}

function SqrtFuncMathValue(MathValue : TMathValue; var FuncOutDomainError : boolean) : TMathValue;
begin {function SqrtFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  FuncOutDomainError := False;
  Result.ValueType := vtFloat;
  if MathValue.FloatValue < 0 then
    begin
      FuncOutDomainError := True;
      Result.ValueType := vtEmpty;
    end
  else
    begin
      Result.FloatValue := Sqrt(MathValue.FloatValue);
      //is it a fraction
      try
        if MathValue.ValueType = vtFraction then
          if (MathValue.Numerator >= 0) and (MathValue.Denominator > 0) then
            if SafeIntSqrt(MathValue.Numerator, Result.Numerator) then
              if SafeIntSqrt(MathValue.Denominator, Result.Denominator) then
                begin
                  Result.ValueType := vtFraction;
                  ReduceValidateMathFraction(MathValue);
                end; {if}
      except
        Result.ValueType := vtFloat;
      end; {try..except}
    end; {else}
end; {function SqrtFuncMathValue}

{----------------------------------------------------------------------}

function SquareFuncMathValue(MathValue : TMathValue) : TMathValue;
begin {function SquareFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  Result := MulMathValue(MathValue, MathValue);
end; {function SquareFuncMathValue}

{----------------------------------------------------------------------}

function RoundFuncMathValue(MathValue : TMathValue) : TMathValue;
begin {function RoundFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  if abs(Frac(MathValue.FloatValue)) >= 0.5 then
    begin
      if MathValue.FloatValue >= 0 then
        Result.FloatValue := int(MathValue.FloatValue) + 1
      else
        Result.FloatValue := int(MathValue.FloatValue) - 1;
    end
  else
    Result.FloatValue := int(MathValue.FloatValue);
  //test to see if result is a fraction
  SetFloatToFracIfPossible(Result);
end; {function RoundFuncMathValue}

{----------------------------------------------------------------------}

function FracFuncMathValue(MathValue : TMathValue) : TMathValue;
begin {function FracFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  Result.FloatValue := Frac(MathValue.FloatValue);
  Result.ValueType := vtFloat;
  //test to see if result is a fraction
  if Result.FloatValue = 0 then
    if MathValue.ValueType = vtFraction then
      if MathValue.Denominator = 1 then //is an integer divided by 1
        begin
          with Result do
            begin
              Numerator := 0;
              Denominator := 1;
              ValueType := vtFraction;
            end; {with}
        end; {if}
end; {function FracFuncMathValue}

{----------------------------------------------------------------------}

function AbsFuncMathValue(MathValue : TMathValue) : TMathValue;
begin {function AbsFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  Result.ValueType := MathValue.ValueType;
  Result.FloatValue := Abs(MathValue.FloatValue);
  if MathValue.ValueType = vtFraction then
    begin
      if (MathValue.Numerator = Low(Int64)) or (MathValue.Denominator = Low(Int64)) then
        Result.ValueType := vtFloat
      else
        begin
          Result.Numerator := Abs(MathValue.Numerator);
          Result.Denominator := Abs(MathValue.Denominator);
        end; {else}
    end; {if}
end; {function AbsFuncMathValue}

{----------------------------------------------------------------------}

function FactFuncMathValue(MathValue : TMathValue;
                           var FuncOutDomainError : boolean;
                           var FactToBigError : boolean) : TMathValue;
var
  TmpR, FactAnswer  : TMathFloat;
begin {function FactFuncMathValue}
  Assert((MathValue.ValueType <> vtEmpty), 'Empty MathValue in Function');
  Assert(((MathValue.ValueType <> vtFraction) or (MathValue.Denominator > 0)), 'Invaid fraction denominator');
  FuncOutDomainError := False;
  FactToBigError := False;
  Result.ValueType := vtEmpty;
  if MathValue.FloatValue < 0 then
    begin
      FuncOutDomainError := True;
      Exit;
    end; {if}
  if MathValue.FloatValue > 20 then
    begin
      FactToBigError := True;
      Exit;
    end; {if}
  if IsMathFractionEqualToZero(MathValue) then
    begin
      with Result do
        begin
          FloatValue := 1;
          Numerator := 1;
          Denominator := 1;
          ValueType := vtFraction;
        end; {with}
      Exit;
    end; {if}
  if Frac(MathValue.FloatValue) = 0 then
    TmpR := MathValue.FloatValue
  else
    begin
      if MathValue.ValueType = vtFraction then
        begin
          if (MathValue.Denominator = 1) and (MathValue.Numerator > 0) and
             (MathValue.Numerator <= 20) then
            TmpR := MathValue.Numerator
          else
            begin
              FuncOutDomainError := True;
              Exit;
            end; {else}
        end
      else
        begin
          FuncOutDomainError := True;
          Exit;
        end; {else}
    end; {else}
  FactAnswer := TmpR;
  while TmpR > 2 do
    begin
      TmpR := TmpR - 1;
      FactAnswer := FactAnswer * TmpR;
    end; {while}
  //Assign Result
  with Result do
    begin
      FloatValue := FactAnswer;
      if IsFloatValidFracValue(FactAnswer) then
        begin
          Numerator := Trunc(FactAnswer);
          if Numerator = Low(Int64) then
            ValueType := vtFloat
          else
            begin
              Denominator := 1;
              ValueType := vtFraction;
            end; {else}
        end
      else
        ValueType := vtFloat;
    end; {with}
end; {function FactFuncMathValue}

{----------------------------------------------------------------------}

function VerifyCleanRealStr(var RealNumString : string; var ErrorPos : integer) : boolean;
var
  I : integer;
  FirstDigitNdx : integer;
  GetDecSepar, GotDigit, ContinueLooping : boolean;
begin {function VerifyCleanRealStr}
  Result := false;
  ErrorPos := 0;
  if RealNumString = '' then
    Exit;
  FirstDigitNdx := 1;
  GetDecSepar := False;
  GotDigit := False;
  for I := 1 to Length(RealNumString) do
    begin
      if RealNumString[I] in ['-','+'] then
        begin
          if I = 1 then
            FirstDigitNdx := 2
          else
            begin
              ErrorPos := I;
              Exit;
            end; {else}
        end
      else
        if RealNumString[I] = DecimalSeparator then
          begin
            if GetDecSepar then
              begin
                ErrorPos := I;
                Exit;
              end
            else
              GetDecSepar := True;
          end
        else
          if RealNumString[I] in ['0'..'9'] then
            begin
              GotDigit := True;
            end
          else
            begin
              ErrorPos := I;
              Exit;
            end; {else}
    end; {for}
  //test results
  if not GotDigit then
    Exit;
  //Remove Leading Plus Sign
  if RealNumString[1] = '+' then
    begin
      Delete(RealNumString, 1, 1);
      FirstDigitNdx := 1;
    end; {if}
  //Remove Leading Zeros
  ContinueLooping := true;
  while ContinueLooping and (FirstDigitNdx < Length(RealNumString)) do
    begin
      ContinueLooping := false;
      if RealNumString[FirstDigitNdx] = '0' then
        begin
          if (FirstDigitNdx + 1) <= Length(RealNumString) then
            if RealNumString[FirstDigitNdx + 1] <> DecimalSeparator then
             begin
               ContinueLooping := true;
               Delete(RealNumString, FirstDigitNdx, 1);
             end; {if}
        end; {if}
    end; {while}
  if FirstDigitNdx > Length(RealNumString) then
    Exit; //unknown error
  //make sure it does not begin with a DecimalSeparator
  if RealNumString[FirstDigitNdx] = DecimalSeparator then
    Insert('0', RealNumString, FirstDigitNdx);
  //Remove Trailing Zeros after DecimalSeparator
  if GetDecSepar then
    begin
      while RealNumString[Length(RealNumString)] = '0' do
        Delete(RealNumString, Length(RealNumString), 1);
      //Remove Trailing DecimalSeparator
      if RealNumString[Length(RealNumString)] = DecimalSeparator then
        Delete(RealNumString, Length(RealNumString), 1);
    end; {if}
  //return successful
  Result := true;
end; {function VerifyCleanRealStr}

{----------------------------------------------------------------------}

function StringToMathValue(RealNumString : string;
                           var ErrorPos : integer;
                           var MathValue : TMathValue) : boolean;
var
  I, DecSepPos : integer;
  DigitCount : int64;
begin {function StringToMathValue}
  Result := false;
  ErrorPos := 0;
  try
    with MathValue do
      begin
        FloatValue := 0;
        Numerator := 0;
        Denominator := 1;
        ValueType := vtEmpty;
      end; {with}
    if not VerifyCleanRealStr(RealNumString, ErrorPos) then
      Exit;
    try //got a real number.
      MathValue.FloatValue := StrToFloat(RealNumString);
      MathValue.ValueType := vtFloat;
    except
      InitMathValue(MathValue);
      Exit;
    end; {try..except}
    Result := true; //successfully converted a real number
    DecSepPos := Pos(DecimalSeparator, RealNumString);
    if DecSepPos = 0 then
      begin
        try //got an int64 number.
          MathValue.Numerator := StrToInt64(RealNumString);
          if MathValue.Numerator = Low(Int64) then
            Exit;
          MathValue.ValueType := vtFraction;
        except
          MathValue.Numerator := 0;
          MathValue.ValueType := vtFloat;
        end; {try..except}
      end
    else
      begin
        try //got a real number.
          //Calc Denominator
          DigitCount := Length(RealNumString) - DecSepPos;
          if DigitCount > MaxFracDecimalDigits then
            Exit;
          MathValue.Denominator := 1;
          for I := 1 to DigitCount do
            MathValue.Denominator := MathValue.Denominator * 10;
          //Calc Numerator
          Delete(RealNumString, DecSepPos, 1);
          MathValue.Numerator := StrToInt64(RealNumString);
          if MathValue.Numerator = Low(Int64) then
            Exit;
          //Reduce
          MathValue.ValueType := vtFraction;
          ReduceValidateMathFraction(MathValue);
        except
          MathValue.Numerator := 0;
          MathValue.Denominator := 1;
          MathValue.ValueType := vtFloat;
        end; {try..except}
      end; {else}
  except
    InitMathValue(MathValue);
    Result := false;
  end; {try..except}
end; {function StringToMathValue}

{----------------------------------------------------------------------}

end.
