unit mathfunc;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Sysutils, MathStak, MathCalc;

const
  MaxMathStringLength = 1000; {Memory useage is affected by this setting}

type
  TMathErrorCode = (
      {Function store and verify errors}
      meNoError, meNulExp, meStringToBig, meIlleagleChar,
      meExprError, meInvalidNum, meGroupingError,
      meFuncWithOutOpenGroup, meUnKnownSymbol,
      {Miscelanous errors}
      meOutOfMem,
      meUnknownException, //Generic Exception, should never happen
      {Evaluation Errors}
      meDivZero, meFuncOutDomain, meInvalidPower, meFactToBig,
      meOverFlow, meUnderFlow, meInvalidOp);

  TMathFuncStatus = (fsNew, fsValidateError, fsValid);

  TMathFunction = class
  private
    { Private declarations }
    MStkI : integer;
    MEStk : TMathEvalStack;
    PostFixCurrEvalRec : integer;
    FFuncString : string;
    FVarCount : integer;
    FVarStrings : TMathStringsArray;
    FVarValues : TMathValuesArray;
    FPostFixArray : TMathTokenArray;
    FStatus : TMathFuncStatus;
    FResultValue : TMathValue;
    FFoundVarInFunc : boolean;
    FErrorCode : TMathErrorCode;
    FErrorPosition : integer;
    function GetFuncString : string;
    procedure SetFuncString(const Value : string);
    procedure SetVarCount(Value : integer);
    function GetVarStrings(Ndx : integer) : string;
    procedure SetVarStrings(Ndx : integer; const Value : string);
    procedure TokenizeVerifyMathStr(TokenObj : TMathTokenArrayStackObj);
    procedure ConvInfixToPostFix(TokenObj, DescObj, PostObj : TMathTokenArrayStackObj);
    function GetMathProcedure(TheDesc : integer) : PMathProcedure;
    procedure InitProcedurePointers(PostObj : TMathTokenArrayStackObj);
    procedure EvalConstantPostFix(PostObj : TMathTokenArrayStackObj);
    procedure SimplifyPostFix(PostObj : TMathTokenArrayStackObj);
    procedure PMath_PushVar;
    procedure PMath_sy_Number;
    procedure PMath_sy_Add;
    procedure PMath_sy_Sub;
    procedure PMath_sy_Mul;
    procedure PMath_sy_Div;
    procedure PMath_sy_Power;
    procedure PMath_sy_DegToRad;
    procedure PMath_sy_RadToDeg;
    procedure PMath_sy_GradToRad;
    procedure PMath_sy_RadToGrad;
    procedure PMath_sy_CycleToRad;
    procedure PMath_sy_RadToCycle;
    function MathCalc_ArcSin(X : TMathFloat) : TMathFloat;
    function MathCalc_ArcCos(X : TMathFloat) : TMathFloat;
    procedure PMath_sy_ARCSINH;
    procedure PMath_sy_ARCCOSH;
    procedure PMath_sy_ARCTANH;
    procedure PMath_sy_ARCCSCH;
    procedure PMath_sy_ARCSECH;
    procedure PMath_sy_ARCCOTH;
    procedure PMath_sy_ARCSIN;
    procedure PMath_sy_ARCCOS;
    procedure PMath_sy_ARCTAN;
    procedure PMath_sy_ARCCSC;
    procedure PMath_sy_ARCSEC;
    procedure PMath_sy_ARCCOT;
    procedure PMath_sy_SINH;
    procedure PMath_sy_COSH;
    procedure PMath_sy_TANH;
    procedure PMath_sy_CSCH;
    procedure PMath_sy_SECH;
    procedure PMath_sy_COTH;
    procedure PMath_sy_SIN;
    procedure PMath_sy_COS;
    procedure PMath_sy_TAN;
    procedure PMath_sy_CSC;
    procedure PMath_sy_SEC;
    procedure PMath_sy_COT;
    procedure PMath_sy_EXP;
    procedure PMath_sy_GRINT;
    procedure PMath_sy_INT;
    procedure PMath_sy_SQRT;
    procedure PMath_sy_SQUARE;
    procedure PMath_sy_TRUNC;
    procedure PMath_sy_ROUND;
    procedure PMath_sy_FRAC;
    procedure PMath_sy_ABS;
    procedure PMath_sy_LN;
    procedure PMath_sy_LOG;
    procedure PMath_sy_FACT;
    function ComputeAOperator(TheMathProc : PMathProcedure;
                              Val1, Val2 : TMathValue): TMathValue;
    function ComputeAFunc(TheMathProc : PMathProcedure;
                          Val1 : TMathValue): TMathValue;
    function GetFloatResult : TMathFloat;
    function GetFloatVariables(Ndx : integer) : TMathFloat;
    procedure SetFloatVariables(Ndx : integer; const Value : TMathFloat);
    function GetMathVariables(Ndx : integer) : TMathValue;
    procedure SetMathVariables(Ndx : integer; const Value : TMathValue);
  public
    { Public declarations }
    property FuncString : string read GetFuncString write SetFuncString;
    property VarCount : integer read FVarCount write SetVarCount;
    property VarStrings[Ndx : integer] : string read GetVarStrings write SetVarStrings;
    property FloatVarValues[Ndx : integer] : TMathFloat read GetFloatVariables write SetFloatVariables;
    property MathVarValues[Ndx : integer] : TMathValue read GetMathVariables write SetMathVariables;
    property Status : TMathFuncStatus read FStatus;
    property FloatResultValue : TMathFloat read GetFloatResult;
    property MathResultValue : TMathValue read FResultValue;
    property FoundVarInFunc : boolean read FFoundVarInFunc;
    property ErrorCode : TMathErrorCode read FErrorCode;
    property ErrorPosition : integer read FErrorPosition;
    constructor Create;
    destructor Destroy; override;
    procedure Assign(SourceFunc : TMathFunction); virtual;
    function GetLastErrorString : string;
    function VerifyMathFunc : boolean;
    function EvaluateMathFunc : boolean;
  end; {TMathFunction}

function GetMLErrorStrFunc(TheErrorCode : TMathErrorCode;
                           TheErrorPosition : integer) : string;
function IsValidVariableString(TheVarStr : string) : boolean;

Type
  TMathTrigFixupType = (tfNormal, tfThou, tfTenThou,
                        tfHundThou, tfMill, tfTenMill);

function SinNormal(X : TMathFloat) : TMathFloat;
function SinRoundThou(X : TMathFloat) : TMathFloat;
function SinRoundTenThou(X : TMathFloat) : TMathFloat;
function SinRoundHundThou(X : TMathFloat) : TMathFloat;
function SinRoundMill(X : TMathFloat) : TMathFloat;
function SinRoundTenMill(X : TMathFloat) : TMathFloat;
function SinExactThou(X : TMathFloat) : TMathFloat;
function SinExactTenThou(X : TMathFloat) : TMathFloat;
function SinExactHundThou(X : TMathFloat) : TMathFloat;
function SinExactMill(X : TMathFloat) : TMathFloat;
function SinExactTenMill(X : TMathFloat) : TMathFloat;

function CosNormal(X : TMathFloat) : TMathFloat;
function CosRoundThou(X : TMathFloat) : TMathFloat;
function CosRoundTenThou(X : TMathFloat) : TMathFloat;
function CosRoundHundThou(X : TMathFloat) : TMathFloat;
function CosRoundMill(X : TMathFloat) : TMathFloat;
function CosRoundTenMill(X : TMathFloat) : TMathFloat;
function CosExactThou(X : TMathFloat) : TMathFloat;
function CosExactTenThou(X : TMathFloat) : TMathFloat;
function CosExactHundThou(X : TMathFloat) : TMathFloat;
function CosExactMill(X : TMathFloat) : TMathFloat;
function CosExactTenMill(X : TMathFloat) : TMathFloat;

procedure SetGlobalMathTrigFixup(FixupType : TMathTrigFixupType;
                                 RequireExact : boolean); //procedure not thread safe
procedure GetGlobalMathTrigFixup(var FixupType : TMathTrigFixupType;
                                 var RequireExact : boolean); //procedure not thread safe
procedure DefaultsGlobalMathTrigFixup; //procedure not thread safe

implementation

const
  LegalCharSet = ['A'..'Z', '0'..'9',
                  '{', '}', '[', ']', '(', ')',
                  '+', '-', '*', '/', '^' ];  //also DecimalSeparator

  sy_BeginSymbol = 1;
  sy_EndSymbol   = 51;
  sy_CountSymbol = 51;
  sy_SymbolSet = [sy_BeginSymbol..sy_EndSymbol];

  sy_BeginOperator   = 1;
  sy_EndOperator     = 7;
  sy_CountOperator   = 7;
  sy_OperatorSet = [sy_BeginOperator..sy_EndOperator];

  sy_BeginAddMulEtc  = 3;
  sy_EndAddMulEtc    = 7;
  sy_CountAddMulEtc  = 5;
  sy_AddMulEtcSet = [sy_BeginAddMulEtc..sy_EndAddMulEtc];

  sy_OpenGroup  = 1;
  sy_CloseGroup = 2;
  sy_Add        = 3;
  sy_Sub        = 4;
  sy_Mul        = 5;
  sy_Div        = 6;
  sy_Power      = 7;

  sy_BeginFunct = 8;
  sy_EndFunct   = 49;
  sy_CountFunct = 41;
  sy_FunctSet = [sy_BeginFunct..sy_EndFunct];

  sy_DegToRad   = 8;
  sy_RadToDeg   = 9;
  sy_GradToRad  = 10;
  sy_RadToGrad  = 11;
  sy_CycleToRad = 12;
  sy_RadToCycle = 13;
  sy_ARCSINH    = 14;
  sy_ARCCOSH    = 15;
  sy_ARCTANH    = 16;
  sy_ARCCSCH    = 17;
  sy_ARCSECH    = 18;
  sy_ARCCOTH    = 19;
  sy_ARCSIN     = 20;
  sy_ARCCOS     = 21;
  sy_ARCTAN     = 22;
  sy_ARCCSC     = 23;
  sy_ARCSEC     = 24;
  sy_ARCCOT     = 25;
  sy_SINH       = 26;
  sy_COSH       = 27;
  sy_TANH       = 28;
  sy_CSCH       = 29;
  sy_SECH       = 30;
  sy_COTH       = 31;
  sy_SIN        = 32;
  sy_COS        = 33;
  sy_TAN        = 34;
  sy_CSC        = 35;
  sy_SEC        = 36;
  sy_COT        = 37;
  sy_EXP        = 38;
  sy_GRINT      = 39;
  sy_INT        = 40;
  sy_SQRT       = 41;
  sy_SQUARE     = 42;
  sy_TRUNC      = 43;
  sy_ROUND      = 44;
  sy_FRAC       = 45;
  sy_ABS        = 46;
  sy_LN         = 47;
  sy_LOG        = 48;
  sy_FACT       = 49;

  sy_BeginConst = 50;
  sy_EndConst   = 51;
  sy_CountConst = 2;

  sy_PI         = 50;
  sy_E          = 51;

  SymbolArray : array [1..sy_CountSymbol] of AnsiString =
     ('(', ')', '+', '-', '*', '/', '^',
      'DEGTORAD',           'RADTODEG',        'GRADTORAD',
      'RADTOGRAD',          'CYCLETORAD',      'RADTOCYCLE',
      'ARCSINH',            'ARCCOSH',         'ARCTANH',
      'ARCCSCH',            'ARCSECH',         'ARCCOTH',
      'ARCSIN',             'ARCCOS',          'ARCTAN',
      'ARCCSC',             'ARCSEC',          'ARCCOT',
      'SINH',               'COSH',            'TANH',
      'CSCH',               'SECH',            'COTH',
      'SIN',                'COS',             'TAN',
      'CSC',                'SEC',             'COT',
      'EXP',                'GRINT',           'INT',
      'SQRT',               'SQUARE',          'TRUNC',
      'ROUND',              'FRAC',            'ABS',
      'LN',                 'LOG',             'FACT',
      'PI',                 'E');

  sy_Number = 1000;
  sy_Variable = 5000;

{=========================================================================}

Type
  TTrigFixupFunc = function(X : TMathFloat) : TMathFloat;

var
  SinFixup : TTrigFixupFunc = SinExactTenThou;
  CosFixup : TTrigFixupFunc = CosExactTenThou;
  GlobalFixupType : TMathTrigFixupType = tfTenThou;
  GlobalRequireExact : boolean = false;

{=========================================================================}

procedure SetGlobalMathTrigFixup(FixupType : TMathTrigFixupType;
                                 RequireExact : boolean);
begin {procedure SetGlobalMathTrigFixup}
  if (FixupType <> GlobalFixupType) or (RequireExact <> GlobalRequireExact) then
    begin
      GlobalFixupType := FixupType;
      GlobalRequireExact := RequireExact;
      if GlobalFixupType = tfNormal then
        begin
          SinFixup := SinNormal;
          CosFixup := CosNormal;
          Exit;
        end; {if}
      if RequireExact then
        begin
          case GlobalFixupType of
            tfThou:
              begin
                SinFixup := SinExactThou;
                CosFixup := CosExactThou;
              end;
            tfTenThou:
              begin
                SinFixup := SinExactTenThou;
                CosFixup := CosExactTenThou;
              end;
            tfHundThou:
              begin
                SinFixup := SinExactHundThou;
                CosFixup := CosExactHundThou;
              end;
            tfMill:
              begin
                SinFixup := SinExactMill;
                CosFixup := CosExactMill;
              end;
            tfTenMill:
              begin
                SinFixup := SinExactTenMill;
                CosFixup := CosExactTenMill;
              end;
          end; {case}
        end
      else
        begin
          case GlobalFixupType of
            tfThou:
              begin
                SinFixup := SinRoundThou;
                CosFixup := CosRoundThou;
              end;
            tfTenThou:
              begin
                SinFixup := SinRoundTenThou;
                CosFixup := CosRoundTenThou;
              end;
            tfHundThou:
              begin
                SinFixup := SinRoundHundThou;
                CosFixup := CosRoundHundThou;
              end;
            tfMill:
              begin
                SinFixup := SinRoundMill;
                CosFixup := CosRoundMill;
              end;
            tfTenMill:
              begin
                SinFixup := SinRoundTenMill;
                CosFixup := CosRoundTenMill;
              end;
          end; {case}
        end; {else}
    end; {if}
end; {procedure SetGlobalMathTrigFixup}

{=========================================================================}

procedure GetGlobalMathTrigFixup(var FixupType : TMathTrigFixupType;
                                 var RequireExact : boolean);
begin {procedure GetGlobalMathTrigFixup}
  FixupType := GlobalFixupType;
  RequireExact := GlobalRequireExact;
end; {procedure GetGlobalMathTrigFixup}

{=========================================================================}

procedure DefaultsGlobalMathTrigFixup;
begin {procedure DefaultsGlobalMathTrigFixup;}
  SetGlobalMathTrigFixup(tfTenThou, true);
end; {procedure DefaultsGlobalMathTrigFixup;}

{=========================================================================}

function GetMLErrorStrFunc(TheErrorCode : TMathErrorCode;
                           TheErrorPosition : integer) : string;
begin {function GetMLErrorStrFunc}
  case TheErrorCode of
    {Function store and verify errors}
    meNoError:
      Result := 'No Error';
    meNulExp:
      Result := 'Empty Expression';
    meStringToBig:
      Result := 'Expressions May Not Be Longer Than ' +
                     IntToStr(MaxMathStringLength) + ' Characters';
    meIlleagleChar:
      Result := 'Illegal Char';
    meExprError:
      Result := 'Expression Syntax Error';
    meInvalidNum:
      Result := 'Invalid Numeric Floating Point Number';
    meGroupingError:
      Result := 'Invalid Parentheses Grouping';
    meFuncWithOutOpenGroup:
      Result := 'Function Not Followed by Open Parentheses Character';
    meUnKnownSymbol:
      Result := 'Unknown Symbol';
    meOutOfMem:
      Result := 'Out of Memory';
    meUnknownException:
      Result := 'Unknown Exception'; //should never happen
    {Evaluation Errors}
    meDivZero:
      Result := 'Division by Zero';
    meFuncOutDomain:
      Result := 'Invalid Function Domain';
    meInvalidPower:
      Result := 'Invalid Exponential';
    meFactToBig:
      Result := 'Factorial to Big';
    meOverFlow:
      Result := 'Floating Point Over Flow';
    meUnderFlow:
      Result := 'Floating Point Under Flow';
    meInvalidOp:
      Result := 'Invalid Floating Point Operation';
  else
    Result := 'Undetermined Error'; //should never happen
  end; {case}
  if TheErrorPosition >= 1 then
    Result := Result + ' at Character ' + IntToStr(TheErrorPosition);
end; {function GetMLErrorStrFunc}

{=========================================================================}

function IsValidVariableString(TheVarStr : string) : boolean;
var
  I : integer;
begin {function IsValidVariableString}
  Result := False;
  if TheVarStr = '' then
    Exit;
  if Length(TheVarStr) > MaxMathStringLength then
    Exit;
  TheVarStr := UpperCase(TheVarStr);
  for I := 1 to Length(TheVarStr) do
    if not (TheVarStr[I] in ['A'..'Z']) then
      Exit;
  for I := 1 to sy_CountSymbol do
    if SymbolArray[I] = TheVarStr then
      Exit;
  Result := True;
end; {function IsValidVariableString}

{=========================================================================}

function TMathFunction.GetFuncString: string;
begin {function TMathFunction.GetFuncString}
  Result := FFuncString;
  UniqueString(Result);
end; {function TMathFunction.GetFuncString}

{=========================================================================}

procedure TMathFunction.SetFuncString(const Value: string);
begin {procedure TMathFunction.SetFuncString}
  FStatus := fsNew;
  FErrorCode := meNoError;
  MEStk := nil;
  FPostFixArray := nil;
  FFuncString := Value;
  UniqueString(FFuncString);
end; {procedure TMathFunction.SetFuncString}

{=========================================================================}

procedure TMathFunction.SetVarCount(Value: integer);
var
  I : integer;
begin {procedure TMathFunction.SetVarCount}
  if FVarCount <> Value then
    begin
      FStatus := fsNew;
      FErrorCode := meNoError;
      MEStk := nil;
      FPostFixArray := nil;
      if Value <= 0 then
        begin
          FVarStrings := nil;
          FVarValues := nil;
          FVarCount := 0;
        end
      else
        begin
          SetLength(FVarStrings, Value);
          for I := FVarCount to (Value-1) do
            FVarStrings[I] := '';
          SetLength(FVarValues, Value);
          for I := FVarCount to (Value-1) do
            InitMathValue(FVarValues[I]); //FVarValues[I] := 0;
          FVarCount := Value;
        end; {else}
    end; {if}
end; {procedure TMathFunction.SetVarCount}

{=========================================================================}

function TMathFunction.GetVarStrings(Ndx: integer): string;
begin {function TMathFunction.GetVarStrings}
  Result := FVarStrings[Ndx];
  UniqueString(Result);
end; {function TMathFunction.GetVarStrings}

{=========================================================================}

procedure TMathFunction.SetVarStrings(Ndx: integer; const Value: string);
begin {procedure TMathFunction.SetVarStrings}
  FStatus := fsNew;
  FErrorCode := meNoError;
  MEStk := nil;
  FPostFixArray := nil;
  FVarStrings[Ndx] := Value;
  UniqueString(FVarStrings[Ndx]);
end; {procedure TMathFunction.SetVarStrings}

{=========================================================================}

constructor TMathFunction.Create;
begin {constructor TMathFunction.Create;}
  inherited create;
  FFuncString := '';
  FVarCount := 0;
  MEStk := nil;
  FVarStrings := nil;
  FVarValues := nil;
  FPostFixArray := nil;
  FStatus := fsNew;
  InitMathValue(FResultValue); //FResultValue := 0;
  FFoundVarInFunc := false;
  FErrorCode := meNoError;
  FErrorPosition := 0;
end; {constructor TMathFunction.Create;}

{=========================================================================}

destructor TMathFunction.Destroy;
var
  I : integer;
begin {destructor TMathFunction.Destroy;}
  FFuncString := '';
  MEStk := nil;
  FVarValues := nil;
  FPostFixArray := nil;
  for I := 0 to (Length(FVarStrings) - 1) do
    FVarStrings[I] := '';
  FVarStrings := nil;
  inherited;
end; {destructor TMathFunction.Destroy;}

{=========================================================================}

procedure TMathFunction.Assign(SourceFunc : TMathFunction);
var
  I : integer;
begin {procedure TMathFunction.Assign}
  FFuncString := SourceFunc.FFuncString;
  FVarCount := SourceFunc.FVarCount;
  SetLength(MEStk, Length(SourceFunc.MEStk));
  if SourceFunc.FVarStrings = nil then
    FVarStrings := nil
  else
    FVarStrings := Copy(SourceFunc.FVarStrings);
  if SourceFunc.FVarValues = nil then
    FVarValues := nil
  else
    FVarValues := Copy(SourceFunc.FVarValues);
  if SourceFunc.FPostFixArray = nil then
    FPostFixArray := nil
  else
    begin
      FPostFixArray := Copy(SourceFunc.FPostFixArray);
      for I := Low(FPostFixArray) to High(FPostFixArray) do
        FPostFixArray[I].MathProc := GetMathProcedure(FPostFixArray[I].Desc);
    end; {else}
  FStatus := SourceFunc.FStatus;
  FResultValue := SourceFunc.FResultValue;
  FFoundVarInFunc := SourceFunc.FFoundVarInFunc;
  FErrorCode := SourceFunc.FErrorCode;
  FErrorPosition := SourceFunc.FErrorPosition;
  //Make Strings Unique
  UniqueString(FFuncString);
  for I := 0 to (Length(FVarStrings) - 1) do
    UniqueString(FVarStrings[I]);
end; {procedure TMathFunction.Assign}

{=========================================================================}

function TMathFunction.GetLastErrorString : string;
begin {function TMathFunction.GetLastErrorString}
  Result := GetMLErrorStrFunc(FErrorCode, FErrorPosition);
end; {function TMathFunction.GetLastErrorString}

{=========================================================================}

function TMathFunction.VerifyMathFunc : boolean;
{returns true if no error occured}
var
  TokenArrayStackObj : TMathTokenArrayStackObj;
  DescStackArrayStackObj : TMathTokenArrayStackObj;
  PostFixArrayStackObj : TMathTokenArrayStackObj;
begin {function TMathFunction.VerifyMathFunc}
  {begin set default values}
  FFoundVarInFunc := false;
  FErrorPosition := 0;
  FErrorCode := meNoError;
  {end set default values}
  try
    MEStk := nil;
    FPostFixArray := nil;
    TokenArrayStackObj := TMathTokenArrayStackObj.Create;
    DescStackArrayStackObj := TMathTokenArrayStackObj.Create;
    PostFixArrayStackObj := TMathTokenArrayStackObj.Create;
    try
      InitMathValue(FResultValue); //FResultValue := 0;
      TokenizeVerifyMathStr(TokenArrayStackObj);
      if FErrorCode = meNoError then
        begin
          ConvInfixToPostFix(TokenArrayStackObj, DescStackArrayStackObj,
                             PostFixArrayStackObj);
          DescStackArrayStackObj.Clear;
          TokenArrayStackObj.Clear;
          InitProcedurePointers(PostFixArrayStackObj);
          if PostFixArrayStackObj.Count <= 0 then
            MEStk := nil
          else
            SetLength(MEStk, PostFixArrayStackObj.Count);
          if FFoundVarInFunc then
            begin
              SimplifyPostFix(PostFixArrayStackObj);
              if FErrorCode = meNoError then
                PostFixArrayStackObj.CopyToTMathTokenArray(FPostFixArray);
            end
          else
            begin
              EvalConstantPostFix(PostFixArrayStackObj);
              FPostFixArray := nil;
            end; {else}
          if Length(MEStk) <> Length(FPostFixArray) then
            SetLength(MEStk, Length(FPostFixArray));
        end; {if}
    finally
      TokenArrayStackObj.Free;
      DescStackArrayStackObj.Free;
      PostFixArrayStackObj.Free;
    end; {try..finally}
  except
    on EOutOfMemory do
      FErrorCode := meOutOfMem;
    else
      FErrorCode := meUnknownException;
  end; {try..execpt}
  if FErrorCode = meNoError then
    begin
      Result := true;
      FStatus := fsValid;
    end
  else
    begin
      Result:= false;
      FStatus := fsValidateError;
      MEStk := nil;
      FPostFixArray := nil;
    end; {else}
end; {function TMathFunction.VerifyMathFunc}

{=========================================================================}

function TMathFunction.EvaluateMathFunc : boolean;
{returns true if no error occured}
begin {function TMathFunction.EvaluateMathFunc}
  if FStatus = fsValid then
    begin
      Assert((Length(MEStk) >= Length(FPostFixArray)), 'Incorrect array size TMathFunction.EvaluateMathFunc');
      FErrorCode := meNoError;
      if FFoundVarInFunc then
        begin
          {begin evaluation code}
          try
            MStkI := -1;
            PostFixCurrEvalRec := 0;
            while PostFixCurrEvalRec < Length(FPostFixArray) do
              begin
                FPostFixArray[PostFixCurrEvalRec].MathProc;
                if FErrorCode <> meNoError then
                  break;
                inc(PostFixCurrEvalRec);
              end; {while}
          except
            on EOverFlow do   FErrorCode := meOverFlow;
            on EUnderFlow do  FErrorCode := meUnderFlow;
            on EZeroDivide do FErrorCode := meDivZero;
            on EInvalidOp do  FErrorCode := meInvalidOp;
            else FErrorCode := meUnknownException;
          end; {except}
          if MStkI = 0 then
            FResultValue := MEStk[0]
          else
            FErrorCode := meUnknownException;
          {end evaluation code}
        end; {if}
    end
  else
    FErrorCode := meUnknownException;
  Result := (FErrorCode = meNoError);
  if not Result then
    InitMathValue(FResultValue); //FResultValue := 0;{set FResultValue to empty if an error occured}
end; {function TMathFunction.EvaluateMathFunc}

{=========================================================================}

procedure TMathFunction.TokenizeVerifyMathStr(TokenObj : TMathTokenArrayStackObj);
var
  MLErPos : integer;
  MLError : TMathErrorCode;

{-------------------------------------------------------------------------}

procedure TokenizeAndStore;
var
  Index : integer;
  PrevBlankSpaceCount, PrevCharCount : integer;
  LocalStr : string;

{-----------------------------------------}

function FindFuncStrPos(LocalIndex : integer) : integer;
var
  BlankSpaceCount, CharCount : integer;
begin {function FindFuncStrPos}
  if (PrevCharCount - PrevBlankSpaceCount) < LocalIndex then
    begin
      BlankSpaceCount := PrevBlankSpaceCount;
      CharCount := PrevCharCount;
    end
  else
    begin
      BlankSpaceCount := 0;
      CharCount := 0;
    end; {else}
  while (CharCount - BlankSpaceCount) < LocalIndex do
    begin
      inc(CharCount);
      if FFuncString[CharCount] = ' ' then
        inc(BlankSpaceCount);
    end; {while}
  PrevBlankSpaceCount := BlankSpaceCount;
  PrevCharCount := CharCount;
  Result := LocalIndex + BlankSpaceCount;
end; {function FindFuncStrPos}

{-----------------------------------------}

procedure CleanAndStoreFFuncString;
var
  I : integer;
begin {procedure CleanAndStoreFFuncString;}
  LocalStr := '';
  for I := 1 to length(FFuncString) do
    if FFuncString[I] <> ' ' then
      if FFuncString[I] in ['{', '['] then
        LocalStr := LocalStr + '('
      else
        if FFuncString[I] in ['}', ']'] then
          LocalStr := LocalStr + ')'
        else
          LocalStr := LocalStr + upcase(FFuncString[I]);
end; {procedure CleanAndStoreFFuncString;}

{-----------------------------------------}

procedure VerifyChars;
var
  I : integer;
begin {procedure VerifyChars;}
  if LocalStr = '' then
    MLError := meNulExp
  else
    begin
      for I := 1 to length(LocalStr) do
        if not ((LocalStr[I] in LegalCharSet) or (LocalStr[I] = DecimalSeparator)) then
          begin
            MLError := meIlleagleChar;
            MLErPos := FindFuncStrPos(I);
            break;
          end; {if}
    end; {else}
end; {procedure VerifyChars;}

{-----------------------------------------}

function FindNumMatch : boolean;
var
  I : integer;
  ConvertSuccessful : boolean;
  TmpNumber : TMathValue;
  ConvertErrorPos : integer;
begin {function FindNumMatch}
  Result := false;
  if (LocalStr[Index] in ['0'..'9']) or (LocalStr[Index] = DecimalSeparator) then
    begin
      Result := true;
      InitMathValue(TmpNumber); //TmpNumber := 0;
      ConvertSuccessful := false;
      ConvertErrorPos := 0;
      I := Index;
      while (I <= length(LocalStr)) and ((LocalStr[I] in ['0'..'9']) or (LocalStr[I] = DecimalSeparator)) do
        begin
          if LocalStr[I] in ['0'..'9'] then
            ConvertSuccessful := true;
          inc(I);
        end; {while}
      if ConvertSuccessful then
        begin
          try
            if not StringToMathValue(copy(LocalStr, Index, (I - Index)),
                                     ConvertErrorPos, TmpNumber) then
              ConvertSuccessful := false;
          except
            ConvertSuccessful := false;
          end; {try..except}
        end; {if}
      if ConvertSuccessful then
        begin
          TokenObj.AddFields(Empty_MathProc,
                             sy_Number,
                             Empty_VarNum,
                             FindFuncStrPos(Index),
                             TmpNumber);
        end
      else
        begin
          MLError := meInvalidNum;
          if ConvertErrorPos > 0 then
            MLErPos := FindFuncStrPos(Index + (ConvertErrorPos - 1))
          else
            MLErPos := FindFuncStrPos(Index);
        end; {else}
      Index := I;
    end; {if}
end; {function FindNumMatch}

{-----------------------------------------}

function FindSymbolMatch : boolean;
var
  I : integer;
begin {function FindSymbolMatch}
  Result := false;
  I := 0;
  while (I < sy_CountSymbol) and (not Result) do
    begin
      inc(I);
      if SymbolArray[I] = Copy(LocalStr, Index, Length(SymbolArray[I])) then
        begin
          Result := true;
          TokenObj.AddBlankRecord;
          with TokenObj.MathTokenArray[TokenObj.EndRecord] do
            begin
              FuncChar := FindFuncStrPos(Index);
              if I = sy_PI then
                begin
                  Desc := sy_Number;
                  Value := FloatToMathValue(pi);
                end
              else
                if I = sy_E then
                  begin
                    Desc := sy_Number;
                    Value := FloatToMathValue(Exp(1.0));
                  end
                else
                  begin
                    Desc := I;
                    InitMathValue(Value); //Value := 0;
                  end; {else}
            end; {with}
          Index := Index + Length(SymbolArray[I]);
        end; {if}
    end; {while}
end; {function FindSymbolMatch}

{-----------------------------------------}

procedure FindVarMatch;
var
  I : integer;
  DidFind : boolean;
begin {procedure FindVarMatch}
  DidFind := false;
  for I := 0 to (FVarCount - 1) do
    begin
      if (FVarStrings[I] <> '') and
         (UpperCase(FVarStrings[I]) = copy(LocalStr, Index, Length(FVarStrings[I]))) then
        begin
          DidFind := true;
          TokenObj.AddFields(Empty_MathProc,
                             sy_Variable,
                             I,
                             FindFuncStrPos(Index),
                             ReturnInitMathValue);
          Index := Index + length(FVarStrings[I]);
          Break;
        end; {if}
    end; {for}
  if DidFind then
    FFoundVarInFunc := true
  else
    begin
      MLErPos := FindFuncStrPos(Index);
      MLError := meUnKnownSymbol;
    end; {else}
end; {procedure FindVarMatch}

{-----------------------------------------}

begin {procedure TokenizeAndStore;}
  PrevBlankSpaceCount := 0;
  PrevCharCount := 0;
  CleanAndStoreFFuncString;
  VerifyChars;
  if MLError = meNoError then
    begin
      Index := 1;
      while (MLError = meNoError) and (Index <= length(LocalStr)) do
        begin
          {The following 2 funcs return true }
          {  if an error occured or if found.}
          if not FindNumMatch then
            if not FindSymbolMatch then
              FindVarMatch;
        end; {while}
    end; {if}
end; {procedure TokenizeAndStore;}

{-------------------------------------------------------------------------}

procedure VerifyGrouping;
var
  I : integer;
  GroupCount : integer;
  CurrDesc : integer;
begin {procedure VerifyGrouping;}
  GroupCount := 0;
  for I := 0 to (TokenObj.Count - 1) do
    begin
      CurrDesc := TokenObj.MathTokenArray[I].Desc;
      if CurrDesc = sy_OpenGroup then
        inc(GroupCount)
      else
        if CurrDesc = sy_CloseGroup then
          begin
            dec(GroupCount);
            if GroupCount < 0 then
              begin
                MLErPos := TokenObj.MathTokenArray[I].FuncChar;
                break;
              end; {if}
          end;
    end; {for}
  if GroupCount <> 0 then
    MLError := meGroupingError;
end; {procedure VerifyGrouping;}

{-------------------------------------------------------------------------}

procedure RemoveDoubleAddSub;
var
  FoundDouble : boolean;
  I : integer;
  PrevDesc, CurrDesc : integer;
begin {procedure RemoveDoubleAddSub;}
  FoundDouble := true;
  while FoundDouble do
    begin
      FoundDouble := false;
      PrevDesc := sy_Number;  { set PrevDesc to something other than }
                              { sy_Add or sy_Sub.                    }
      for I := 0 to (TokenObj.Count - 1) do
        begin
          CurrDesc := TokenObj.MathTokenArray[I].Desc;
          if (PrevDesc in [sy_Add,sy_Sub]) and
             (CurrDesc in [sy_Add,sy_Sub]) then
            begin
              FoundDouble := true;
              if PrevDesc = CurrDesc then
                TokenObj.MathTokenArray[I-1].Desc := sy_Add
              else {if PrevDesc <> CurrDesc}
                TokenObj.MathTokenArray[I-1].Desc := sy_Sub;
              TokenObj.Delete(I, 1);
              Break;
            end; {if}
          PrevDesc := CurrDesc;
        end; {for}
    end; {while}
end; {procedure RemoveDoubleAddSub;}

{-------------------------------------------------------------------------}

procedure FixImpliedMuls;
var
  DidFind : boolean;
  I : integer;
  PrevDesc, CurrDesc : integer;
begin {procedure FixImpliedMuls;}
  DidFind := true;
  while DidFind do
    begin
      DidFind := false;
      PrevDesc := sy_Add;   { Set PrevDesc to a value other than }
                            { the value being searched for.      }
      for I := 0 to (TokenObj.Count - 1) do
        begin
          CurrDesc := TokenObj.MathTokenArray[I].Desc;
          if (CurrDesc = sy_Number) or (CurrDesc = sy_Variable) or
             (CurrDesc in sy_FunctSet) or (CurrDesc = sy_OpenGroup) then
             if (PrevDesc = sy_Number) or (PrevDesc = sy_Variable) or
                (PrevDesc = sy_CloseGroup) then
               begin
                 DidFind := true;
                 TokenObj.InsertFields(I,
                                       Empty_MathProc,
                                       sy_Mul,
                                       Empty_VarNum,
                                       TokenObj.MathTokenArray[I].FuncChar,
                                       ReturnInitMathValue);
                 Break;
               end;
          PrevDesc := CurrDesc;
        end; {for}
    end; {while}
end; {procedure FixImpliedMuls;}

{-------------------------------------------------------------------------}

procedure FixOpposites;
var
  OppositeFuncChar : integer;
  DidFind : boolean;
  I : integer;
  PrevPrevDesc, PrevDesc, CurrDesc : integer;

{----------------------------------------}

procedure MulNegOneWithGrouping;
var
  CurrFuncChar : integer;
  GroupLevel : integer;
begin {procedure MulNegOneWithGrouping;}
  TokenObj.InsertBlankRecord(I);
  TokenObj.InsertBlankRecord(I);
  with TokenObj.MathTokenArray[I-1] do
    begin
      Desc := sy_OpenGroup;
      FuncChar := OppositeFuncChar;
      InitMathValue(Value); //Value := 0;
    end; {with}
  with TokenObj.MathTokenArray[I] do
    begin
      Desc := sy_Number;
      FuncChar := OppositeFuncChar;
      Value := LongintToMathValue(-1);
    end; {with}
  inc(I);
  with TokenObj.MathTokenArray[I] do
    begin
      Desc := sy_Mul;
      FuncChar := OppositeFuncChar;
      InitMathValue(Value); //Value := 0;
    end; {with}
  inc(I);
  CurrFuncChar := TokenObj.MathTokenArray[I].FuncChar;
  CurrDesc := TokenObj.MathTokenArray[I].Desc;
  if (CurrDesc in sy_FunctSet) then
    begin
      if I >= TokenObj.EndRecord then
        begin
          MLError := meFuncWithOutOpenGroup;
          MLErPos := CurrFuncChar;
        end
      else
        begin
          inc(I);
          CurrDesc := TokenObj.MathTokenArray[I].Desc;
          if CurrDesc <> sy_OpenGroup then
            begin
              MLError := meFuncWithOutOpenGroup;
              MLErPos := CurrFuncChar;
            end; {if}
        end; {else}
    end; {if}
  if MLError = meNoError then
    begin
      if (CurrDesc = sy_OpenGroup) then
        begin
          GroupLevel := 1;
          while (I < TokenObj.EndRecord) and (GroupLevel <> 0) do
            begin
              inc(I);
              CurrDesc := TokenObj.MathTokenArray[I].Desc;
              if CurrDesc = sy_OpenGroup then
                inc(GroupLevel)
              else
                if CurrDesc = sy_CloseGroup then
                  dec(GroupLevel);
            end; {while}
        end; {if}
      TokenObj.InsertFields((I+1),
                            Empty_MathProc,
                            sy_CloseGroup,
                            Empty_VarNum,
                            0,
                            ReturnInitMathValue);
    end; {if}
end; {procedure MulNegOneWithGrouping;}

{----------------------------------------}

procedure MulNegOneNoGrouping;
begin {procedure MulNegOneNoGrouping;}
  TokenObj.InsertBlankRecord(I);
  with TokenObj.MathTokenArray[I-1] do
    begin
      Desc := sy_Number;
      FuncChar := OppositeFuncChar;
      Value := LongintToMathValue(-1);
    end; {with}
  with TokenObj.MathTokenArray[I] do
    begin
      Desc := sy_Mul;
      FuncChar := OppositeFuncChar;
      InitMathValue(Value); //Value := 0;
    end; {with}
end; {procedure MulNegOneNoGrouping;}

{----------------------------------------}

begin {procedure FixOpposites;}
  DidFind := true;
  while DidFind do
    begin
      DidFind := false;
      PrevPrevDesc := sy_Number;  { Set PrevDesc's to values other than }
      PrevDesc := sy_Number;      { the values being searched for. }
      for I := 0 to (TokenObj.Count - 1) do
        begin
          CurrDesc := TokenObj.MathTokenArray[I].Desc;
          if (CurrDesc = sy_Variable) or (CurrDesc in sy_FunctSet) or
             (CurrDesc = sy_Number) or (CurrDesc = sy_OpenGroup) then
            if (PrevDesc in [sy_Add,sy_Sub]) then
              if ((PrevPrevDesc in sy_OperatorSet) and
                 (PrevPrevDesc <> sy_CloseGroup)) or (I = 1) then
                begin
                  if PrevDesc = sy_Add then
                    begin
                      TokenObj.Delete(I-1, 1);
                    end
                  else {if PrevDesc = sy_Sub then}
                    begin
                      OppositeFuncChar := TokenObj.MathTokenArray[I-1].FuncChar;
                      if PrevPrevDesc in [sy_Mul, sy_Div, sy_Power] then
                        MulNegOneWithGrouping
                      else
                        MulNegOneNoGrouping;
                      if MLError <> meNoError then
                        begin
                          DidFind := false;
                          Break;
                        end; {if}
                    end; {else}
                  DidFind := true;
                  Break;
                end; {if}
          PrevPrevDesc := PrevDesc;
          PrevDesc := CurrDesc;
        end; {for}
    end; {while}
end; {procedure FixOpposites;}

{-------------------------------------------------------------------------}

procedure VerifyOrder;
{ This procedure searches for illeagle token sequences. }
{ It takes the place of a lexical analizer.             }

{--------------------------------}

procedure DoLoopSearch;
var
  I : integer;
  PrevDesc, CurrDesc : integer;
begin {procedure DoLoopSearch;}
  PrevDesc := sy_Number; { Set PrevDesc to a value other than }
                         { the value being searched for.      }
  I := 0;
  while (I < TokenObj.Count) and (MLError = meNoError) do
    begin
      CurrDesc := TokenObj.MathTokenArray[I].Desc;
      if (PrevDesc in sy_AddMulEtcSet) or (PrevDesc = sy_OpenGroup) then
        begin
          if (CurrDesc in sy_AddMulEtcSet)or(CurrDesc = sy_CloseGroup) then
            begin
              MLError := meExprError;
              MLErPos := TokenObj.MathTokenArray[I].FuncChar;
            end; {if}
        end
      else
        begin
          if (PrevDesc in sy_FunctSet) then
            begin
              if CurrDesc <> sy_OpenGroup then
                begin
                  MLError := meFuncWithOutOpenGroup;
                  //next line assumes I > 0
                  MLErPos := TokenObj.MathTokenArray[I-1].FuncChar;
                end; {if}
            end; {if}
        end; {else}
      PrevDesc := CurrDesc;
      inc(I);
    end; {while}
end; {procedure DoLoopSearch;}

{--------------------------------}

begin {procedure VerifyOrder;}
  if TokenObj.MathTokenArray[0].Desc in sy_AddMulEtcSet then
    begin
      MLError := meExprError;
      MLErPos := TokenObj.MathTokenArray[0].FuncChar;
    end
  else
    begin
      if TokenObj.MathTokenArray[TokenObj.EndRecord].Desc in sy_AddMulEtcSet then
        begin
          MLError := meExprError;
          MLErPos := TokenObj.MathTokenArray[TokenObj.EndRecord].FuncChar;
        end
      else
        if TokenObj.MathTokenArray[TokenObj.EndRecord].Desc in sy_FunctSet then
          begin
            MLError := meFuncWithOutOpenGroup;
            MLErPos := TokenObj.MathTokenArray[TokenObj.EndRecord].FuncChar;
          end
        else
          DoLoopSearch;
    end; {else}
end; {procedure VerifyOrder;}

{-------------------------------------------------------------------------}

begin {procedure TMathFunction.TokenizeVerifyMathStr}
  FFoundVarInFunc := false;
  MLErPos := 0;
  MLError := meNoError;
  if length(FFuncString) > MaxMathStringLength then
    MLError := meStringToBig
  else
    begin
      TokenObj.SetCapacity(length(FFuncString) + 8);
      TokenizeAndStore;
      if MLError = meNoError then
        begin
          VerifyGrouping;
          if MLError = meNoError then
            begin
              RemoveDoubleAddSub;
              FixImpliedMuls;
              FixOpposites;
              if MLError = meNoError then
                VerifyOrder;
            end; {if}
        end; {if}
    end; {else}
  FErrorPosition := MLErPos;
  FErrorCode := MLError;
end; {procedure TMathFunction.TokenizeVerifyMathStr}

{=========================================================================}

procedure TMathFunction.ConvInfixToPostFix(TokenObj, DescObj, PostObj : TMathTokenArrayStackObj);
var
  QueCount : integer;
  QueRec1, QueRec2 : TMathTokenRec;
  TopDescRec : TMathTokenRec;
  I : integer;

{-------------------------------------}

procedure PostFixAddNode(const TheTokenRec : TMathTokenRec);
begin {procedure PostFixAddNode}
  PostObj.AddTokenRec(TheTokenRec);
end; {procedure PostFixAddNode}

{-------------------------------------}

function DescStackPop : TMathTokenRec;
begin {function DescStackPop}
  Result := DescObj.PopTokenRec;
end; {function DescStackPop}

{-------------------------------------}

procedure DescStackPush(const TheTokenRec : TMathTokenRec);
begin {procedure DescStackPush}
  DescObj.PushTokenRec(TheTokenRec);
end; {procedure DescStackPush}

{-------------------------------------}

procedure AddToQue(const TheTokenRec : TMathTokenRec);
begin {procedure AddToQue}
  case QueCount of
    0: begin
         QueRec1 := TheTokenRec;
         QueCount := 1;
       end; {0:}
    1: begin
         QueRec2 := TheTokenRec;
         QueCount := 2;
       end; {1:}
    2: begin
         PostFixAddNode(QueRec1);
         QueRec1 := QueRec2;
         QueRec2 := TheTokenRec;
       end; {2:}
  end; {case}
end; {procedure AddToQue}

{-------------------------------------}

procedure QueRemoveTwo;
begin {procedure QueRemoveTwo;}
  case QueCount of
    1: begin
         PostFixAddNode(QueRec1);
         QueCount := 0;
       end; {1:}
    2: begin
         PostFixAddNode(QueRec1);
         PostFixAddNode(QueRec2);
         QueCount := 0;
       end; {2:}
  end; {case}
end; {procedure QueRemoveTwo;}

{-------------------------------------}

procedure QueRemoveOne;
begin {procedure QueRemoveOne;}
  case QueCount of
    1: begin
         PostFixAddNode(QueRec1);
         QueCount := 0;
       end; {1:}
    2: begin
         PostFixAddNode(QueRec2);
         QueCount := 1;
       end; {2:}
  end; {case}
end; {procedure QueRemoveOne;}

{-------------------------------------}

procedure QueRemoveAndAdd(const TheTokenRec : TMathTokenRec);
begin {procedure QueRemoveAndAdd}
  if TheTokenRec.Desc in sy_FunctSet then
    QueRemoveOne
  else
    QueRemoveTwo;
  PostFixAddNode(TheTokenRec);
end; {procedure QueRemoveAndAdd}

{-------------------------------------}

function GetOperOrder(TheDesc : integer) : integer;
begin {function GetOperOrder}
  if TheDesc in sy_FunctSet then
    Result := 4
  else
    begin
      case TheDesc of
        sy_Add        : Result := 1;
        sy_Sub        : Result := 1;
        sy_Mul        : Result := 2;
        sy_Div        : Result := 2;
        sy_Power      : Result := 3;
        sy_OpenGroup  : Result := 5;
        sy_CloseGroup : Result := 5;
      else
        Result := 0;
      end; {case}
    end; {else}
end; {function GetOperOrder}

{-------------------------------------}

procedure GotAddMulEtc(const TheInfixTokenRec : TMathTokenRec);
var
  TmpTopDesc : integer;
  LoopAgain : boolean;
begin {procedure GotAddMulEtc}
  repeat
    LoopAgain := false;
    if DescObj.Count <> 0 then
      begin
        TmpTopDesc := DescObj.MathTokenArray[DescObj.EndRecord].Desc;
        if TmpTopDesc <> sy_OpenGroup then
          if GetOperOrder(TheInfixTokenRec.Desc) <= GetOperOrder(TmpTopDesc) then
            begin
              QueRemoveAndAdd(DescStackPop);
              LoopAgain := true;
            end; {if}
      end; {if}
  until LoopAgain = false;
  DescStackPush(TheInfixTokenRec);
end; {procedure GotAddMulEtc}

{-------------------------------------}

begin {procedure TMathFunction.ConvInfixToPostFix}
  DescObj.SetCapacity(TokenObj.Count);
  PostObj.SetCapacity(TokenObj.Count);
  QueCount := 0; {intiialize Que}
  for I := 0 to (TokenObj.Count - 1) do
    begin
      if (TokenObj.MathTokenArray[I].Desc = sy_Number) or (TokenObj.MathTokenArray[I].Desc = sy_Variable) then
        AddToQue(TokenObj.MathTokenArray[I])
      else
        if TokenObj.MathTokenArray[I].Desc in sy_FunctSet then
          DescStackPush(TokenObj.MathTokenArray[I])
        else
          if TokenObj.MathTokenArray[I].Desc = sy_OpenGroup then
            begin
              DescStackPush(TokenObj.MathTokenArray[I]);
              QueRemoveTwo;
            end
          else
            if TokenObj.MathTokenArray[I].Desc = sy_CloseGroup then
              begin
                TopDescRec := DescStackPop;
                while TopDescRec.Desc <> sy_OpenGroup do
                  begin
                    QueRemoveAndAdd(TopDescRec);
                    TopDescRec := DescStackPop;
                  end; {while}
              end {if}
            else {if (InfixDesc in sy_AddMulEtcSet) then}
              GotAddMulEtc(TokenObj.MathTokenArray[I]);
    end; {for}
  while DescObj.Count <> 0 do
    QueRemoveAndAdd(DescStackPop);
  QueRemoveTwo;
end; {procedure TMathFunction.ConvInfixToPostFix}

{=========================================================================}

function TMathFunction.GetMathProcedure(TheDesc : integer) : PMathProcedure;
begin {function TMathFunction.GetMathProcedure}
  if TheDesc = sy_Variable then
    Result := PMath_PushVar
  else
  if TheDesc = sy_Number then
    Result := PMath_sy_Number
  else
    case TheDesc of
      sy_Add : Result := PMath_sy_Add;
      sy_Sub : Result := PMath_sy_Sub;
      sy_Mul : Result := PMath_sy_Mul;
      sy_Div : Result := PMath_sy_Div;
      sy_Power : Result := PMath_sy_Power;
      sy_DegToRad : Result := PMath_sy_DegToRad;
      sy_RadToDeg : Result := PMath_sy_RadToDeg;
      sy_GradToRad : Result := PMath_sy_GradToRad;
      sy_RadToGrad : Result := PMath_sy_RadToGrad;
      sy_CycleToRad : Result := PMath_sy_CycleToRad;
      sy_RadToCycle : Result := PMath_sy_RadToCycle;
      sy_ARCSINH : Result := PMath_sy_ARCSINH;
      sy_ARCCOSH : Result := PMath_sy_ARCCOSH;
      sy_ARCTANH : Result := PMath_sy_ARCTANH;
      sy_ARCCSCH : Result := PMath_sy_ARCCSCH;
      sy_ARCSECH : Result := PMath_sy_ARCSECH;
      sy_ARCCOTH : Result := PMath_sy_ARCCOTH;
      sy_ARCSIN : Result := PMath_sy_ARCSIN;
      sy_ARCCOS : Result := PMath_sy_ARCCOS;
      sy_ARCTAN : Result := PMath_sy_ARCTAN;
      sy_ARCCSC : Result := PMath_sy_ARCCSC;
      sy_ARCSEC : Result := PMath_sy_ARCSEC;
      sy_ARCCOT : Result := PMath_sy_ARCCOT;
      sy_SINH : Result := PMath_sy_SINH;
      sy_COSH : Result := PMath_sy_COSH;
      sy_TANH : Result := PMath_sy_TANH;
      sy_CSCH : Result := PMath_sy_CSCH;
      sy_SECH : Result := PMath_sy_SECH;
      sy_COTH : Result := PMath_sy_COTH;
      sy_SIN : Result := PMath_sy_SIN;
      sy_COS : Result := PMath_sy_COS;
      sy_TAN : Result := PMath_sy_TAN;
      sy_CSC : Result := PMath_sy_CSC;
      sy_SEC : Result := PMath_sy_SEC;
      sy_COT : Result := PMath_sy_COT;
      sy_EXP : Result := PMath_sy_EXP;
      sy_GRINT : Result := PMath_sy_GRINT;
      sy_INT : Result := PMath_sy_INT;
      sy_SQRT : Result := PMath_sy_SQRT;
      sy_SQUARE : Result := PMath_sy_SQUARE;
      sy_TRUNC : Result := PMath_sy_TRUNC;
      sy_ROUND : Result := PMath_sy_ROUND;
      sy_FRAC : Result := PMath_sy_FRAC;
      sy_ABS : Result := PMath_sy_ABS;
      sy_LN : Result := PMath_sy_LN;
      sy_LOG : Result := PMath_sy_LOG;
      sy_FACT : Result := PMath_sy_FACT;
      else Result := nil;
    end; {case}
end; {function TMathFunction.GetMathProcedure}

{=========================================================================}

procedure TMathFunction.InitProcedurePointers(PostObj : TMathTokenArrayStackObj);
var
  I : integer;
begin {procedure TMathFunction.InitProcedurePointers;}
  for I := 0 to (PostObj.Count - 1) do
    PostObj.MathTokenArray[I].MathProc := GetMathProcedure(PostObj.MathTokenArray[I].Desc);
end; {procedure TMathFunction.InitProcedurePointers;}

{=========================================================================}

procedure TMathFunction.EvalConstantPostFix(PostObj : TMathTokenArrayStackObj);
begin {procedure TMathFunction.EvalConstantPostFix;}
  Assert((Length(MEStk) >= PostObj.Count), 'Incorrect array size TMathFunction.EvalConstantPostFix');
  FErrorCode := meNoError;
  try
    MStkI := -1;
    PostFixCurrEvalRec := 0;
    while PostFixCurrEvalRec < PostObj.Count do
      begin
        with PostObj.MathTokenArray[PostFixCurrEvalRec] do
          begin
            if Desc = sy_Number then
              begin
                inc(MStkI);
                MEStk[MStkI] := Value;
              end
            else
              if Desc = sy_Variable then
                FErrorCode := meUnknownException
              else
                MathProc;
          end; {with}
        if FErrorCode <> meNoError then
          break;
        inc(PostFixCurrEvalRec);
      end; {while}
  except
    on EOverFlow do   FErrorCode := meOverFlow;
    on EUnderFlow do  FErrorCode := meUnderFlow;
    on EZeroDivide do FErrorCode := meDivZero;
    on EInvalidOp do  FErrorCode := meInvalidOp;
    else FErrorCode := meUnknownException;
  end; {except}
  if FErrorCode = meNoError then
    begin
      if MStkI = 0 then
        FResultValue := MEStk[0]
      else
        begin
          FErrorCode := meUnknownException;
          InitMathValue(FResultValue); //FResultValue := 0;
        end; {else}
    end
  else
    InitMathValue(FResultValue); //FResultValue := 0;
end; {procedure TMathFunction.EvalConstantPostFix;}

{=========================================================================}

procedure TMathFunction.SimplifyPostFix(PostObj : TMathTokenArrayStackObj);
var
  DidFind : boolean;
  I : integer;
  PrevPrevDesc,  PrevDesc,  CurrDesc : integer;
  SimpValue : TMathValue;
begin {procedure TMathFunction.SimplifyPostFix}
  Assert((Length(MEStk) >= PostObj.Count), 'Incorrect array size TMathFunction.SimplifyPostFix');
  FErrorCode := meNoError;
  FErrorPosition := 0;
  DidFind := true;
  while DidFind do
    begin
      DidFind := false;
      PrevPrevDesc := sy_Add;  { Set PrevDesc's to values other than }
      PrevDesc := sy_Add;      { the values being searched for. }
      for I := 0 to (PostObj.Count - 1) do
        begin
          CurrDesc := PostObj.MathTokenArray[I].Desc;
          if CurrDesc in sy_FunctSet then
            begin
              if PrevDesc = sy_Number then
                begin
                  SimpValue := ComputeAFunc(PostObj.MathTokenArray[I].MathProc,
                                            PostObj.MathTokenArray[I-1].Value);
                  if FErrorCode <> meNoError then
                    FErrorPosition := PostObj.MathTokenArray[I].FuncChar
                  else
                    begin
                      with PostObj.MathTokenArray[I-1] do
                        begin
                          MathProc := PMath_sy_Number;
                          Desc := sy_Number;
                          FuncChar := 0;
                          Value := SimpValue;
                        end; {with}
                      PostObj.Delete(I, 1);
                      DidFind := true;
                    end; {if}
                  Break;
                end; {if}
            end {if}
          else
            if CurrDesc in sy_AddMulEtcSet then
              begin
                if (PrevDesc = sy_Number) and (PrevPrevDesc = sy_Number) then
                  begin
                    SimpValue := ComputeAOperator(PostObj.MathTokenArray[I].MathProc,
                        PostObj.MathTokenArray[I-2].Value, PostObj.MathTokenArray[I-1].Value);
                      if FErrorCode <> meNoError then
                        FErrorPosition := PostObj.MathTokenArray[I].FuncChar
                      else
                        begin
                          with PostObj.MathTokenArray[I] do
                            begin
                              MathProc := PMath_sy_Number;
                              Desc := sy_Number;
                              FuncChar := 0;
                              Value := SimpValue;
                            end; {with}
                          PostObj.Delete(I-2, 2);
                          DidFind := true;
                        end; {if}
                    Break;
                  end; {if}
              end; {if}
          PrevPrevDesc := PrevDesc;
          PrevDesc := CurrDesc;
        end; {for}
    end; {while}
end; {procedure TMathFunction.SimplifyPostFix}

{=========================================================================}

procedure TMathFunction.PMath_PushVar;
begin {procedure TMathFunction.PMath_PushVar;}
  inc(MStkI);
  MEStk[MStkI] := FVarValues[FPostFixArray[PostFixCurrEvalRec].VarNum];
end; {procedure TMathFunction.PMath_PushVar;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_Number;
begin {procedure TMathFunction.PMath_sy_Number;}
  inc(MStkI);
  MEStk[MStkI] := FPostFixArray[PostFixCurrEvalRec].Value;
end; {procedure TMathFunction.PMath_sy_Number;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_Add;
begin {procedure TMathFunction.PMath_sy_Add;}
  dec(MStkI);
  MEStk[MStkI] := AddMathValue(MEStk[MStkI], MEStk[MStkI+1]);
end; {procedure TMathFunction.PMath_sy_Add;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_Sub;
begin {procedure TMathFunction.PMath_sys_Sub;}
  dec(MStkI);
  MEStk[MStkI] := SubMathValue(MEStk[MStkI], MEStk[MStkI+1]);
end; {procedure TMathFunction.PMath_sy_Sub;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_Mul;
begin {procedure TMathFunction.PMath_sy_Mul;}
  dec(MStkI);
  MEStk[MStkI] := MulMathValue(MEStk[MStkI], MEStk[MStkI+1]);
end; {procedure TMathFunction.PMath_sy_Mul;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_Div;
var
  DivByZeroError : boolean;
begin {procedure TMathFunction.PMath_sy_Div;}
  dec(MStkI);
  MEStk[MStkI] := DivMathValue(MEStk[MStkI], MEStk[MStkI+1], DivByZeroError);
  if DivByZeroError then
    FErrorCode := meDivZero;
end; {procedure TMathFunction.PMath_sy_Div;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_Power;
var
  InvalidPowerError : boolean;
begin {procedure TMathFunction.PMath_sy_Power;}
  dec(MStkI);
  MEStk[MStkI] := PowerMathValue(MEStk[MStkI], MEStk[MStkI+1], InvalidPowerError);
  if InvalidPowerError then
    FErrorCode := meInvalidPower;
end; {procedure TMathFunction.PMath_sy_Power;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_DegToRad;
{ Radians := Degrees * PI / 180 }
begin {procedure TMathFunction.PMath_sy_DegToRad;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := MEStk[MStkI].FloatValue * (PI / 180);
end; {procedure TMathFunction.PMath_sy_DegToRad;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_RadToDeg;
{ Degrees := Radians * 180 / PI }
begin {procedure TMathFunction.PMath_sy_RadToDeg;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := MEStk[MStkI].FloatValue * (180 / PI);
end; {procedure TMathFunction.PMath_sy_RadToDeg;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_GradToRad;
{ Radians := Grads * PI / 200 }
begin {procedure TMathFunction.PMath_sy_GradToRad;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := MEStk[MStkI].FloatValue * (PI / 200);
end; {procedure TMathFunction.PMath_sy_GradToRad;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_RadToGrad;
{ Grads := Radians * 200 / PI }
begin {procedure TMathFunction.PMath_sy_RadToGrad;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := MEStk[MStkI].FloatValue * (200 / PI);
end; {procedure TMathFunction.PMath_sy_RadToGrad;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_CycleToRad;
{ Radians := Cycles * 2PI }
begin {procedure TMathFunction.PMath_sy_CycleToRad;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := MEStk[MStkI].FloatValue * (2 * PI);
end; {procedure TMathFunction.PMath_sy_CycleToRad;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_RadToCycle;
{ Cycles := Radians / 2PI }
begin {procedure TMathFunction.PMath_sy_RadToCycle;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := MEStk[MStkI].FloatValue / (2 * PI);
end; {procedure TMathFunction.PMath_sy_RadToCycle;}

{----------------------------------------------------------------------}

function TMathFunction.MathCalc_ArcSin(X : TMathFloat) : TMathFloat;
begin {ArcSin(x) = ArcTan (x/sqrt (1-sqr (x)))}
  if X = 1 then
    MathCalc_ArcSin := (pi/2)
  else
    begin
      if X = -1 then
        MathCalc_ArcSin := -(pi/2)
      else
        begin
          if (-1 <= X) and (X <= 1) then
            MathCalc_ArcSin := ArcTan (X / sqrt (1-sqr (X)))
          else
            begin
              FErrorCode := meFuncOutDomain;
              {guarantee valid floating point value is returned}
              MathCalc_ArcSin := 0;
            end; {else}
        end; {else}
    end; {else}
end; {function TMathFunction.MathCalc_ArcSin}

{----------------------------------------------------------------------}

function TMathFunction.MathCalc_ArcCos(X : TMathFloat) : TMathFloat;
begin {ArcCos(x) = ArcTan (-x / sqrt(1 - sqr(x))) + (pi/2)}
  if X = 1 then
    MathCalc_ArcCos := 0
  else
    begin
      if X = (-1) then
        MathCalc_ArcCos := pi
      else
        begin
          if (-1 <= X) and (X <= 1) then
            MathCalc_ArcCos := ArcTan ((-X) / sqrt(1 - sqr(X))) + (pi/2)
          else
            begin
              FErrorCode := meFuncOutDomain;
              {guarantee valid floating point value is returned}
              MathCalc_ArcCos := 0;
            end; {else}
        end; {else}
    end; {else}
end; {function TMathFunction.MathCalc_ArcCos}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCSINH;
begin {procedure TMathFunction.PMath_sy_ARCSINH;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := ln(MEStk[MStkI].FloatValue + Sqrt(Sqr(MEStk[MStkI].FloatValue) + 1));
end; {procedure TMathFunction.PMath_sy_ARCSINH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCCOSH;
begin {procedure TMathFunction.PMath_sy_ARCCOSH;}
  MEStk[MStkI].ValueType := vtFloat;
  if MEStk[MStkI].FloatValue >= 1 then
    MEStk[MStkI].FloatValue := ln(MEStk[MStkI].FloatValue + Sqrt(Sqr(MEStk[MStkI].FloatValue) - 1))
  else
    FErrorCode := meFuncOutDomain
end; {procedure TMathFunction.PMath_sy_ARCCOSH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCTANH;
begin {procedure TMathFunction.PMath_sy_ARCTANH;}
  MEStk[MStkI].ValueType := vtFloat;
  if abs(MEStk[MStkI].FloatValue) < 1 then
    MEStk[MStkI].FloatValue := ln((1+MEStk[MStkI].FloatValue) / (1-MEStk[MStkI].FloatValue)) / 2
  else
    FErrorCode := meFuncOutDomain
end; {procedure TMathFunction.PMath_sy_ARCTANH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCCSCH;
begin {procedure TMathFunction.PMath_sy_ARCCSCH;}
  MEStk[MStkI].ValueType := vtFloat;
  if MEStk[MStkI].FloatValue <> 0 then
    MEStk[MStkI].FloatValue := ln( (1/MEStk[MStkI].FloatValue) + (Sqrt(1+Sqr(MEStk[MStkI].FloatValue))/Abs(MEStk[MStkI].FloatValue)) )
  else
    FErrorCode := meFuncOutDomain
end; {procedure TMathFunction.PMath_sy_ARCCSCH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCSECH;
begin {procedure TMathFunction.PMath_sy_ARCSECH;}
  MEStk[MStkI].ValueType := vtFloat;
  if (0 < MEStk[MStkI].FloatValue) and (MEStk[MStkI].FloatValue <= 1) then
    MEStk[MStkI].FloatValue := ln((1+Sqrt(1-sqr(MEStk[MStkI].FloatValue))) / MEStk[MStkI].FloatValue)
  else
    FErrorCode := meFuncOutDomain
end; {procedure TMathFunction.PMath_sy_ARCSECH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCCOTH;
begin {procedure TMathFunction.PMath_sy_ARCCOTH;}
  MEStk[MStkI].ValueType := vtFloat;
  if abs(MEStk[MStkI].FloatValue) > 1 then
    MEStk[MStkI].FloatValue := ln((MEStk[MStkI].FloatValue+1) / (MEStk[MStkI].FloatValue-1)) / 2
  else
    FErrorCode := meFuncOutDomain
end; {procedure TMathFunction.PMath_sy_ARCCOTH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCSIN;
begin {procedure TMathFunction.PMath_sy_ARCSIN;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := MathCalc_ArcSin(MEStk[MStkI].FloatValue);
end; {procedure TMathFunction.PMath_sy_ARCSIN;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCCOS;
begin {procedure TMathFunction.PMath_sy_ARCCOS;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := MathCalc_ArcCos(MEStk[MStkI].FloatValue);
end; {procedure TMathFunction.PMath_sy_ARCCOS;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCTAN;
begin {procedure TMathFunction.PMath_sy_ARCTAN;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := ArcTan(MEStk[MStkI].FloatValue);
end; {procedure TMathFunction.PMath_sy_ARCTAN;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCCSC;
begin {procedure TMathFunction.PMath_sy_ARCCSC;}
  MEStk[MStkI].ValueType := vtFloat;
  if (MEStk[MStkI].FloatValue >= 1)or(MEStk[MStkI].FloatValue <= (-1)) then
    MEStk[MStkI].FloatValue := MathCalc_ArcSin(1 / MEStk[MStkI].FloatValue)
  else
    FErrorCode := meFuncOutDomain;
end; {procedure TMathFunction.PMath_sy_ARCCSC;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCSEC;
begin {procedure TMathFunction.PMath_sy_ARCSEC;}
  MEStk[MStkI].ValueType := vtFloat;
  if (MEStk[MStkI].FloatValue >= 1)or(MEStk[MStkI].FloatValue <= (-1)) then
    MEStk[MStkI].FloatValue := MathCalc_ArcCos(1 / MEStk[MStkI].FloatValue)
  else
    FErrorCode := meFuncOutDomain;
end; {procedure TMathFunction.PMath_sy_ARCSEC;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ARCCOT;
begin {procedure TMathFunction.PMath_sy_ARCCOT;}
  MEStk[MStkI].ValueType := vtFloat;
  if MEStk[MStkI].FloatValue > 0 then
    MEStk[MStkI].FloatValue := arctan(1/MEStk[MStkI].FloatValue)
  else
    begin
      if MEStk[MStkI].FloatValue < 0 then
        MEStk[MStkI].FloatValue := pi + arctan(1/MEStk[MStkI].FloatValue)
      else {if MEStk[MStkI].FloatValue = 0 then}
        MEStk[MStkI].FloatValue := (pi/2);
    end; {else}
end; {procedure TMathFunction.PMath_sy_ARCCOT;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_SINH;
begin {procedure TMathFunction.PMath_sy_SINH;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := (Exp(MEStk[MStkI].FloatValue) - Exp(-MEStk[MStkI].FloatValue)) / 2;
end; {procedure TMathFunction.PMath_sy_SINH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_COSH;
begin {procedure TMathFunction.PMath_sy_COSH;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := (Exp(MEStk[MStkI].FloatValue) + Exp(-MEStk[MStkI].FloatValue)) / 2;
end; {procedure TMathFunction.PMath_sy_COSH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_TANH;
begin {procedure TMathFunction.PMath_sy_TANH;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := (Exp(MEStk[MStkI].FloatValue) - Exp(-MEStk[MStkI].FloatValue)) /
                  (Exp(MEStk[MStkI].FloatValue) + Exp(-MEStk[MStkI].FloatValue));
end; {procedure TMathFunction.PMath_sy_TANH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_CSCH;
begin {procedure TMathFunction.PMath_sy_CSCH;}
  MEStk[MStkI].ValueType := vtFloat;
  if MEStk[MStkI].FloatValue = 0 then
    FErrorCode := meFuncOutDomain
  else
    MEStk[MStkI].FloatValue := 2 / (Exp(MEStk[MStkI].FloatValue) - Exp(-MEStk[MStkI].FloatValue));
end; {procedure TMathFunction.PMath_sy_CSCH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_SECH;
begin {procedure TMathFunction.PMath_sy_SECH;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := 2 / (Exp(MEStk[MStkI].FloatValue) + Exp(-MEStk[MStkI].FloatValue));
end; {procedure TMathFunction.PMath_sy_SECH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_COTH;
begin {procedure TMathFunction.PMath_sy_COTH;}
  MEStk[MStkI].ValueType := vtFloat;
  if MEStk[MStkI].FloatValue = 0 then
    FErrorCode := meFuncOutDomain
  else
    MEStk[MStkI].FloatValue := (Exp(MEStk[MStkI].FloatValue) + Exp(-MEStk[MStkI].FloatValue)) /
                    (Exp(MEStk[MStkI].FloatValue) - Exp(-MEStk[MStkI].FloatValue));
end; {procedure TMathFunction.PMath_sy_COTH;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_SIN;
begin {procedure TMathFunction.PMath_sy_SIN;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := SinFixup(MEStk[MStkI].FloatValue);
end; {procedure TMathFunction.PMath_sy_SIN;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_COS;
begin {procedure TMathFunction.PMath_sy_COS;}
  MEStk[MStkI].ValueType := vtFloat;
  MEStk[MStkI].FloatValue := CosFixup(MEStk[MStkI].FloatValue);
end; {procedure TMathFunction.PMath_sy_COS;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_TAN;
var
  TmpR : TMathFloat;
begin {procedure TMathFunction.PMath_sy_TAN;}
  MEStk[MStkI].ValueType := vtFloat;
  TmpR := CosFixup(MEStk[MStkI].FloatValue);
  if TmpR = 0 then
    FErrorCode := meFuncOutDomain
  else
    MEStk[MStkI].FloatValue := SinFixup(MEStk[MStkI].FloatValue) / TmpR;
end; {procedure TMathFunction.PMath_sy_TAN;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_CSC;
var
  TmpR : TMathFloat;
begin {procedure TMathFunction.PMath_sy_CSC;}
  MEStk[MStkI].ValueType := vtFloat;
  TmpR := SinFixup(MEStk[MStkI].FloatValue);
  if TmpR = 0 then
    FErrorCode := meFuncOutDomain
  else
    MEStk[MStkI].FloatValue := 1 / TmpR;
end; {procedure TMathFunction.PMath_sy_CSC;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_SEC;
var
  TmpR : TMathFloat;
begin {procedure TMathFunction.PMath_sy_SEC;}
  MEStk[MStkI].ValueType := vtFloat;
  TmpR := CosFixup(MEStk[MStkI].FloatValue);
  if TmpR = 0 then
    FErrorCode := meFuncOutDomain
  else
    MEStk[MStkI].FloatValue := 1 / TmpR;
end; {procedure TMathFunction.PMath_sy_SEC;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_COT;
var
  TmpR : TMathFloat;
begin {procedure TMathFunction.PMath_sy_COT;}
  MEStk[MStkI].ValueType := vtFloat;
  TmpR := SinFixup(MEStk[MStkI].FloatValue);
  if TmpR = 0 then
    FErrorCode := meFuncOutDomain
  else
    MEStk[MStkI].FloatValue := CosFixup(MEStk[MStkI].FloatValue) / TmpR;
end; {procedure TMathFunction.PMath_sy_COT;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_EXP;
begin {procedure TMathFunction.PMath_sy_EXP;}
  MEStk[MStkI] := ExpFuncMathValue(MEStk[MStkI]);
end; {procedure TMathFunction.PMath_sy_EXP;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_GRINT;
begin {procedure TMathFunction.PMath_sy_GRINT;}
  MEStk[MStkI] := GrintFuncMathValue(MEStk[MStkI]);
end; {procedure TMathFunction.PMath_sy_GRINT;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_INT;
begin {procedure TMathFunction.PMath_sy_INT;}
  MEStk[MStkI] := IntFuncMathValue(MEStk[MStkI]);
end; {procedure TMathFunction.PMath_sy_INT;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_SQRT;
var
  FuncOutDomainError : boolean;
begin {procedure TMathFunction.PMath_sy_SQRT;}
  MEStk[MStkI] := SqrtFuncMathValue(MEStk[MStkI], FuncOutDomainError);
  if FuncOutDomainError then
    FErrorCode := meFuncOutDomain;
end; {procedure TMathFunction.PMath_sy_SQRT;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_SQUARE;
begin {procedure TMathFunction.PMath_sy_SQUARE;}
  MEStk[MStkI] := SquareFuncMathValue(MEStk[MStkI]);
end; {procedure TMathFunction.PMath_sy_SQUARE;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_TRUNC;
begin {procedure TMathFunction.PMath_sy_TRUNC;}
  MEStk[MStkI] := IntFuncMathValue(MEStk[MStkI]); {int is same as trunc}
end; {procedure TMathFunction.PMath_sy_TRUNC;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ROUND;
begin {procedure TMathFunction.PMath_sy_ROUND;}
  MEStk[MStkI] := RoundFuncMathValue(MEStk[MStkI]);
end; {procedure TMathFunction.PMath_sy_ROUND;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_FRAC;
begin {procedure TMathFunction.PMath_sy_FRAC;}
  MEStk[MStkI] := FracFuncMathValue(MEStk[MStkI]);
end; {procedure TMathFunction.PMath_sy_FRAC;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_ABS;
begin {procedure TMathFunction.PMath_sy_ABS;}
  MEStk[MStkI] := AbsFuncMathValue(MEStk[MStkI]);
end; {procedure TMathFunction.PMath_sy_ABS;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_LN;
begin {procedure TMathFunction.PMath_sy_LN;}
  MEStk[MStkI].ValueType := vtFloat;
  if MEStk[MStkI].FloatValue <= 0 then
    FErrorCode := meFuncOutDomain
  else
    MEStk[MStkI].FloatValue := Ln(MEStk[MStkI].FloatValue);
end; {procedure TMathFunction.PMath_sy_LN;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_LOG;
begin {procedure TMathFunction.PMath_sy_LOG;}
  MEStk[MStkI].ValueType := vtFloat;
  if MEStk[MStkI].FloatValue <= 0 then
    FErrorCode := meFuncOutDomain
  else
    MEStk[MStkI].FloatValue := ln(MEStk[MStkI].FloatValue) / Ln(10);
end; {procedure TMathFunction.PMath_sy_LOG;}

{----------------------------------------------------------------------}

procedure TMathFunction.PMath_sy_FACT;
var
  FuncOutDomainError : boolean;
  FactToBigError : boolean;
begin {procedure TMathFunction.PMath_sy_FACT;}
  MEStk[MStkI] := FactFuncMathValue(MEStk[MStkI], FuncOutDomainError, FactToBigError);
  if FuncOutDomainError then
    FErrorCode := meFuncOutDomain
  else
    if FactToBigError then
      FErrorCode := meFactToBig;
end; {procedure TMathFunction.PMath_sy_FACT;}

{=========================================================================}

function TMathFunction.ComputeAOperator(TheMathProc : PMathProcedure;
                            Val1, Val2 : TMathValue): TMathValue;
begin {function TMathFunction.ComputeAOperator}
  Assert((Length(MEStk) >= 2), 'Incorrect array size TMathFunction.ComputeAOperator');
  try
    FErrorCode := meNoError;
    MStkI := 1;
    MEStk[0] := Val1;
    MEStk[1] := Val2;
    TheMathProc;
    Result := MEStk[0];
  except
    on EOverFlow do begin
                    FErrorCode := meOverFlow;
                    Result := FloatToMathValue(0);
                  end;
    on EUnderFlow do begin
                    FErrorCode := meUnderFlow;
                    Result := FloatToMathValue(0);
                  end;
    on EZeroDivide do begin
                    FErrorCode := meDivZero;
                    Result := FloatToMathValue(0);
                  end;
    on EInvalidOp do begin
                    FErrorCode := meInvalidOp;
                    Result := FloatToMathValue(0);
                  end;
    else begin
                    FErrorCode := meUnknownException;
                    Result := FloatToMathValue(0);
                  end;
  end; {except}
end; {function TMathFunction.ComputeAOperator}

{=========================================================================}

function TMathFunction.ComputeAFunc(TheMathProc : PMathProcedure; Val1 : TMathValue): TMathValue;
begin {function TMathFunction.ComputeAFunc}
  Assert((Length(MEStk) >= 1), 'Incorrect array size TMathFunction.ComputeAFunc');
  try
    FErrorCode := meNoError;
    MStkI := 0;
    MEStk[0] := Val1;
    TheMathProc;
    Result := MEStk[0];
  except
    on EOverFlow do begin
                    FErrorCode := meOverFlow;
                    Result := FloatToMathValue(0);
                  end;
    on EUnderFlow do begin
                    FErrorCode := meUnderFlow;
                    Result := FloatToMathValue(0);
                  end;
    on EZeroDivide do begin
                    FErrorCode := meDivZero;
                    Result := FloatToMathValue(0);
                  end;
    on EInvalidOp do begin
                    FErrorCode := meInvalidOp;
                    Result := FloatToMathValue(0);
                  end;
    else begin
                    FErrorCode := meUnknownException;
                    Result := FloatToMathValue(0);
                  end;
  end; {except}
end; {function TMathFunction.ComputeAFunc}

{=========================================================================}

function SinNormal(X : TMathFloat) : TMathFloat;
begin
  Result := Sin(X);
end;

function SinRoundThou(X : TMathFloat) : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.60461921527854657E-16) and (abs(X) <= (1000 * pi)) then
    Result := 0;
end;

function SinRoundTenThou(X : TMathFloat) : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.38533932589335330E-15) and (abs(X) <= (10000 * pi)) then
    Result := 0;
end;

function SinRoundHundThou(X : TMathFloat) : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.93736086201434788E-14) and (abs(X) <= (100000 * pi)) then
    Result := 0;
end;

function SinRoundMill(X : TMathFloat) : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.67272391684430977E-13) and (abs(X) <= (1000000 * pi)) then
    Result := 0;
end;

function SinRoundTenMill(X : TMathFloat) : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.45127985150261818E-12) and (abs(X) <= (10000000 * pi)) then
    Result := 0;
end;

{----------------------------------------------------------------------}

function SinExactThou(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.60461921527854657E-16) and (abs(X) <= (1000 * pi)) then
    begin
      try
        MullAmt := Round(X / Pi);
        if X = (MullAmt * Pi) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

function SinExactTenThou(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.38533932589335330E-15) and (abs(X) <= (10000 * pi)) then
    begin
      try
        MullAmt := Round(X / Pi);
        if X = (MullAmt * Pi) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

function SinExactHundThou(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.93736086201434788E-14) and (abs(X) <= (100000 * pi)) then
    begin
      try
        MullAmt := Round(X / Pi);
        if X = (MullAmt * Pi) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

function SinExactMill(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.67272391684430977E-13) and (abs(X) <= (1000000 * pi)) then
    begin
      try
        MullAmt := Round(X / Pi);
        if X = (MullAmt * Pi) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

function SinExactTenMill(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Sin(X);
  if (abs(Result) <= 1.45127985150261818E-12) and (abs(X) <= (10000000 * pi)) then
    begin
      try
        MullAmt := Round(X / Pi);
        if X = (MullAmt * Pi) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

{----------------------------------------------------------------------}

function CosNormal(X : TMathFloat) : TMathFloat;
begin
  Result := Cos(X);
end;

function CosRoundThou(X : TMathFloat) : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 7.95533344061238879E-17) and (abs(X) <= (1001 * (pi / 2))) then
    Result := 0;
end;

function CosRoundTenThou(X : TMathFloat) : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 6.92669662946676646E-16) and (abs(X) <= (10001 * (pi / 2))) then
    Result := 0;
end;

function CosRoundHundThou(X : TMathFloat) : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 9.68612668371393593E-15) and (abs(X) <= (100001 * (pi / 2))) then
    Result := 0;
end;

function CosRoundMill(X : TMathFloat) : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 8.36361958422154884E-14) and (abs(X) <= (1000001 * (pi / 2))) then
    Result := 0;
end;

function CosRoundTenMill(X : TMathFloat) : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 7.25141274067128694E-13) and (abs(X) <= (10000001 * (pi / 2))) then
    Result := 0;
end;

{----------------------------------------------------------------------}

function CosExactThou(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 7.95533344061238879E-17) and (abs(X) <= (1001 * (pi / 2))) then
    begin
      try
        MullAmt := Round(X * (2 / Pi));
        if X = (MullAmt * (Pi / 2)) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

function CosExactTenThou(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 6.92669662946676646E-16) and (abs(X) <= (10001 * (pi / 2))) then
    begin
      try
        MullAmt := Round(X * (2 / Pi));
        if X = (MullAmt * (Pi / 2)) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

function CosExactHundThou(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 9.68612668371393593E-15) and (abs(X) <= (100001 * (pi / 2))) then
    begin
      try
        MullAmt := Round(X * (2 / Pi));
        if X = (MullAmt * (Pi / 2)) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

function CosExactMill(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 8.36361958422154884E-14) and (abs(X) <= (1000001 * (pi / 2))) then
    begin
      try
        MullAmt := Round(X * (2 / Pi));
        if X = (MullAmt * (Pi / 2)) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

function CosExactTenMill(X : TMathFloat) : TMathFloat;
var
  MullAmt : TMathFloat;
begin
  Result := Cos(X);
  if (abs(Result) <= 7.25141274067128694E-13) and (abs(X) <= (10000001 * (pi / 2))) then
    begin
      try
        MullAmt := Round(X * (2 / Pi));
        if X = (MullAmt * (Pi / 2)) then
          Result := 0;
      except
      end; {Try..Except}
    end; {if}
end;

{=========================================================================}

function TMathFunction.GetFloatResult: TMathFloat;
begin
  Result := FResultValue.FloatValue;
end;

function TMathFunction.GetFloatVariables(Ndx: integer): TMathFloat;
begin
  if FVarValues[Ndx].ValueType = vtEmpty then
    Result := 0
  else
    Result := FVarValues[Ndx].FloatValue;
end;

function TMathFunction.GetMathVariables(Ndx: integer): TMathValue;
begin
  Result := FVarValues[Ndx];
end;

procedure TMathFunction.SetFloatVariables(Ndx: integer;
  const Value: TMathFloat);
begin
  FVarValues[Ndx].ValueType := vtFloat;
  FVarValues[Ndx].FloatValue := Value;
end;

procedure TMathFunction.SetMathVariables(Ndx: integer;
  const Value: TMathValue);
begin
  FVarValues[Ndx] := Value;
end;

end.
