unit MathStak;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Sysutils, MathCalc;

type
  PMathProcedure = procedure of object;

const
  Empty_MathProc = nil;
  Empty_Desc = 0;
  Empty_VarNum = 0;
  Empty_FuncChar = 0;

type
  TMathTokenRec = record
    MathProc : PMathProcedure;
    Desc : integer;
    VarNum : integer;
    FuncChar : integer;
    Value : TMathValue;
  end; {TMathTokenRec}

  TMathTokenArray = array of TMathTokenRec;

  TMathEvalStack = array of TMathValue;

  TMathStringsArray = array of string;
  TMathValuesArray = array of TMathValue;

  TMathTokenArrayStackObj = class
  private
    { Private declarations }
    FCount : integer;
    FMathTokenArray : TMathTokenArray;
    function GetStackTop : integer;
    procedure ExpandArrayIfNecessary;
  public
    { Public declarations }
    property Count : integer read FCount;
    property StackTop : integer read GetStackTop;
    property EndRecord : integer read GetStackTop;
    property MathTokenArray : TMathTokenArray read FMathTokenArray;
    constructor Create;
    destructor Destroy; override;
    procedure SetCapacity(InitialSize : integer);
    procedure Assign(Source : TMathTokenArrayStackObj);
    procedure CopyToTMathTokenArray(var MathTokenArray : TMathTokenArray);
    procedure PushBlankRecord;
    procedure AddBlankRecord;
    procedure PopBlankRecord;
    procedure DeleteLastBlankRecord;
    procedure InsertBlankRecord(Position : integer);
    procedure PushFields(TheMathProc : PMathProcedure;
                         TheDesc : integer;
                         TheVarNum : integer;
                         TheFuncChar : integer;
                         const TheValue : TMathValue);
    procedure PushTokenRec(const TheRec : TMathTokenRec);
    procedure PopFields(var TheMathProc : PMathProcedure;
                         var TheDesc : integer;
                         var TheVarNum : integer;
                         var TheFuncChar : integer;
                         var TheValue : TMathValue);
    function PopTokenRec : TMathTokenRec;
    procedure AddFields(TheMathProc : PMathProcedure;
                         TheDesc : integer;
                         TheVarNum : integer;
                         TheFuncChar : integer;
                         const TheValue : TMathValue);
    procedure AddTokenRec(const TheRec : TMathTokenRec);
    procedure InsertFields(Position : integer;
                         TheMathProc : PMathProcedure;
                         TheDesc : integer;
                         TheVarNum : integer;
                         TheFuncChar : integer;
                         const TheValue : TMathValue);
    procedure InsertTokenRec(Position : integer; const TheRec : TMathTokenRec);
    procedure Delete(Position, RecCount : integer);
    procedure Clear;
  end; {TMathTokenArrayStackObj}

procedure InitTMathTokenRec(var TheRec : TMathTokenRec);

implementation

{----------------------------------------------------------------------}

procedure InitTMathTokenRec(var TheRec : TMathTokenRec);
begin {procedure InitTMathTokenRec}
  with TheRec do
    begin
      MathProc := Empty_MathProc;
      Desc := Empty_Desc;
      VarNum := Empty_VarNum;
      FuncChar := Empty_FuncChar;
      InitMathValue(Value);
    end; {with}
end; {procedure InitTMathTokenRec}

{----------------------------------------------------------------------}

{ TMathTokenArrayStackObj }

procedure TMathTokenArrayStackObj.AddBlankRecord;
begin
  ExpandArrayIfNecessary;
  InitTMathTokenRec(FMathTokenArray[FCount]);
  inc(FCount);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.AddFields(TheMathProc: PMathProcedure; TheDesc,
  TheVarNum, TheFuncChar: integer; const TheValue: TMathValue);
begin
  ExpandArrayIfNecessary;
  with FMathTokenArray[FCount] do
    begin
      MathProc := TheMathProc;
      Desc := TheDesc;
      VarNum := TheVarNum;
      FuncChar := TheFuncChar;
      Value := TheValue;
    end; {with}
  inc(FCount);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.AddTokenRec(const TheRec: TMathTokenRec);
begin
  ExpandArrayIfNecessary;
  FMathTokenArray[FCount] := TheRec;
  inc(FCount);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.Assign(Source: TMathTokenArrayStackObj);
begin
  FCount := Source.FCount;
  FMathTokenArray := Copy(Source.FMathTokenArray);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.Clear;
begin
  FCount := 0;
  FMathTokenArray := nil;
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.CopyToTMathTokenArray(
  var MathTokenArray: TMathTokenArray);
begin
  MathTokenArray := nil;
  if FCount > 0 then
    MathTokenArray := Copy(FMathTokenArray, 0, FCount);
end;

{----------------------------------------------------------------------}

constructor TMathTokenArrayStackObj.Create;
begin
  inherited;
  FCount := 0;
  FMathTokenArray := nil;
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.Delete(Position, RecCount: integer);
var
  I : integer;
begin
  Assert(((Position >= 0) and (Position < FCount)), 'invalid Position parameter in procedure TMathTokenArrayStackObj.Delete');
  Assert((RecCount >= 1), 'invalid RecCount parameter in procedure TMathTokenArrayStackObj.Delete');
  if (Position + RecCount) >= FCount then
    FCount := Position
  else
    begin
      FCount := FCount - RecCount;
      for I := Position to (FCount - 1) do
        FMathTokenArray[I] := FMathTokenArray[I + RecCount];
    end; {else}
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.DeleteLastBlankRecord;
begin
  Assert((FCount >= 1), 'Error in procedure TMathTokenArrayStackObj.DeleteLastBlankRecord');
  dec(FCount);
end;

{----------------------------------------------------------------------}

destructor TMathTokenArrayStackObj.Destroy;
begin
  FMathTokenArray := nil;
  inherited;
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.ExpandArrayIfNecessary;
var
  TheLength : integer;
begin
  TheLength := Length(FMathTokenArray);
  if FCount >= TheLength then
    begin
      if TheLength <= 64 then
        SetLength(FMathTokenArray, (TheLength + 8))
      else
        if TheLength <= 128 then
          SetLength(FMathTokenArray, (TheLength + 16))
        else
          if TheLength <= 512 then
            SetLength(FMathTokenArray, (TheLength + 32))
          else
            if TheLength <= 1024 then
              SetLength(FMathTokenArray, (TheLength + 64))
            else
              SetLength(FMathTokenArray, (TheLength + 128));
    end; {if}
end;

{----------------------------------------------------------------------}

function TMathTokenArrayStackObj.GetStackTop: integer;
begin
  Result := FCount - 1;
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.InsertBlankRecord(Position: integer);
var
  I : integer;
begin
  Assert(((Position >= 0) and (Position <= FCount)), 'invalid position parameter in procedure TMathTokenArrayStackObj.InsertBlankRecord');
  ExpandArrayIfNecessary;
  for I := (FCount - 1) downto Position do
    FMathTokenArray[I+1] := FMathTokenArray[I];
  inc(FCount);
  InitTMathTokenRec(FMathTokenArray[Position]);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.InsertFields(Position: integer;
  TheMathProc: PMathProcedure; TheDesc, TheVarNum, TheFuncChar: integer;
  const TheValue: TMathValue);
var
  I : integer;
begin
  Assert(((Position >= 0) and (Position <= FCount)), 'invalid position parameter in procedure TMathTokenArrayStackObj.InsertFields');
  ExpandArrayIfNecessary;
  for I := (FCount - 1) downto Position do
    FMathTokenArray[I+1] := FMathTokenArray[I];
  inc(FCount);
  with FMathTokenArray[Position] do
    begin
      MathProc := TheMathProc;
      Desc := TheDesc;
      VarNum := TheVarNum;
      FuncChar := TheFuncChar;
      Value := TheValue;
    end; {with}
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.InsertTokenRec(Position: integer;
  const TheRec: TMathTokenRec);
var
  I : integer;
begin
  Assert(((Position >= 0) and (Position <= FCount)), 'invalid position parameter in procedure TMathTokenArrayStackObj.InsertTokenRec');
  ExpandArrayIfNecessary;
  for I := (FCount - 1) downto Position do
    FMathTokenArray[I+1] := FMathTokenArray[I];
  inc(FCount);
  FMathTokenArray[Position] := TheRec;
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.PopBlankRecord;
begin
  Assert((FCount >= 1), 'Error in procedure TMathTokenArrayStackObj.PopBlankRecord');
  dec(FCount);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.PopFields(var TheMathProc: PMathProcedure;
  var TheDesc, TheVarNum, TheFuncChar: integer; var TheValue: TMathValue);
begin
  Assert((FCount >= 1), 'Error in procedure TMathTokenArrayStackObj.PopFields');
  dec(FCount);
  with FMathTokenArray[FCount] do
    begin
      TheMathProc := MathProc;
      TheDesc := Desc;
      TheVarNum := VarNum;
      TheFuncChar := FuncChar;
      TheValue := Value;
    end; {with}
end;

{----------------------------------------------------------------------}

function TMathTokenArrayStackObj.PopTokenRec: TMathTokenRec;
begin
  Assert((FCount >= 1), 'Error in procedure TMathTokenArrayStackObj.PopTokenRec');
  dec(FCount);
  Result := FMathTokenArray[FCount];
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.PushBlankRecord;
begin
  ExpandArrayIfNecessary;
  InitTMathTokenRec(FMathTokenArray[FCount]);
  inc(FCount);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.PushFields(TheMathProc: PMathProcedure;
  TheDesc, TheVarNum, TheFuncChar: integer; const TheValue: TMathValue);
begin
  ExpandArrayIfNecessary;
  with FMathTokenArray[FCount] do
    begin
      MathProc := TheMathProc;
      Desc := TheDesc;
      VarNum := TheVarNum;
      FuncChar := TheFuncChar;
      Value := TheValue;
    end; {with}
  inc(FCount);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.PushTokenRec(
  const TheRec: TMathTokenRec);
begin
  ExpandArrayIfNecessary;
  FMathTokenArray[FCount] := TheRec;
  inc(FCount);
end;

{----------------------------------------------------------------------}

procedure TMathTokenArrayStackObj.SetCapacity(InitialSize: integer);
begin
  Assert(InitialSize >= 0, 'Invalid parameter in procedure TMathTokenArrayStackObj.SetInitialCapacity()');
  if (length(FMathTokenArray) <> InitialSize) and (InitialSize >= FCount) then
    SetLength(FMathTokenArray, InitialSize);
end;

{----------------------------------------------------------------------}

end.
