unit PtBuffer;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  SyncObjs, Graphics, PtShared;

type
  TSafeBmpBuffer = class
  private
    { Private declarations }
    FBufferBitmap : TBitmap;
    CriticalSection : TCriticalSection;
  protected
    { Protected declarations }
  public
    { Public declarations }
    //begin non-thread safe methods
    constructor Create;
    destructor Destroy; override;
    procedure InitalizeNew(TheColorDepth : TPixelFormat;
                           TheDefaultBmpWidth, TheDefaultBmpHeight : integer);
    procedure SetColorDepth(TheColorDepth : TPixelFormat);
    procedure SetSize(TheBmpWidth, TheBmpHeight : integer);
    //begin thread safe methods
    procedure SetBufferBitmap(TheSourceBmp : TBitmap);
    procedure GetBufferBitmap(TheDestinationBmp : TBitmap);
  end; {TSafeBmpBuffer}

implementation

{---------------------------------------------------------------------}

{ TSafeBmpBuffer }

constructor TSafeBmpBuffer.Create;
begin {constructor TSafeBmpBuffer.Create;}
  inherited;
  CriticalSection := TCriticalSection.Create;
  FBufferBitmap := TBitmap.Create;
  FBufferBitmap.HandleType := bmDIB;
end; {constructor TSafeBmpBuffer.Create;}

{---------------------------------------------------------------------}

destructor TSafeBmpBuffer.Destroy;
begin {destructor TSafeBmpBuffer.Destroy;}
  FBufferBitmap.free;
  CriticalSection.Free;
  inherited;
end; {destructor TSafeBmpBuffer.Destroy;}

{---------------------------------------------------------------------}

procedure TSafeBmpBuffer.GetBufferBitmap(TheDestinationBmp: TBitmap);
begin {procedure TSafeBmpBuffer.GetBufferBitmap}
  CriticalSection.Acquire;
  try
    Assert(FBufferBitmap.Width = TheDestinationBmp.Width, 'Width not equal in procedure TSafeBmpBuffer.GetBufferBitmap');
    Assert(FBufferBitmap.Height = TheDestinationBmp.Height, 'Width not equal in procedure TSafeBmpBuffer.GetBufferBitmap');
    DoCopyBitmapBuffers(TheDestinationBmp, FBufferBitmap);
  finally
    CriticalSection.Release;
  end; {try..finally}
end; {procedure TSafeBmpBuffer.GetBufferBitmap}

{---------------------------------------------------------------------}

procedure TSafeBmpBuffer.InitalizeNew(TheColorDepth: TPixelFormat;
  TheDefaultBmpWidth, TheDefaultBmpHeight: integer);
begin {procedure TSafeBmpBuffer.InitalizeNew}
  SetColorDepth(TheColorDepth);
  SetSize(TheDefaultBmpWidth, TheDefaultBmpHeight);
end; {procedure TSafeBmpBuffer.InitalizeNew}

{---------------------------------------------------------------------}

procedure TSafeBmpBuffer.SetBufferBitmap(TheSourceBmp: TBitmap);
begin {procedure TSafeBmpBuffer.SetBufferBitmap}
  CriticalSection.Acquire;
  try
    Assert(FBufferBitmap.Width = TheSourceBmp.Width, 'Width not equal in procedure TSafeBmpBuffer.SetBufferBitmap');
    Assert(FBufferBitmap.Height = TheSourceBmp.Height, 'Width not equal in procedure TSafeBmpBuffer.SetBufferBitmap');
    DoCopyBitmapBuffers(FBufferBitmap, TheSourceBmp);
  finally
    CriticalSection.Release;
  end; {try..finally}
end; {procedure TSafeBmpBuffer.SetBufferBitmap}

{---------------------------------------------------------------------}

procedure TSafeBmpBuffer.SetColorDepth(TheColorDepth: TPixelFormat);
begin {procedure TSafeBmpBuffer.SetColorDepth}
  FBufferBitmap.PixelFormat := TheColorDepth;
  FBufferBitmap.HandleType := bmDIB;
end; {procedure TSafeBmpBuffer.SetColorDepth}

{---------------------------------------------------------------------}

procedure TSafeBmpBuffer.SetSize(TheBmpWidth, TheBmpHeight : integer);
begin {procedure TSafeBmpBuffer.SetSize}
  FBufferBitmap.SetSize(TheBmpWidth, TheBmpHeight);
  FBufferBitmap.HandleType := bmDIB;
end; {procedure TSafeBmpBuffer.SetSize}

{---------------------------------------------------------------------}

end.
