unit PtGr3d;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Windows, Classes, Graphics, GrGlobal, MathCalc, MathFunc,
  PtGrList, Vw3dObj, Set3dObj, Func3d, Fg3DLine;

const
  Max3DFuncCount = Fg3DLine.Max3DFuncCount;

type
  TPixOwnerArray = array of array of word;

type
  TPlotGraph3D = class(TPlotGraphFuncList)
  private
    { Private declarations }
    FSPScanLines : packed array[0..MaxBmpDim] of pointer;
    FSPPaintDC : HDC;
    FSPCurrLineColor : TColor;
    FSPCurrLineRGB : Cardinal;
    FSPPalette4BitsLo : byte;
    FSPPalette4BitsHi : byte;
    FSPPalette8Bits : byte;
    FSPPalette15And16Bits : word;
    FSPPalette24Red : byte;
    FSPPalette24Green : byte;
    FSPPalette24Blue : byte;
    FSPPalette32Bits : Cardinal;
    {Saved to Disk}
    FView3DZoomObj : TView3DZoomObj;
    FSetting3DObj : TSetting3DObj;
    {NOT Saved to Disk}
    LineObj : TLineObj;
    PixOwnerArray : TPixOwnerArray;
    AxisLabelBufferBmp : TBitmap;
    OldBmpHeight : integer;
    OldBmpWidth : integer;
    function TColorTo32BitPixel(TheColor : TColor) : Cardinal;
    procedure FastSetPixDefault(X, Y : integer);
    procedure FastSetPix4Bits(X, Y : integer);
    procedure FastSetPix8Bits(X, Y : integer);
    procedure FastSetPix15And16Bits(X, Y : integer);
    procedure FastSetPix24Bits(X, Y : integer);
    procedure FastSetPix32Bits(X, Y : integer);
    procedure GraphWinPaint;
  protected
    { Protected declarations }
  public
    { Public declarations }
    property View3DZoomObj : TView3DZoomObj read FView3DZoomObj;
    property Setting3DObj : TSetting3DObj read FSetting3DObj;
    constructor Create;
    destructor Destroy; override;
    function FullGraphPaint : string; override;
    procedure RestartAnimatedGraphPaint; override;
    function StepAnimatedGraphPaint : boolean; override;
  end; {TPlotGraph3D}

implementation

{ TPlotGraph3D }

{---------------------------------------------------------------------}

constructor TPlotGraph3D.Create;
begin {constructor TPlotGraph3D.Create}
  inherited;
  FSPPaintDC := 0;
  FSPCurrLineColor := clBlack;
  FSPCurrLineRGB := 0;
  FSPPalette4BitsLo := 0;
  FSPPalette4BitsHi := 0;
  FSPPalette8Bits := 0;
  FSPPalette15And16Bits := 0;
  FSPPalette24Red := 0;
  FSPPalette24Green := 0;
  FSPPalette24Blue := 0;
  FSPPalette32Bits := 0;
  PixOwnerArray := nil;
  OldBmpHeight := -1;
  OldBmpWidth := -1;
  AxisLabelBufferBmp := TBitmap.Create;
  with AxisLabelBufferBmp do
    begin
      HandleType := bmDIB;
      PixelFormat := pf8bit;
      Height := 32;
      Width := 32;
    end; {with}
  FView3DZoomObj := TView3DZoomObj.Create;
  FSetting3DObj := TSetting3DObj.Create;
  LineObj := TLineObj.Create;
end; {constructor TPlotGraph3D.Create}

{---------------------------------------------------------------------}

destructor TPlotGraph3D.Destroy;
begin {destructor TPlotGraph3D.Destroy;}
  AxisLabelBufferBmp.Free;
  FView3DZoomObj.free;
  FSetting3DObj.free;
  LineObj.Free;
  PixOwnerArray := nil;
  inherited;
end; {destructor TPlotGraph3D.Destroy;}

{---------------------------------------------------------------------}

function TPlotGraph3D.FullGraphPaint : string;
var
  I : integer;
begin {function TPlotGraph3D.FullGraphPaint}
  Result := '';
  if FunctionListDidChange then
  begin
    LineObj.ReInit;
    if FuncList.Count > 0 then
      for I := 0 to FuncList.Count -1 do
        Result := LineListErrorToString(TBaseFunc3DObj(FuncList[I]).AddLinesToList(I,LineObj));
  end; {if}
  if Result = '' then
  begin
    FunctionListDidChange := false;
    GraphWinPaint;
  end; {if}
end; {function TPlotGraph3D.FullGraphPaint}

{=====================================================================}

{begin declarations used by procedure TPlotGraph3D.GraphWinPaint;}

type
  TByteCast = packed array[0..(MaxBmpDim*16)] of byte;
  PByteCast = ^TByteCast;
  TBytePtr = ^byte;
  TWordCast = packed array[0..(MaxBmpDim*16)] of word;
  PWordCast = ^TWordCast;
  TCardinalCast = packed array[0..(MaxBmpDim*16)] of Cardinal;
  PCardinalCast = ^TCardinalCast;

{-----------------------------------------}

procedure ExtractRGBFromTColor(TheColor : TColor; var Red, Green, Blue : byte);
begin {procedure ExtractRGBFromTColor}
  Red := Cardinal(TheColor) and $000000FF;
  Green := (Cardinal(TheColor) and $0000FF00) shr 8;
  Blue := (Cardinal(TheColor) and $00FF0000) shr 16;
end; {procedure ExtractRGBFromTColor}

{-----------------------------------------}

function TPlotGraph3D.TColorTo32BitPixel(TheColor : TColor) : Cardinal;
var
  TmpRed, TmpGreen, TmpBlue : Cardinal;
begin {function TPlotGraph3D.TColorTo32BitPixel}
  ExtractRGBFromTColor(TheColor, FSPPalette24Red,
                       FSPPalette24Green, FSPPalette24Blue);
  TmpRed := FSPPalette24Red;
  TmpGreen := FSPPalette24Green;
  TmpBlue := FSPPalette24Blue;
  Result := TmpBlue or (TmpGreen shl 8) or (TmpRed shl 16);
end; {function TPlotGraph3D.TColorTo32BitPixel}

{-----------------------------------------}

procedure TPlotGraph3D.FastSetPixDefault(X, Y : integer);
begin {procedure TPlotGraph3D.FastSetPixDefault}
  SetPixelV(FSPPaintDC, X, Y, FSPCurrLineRGB);
end; {procedure TPlotGraph3D.FastSetPixDefault}

{-----------------------------------------}

procedure TPlotGraph3D.FastSetPix4Bits(X, Y : integer);
var
  BytePtr : TBytePtr;
begin {procedure TPlotGraph3D.FastSetPix4Bits}
  BytePtr := addr(PByteCast(FSPScanLines[Y])^[X div 2]);
  if (X mod 2) <> 0 then //if pixel is low nibble
    BytePtr^ := (BytePtr^ and $F0) or FSPPalette4BitsLo
  else //if pixel is high nibble
    BytePtr^ := (BytePtr^ and $0F) or FSPPalette4BitsHi;
end; {procedure TPlotGraph3D.FastSetPix4Bits}

{-----------------------------------------}

procedure TPlotGraph3D.FastSetPix8Bits(X, Y : integer);
begin {procedure TPlotGraph3D.FastSetPix8Bits}
  PByteCast(FSPScanLines[Y])^[X] := FSPPalette8Bits;
end; {procedure TPlotGraph3D.FastSetPix8Bits}

{-----------------------------------------}

procedure TPlotGraph3D.FastSetPix15And16Bits(X, Y : integer);
begin {procedure TPlotGraph3D.FastSetPix15And16Bits}
  PWordCast(FSPScanLines[Y])^[X] := FSPPalette15And16Bits;
end; {procedure TPlotGraph3D.FastSetPix15And16Bits}

{-----------------------------------------}

procedure TPlotGraph3D.FastSetPix24Bits(X, Y : integer);
var
  ByteRow : PByteCast;
begin {procedure TPlotGraph3D.FastSetPix24Bits}
  ByteRow := FSPScanLines[Y];
  X := X * 3;
  ByteRow^[X] := FSPPalette24Blue;
  ByteRow^[X + 1] := FSPPalette24Green;
  ByteRow^[X + 2] := FSPPalette24Red;
end; {procedure TPlotGraph3D.FastSetPix24Bits}

{-----------------------------------------}

procedure TPlotGraph3D.FastSetPix32Bits(X, Y : integer);
begin {procedure TPlotGraph3D.FastSetPix32Bits}
  PCardinalCast(FSPScanLines[Y])^[X] := FSPPalette32Bits;
end; {procedure TPlotGraph3D.FastSetPix32Bits}

{-----------------------------------------}

{end declarations used by procedure TPlotGraph3D.GraphWinPaint;}

procedure TPlotGraph3D.GraphWinPaint;
type
  PFastSetPixel = procedure(X, Y : integer) of object;
var
  FastSetPixel : PFastSetPixel;
  TrigAngles : TTrigAngles;
  HalfBMPHeight,
  HalfBmpWidth : integer;
  BmpHeightSub1,
  BmpWidthSub1 : integer;
  CurrPlot0, CurrPlot1 : TPlotRec;
  CurrLowZ, CurrHighZ : TFixedPoint;
  CurrFuncNum : byte;
  CurrLineNumber : word;
  GraphPixelFormat : TPixelFormat;

{-----------------------------------------}

procedure InitPixOwnerArray;
var
  I, J : integer;
begin {procedure InitPixOwnerArray;}
  if DoIsTerminated then Exit;
  //the next line could cause an out of memory error
  if (OldBmpWidth <> BmpWidth) or (OldBmpHeight <> BmpHeight) then
  begin
    SetLength(PixOwnerArray, BmpWidth, BmpHeight);
    OldBmpWidth := BmpWidth;
    OldBmpHeight := BmpHeight;
  end; {if}
  if DoIsTerminated then Exit;
  for I := 0 to (BmpWidth - 1) do
    for J := 0 to (BmpHeight - 1) do
      PixOwnerArray[I, J] := NilLineNumber;
end; {procedure InitPixOwnerArray;}

{-----------------------------------------}

procedure InitialCalculations;
var
  RadRotateX,
  RadRotateY,
  RadRotateZ : TMathFloat;
  I : integer;
begin {procedure InitialCalculations}
  if DoIsTerminated then Exit;
  with View3DZoomObj do
    begin
      RadRotateX := RotateX*(pi/180);
      RadRotateY := RotateY*(pi/180);
      RadRotateZ := RotateZ*(pi/180);
    end; {with}
  HalfBmpHeight := BMPHeight div 2;
  HalfBmpWidth := BMPWidth div 2;
  BmpHeightSub1 := BMPHeight -1;
  BmpWidthSub1 := BMPWidth -1;
  if DoIsTerminated then Exit;
  with TrigAngles do
    begin
      CosRadX := trunc(Cos(RadRotateX)*FixedMul);
      CosRadY := trunc(Cos(RadRotateY)*FixedMul);
      CosRadZ := trunc(Cos(RadRotateZ)*FixedMul);
      SinRadX := trunc(Sin(RadRotateX)*FixedMul);
      SinRadY := trunc(Sin(RadRotateY)*FixedMul);
      SinRadZ := trunc(Sin(RadRotateZ)*FixedMul);
    end; {with}
  if DoIsTerminated then Exit;
  FSPPaintDC := GraphBitmap.Canvas.Handle;
  GraphPixelFormat := GraphBitmap.PixelFormat;
  if DoIsTerminated then Exit;
  case GraphPixelFormat of
    pf4bit: FastSetPixel := FastSetPix4Bits;
    pf8bit: FastSetPixel := FastSetPix8Bits;
    pf15bit, pf16bit: FastSetPixel := FastSetPix15And16Bits;
    pf24bit: FastSetPixel := FastSetPix24Bits;
    pf32bit: FastSetPixel := FastSetPix32Bits;
    else
      begin
        FastSetPixel := FastSetPixDefault;
        Exit; {don't run ScanLines code}
      end; {else}
  end; {case}
  if DoIsTerminated then Exit;
  for I := 0 to (BmpHeight-1) do
    FSPScanLines[I] := GraphBitmap.ScanLine[I];
  if DoIsTerminated then Exit;
  for I := BmpHeight to High(FSPScanLines) do
    FSPScanLines[I] := nil;
end; {procedure InitialCalculations}

{-----------------------------------------}

procedure SetNewPixelColor(TheNewColor : TColor);
var
  TmpWordPixel : word;
begin {procedure SetNewPixelColor}
  FSPCurrLineColor := TheNewColor;
  FSPCurrLineRGB := ColorToRGB(FSPCurrLineColor);
  try
    case GraphPixelFormat of
      pf4bit:
        begin
          FSPPalette4BitsLo := $0F and
            GetNearestPaletteIndex(GraphBitmap.Palette, FSPCurrLineRGB);
          FSPPalette4BitsHi := FSPPalette4BitsLo shl 4;
        end;
      pf8bit: FSPPalette8Bits :=
                GetNearestPaletteIndex(GraphBitmap.Palette, FSPCurrLineRGB);
      pf15bit, pf16bit:
        begin
          TmpWordPixel := PWordCast(FSPScanLines[0])^[0];
          if SetPixelV(FSPPaintDC, 0, 0, FSPCurrLineRGB) then
            FSPPalette15And16Bits := PWordCast(FSPScanLines[0])^[0]
          else
            FSPPalette15And16Bits := 0;
          PWordCast(FSPScanLines[0])^[0] := TmpWordPixel;
        end;
      pf24bit: ExtractRGBFromTColor(TheNewColor, FSPPalette24Red,
                                    FSPPalette24Green, FSPPalette24Blue);
      pf32bit: FSPPalette32Bits := TColorTo32BitPixel(TheNewColor);
    end; {case}
  except
    FastSetPixel := FastSetPixDefault;
  end; {try..except}
end; {procedure SetNewPixelColor}

{-----------------------------------------}

function CalcZValOnLineWithOctant0(const ThePlot0, ThePlot1 : TPlotRec;
                                   TheXVal : integer) : TFixedPoint;
var
  XDist : TMathFloat;
  ZDist : TMathFloat;
  PixXDist : TMathFloat;
begin {function CalcZValOnLineWithOctant0}
  XDist := abs(ThePlot0.XPlot - ThePlot1.XPlot);
  if (XDist = 0) then
    begin
      if (ThePlot0.Z <= ThePlot1.Z) then
        Result := ThePlot1.Z
      else
        Result := ThePlot0.Z;
    end
  else
    begin
      ZDist := abs(ThePlot0.Z - ThePlot1.Z) / FixedMul;
      if (ThePlot0.Z <= ThePlot1.Z) then
        begin
          PixXDist := abs(TheXVal - ThePlot0.XPlot);
          Result := trunc(((PixXDist/XDist)*ZDist)*FixedMul);
          Result := Result + ThePlot0.Z;
        end
      else
        begin
          PixXDist := abs(TheXVal - ThePlot1.XPlot);
          Result := trunc(((PixXDist/XDist)*ZDist)*FixedMul);
          Result := Result + ThePlot1.Z;
        end; {else}
    end; {else}
end; {function CalcZValOnLineWithOctant0}

{-----------------------------------------}

function CalcZValOnLineWithOctant1(const ThePlot0, ThePlot1 : TPlotRec;
                                   TheYVal : integer) : TFixedPoint;
var
  YDist : TMathFloat;
  ZDist : TMathFloat;
  PixYDist : TMathFloat;
begin {function CalcZValOnLineWithOctant1}
  YDist := abs(ThePlot0.YPlot - ThePlot1.YPlot);
  if (YDist = 0) then
    begin
      if (ThePlot0.Z <= ThePlot1.Z) then
        Result := ThePlot1.Z
      else
        Result := ThePlot0.Z;
    end
  else
    begin
      ZDist := abs(ThePlot0.Z - ThePlot1.Z) / FixedMul;
      if (ThePlot0.Z <= ThePlot1.Z) then
        begin
          PixYDist := abs(TheYVal - ThePlot0.YPlot);
          Result := trunc(((PixYDist/YDist)*ZDist)*FixedMul);
          Result := Result + ThePlot0.Z;
        end
      else
        begin
          PixYDist := abs(TheYVal - ThePlot1.YPlot);
          Result := trunc(((PixYDist/YDist)*ZDist)*FixedMul);
          Result := Result + ThePlot1.Z;
        end; {else}
    end; {else}
end; {function CalcZValOnLineWithOctant1}

{-----------------------------------------}

procedure BresenhamLine(X0, Y0, X1, Y1 : integer);
{
X0, Y0;     coordinates of one end of the line
X1, Y1;     coordinates of the other end of the line
}
var
  DeltaX, DeltaY, Temp : integer;
  DeltaYx2, DeltaYx2MinusDeltaXx2 : integer; {Var's used by Octant0}
  DeltaXx2, DeltaXx2MinusDeltaYx2 : integer; {Var's used by Octant1}
  ErrorTerm : integer;

{-----------------------------------}

procedure BitMapDotOct0;
var
  CompLineRecPtr : PLineRec;
  CompPoint1Rec, CompPoint2Rec : TPlotRec;
  LowCompPointZ, HighCompPointZ : TFixedPoint;
  CurrZ, CompZ : TFixedPoint;
  PixOwnerNumber : integer;
begin {procedure BitMapDotOct0}
  PixOwnerNumber := PixOwnerArray[X0, Y0];
  if PixOwnerNumber = NilLineNumber then
    begin
      PixOwnerArray[X0, Y0] := CurrLineNumber;
      FastSetPixel(X0, Y0);
    end
  else
    begin
      CompLineRecPtr := addr(LineObj.LinePtrArray
                             [(PixOwnerNumber div TLineArraySize)]^
                             [(PixOwnerNumber mod TLineArraySize)]);
      CompPoint1Rec := CompLineRecPtr^.P1^.Rotate;
      CompPoint2Rec := CompLineRecPtr^.P2^.Rotate;
      if CompPoint1Rec.Z <= CompPoint2Rec.Z then
        begin
          LowCompPointZ := CompPoint1Rec.Z;
          HighCompPointZ := CompPoint2Rec.Z;
        end
      else
        begin
          LowCompPointZ := CompPoint2Rec.Z;
          HighCompPointZ := CompPoint1Rec.Z;
        end;
      if CurrHighZ < LowCompPointZ then
        Exit {don't draw point}
      else
        if CurrLowZ > HighCompPointZ then
          begin
            PixOwnerArray[X0, Y0] := CurrLineNumber;
            FastSetPixel(X0, Y0);
            Exit;
          end; {if}
      if abs(CompPoint1Rec.XPlot - CompPoint2Rec.XPlot) >=
         abs(CompPoint1Rec.YPlot - CompPoint2Rec.YPlot) then
        CompZ := CalcZValOnLineWithOctant0(CompPoint1Rec, CompPoint2Rec, X0)
      else
        CompZ := CalcZValOnLineWithOctant1(CompPoint1Rec, CompPoint2Rec, Y0);
      {the following line is different depending on line octant}
      CurrZ := CalcZValOnLineWithOctant0(CurrPlot0, CurrPlot1, X0);
      if CurrZ > CompZ then
        begin
          PixOwnerArray[X0, Y0] := CurrLineNumber;
          FastSetPixel(X0, Y0);
        end
      else
        if CurrZ = CompZ then
          begin
            if CurrFuncNum > CompLineRecPtr^.FuncNum then
              begin
                PixOwnerArray[X0, Y0] := CurrLineNumber;
                FastSetPixel(X0, Y0);
              end
          end; {if}
      {GraphBitmap.picture.bitmap.Canvas.Pixels[X0,Y0] := CurrLineColor;}
    end; {else}
end; {procedure BitMapDotOct0}

{-----------------------------------}

procedure BitMapDotOct1;
var
  CompLineRecPtr : PLineRec;
  CompPoint1Rec, CompPoint2Rec : TPlotRec;
  LowCompPointZ, HighCompPointZ : TFixedPoint;
  CurrZ, CompZ : TFixedPoint;
  PixOwnerNumber : integer;
begin {procedure BitMapDotOct1}
  PixOwnerNumber := PixOwnerArray[X0, Y0];
  if PixOwnerNumber = NilLineNumber then
    begin
      PixOwnerArray[X0, Y0] := CurrLineNumber;
      FastSetPixel(X0, Y0);
    end
  else
    begin
      CompLineRecPtr := addr(LineObj.LinePtrArray
                             [(PixOwnerNumber div TLineArraySize)]^
                             [(PixOwnerNumber mod TLineArraySize)]);
      CompPoint1Rec := CompLineRecPtr^.P1^.Rotate;
      CompPoint2Rec := CompLineRecPtr^.P2^.Rotate;
      if CompPoint1Rec.Z <= CompPoint2Rec.Z then
        begin
          LowCompPointZ := CompPoint1Rec.Z;
          HighCompPointZ := CompPoint2Rec.Z;
        end
      else
        begin
          LowCompPointZ := CompPoint2Rec.Z;
          HighCompPointZ := CompPoint1Rec.Z;
        end;
      if CurrHighZ < LowCompPointZ then
        Exit {don't draw point}
      else
        if CurrLowZ > HighCompPointZ then
          begin
            PixOwnerArray[X0, Y0] := CurrLineNumber;
            FastSetPixel(X0, Y0);
            Exit;
          end; {if}
      if abs(CompPoint1Rec.XPlot - CompPoint2Rec.XPlot) >=
         abs(CompPoint1Rec.YPlot - CompPoint2Rec.YPlot) then
        CompZ := CalcZValOnLineWithOctant0(CompPoint1Rec, CompPoint2Rec, X0)
      else
        CompZ := CalcZValOnLineWithOctant1(CompPoint1Rec, CompPoint2Rec, Y0);
      {the following line is different depending on line octant}
      CurrZ := CalcZValOnLineWithOctant1(CurrPlot0, CurrPlot1, Y0);
      if CurrZ > CompZ then
        begin
          PixOwnerArray[X0, Y0] := CurrLineNumber;
          FastSetPixel(X0, Y0);
        end
      else
        if CurrZ = CompZ then
          begin
            if CurrFuncNum > CompLineRecPtr^.FuncNum then
              begin
                PixOwnerArray[X0, Y0] := CurrLineNumber;
                FastSetPixel(X0, Y0);
              end
          end; {if}
      {GraphBitmap.picture.bitmap.Canvas.Pixels[X0,Y0] := CurrLineColor;}
    end; {else}
end; {procedure BitMapDotOct1}

{-----------------------------------}

begin {procedure BresenhamLine}
  { Save half the line-drawing cases by swapping Y0 with Y1
     and X0 with X1 if Y0 is greater than Y1. As a result, DeltaY
     is always > 0, and only the octant 0-3 cases need to be
     handled. }
  if Y0 > Y1 then
    begin
      Temp := Y0;
      Y0 := Y1;
      Y1 := Temp;
      Temp := X0;
      X0 := X1;
      X1 := Temp;
    end; {if}
  { Handle as four separate cases, for the four octants in which
     Y1 is greater than Y0 }
  DeltaX := X1 - X0;    { calculate the length of the line
                          in each coordinate }
  DeltaY := Y1 - Y0;
  if DeltaX > 0 then
    begin
      if DeltaX > DeltaY then
        begin {Octant0(X0, Y0, DeltaX, DeltaY, 1)}
          { Set up initial error term and values used inside drawing loop }
          DeltaYx2 := DeltaY * 2;
          DeltaYx2MinusDeltaXx2 := DeltaYx2 - integer( DeltaX * 2 );
          ErrorTerm := DeltaYx2 - integer(DeltaX);
          { Draw the line }
          BitMapDotOct0;              { draw the first pixel }
          while DeltaX > 0 do
            begin
              dec(DeltaX);
              { See if it's time to advance the Y coordinate }
              if  ErrorTerm >= 0 then
                begin
                  { Advance the Y coordinate & adjust the error term
                     back down }
                  inc(Y0);
                  ErrorTerm := ErrorTerm + DeltaYx2MinusDeltaXx2;
                end
              else
                begin
                  { Add to the error term }
                  ErrorTerm := ErrorTerm + DeltaYx2;
                end; {else}
              inc(X0);                 { advance the X coordinate }
              BitMapDotOct0;           { draw a pixel }
            end; {while}
        end
      else
        begin {Octant1(X0, Y0, DeltaX, DeltaY, 1);}
          { Set up initial error term and values used inside drawing loop }
          DeltaXx2 := DeltaX * 2;
          DeltaXx2MinusDeltaYx2 := DeltaXx2 - integer( DeltaY * 2 );
          ErrorTerm := DeltaXx2 - integer(DeltaY);
          BitMapDotOct1;           { draw the first pixel }
          while DeltaY > 0 do
            begin
               dec(DeltaY);
               { See if it's time to advance the X coordinate }
               if ErrorTerm >= 0 then
                 begin
                   { Advance the X coordinate & adjust the error term
                     back down }
                   inc(X0);
                   ErrorTerm := ErrorTerm + DeltaXx2MinusDeltaYx2;
                 end
               else
                 begin
                   { Add to the error term }
                   ErrorTerm := ErrorTerm + DeltaXx2;
                 end; {else}
              inc(Y0);                   { advance the Y coordinate }
              BitMapDotOct1;        { draw a pixel }
            end; {while}
        end; {else}
    end
  else
    begin
      DeltaX := -DeltaX;             { absolute value of DeltaX }
      if DeltaX > DeltaY then
        begin {Octant0(X0, Y0, DeltaX, DeltaY, -1)}
          { Set up initial error term and values used inside drawing loop }
          DeltaYx2 := DeltaY * 2;
          DeltaYx2MinusDeltaXx2 := DeltaYx2 - integer( DeltaX * 2 );
          ErrorTerm := DeltaYx2 - integer(DeltaX);
          { Draw the line }
          BitMapDotOct0;              { draw the first pixel }
          while DeltaX > 0 do
            begin
              dec(DeltaX);
              { See if it's time to advance the Y coordinate }
              if  ErrorTerm >= 0 then
                begin
                  { Advance the Y coordinate & adjust the error term
                     back down }
                  inc(Y0);
                  ErrorTerm := ErrorTerm + DeltaYx2MinusDeltaXx2;
                end
              else
                begin
                  { Add to the error term }
                  ErrorTerm := ErrorTerm + DeltaYx2;
                end; {else}
              dec(X0);                 { advance the X coordinate }
              BitMapDotOct0;           { draw a pixel }
            end; {while}
        end
      else
        begin {Octant1(X0, Y0, DeltaX, DeltaY, -1);}
          { Set up initial error term and values used inside drawing loop }
          DeltaXx2 := DeltaX * 2;
          DeltaXx2MinusDeltaYx2 := DeltaXx2 - integer( DeltaY * 2 );
          ErrorTerm := DeltaXx2 - integer(DeltaY);
          BitMapDotOct1;           { draw the first pixel }
          while DeltaY > 0 do
            begin
               dec(DeltaY);
               { See if it's time to advance the X coordinate }
               if ErrorTerm >= 0 then
                 begin
                   { Advance the X coordinate & adjust the error term
                     back down }
                   dec(X0);
                   ErrorTerm := ErrorTerm + DeltaXx2MinusDeltaYx2;
                 end
               else
                 begin
                   { Add to the error term }
                   ErrorTerm := ErrorTerm + DeltaXx2;
                 end; {else}
              inc(Y0);                   { advance the Y coordinate }
              BitMapDotOct1;        { draw a pixel }
            end; {while}
        end; {else}
    end; {else}
end; {procedure BresenhamLine}

{-----------------------------------------}

procedure ClipAndDrawCurrLine;
const
  TopLeftX = 0;
  TopLeftY = 0;
type
  TClipRec = record
    X, Y : longint;
  end; {TClipRec}
var
  IsOnBmpP0, IsOnBmpP1 : boolean;
  ClipP0, ClipP1 : TClipRec;

{------------------------}

procedure ClipAginstVert(var ClipedPoint : TClipRec; RectXVal: longint);
var
  MSlope : Extended;
begin
  try
    MSlope := (ClipP0.Y-ClipP1.Y)/(ClipP0.X-ClipP1.X);
    with ClipedPoint do
      begin
        Y := trunc(Y + (RectXVal - X)*(MSlope));
        X := RectXVal;
      end; {with}
  except
    ClipedPoint.Y := 0;
    ClipedPoint.X := RectXVal;
  end; {try..except}
end; {procedure ClipAginstVert}

{------------------------}

procedure ClipAginstHorz(var ClipedPoint : TClipRec; RectYVal: longint);
var
  MSlope : Extended;
begin
  try
    MSlope := (ClipP0.X-ClipP1.X)/(ClipP0.Y-ClipP1.Y);
    with ClipedPoint do
      begin
        X := trunc(X + (RectYVal - Y)*(MSlope));
        Y := RectYVal;
      end; {with}
  except
    ClipedPoint.X := 0;
    ClipedPoint.Y := RectYVal;
  end; {try..except}
end; {procedure ClipAginstHorz}

{------------------------}

begin {procedure ClipAndDrawCurrLine;}
  IsOnBmpP0 := (((CurrPlot0.XPlot >= TopLeftX)and(CurrPlot0.XPlot <= BmpWidthSub1))and
                ((CurrPlot0.YPlot >= TopLeftY)and(CurrPlot0.YPlot <= BmpHeightSub1)));
  IsOnBmpP1 := (((CurrPlot1.XPlot >= TopLeftX)and(CurrPlot1.XPlot <= BmpWidthSub1))and
                ((CurrPlot1.YPlot >= TopLeftY)and(CurrPlot1.YPlot <= BmpHeightSub1)));
  if (IsOnBmpP0 and IsOnBmpP1) then
    begin
      BresenhamLine(CurrPlot0.XPlot,CurrPlot0.YPlot,
                    CurrPlot1.XPlot,CurrPlot1.YPlot);
      exit;
    end
  else
    if ((not IsOnBmpP0) and (not IsOnBmpP1)) then
      begin
        if ((CurrPlot0.XPlot < TopLeftX)and(CurrPlot1.XPlot < TopLeftX))or
           ((CurrPlot0.XPlot > BmpWidthSub1)and(CurrPlot1.XPlot > BmpWidthSub1))or
           ((CurrPlot0.YPlot < TopLeftY)and(CurrPlot1.YPlot < TopLeftY))or
           ((CurrPlot0.YPlot > BmpHeightSub1)and(CurrPlot1.YPlot > BmpHeightSub1))then
          exit; {Line does not cross BMP, do not draw}
      end; {else}
  ClipP0.X := CurrPlot0.XPlot;
  ClipP0.Y := CurrPlot0.YPlot;
  ClipP1.X := CurrPlot1.XPlot;
  ClipP1.Y := CurrPlot1.YPlot;
  if ClipP0.X = ClipP1.X then
    begin
      if ClipP0.Y < ClipP1.Y then
        begin
          if (ClipP1.Y >= TopLeftY)and(ClipP0.Y < TopLeftY) then
            ClipP0.Y := TopLeftY;
          if (ClipP1.Y > BmpHeightSub1)and(ClipP0.Y <= BmpHeightSub1) then
            ClipP1.Y := BmpHeightSub1;
        end
      else
        begin
          if (ClipP0.Y >= TopLeftY)and(ClipP1.Y < TopLeftY) then
            ClipP1.Y := TopLeftY;
          if (ClipP0.Y > BmpHeightSub1)and(ClipP1.Y <= BmpHeightSub1) then
            ClipP0.Y := BmpHeightSub1;
        end; {else}
    end
  else
  if ClipP0.Y = ClipP1.Y then
    begin
      if ClipP0.X < ClipP1.X then
        begin
          if (ClipP1.X >= TopLeftX)and(ClipP0.X < TopLeftX) then
            ClipP0.X := TopLeftX;
          if (ClipP1.X > BmpWidthSub1)and(ClipP0.X <= BmpWidthSub1) then
            ClipP1.X := BmpWidthSub1;
        end
      else
        begin
          if (ClipP0.X >= TopLeftX)and(ClipP1.X < TopLeftX) then
            ClipP1.X := TopLeftX;
          if (ClipP0.X > BmpWidthSub1)and(ClipP1.X <= BmpWidthSub1) then
            ClipP0.X := BmpWidthSub1;
        end; {else}
    end
  else
    begin
      if ClipP0.X < ClipP1.X then
        begin
          if (ClipP1.X >= TopLeftX)and(ClipP0.X < TopLeftX) then
            ClipAginstVert(ClipP0, TopLeftX);
          if (ClipP1.X > BmpWidthSub1)and(ClipP0.X <= BmpWidthSub1) then
            ClipAginstVert(ClipP1, BmpWidthSub1);
        end
      else
        begin
          if (ClipP0.X >= TopLeftX)and(ClipP1.X < TopLeftX) then
            ClipAginstVert(ClipP1, TopLeftX);
          if (ClipP0.X > BmpWidthSub1)and(ClipP1.X <= BmpWidthSub1) then
            ClipAginstVert(ClipP0, BmpWidthSub1);
        end; {else}
      if ClipP0.Y < ClipP1.Y then
        begin
          if (ClipP1.Y >= TopLeftY)and(ClipP0.Y < TopLeftY) then
            ClipAginstHorz(ClipP0, TopLeftY);
          if (ClipP1.Y > BmpHeightSub1)and(ClipP0.Y <= BmpHeightSub1) then
            ClipAginstHorz(ClipP1, BmpHeightSub1);
        end
      else
        begin
          if (ClipP0.Y >= TopLeftY)and(ClipP1.Y < TopLeftY) then
            ClipAginstHorz(ClipP1, TopLeftY);
          if (ClipP0.Y > BmpHeightSub1)and(ClipP1.Y <= BmpHeightSub1) then
            ClipAginstHorz(ClipP0, BmpHeightSub1);
        end; {else}
    end; {else}
  if (((ClipP0.X >= TopLeftX)and(ClipP0.X <= BmpWidthSub1))and
      ((ClipP0.Y >= TopLeftY)and(ClipP0.Y <= BmpHeightSub1)))
    and
     (((ClipP1.X >= TopLeftX)and(ClipP1.X <= BmpWidthSub1))and
      ((ClipP1.Y >= TopLeftY)and(ClipP1.Y <= BmpHeightSub1))) then
    begin
      BresenhamLine(ClipP0.X, ClipP0.Y, ClipP1.X, ClipP1.Y);
    end; {if}
end; {procedure ClipAndDrawCurrLine;}

{-----------------------------------------}

procedure PaintFunctions;
var
  I, LinePtrArrayI : integer;
  NumLeftToSet : integer;
  ITimes : integer;
  CurrLineArray : PLineArray;
  BeginLineIndex : integer;

{------------------}

procedure PaintLine(LineNdx : integer);
begin {procedure PaintLine}
  with CurrLineArray^[LineNdx] do
    begin
      CurrPlot0 := P1^.Rotate;
      CurrPlot1 := P2^.Rotate;
      CurrFuncNum := FuncNum;
    end; {with}
  if FSPCurrLineColor <> LineObj.FuncColorArray[CurrFuncNum] then
    SetNewPixelColor(LineObj.FuncColorArray[CurrFuncNum]);
  if CurrPlot0.Z <= CurrPlot1.Z then
    begin
      CurrLowZ := CurrPlot0.Z;
      CurrHighZ := CurrPlot1.Z;
    end
  else
    begin
      CurrLowZ := CurrPlot1.Z;
      CurrHighZ := CurrPlot0.Z;
    end; {else}
  ClipAndDrawCurrLine;
  inc(CurrLineNumber);
end; {procedure PaintLine}

{------------------}

begin {procedure PaintFunctions;}
  if DoIsTerminated then Exit;
  CurrLineNumber := 0;
  NumLeftToSet := LineObj.LineCount;
  {begin draw the axis lines}
  CurrLineArray := LineObj.LinePtrArray[0];
  if Setting3DObj.XShowAxisLine then
    PaintLine(0);
  if DoIsTerminated then Exit;
  if Setting3DObj.YShowAxisLine then
    PaintLine(1);
  if DoIsTerminated then Exit;
  if Setting3DObj.ZShowAxisLine then
    PaintLine(2);
  if DoIsTerminated then Exit;
  {end draw the axis lines}
  {begin draw the function lines}
  BeginLineIndex := 3;
  for LinePtrArrayI := 0 to (LineObj.LinePtrArrayCount - 1) do
    begin
      if DoIsTerminated then Exit;
      CurrLineArray := LineObj.LinePtrArray[LinePtrArrayI];
      if NumLeftToSet >= TLineArraySize then
        begin
          ITimes := TLineArraySize;
          NumLeftToSet := NumLeftToSet - TLineArraySize;
        end {if}
      else
        ITimes := NumLeftToSet;
      for I := BeginLineIndex to (ITimes - 1) do
      begin
        if DoIsTerminated then Exit;
        PaintLine(I);
      end;
      BeginLineIndex := 0;
    end; {for}
  {end draw the function lines}
end; {procedure PaintFunctions;}

{-----------------------------------------}

procedure DrawAxisLabels;
const
  TopLeftX = 0;
  TopLeftY = 0;
  axPosX = 1;
  axPosY = 2;
  axPosZ = 3;
  axNegX = 4;
  axNegY = 5;
  axNegZ = 6;

type
  TAxisLabelRec = record
    AxisEnd : integer;
    PlotRec,
    OtherPointRec : TPlotRec;
  end; {TAxisRec}
  TAxisLabelArray = array[1..6] of TAxisLabelRec;

type
  TBmpByteArray = array[0..100000] of byte;
  PBmpByteArray = ^TBmpByteArray;

var
  AxisLabelArray : TAxisLabelArray;
  AxisLabelCount : integer;
  CurrLabelI : integer;
  CurrTextWidth, CurrTextHeight : integer;
  WhitePaletteNdx : byte;

{-----------------------}

procedure InitalizeLabelSettings;
var
  CmpI : integer;
  SwapRec : TAxisLabelRec;
  DidSwap : boolean;

{-----------}

procedure SetArrayVals(DrawThisLabel : boolean; TheAxisEnd : integer;
                       const ThePlotRec, TheOtherPointRec : TPlotRec);
begin {procedure SetArrayVals}
  if DrawThisLabel then
    begin
      if (((ThePlotRec.XPlot >= TopLeftX)and(ThePlotRec.XPlot <= BmpWidthSub1))and
          ((ThePlotRec.YPlot >= TopLeftY)and(ThePlotRec.YPlot <= BmpHeightSub1)))then
        begin
          inc(AxisLabelCount);
          with AxisLabelArray[AxisLabelCount] do
            begin
              AxisEnd := TheAxisEnd;
              PlotRec := ThePlotRec;
              OtherPointRec := TheOtherPointRec;
            end; {with}
        end; {if}
    end; {if}
end; {procedure SetArrayVals}

{-----------}

begin {procedure InitalizeLabelSettings;}
  AxisLabelCount := 0;
  {set X Lable Values}
  if Setting3DObj.XShowAxisLine and ((LineObj.LowestX <> 0)or
                                      (LineObj.HighestX <> 0)) then
    begin
      SetArrayVals(Setting3DObj.XLabelPos, axPosX,
                   LineObj.PixPtrArray[0]^[1].Rotate,
                   LineObj.PixPtrArray[0]^[0].Rotate);
      SetArrayVals(Setting3DObj.XLabelNeg, axNegX,
                   LineObj.PixPtrArray[0]^[0].Rotate,
                   LineObj.PixPtrArray[0]^[1].Rotate);
    end; {if}
  {set Y Lable Values}
  if Setting3DObj.YShowAxisLine and ((LineObj.LowestY <> 0)or
                                      (LineObj.HighestY <> 0)) then
    begin
      SetArrayVals(Setting3DObj.YLabelPos, axPosY,
                   LineObj.PixPtrArray[0]^[3].Rotate,
                   LineObj.PixPtrArray[0]^[2].Rotate);
      SetArrayVals(Setting3DObj.YLabelNeg, axNegY,
                   LineObj.PixPtrArray[0]^[2].Rotate,
                   LineObj.PixPtrArray[0]^[3].Rotate);
    end; {if}
  {set Z Lable Values}
  if Setting3DObj.ZShowAxisLine and ((LineObj.LowestZ <> 0)or
                                      (LineObj.HighestZ <> 0)) then
    begin
      SetArrayVals(Setting3DObj.ZLabelPos, axPosZ,
                   LineObj.PixPtrArray[0]^[5].Rotate,
                   LineObj.PixPtrArray[0]^[4].Rotate);
      SetArrayVals(Setting3DObj.ZLabelNeg, axNegZ,
                   LineObj.PixPtrArray[0]^[4].Rotate,
                   LineObj.PixPtrArray[0]^[5].Rotate);
    end; {if}
  if AxisLabelCount > 1 then
    begin
      {sort the array it assinding order, bubble sort}
      DidSwap := true;
      while DidSwap do
        begin
          DidSwap := false;
          for CmpI := 1 to (AxisLabelCount-1) do
            begin
              if (AxisLabelArray[CmpI].PlotRec.Z > AxisLabelArray[CmpI+1].PlotRec.Z)or
                 ((AxisLabelArray[CmpI].PlotRec.Z = AxisLabelArray[CmpI+1].PlotRec.Z)and
                  (AxisLabelArray[CmpI].AxisEnd > AxisLabelArray[CmpI+1].AxisEnd))then
                begin
                  DidSwap := true;
                  SwapRec := AxisLabelArray[CmpI];
                  AxisLabelArray[CmpI] := AxisLabelArray[CmpI+1];
                  AxisLabelArray[CmpI+1] := SwapRec;
                end; {if}
            end;
        end; {while}
    end; {if}
end; {procedure InitalizeLabelSettings;}

{-----------------------}

procedure DrawLabelOnBufferBmp;
var
  TheLabelStr : string[3];
begin {procedure DrawLabelOnBufferBmp}
  TheLabelStr := '';
  with Setting3DObj do
    case AxisLabelArray[CurrLabelI].AxisEnd of
      axPosX: begin
                if XShowPlusSigns then
                  TheLabelStr := '+X'
                else
                  TheLabelStr := 'X';
                AxisLabelBufferBmp.Canvas.Font.Assign(XLabelFont);
              end; {axPosX:}
      axPosY: begin
                if YShowPlusSigns then
                  TheLabelStr := '+Y'
                else
                  TheLabelStr := 'Y';
                AxisLabelBufferBmp.Canvas.Font.Assign(YLabelFont);
              end; {axPosY:}
      axPosZ: begin
                if ZShowPlusSigns then
                  TheLabelStr := '+Z'
                else
                  TheLabelStr := 'Z';
                AxisLabelBufferBmp.Canvas.Font.Assign(ZLabelFont);
              end; {axPosZ:}
      axNegX: begin
                if XShowMinusSigns then
                  TheLabelStr := '-X'
                else
                  TheLabelStr := 'X';
                AxisLabelBufferBmp.Canvas.Font.Assign(XLabelFont);
              end; {axNegX:}
      axNegY: begin
                if YShowMinusSigns then
                  TheLabelStr := '-Y'
                else
                  TheLabelStr := 'Y';
                AxisLabelBufferBmp.Canvas.Font.Assign(YLabelFont);
              end; {axNegY:}
      axNegZ: begin
                if ZShowMinusSigns then
                  TheLabelStr := '-Z'
                else
                  TheLabelStr := 'Z';
                AxisLabelBufferBmp.Canvas.Font.Assign(ZLabelFont);
              end; {axNegZ:}
    end; {case}
  CurrTextWidth := AxisLabelBufferBmp.Canvas.TextWidth(TheLabelStr);
  CurrTextHeight := AxisLabelBufferBmp.Canvas.TextHeight(TheLabelStr);
  if CurrTextWidth > AxisLabelBufferBmp.Width then
    AxisLabelBufferBmp.Width := CurrTextWidth;
  if CurrTextHeight > AxisLabelBufferBmp.Height then
    AxisLabelBufferBmp.Height := CurrTextHeight;
  with AxisLabelBufferBmp.Canvas do
    begin
      SetNewPixelColor(Font.Color);
      Pen.Mode := pmCopy;
      Pen.Style := psSolid;
      Pen.Color := clBlack;
      Pen.Width := 1;
      Brush.Style := bsSolid;
      Brush.Color := clWhite;
      FillRect(Rect(0, 0, AxisLabelBufferBmp.Width, AxisLabelBufferBmp.Height));
      WhitePaletteNdx := PBmpByteArray(AxisLabelBufferBmp.ScanLine[0])^[0];
      Font.Color := clBlack;
      TextOut(0, 0, TheLabelStr);
    end; {with}
end; {procedure DrawLabelOnBufferBmp}

{-----------------------}

procedure DrawBufferLabelOnGraph;
var
  AxisPlotX, AxisPlotY : integer;
  AxisPlotZ : TFixedPoint;
  GraphBmpX, GraphBmpY : integer;
  CurrLabelBmpX, CurrLabelBmpY : integer;
  CurrRowPtr : PBmpByteArray;
  CompLineRecPtr : PLineRec;
  CompPoint1Rec, CompPoint2Rec : TPlotRec;
  LowCompPointZ, HighCompPointZ : TFixedPoint;
  CompZ : TFixedPoint;
  PixOwnerNumber : integer;
begin {procedure DrawBufferLabelOnGraph;}
  with AxisLabelArray[CurrLabelI] do
    begin
      AxisPlotZ := PlotRec.Z;
      if PlotRec.XPlot <= OtherPointRec.XPlot then
        begin
          if PlotRec.YPlot <= OtherPointRec.YPlot then
            begin {corner 3}
              AxisPlotX := PlotRec.XPlot - CurrTextWidth;
              AxisPlotY := PlotRec.YPlot - CurrTextHeight;
            end
          else
            begin {corner 2}
              AxisPlotX := PlotRec.XPlot - CurrTextWidth;
              AxisPlotY := PlotRec.YPlot;
            end; {else}
        end
      else
        begin
          if PlotRec.YPlot <= OtherPointRec.YPlot then
            begin {corner 4}
              AxisPlotX := PlotRec.XPlot;
              AxisPlotY := PlotRec.YPlot - CurrTextHeight;
            end
          else
            begin {corner 1}
              AxisPlotX := PlotRec.XPlot;
              AxisPlotY := PlotRec.YPlot;
            end; {else}
        end; {else}
    end; {with}
  for CurrLabelBmpY := 0 to (CurrTextHeight-1) do
    begin
      CurrRowPtr := AxisLabelBufferBmp.ScanLine[CurrLabelBmpY];
      for CurrLabelBmpX := 0 to (CurrTextWidth-1) do
        begin
          if CurrRowPtr^[CurrLabelBmpX] <> WhitePaletteNdx then
            begin
              GraphBmpX := CurrLabelBmpX + AxisPlotX;
              GraphBmpY := CurrLabelBmpY + AxisPlotY;
              if (((GraphBmpX >= TopLeftX)and(GraphBmpX <= BmpWidthSub1))and
                  ((GraphBmpY >= TopLeftY)and(GraphBmpY <= BmpHeightSub1)))then
                begin
                  PixOwnerNumber := PixOwnerArray[GraphBmpX, GraphBmpY];
                  if PixOwnerNumber = NilLineNumber then
                    FastSetPixel(GraphBmpX, GraphBmpY)
                  else
                    begin
                      CompLineRecPtr := addr(LineObj.LinePtrArray
                                             [(PixOwnerNumber div TLineArraySize)]^
                                             [(PixOwnerNumber mod TLineArraySize)]);
                      CompPoint1Rec := CompLineRecPtr^.P1^.Rotate;
                      CompPoint2Rec := CompLineRecPtr^.P2^.Rotate;
                      if CompPoint1Rec.Z <= CompPoint2Rec.Z then
                        begin
                          LowCompPointZ := CompPoint1Rec.Z;
                          HighCompPointZ := CompPoint2Rec.Z;
                        end
                      else
                        begin
                          LowCompPointZ := CompPoint2Rec.Z;
                          HighCompPointZ := CompPoint1Rec.Z;
                        end;
                      if AxisPlotZ < LowCompPointZ then
                        Continue {don't draw point}
                      else
                        if AxisPlotZ > HighCompPointZ then
                          begin
                            FastSetPixel(GraphBmpX, GraphBmpY);
                            Continue;
                          end; {if}
                      if abs(CompPoint1Rec.XPlot - CompPoint2Rec.XPlot) >=
                         abs(CompPoint1Rec.YPlot - CompPoint2Rec.YPlot) then
                        CompZ := CalcZValOnLineWithOctant0(CompPoint1Rec,
                                   CompPoint2Rec, GraphBmpX)
                      else
                        CompZ := CalcZValOnLineWithOctant1(CompPoint1Rec,
                                   CompPoint2Rec, GraphBmpY);
                      if AxisPlotZ >= CompZ then
                          FastSetPixel(GraphBmpX, GraphBmpY);
                      {GraphBitmap.picture.bitmap.Canvas.Pixels[X0,Y0] := CurrLineColor;}
                    end; {else}
                end; {if}
            end; {if}
        end; {for}
    end; {for}
end; {procedure DrawBufferLabelOnGraph;}

{-----------------------}

begin {procedure DrawAxisLabels;}
  if DoIsTerminated then Exit;
  InitalizeLabelSettings;
  if DoIsTerminated then Exit;
  for CurrLabelI := 1 to AxisLabelCount do
    begin
      if DoIsTerminated then Exit;
      try
        DrawLabelOnBufferBmp;
      except
        Continue;
      end; {try..except}
      if DoIsTerminated then Exit;
      DrawBufferLabelOnGraph;
    end; {for}
end; {procedure DrawAxisLabels;}

{-----------------------------------------}

begin {procedure TPlotGraph3D.GraphWinPaint;}
  if DoIsTerminated then Exit;
  FastSetPixel := FastSetPixDefault; //initalize to a safe default
  if DoIsTerminated then Exit;
  FillGraphBmpBackGround;
  if DoIsTerminated then Exit;
  LineObj.RegisterColor(XAxisFuncNum, Setting3DObj.XAxisColor);
  if DoIsTerminated then Exit;
  LineObj.RegisterColor(YAxisFuncNum, Setting3DObj.YAxisColor);
  if DoIsTerminated then Exit;
  LineObj.RegisterColor(ZAxisFuncNum, Setting3DObj.ZAxisColor);
  if DoIsTerminated then Exit;
  InitPixOwnerArray;
  if DoIsTerminated then Exit;
  InitialCalculations;
  if DoIsTerminated then Exit;
  SetNewPixelColor(clBlack);
  if DoIsTerminated then Exit;
  if LineObj.LineCount > 3 then
    begin
      LineObj.RotateSort(TrigAngles, View3DZoomObj.ViewZoom,
                         HalfBmpHeight, HalfBmpWidth, OnIsTerminated);
      if DoIsTerminated then Exit;
      PaintFunctions;
      if DoIsTerminated then Exit;
      //lock the AxisLabelBufferBmp for threaded rendering bug fix
      AxisLabelBufferBmp.Canvas.Lock;
      try
        DrawAxisLabels;
      finally
        AxisLabelBufferBmp.Canvas.Unlock;
      end;
    end; {if}
end; {procedure TPlotGraph3D.GraphWinPaint;}

{=====================================================================}

procedure TPlotGraph3D.RestartAnimatedGraphPaint;
begin
  GraphWinPaint;
end;

{---------------------------------------------------------------------}

function TPlotGraph3D.StepAnimatedGraphPaint : boolean;
begin
  GraphWinPaint;
  result := true;
end;

{---------------------------------------------------------------------}

end.
