unit PtGraph;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Windows, Classes, Graphics, Jpeg, PtShared, GrGlobal;

type
  TPlotGraph = class
  private
    { Private declarations }
    FGraphBitmap : TBitmap;
    FBackgroundBitmap : TBitmap;
    FBackgroundTileMode : TGraphBackgroundTileMode;
    FBackgroundColor : TColor;
    FBmpHeight : integer;
    FBmpWidth : integer;
    FGraphColorDepth : TPixelFormat;
    FOnIsTerminated : TIsTerminatedEvent;
  protected
    { Protected declarations }
    function DoIsTerminated : boolean;
    procedure FillGraphBmpBackground;
  public
    { Public declarations }
    property GraphBitmap : TBitmap read FGraphBitmap;
    property BackgroundBitmap : TBitmap read FBackgroundBitmap;
    property BmpHeight : integer read FBmpHeight;
    property BmpWidth : integer read FBmpWidth;
    property GraphColorDepth : TPixelFormat read FGraphColorDepth;
    property BackgroundTileMode : TGraphBackgroundTileMode read FBackgroundTileMode
             write FBackgroundTileMode;
    property BackgroundColor : TColor read FBackgroundColor
             write FBackgroundColor;
    property OnIsTerminated : TIsTerminatedEvent read FOnIsTerminated
             write FOnIsTerminated;
    constructor Create;
    destructor Destroy; override;
    procedure InitalizeNew(TheColorDepth : TPixelFormat;
                           TheDefaultBmpWidth, TheDefaultBmpHeight : integer);
    procedure SetBackgroundBitmap(TheNewBackground: TBitmap;
                                  TheGraphColorDepth: TPixelFormat);
    procedure SetBackgroundJpeg(TheNewBackground: TJPEGImage;
                                TheGraphColorDepth: TPixelFormat);
    procedure SetBackgroundNoPicture(TheGraphColorDepth: TPixelFormat);
    procedure SetResizeBitmap(TheNewBmpWidth, TheNewBmpHeight : integer);
    function FullGraphPaint : string; virtual; abstract;
    procedure RestartAnimatedGraphPaint; virtual; abstract;
    //StepAnimatedGraphPaint returns true if the graph is done being drawn
    function StepAnimatedGraphPaint : boolean; virtual; abstract;
  end; {TPlotGraph}

implementation

{ TPlotGraph }

{---------------------------------------------------------------------}

constructor TPlotGraph.Create;
begin {constructor TPlotGraph.Create}
  inherited;
  FBackgroundTileMode := btCenter;
  FBackgroundColor := clWhite;
  FBmpHeight := -1;
  FBmpWidth := -1;
  FBackgroundBitmap := nil;
  FOnIsTerminated := nil;
  FGraphBitmap := TBitmap.Create;
  FGraphBitmap.HandleType := bmDIB;
  FGraphColorDepth := pf24bit;
end; {constructor TPlotGraph.Create}

{---------------------------------------------------------------------}

destructor TPlotGraph.Destroy;
begin {destructor TPlotGraph.Destroy;}
  FGraphBitmap.free;
  if Assigned(FBackgroundBitmap) then
    FBackgroundBitmap.free;
  inherited;
end; {destructor TPlotGraph.Destroy;}

{---------------------------------------------------------------------}

function TPlotGraph.DoIsTerminated: boolean;
begin {function TPlotGraph.DoIsTerminated}
  Result := false;
  if Assigned(FOnIsTerminated) then
    FOnIsTerminated(Result);
end; {function TPlotGraph.DoIsTerminated}

{---------------------------------------------------------------------}

procedure TPlotGraph.FillGraphBmpBackground;

{---------------------------}

procedure LockedFillGraphBmpBackground;
var
  SourceWidth, SourceHeight : integer;
  DestWidth, DestHeight : integer;
  SmallWidth, SmallHeight : integer;
  Source: TRect;
  Dest: TRect;
begin {procedure LockedFillGraphBmpBackground;}
  if DoIsTerminated then Exit;
  Assert(((FBmpHeight > 0) and (FBmpHeight = FGraphBitmap.Height)), 'Invalid Graph Height in procedure TPlotGraph.FillGraphBmpBackground');
  Assert(((FBmpWidth > 0) and (FBmpWidth = FGraphBitmap.Width)), 'Invalid Graph Width in procedure TPlotGraph.FillGraphBmpBackground');
  DestHeight := FBmpHeight;
  DestWidth := FBmpWidth;
  with FGraphBitmap.Canvas do
    begin
      Pen.Mode := pmCopy;
      Pen.Style := psSolid;
      Pen.Color := clBlue;
      Pen.Width := 1;
      Brush.Style := bsSolid;
      Brush.Color := FBackgroundColor;
      FillRect(Rect(0,0,DestWidth,DestHeight)); //old version FillRect(Rect(0,0,DestWidth,DestHeight));
      Brush.Style := bsClear;
    end; {with}
  if DoIsTerminated then Exit;
  if Assigned(FBackgroundBitmap) then
    begin
      SourceHeight := FBackgroundBitmap.Height;
      SourceWidth := FBackgroundBitmap.Width;
      Assert((SourceHeight > 0), 'Invalid Background Height in procedure TPlotGraph.FillGraphBmpBackground');
      Assert((SourceWidth > 0), 'Invalid Background Width in procedure TPlotGraph.FillGraphBmpBackground');
      if SourceHeight <= DestHeight then
        SmallHeight := SourceHeight
      else
        SmallHeight := DestHeight;
      if SourceWidth <= DestWidth then
        SmallWidth := SourceWidth
      else
        SmallWidth := DestWidth;
      FGraphBitmap.Canvas.CopyMode := cmSrcCopy;
      if DoIsTerminated then Exit;
      case FBackgroundTileMode of
        btCenter:
          begin
            if SourceHeight <= DestHeight then
              begin
                Source.Top := 0;
                Source.Bottom := SourceHeight - 1;
                Dest.Top := (DestHeight - SourceHeight) div 2;
                Dest.Bottom := Dest.Top + (SourceHeight - 1);
              end
            else {if SourceHeight > DestHeight then}
              begin
                Dest.Top := 0;
                Dest.Bottom := DestHeight - 1;
                Source.Top := (SourceHeight - DestHeight) div 2;
                Source.Bottom := Source.Top + (DestHeight - 1);
              end; {else}
            if DoIsTerminated then Exit;
            if SourceWidth <= DestWidth then
              begin
                Source.Left := 0;
                Source.Right := SourceWidth - 1;
                Dest.Left := (DestWidth - SourceWidth) div 2;
                Dest.Right := Dest.Left + (SourceWidth - 1);
              end
            else {if SourceWidth > DestWidth then}
              begin
                Dest.Left := 0;
                Dest.Right := DestWidth - 1;
                Source.Left := (SourceWidth - DestWidth) div 2;
                Source.Right := Source.Left + (DestWidth - 1);
              end; {else}
            Assert((Source.Top <= Source.Bottom) and (Source.Top >= 0) and (Source.Top <= (SourceHeight - 1)) and (Source.Bottom >= 0) and (Source.Bottom <= (SourceHeight - 1)) and
                   (Source.Left <= Source.Right) and (Source.Left >= 0) and (Source.Left <= (SourceWidth - 1)) and (Source.Right >= 0) and (Source.Right <= (SourceWidth - 1)),
                   'Error in Source rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert((Dest.Top <= Dest.Bottom) and (Dest.Top >= 0) and (Dest.Top <= (DestHeight - 1)) and (Dest.Bottom >= 0) and (Dest.Bottom <= (DestHeight - 1)) and
                   (Dest.Left <= Dest.Right) and (Dest.Left >= 0) and (Dest.Left <= (DestWidth - 1)) and  (Dest.Right >= 0) and (Dest.Right <= (DestWidth - 1)),
                   'Error in Dest rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert(((Dest.Right - Dest.Left) = (Source.Right - Source.Left)) and ((Dest.Bottom - Dest.Top) = (Source.Bottom - Source.Top)),
                   'Error: Dest/Source rect different sizes in procedure TPlotGraph.FillGraphBmpBackground');
            if DoIsTerminated then Exit;
            FGraphBitmap.Canvas.CopyRect(Dest, FBackgroundBitmap.Canvas, Source);
          end; {btCenter:}
        btStretchToFit:
          begin
            if DoIsTerminated then Exit;
            FGraphBitmap.Canvas.StretchDraw(Rect(0,0,DestWidth-1,DestHeight-1), FBackgroundBitmap);
          end; {btStretchToFit:}
        btTile:
          begin
            //start in the upper left
            //calculate top
            Source.Top := 0;
            Dest.Top := 0;
            //calculate left
            Source.Left := 0;
            Dest.Left := 0;
            //calculate Bottom
            Source.Bottom := SmallHeight - 1;
            Dest.Bottom := SmallHeight - 1;
            //calculate right
            Source.Right := SmallWidth - 1;
            Dest.Right := SmallWidth - 1;
            Assert((Source.Top <= Source.Bottom) and (Source.Top >= 0) and (Source.Top <= (SourceHeight - 1)) and (Source.Bottom >= 0) and (Source.Bottom <= (SourceHeight - 1)) and
                   (Source.Left <= Source.Right) and (Source.Left >= 0) and (Source.Left <= (SourceWidth - 1)) and (Source.Right >= 0) and (Source.Right <= (SourceWidth - 1)),
                   'Error in Source rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert((Dest.Top <= Dest.Bottom) and (Dest.Top >= 0) and (Dest.Top <= (DestHeight - 1)) and (Dest.Bottom >= 0) and (Dest.Bottom <= (DestHeight - 1)) and
                   (Dest.Left <= Dest.Right) and (Dest.Left >= 0) and (Dest.Left <= (DestWidth - 1)) and  (Dest.Right >= 0) and (Dest.Right <= (DestWidth - 1)),
                   'Error in Dest rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert(((Dest.Right - Dest.Left) = (Source.Right - Source.Left)) and ((Dest.Bottom - Dest.Top) = (Source.Bottom - Source.Top)),
                   'Error: Dest/Source rect different sizes in procedure TPlotGraph.FillGraphBmpBackground');
            if DoIsTerminated then Exit;
            FGraphBitmap.Canvas.CopyRect(Dest, FBackgroundBitmap.Canvas, Source);
            if DoIsTerminated then Exit;
            repeat
              while Dest.Right < (DestWidth - 1) do
                begin
                  //move to the right
                  Dest.Left := Dest.Right + 1;
                  Dest.Right := Dest.Left + SourceWidth - 1;
                  if Dest.Right > (DestWidth - 1) then  //clip right destination side
                    Dest.Right := (DestWidth - 1);
                  //calculate source rect
                  Source.Right := abs(Dest.Right - Dest.Left);
                  Source.Bottom := abs(Dest.Bottom - Dest.Top);
                  Assert((Source.Top <= Source.Bottom) and (Source.Top >= 0) and (Source.Top <= (SourceHeight - 1)) and (Source.Bottom >= 0) and (Source.Bottom <= (SourceHeight - 1)) and
                         (Source.Left <= Source.Right) and (Source.Left >= 0) and (Source.Left <= (SourceWidth - 1)) and (Source.Right >= 0) and (Source.Right <= (SourceWidth - 1)),
                         'Error in Source rect in procedure TPlotGraph.FillGraphBmpBackground');
                  Assert((Dest.Top <= Dest.Bottom) and (Dest.Top >= 0) and (Dest.Top <= (DestHeight - 1)) and (Dest.Bottom >= 0) and (Dest.Bottom <= (DestHeight - 1)) and
                         (Dest.Left <= Dest.Right) and (Dest.Left >= 0) and (Dest.Left <= (DestWidth - 1)) and  (Dest.Right >= 0) and (Dest.Right <= (DestWidth - 1)),
                         'Error in Dest rect in procedure TPlotGraph.FillGraphBmpBackground');
                  Assert(((Dest.Right - Dest.Left) = (Source.Right - Source.Left)) and ((Dest.Bottom - Dest.Top) = (Source.Bottom - Source.Top)),
                         'Error: Dest/Source rect different sizes in procedure TPlotGraph.FillGraphBmpBackground');
                  if DoIsTerminated then Exit;
                  FGraphBitmap.Canvas.CopyRect(Dest, FBackgroundBitmap.Canvas, Source);
                  if DoIsTerminated then Exit;
                end; {while}
              if DoIsTerminated then Exit;
              Dest.Top := Dest.Bottom + 1;
              if Dest.Top <= (DestHeight - 1) then
                begin
                  Dest.Bottom := Dest.Top + SourceHeight - 1;
                  Dest.Left := 0;
                  Dest.Right := SmallWidth - 1;
                  //calculate source rect
                  Source.Right := abs(Dest.Right - Dest.Left);
                  Source.Bottom := abs(Dest.Bottom - Dest.Top);
                  Assert((Source.Top <= Source.Bottom) and (Source.Top >= 0) and (Source.Top <= (SourceHeight - 1)) and (Source.Bottom >= 0) and (Source.Bottom <= (SourceHeight - 1)) and
                         (Source.Left <= Source.Right) and (Source.Left >= 0) and (Source.Left <= (SourceWidth - 1)) and (Source.Right >= 0) and (Source.Right <= (SourceWidth - 1)),
                         'Error in Source rect in procedure TPlotGraph.FillGraphBmpBackground');
                  Assert((Dest.Top <= Dest.Bottom) and (Dest.Top >= 0) and (Dest.Top <= (DestHeight - 1)) and (Dest.Bottom >= 0) and (Dest.Bottom <= (DestHeight - 1)) and
                         (Dest.Left <= Dest.Right) and (Dest.Left >= 0) and (Dest.Left <= (DestWidth - 1)) and  (Dest.Right >= 0) and (Dest.Right <= (DestWidth - 1)),
                         'Error in Dest rect in procedure TPlotGraph.FillGraphBmpBackground');
                  Assert(((Dest.Right - Dest.Left) = (Source.Right - Source.Left)) and ((Dest.Bottom - Dest.Top) = (Source.Bottom - Source.Top)),
                         'Error: Dest/Source rect different sizes in procedure TPlotGraph.FillGraphBmpBackground');
                  if DoIsTerminated then Exit;
                  FGraphBitmap.Canvas.CopyRect(Dest, FBackgroundBitmap.Canvas, Source);
                end; {if}
              if DoIsTerminated then Exit;
            until Dest.Top > (DestHeight - 1);
          end; {btTile:}
        btUpperLeft:
          begin
            //calculate top
            Source.Top := 0;
            Dest.Top := 0;
            //calculate left
            Source.Left := 0;
            Dest.Left := 0;
            //calculate Bottom
            Source.Bottom := SmallHeight - 1;
            Dest.Bottom := SmallHeight - 1;
            //calculate right
            Source.Right := SmallWidth - 1;
            Dest.Right := SmallWidth - 1;
            Assert((Source.Top <= Source.Bottom) and (Source.Top >= 0) and (Source.Top <= (SourceHeight - 1)) and (Source.Bottom >= 0) and (Source.Bottom <= (SourceHeight - 1)) and
                   (Source.Left <= Source.Right) and (Source.Left >= 0) and (Source.Left <= (SourceWidth - 1)) and (Source.Right >= 0) and (Source.Right <= (SourceWidth - 1)),
                   'Error in Source rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert((Dest.Top <= Dest.Bottom) and (Dest.Top >= 0) and (Dest.Top <= (DestHeight - 1)) and (Dest.Bottom >= 0) and (Dest.Bottom <= (DestHeight - 1)) and
                   (Dest.Left <= Dest.Right) and (Dest.Left >= 0) and (Dest.Left <= (DestWidth - 1)) and  (Dest.Right >= 0) and (Dest.Right <= (DestWidth - 1)),
                   'Error in Dest rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert(((Dest.Right - Dest.Left) = (Source.Right - Source.Left)) and ((Dest.Bottom - Dest.Top) = (Source.Bottom - Source.Top)),
                   'Error: Dest/Source rect different sizes in procedure TPlotGraph.FillGraphBmpBackground');
            if DoIsTerminated then Exit;
            FGraphBitmap.Canvas.CopyRect(Dest, FBackgroundBitmap.Canvas, Source);
          end; {btUpperLeft:}
        btUpperRight:
          begin
            //calculate top
            Source.Top := 0;
            Dest.Top := 0;
            //calculate right
            Source.Right := SourceWidth - 1;
            Dest.Right := DestWidth - 1;
            //calculate bottom
            Source.Bottom := SmallHeight - 1;
            Dest.Bottom := SmallHeight - 1;
            //calculate left
            Source.Left := Abs(SourceWidth - SmallWidth);
            Dest.Left := Abs(DestWidth - SmallWidth);
            Assert((Source.Top <= Source.Bottom) and (Source.Top >= 0) and (Source.Top <= (SourceHeight - 1)) and (Source.Bottom >= 0) and (Source.Bottom <= (SourceHeight - 1)) and
                   (Source.Left <= Source.Right) and (Source.Left >= 0) and (Source.Left <= (SourceWidth - 1)) and (Source.Right >= 0) and (Source.Right <= (SourceWidth - 1)),
                   'Error in Source rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert((Dest.Top <= Dest.Bottom) and (Dest.Top >= 0) and (Dest.Top <= (DestHeight - 1)) and (Dest.Bottom >= 0) and (Dest.Bottom <= (DestHeight - 1)) and
                   (Dest.Left <= Dest.Right) and (Dest.Left >= 0) and (Dest.Left <= (DestWidth - 1)) and  (Dest.Right >= 0) and (Dest.Right <= (DestWidth - 1)),
                   'Error in Dest rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert(((Dest.Right - Dest.Left) = (Source.Right - Source.Left)) and ((Dest.Bottom - Dest.Top) = (Source.Bottom - Source.Top)),
                   'Error: Dest/Source rect different sizes in procedure TPlotGraph.FillGraphBmpBackground');
            if DoIsTerminated then Exit;
            FGraphBitmap.Canvas.CopyRect(Dest, FBackgroundBitmap.Canvas, Source);
          end; {btUpperRight:}
        LowerLeft:
          begin
            //calculate bottom
            Source.Bottom := SourceHeight - 1;
            Dest.Bottom := DestHeight - 1;
            //calculate left
            Source.Left := 0;
            Dest.Left := 0;
            //calculate top
            Source.Top := Abs(SourceHeight - SmallHeight);
            Dest.Top := Abs(DestHeight - SmallHeight);
            //calculate right
            Source.Right := SmallWidth - 1;
            Dest.Right := SmallWidth - 1;
            Assert((Source.Top <= Source.Bottom) and (Source.Top >= 0) and (Source.Top <= (SourceHeight - 1)) and (Source.Bottom >= 0) and (Source.Bottom <= (SourceHeight - 1)) and
                   (Source.Left <= Source.Right) and (Source.Left >= 0) and (Source.Left <= (SourceWidth - 1)) and (Source.Right >= 0) and (Source.Right <= (SourceWidth - 1)),
                   'Error in Source rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert((Dest.Top <= Dest.Bottom) and (Dest.Top >= 0) and (Dest.Top <= (DestHeight - 1)) and (Dest.Bottom >= 0) and (Dest.Bottom <= (DestHeight - 1)) and
                   (Dest.Left <= Dest.Right) and (Dest.Left >= 0) and (Dest.Left <= (DestWidth - 1)) and  (Dest.Right >= 0) and (Dest.Right <= (DestWidth - 1)),
                   'Error in Dest rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert(((Dest.Right - Dest.Left) = (Source.Right - Source.Left)) and ((Dest.Bottom - Dest.Top) = (Source.Bottom - Source.Top)),
                   'Error: Dest/Source rect different sizes in procedure TPlotGraph.FillGraphBmpBackground');
            if DoIsTerminated then Exit;
            FGraphBitmap.Canvas.CopyRect(Dest, FBackgroundBitmap.Canvas, Source);
          end; {LowerLeft:}
        LowerRight:
          begin
            //calculate bottom
            Source.Bottom := SourceHeight - 1;
            Dest.Bottom := DestHeight - 1;
            //calculate right
            Source.right := SourceWidth - 1;
            Dest.right := DestWidth - 1;
            //calculate top
            Source.Top := Abs(SourceHeight - SmallHeight);
            Dest.Top := Abs(DestHeight - SmallHeight);
            //calculate left
            Source.Left := Abs(SourceWidth - SmallWidth);
            Dest.Left := Abs(DestWidth - SmallWidth);
            Assert((Source.Top <= Source.Bottom) and (Source.Top >= 0) and (Source.Top <= (SourceHeight - 1)) and (Source.Bottom >= 0) and (Source.Bottom <= (SourceHeight - 1)) and
                   (Source.Left <= Source.Right) and (Source.Left >= 0) and (Source.Left <= (SourceWidth - 1)) and (Source.Right >= 0) and (Source.Right <= (SourceWidth - 1)),
                   'Error in Source rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert((Dest.Top <= Dest.Bottom) and (Dest.Top >= 0) and (Dest.Top <= (DestHeight - 1)) and (Dest.Bottom >= 0) and (Dest.Bottom <= (DestHeight - 1)) and
                   (Dest.Left <= Dest.Right) and (Dest.Left >= 0) and (Dest.Left <= (DestWidth - 1)) and  (Dest.Right >= 0) and (Dest.Right <= (DestWidth - 1)),
                   'Error in Dest rect in procedure TPlotGraph.FillGraphBmpBackground');
            Assert(((Dest.Right - Dest.Left) = (Source.Right - Source.Left)) and ((Dest.Bottom - Dest.Top) = (Source.Bottom - Source.Top)),
                   'Error: Dest/Source rect different sizes in procedure TPlotGraph.FillGraphBmpBackground');
            if DoIsTerminated then Exit;
            FGraphBitmap.Canvas.CopyRect(Dest, FBackgroundBitmap.Canvas, Source);
          end; {LowerRight:}
      end; {case}
    end; {if}
end; {procedure LockedFillGraphBmpBackground;}

{---------------------------}

begin {procedure TPlotGraph.FillGraphBmpBackground;}
  if Assigned(FGraphBitmap) then FGraphBitmap.Canvas.Lock;
  try
    if Assigned(FBackgroundBitmap) then FBackgroundBitmap.Canvas.Lock;
    try
      LockedFillGraphBmpBackground; //do the work now that the canvases are locked
    finally
      if Assigned(FBackgroundBitmap) then FBackgroundBitmap.Canvas.Unlock;
    end;
  finally
    if Assigned(FGraphBitmap) then FGraphBitmap.Canvas.Unlock;
  end;
end; {procedure TPlotGraph.FillGraphBmpBackground;}

{---------------------------------------------------------------------}

procedure TPlotGraph.InitalizeNew(TheColorDepth: TPixelFormat;
  TheDefaultBmpWidth, TheDefaultBmpHeight: integer);
begin {procedure TPlotGraph.InitalizeNew}
  Assert((TheColorDepth in SupportedBmpColorDepths), 'Invalid picture color depth');
  SetBackgroundNoPicture(TheColorDepth);
  SetResizeBitmap(TheDefaultBmpWidth, TheDefaultBmpHeight);
end; {procedure TPlotGraph.InitalizeNew}

{---------------------------------------------------------------------}

procedure TPlotGraph.SetBackgroundBitmap(TheNewBackground: TBitmap;
  TheGraphColorDepth: TPixelFormat);
begin {procedure TPlotGraph.SetBackgroundBitmap}
  Assert(Assigned(TheNewBackground), 'Nil pointer in set bitmap background');
  Assert((TheGraphColorDepth in SupportedBmpColorDepths), 'Invalid picture color depth');
  if not Assigned(FBackgroundBitmap) then
    begin
      FBackgroundBitmap := TBitmap.Create;
    end; {if}
  FBackgroundBitmap.Canvas.Lock;
  try
    FGraphBitmap.Canvas.Lock;
    try
      TheNewBackground.Canvas.Lock;
      try
        FBackgroundBitmap.HandleType := bmDIB;
        FBackgroundBitmap.PixelFormat := TheGraphColorDepth;
        FGraphBitmap.PixelFormat := TheGraphColorDepth;
        FGraphColorDepth := TheGraphColorDepth;
        FBackgroundBitmap.Height := TheNewBackground.Height;
        FBackgroundBitmap.Width := TheNewBackground.Width;
        DoCopyBitmapBuffers(FBackgroundBitmap, TheNewBackground);
      finally
        TheNewBackground.Canvas.Unlock;
      end; {try..finally}
    finally
      FGraphBitmap.Canvas.Unlock;
    end; {try..finally}
  finally
    FBackgroundBitmap.Canvas.Unlock;
  end; {try..finally}
end; {procedure TPlotGraph.SetBackgroundBitmap}

{---------------------------------------------------------------------}

procedure TPlotGraph.SetBackgroundJpeg(TheNewBackground: TJPEGImage;
  TheGraphColorDepth: TPixelFormat);
var
  TmpBG : TBitmap;
begin {procedure TPlotGraph.SetBackgroundJpeg}
  Assert(Assigned(TheNewBackground), 'Nil pointer in set jpeg background');
  Assert((TheGraphColorDepth in SupportedBmpColorDepths), 'Invalid picture color depth');
  if not Assigned(FBackgroundBitmap) then
    begin
      FBackgroundBitmap := TBitmap.Create;
    end; {if}
  FBackgroundBitmap.Canvas.Lock;
  try
    FGraphBitmap.Canvas.Lock;
    try
      FBackgroundBitmap.HandleType := bmDIB;
      FBackgroundBitmap.PixelFormat := TheGraphColorDepth;
      FGraphBitmap.PixelFormat := TheGraphColorDepth;
      FGraphColorDepth := TheGraphColorDepth;
      FBackgroundBitmap.Height := TheNewBackground.Height;
      FBackgroundBitmap.Width := TheNewBackground.Width;
      TmpBG := TBitmap.Create;
      TmpBG.Canvas.Lock;
      try
        TmpBG.HandleType := bmDIB;
        TmpBG.PixelFormat := pf24bit;
        TmpBG.Assign(TheNewBackground);
        DoCopyBitmapBuffers(FBackgroundBitmap, TmpBG);
      finally
        TmpBG.Canvas.Unlock;
        TmpBG.free;
      end; {try..finally}
    finally
      FGraphBitmap.Canvas.Unlock;
    end; {try..finally}
  finally
    FBackgroundBitmap.Canvas.Unlock;
  end; {try..finally}
end; {procedure TPlotGraph.SetBackgroundJpeg}

{---------------------------------------------------------------------}

procedure TPlotGraph.SetBackgroundNoPicture(
  TheGraphColorDepth: TPixelFormat);
begin {procedure TPlotGraph.SetBackgroundNoPicture}
  Assert((TheGraphColorDepth in SupportedBmpColorDepths), 'Invalid picture color depth');
  if Assigned(FBackgroundBitmap) then
    FBackgroundBitmap.free;
  FBackgroundBitmap := nil;
  FGraphBitmap.Canvas.Lock;
  try
    FGraphBitmap.PixelFormat := TheGraphColorDepth;
    FGraphColorDepth := TheGraphColorDepth;
  finally
    FGraphBitmap.Canvas.Unlock;
  end;
end; {procedure TPlotGraph.SetBackgroundNoPicture}

{---------------------------------------------------------------------}

procedure TPlotGraph.SetResizeBitmap(TheNewBmpWidth,
  TheNewBmpHeight: integer);
begin {procedure TPlotGraph.SetResizeBitmap}
  Assert((TheNewBmpWidth <= MaxBmpWidth), 'Bitmap width to large');
  Assert((TheNewBmpHeight <= MaxBmpHeight), 'Bitmap height to large');
  Assert((TheNewBmpWidth >= MinBmpWidth), 'Bitmap width to small');
  Assert((TheNewBmpHeight >= MinBmpHeight), 'Bitmap height to small');
  FGraphBitmap.Canvas.Lock;
  try
    try
      FGraphBitmap.Width := TheNewBmpWidth;
      FGraphBitmap.Height := TheNewBmpHeight;
    finally
      FBmpWidth := FGraphBitmap.Width;
      FBmpHeight := FGraphBitmap.Height;
    end; {try..finally}
  finally
    FGraphBitmap.Canvas.Unlock;
  end; {try..finally}
end; {procedure TPlotGraph.SetResizeBitmap}

{---------------------------------------------------------------------}

end.
