unit PtLabel;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  SyncObjs, Classes, Graphics, PtShared, SeqList, GrDraw, DrawObjs, GrGlobal;

type
  TPlotLabelsBuffer = class
  private
    { Private declarations }
    FHiddenLabelIndex : integer;
    FLabelList : TList;
    FBitmap : TBitmap;
    FBitmapAttribRec : TBitmapAttribRec;
  protected
    { Protected declarations }
  public
    { Public declarations }
    property Bitmap : TBitmap read FBitmap;
    property BitmapAttribRec : TBitmapAttribRec read FBitmapAttribRec;
    constructor Create;
    destructor Destroy; override;
    procedure InitalizeNew(TheColorDepth : TPixelFormat;
                           TheDefaultBmpWidth, TheDefaultBmpHeight : integer);
    procedure SetColorDepth(TheColorDepth : TPixelFormat);
    procedure SetSize(TheBmpWidth, TheBmpHeight : integer);
    procedure SetBackgroundColor(TheBackgroundColor : TColor);
    procedure AddNewLabel(TheNewLabel : TDrawBaseClass);
    procedure ClearLabelList;
    procedure AssignLabelList(SourceLabelList : TList; HiddenLabelIndex : integer);
    //label drawing method, do not call from main thread
    procedure PaintDrawLables(IsTerminatedEvent : TIsTerminatedEvent);
  end; {TPlotLabelsBuffer}

implementation

{ TPlotLabelsBuffer }

{---------------------------------------------------------------------}

procedure TPlotLabelsBuffer.AddNewLabel(TheNewLabel: TDrawBaseClass);
begin {procedure TPlotLabelsBuffer.AddNewLabel}
  FLabelList.Add(TheNewLabel);
end; {procedure TPlotLabelsBuffer.AddNewLabel}

{---------------------------------------------------------------------}

procedure TPlotLabelsBuffer.AssignLabelList(SourceLabelList: TList; HiddenLabelIndex : integer);
begin {procedure TPlotLabelsBuffer.AssignLabelList}
  AssignSequenceLists(FLabelList, SourceLabelList);
  if HiddenLabelIndex <> lbHiddenLabelDoNotChange then //do not change
    FHiddenLabelIndex := HiddenLabelIndex;
end; {procedure TPlotLabelsBuffer.AssignLabelList}

{---------------------------------------------------------------------}

procedure TPlotLabelsBuffer.ClearLabelList;
var
  I : integer;
begin {procedure TPlotLabelsBuffer.ClearLabelList;}
  if FLabelList.Count >= 1 then
    for I := 0 to FLabelList.Count -1 do
      if Assigned(FLabelList[I]) then
        TObject(FLabelList[I]).Free;
  FLabelList.Clear;
end; {procedure TPlotLabelsBuffer.ClearLabelList;}

{---------------------------------------------------------------------}

constructor TPlotLabelsBuffer.Create;
begin {constructor TPlotLabelsBuffer.Create;}
  inherited;
  FHiddenLabelIndex := lbHiddenLabelNone;
  FLabelList := TList.Create;
  FBitmap := TBitmap.Create;
  FBitmap.HandleType := bmDIB;
  FBitmapAttribRec.BmpHeight := -1;
  FBitmapAttribRec.BmpWidth := -1;
  FBitmapAttribRec.BackgroundColor := clWhite;
end; {constructor TPlotLabelsBuffer.Create;}

{---------------------------------------------------------------------}

destructor TPlotLabelsBuffer.Destroy;
begin {destructor TPlotLabelsBuffer.Destroy;}
  ClearLabelList;
  FLabelList.free;
  FBitmap.free;
  inherited;
end; {destructor TPlotLabelsBuffer.Destroy;}

{---------------------------------------------------------------------}

procedure TPlotLabelsBuffer.InitalizeNew(TheColorDepth: TPixelFormat;
  TheDefaultBmpWidth, TheDefaultBmpHeight: integer);
begin {procedure TPlotLabelsBuffer.InitalizeNew}
  SetColorDepth(TheColorDepth);
  SetSize(TheDefaultBmpWidth, TheDefaultBmpHeight);
end; {procedure TPlotLabelsBuffer.InitalizeNew}

{---------------------------------------------------------------------}

procedure TPlotLabelsBuffer.PaintDrawLables(IsTerminatedEvent: TIsTerminatedEvent);
{Returns true if all labels were drawn}
var
  I : integer;
  IsTermResult : boolean;
begin {procedure TPlotLabelsBuffer.PaintDrawLables}
  Assert(Assigned(IsTerminatedEvent), 'IsTerminatedEvent = nil in function TPlotLabelsBuffer.PaintDrawLables');
  IsTerminatedEvent(IsTermResult);
  if not IsTermResult then
    begin
      for I := 0 to FLabelList.Count - 1 do
        begin
          IsTerminatedEvent(IsTermResult);
          if IsTermResult then
            Break
          else
            begin
              if I <> FHiddenLabelIndex then
                TDrawBaseClass(FLabelList[I]).Draw(FBitmap, FBitmapAttribRec);
            end; {else}
        end; {for}
    end; {if}
end; {procedure TPlotLabelsBuffer.PaintDrawLables}

{---------------------------------------------------------------------}

procedure TPlotLabelsBuffer.SetBackgroundColor(TheBackgroundColor: TColor);
begin {procedure TPlotLabelsBuffer.SetBackgroundColor}
  FBitmapAttribRec.BackgroundColor := TheBackgroundColor;
end; {procedure TPlotLabelsBuffer.SetBackgroundColor}

{---------------------------------------------------------------------}

procedure TPlotLabelsBuffer.SetColorDepth(TheColorDepth: TPixelFormat);
begin {procedure TPlotLabelsBuffer.SetColorDepth}
  FBitmap.PixelFormat := TheColorDepth;
end; {procedure TPlotLabelsBuffer.SetColorDepth}

{---------------------------------------------------------------------}

procedure TPlotLabelsBuffer.SetSize(TheBmpWidth, TheBmpHeight: integer);
begin {procedure TPlotLabelsBuffer.SetSize}
  Assert((TheBmpWidth <= MaxBmpWidth), 'Bitmap width to large');
  Assert((TheBmpHeight <= MaxBmpHeight), 'Bitmap height to large');
  Assert((TheBmpWidth >= MinBmpWidth), 'Bitmap width to small');
  Assert((TheBmpHeight >= MinBmpHeight), 'Bitmap height to small');
  try
    FBitmap.SetSize(TheBmpWidth, TheBmpHeight);
  finally
    FBitmapAttribRec.BmpWidth := FBitmap.Width;
    FBitmapAttribRec.BmpHeight := FBitmap.Height;
  end; {try..finally}
end; {procedure TPlotLabelsBuffer.SetSize}

{---------------------------------------------------------------------}

end.
