unit PtTh2d;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Classes, Graphics, Sysutils, GrGlobal, PtThList, Vw2dObj, Set2dObj, funcObj, func2d,
  funcpara, PtGr2d, FileSave, CentrDlg, dialogs;

const
  Max2DFuncCount = PtGr2d.Max2DFuncCount;

type
  TPlotThread2D = class(TPlotThreadFuncList)
  private
    { Private declarations }
    FView2DZoomObj : TView2DZoomObj;
    FSetting2DObj : TSetting2DObj;
  protected
    { Protected declarations }
    function IsColorAlreadyInUse(TheColor : TColor) : boolean; override;
    procedure DoAssignFieldObjects; override;
    function WriteFileToDisk(const FString : string) : boolean; override;
    function GetPlotTypeString : string; override;
  public
    { Public declarations }
    property View2DZoomObj : TView2DZoomObj read FView2DZoomObj;
    property Setting2DObj : TSetting2DObj read FSetting2DObj;
    constructor Create;
    destructor Destroy; override;
    function IsFreeFuncListObjSlot : boolean; override;
    function InitalizeFromFile(TheColorDepth : TPixelFormat;
                               const TheFileName : string;
                               ReadGraphFile : TReadGraphFile) : boolean; override;
  end; {TPlotThread2D}

implementation

const
  f2D_FileType_2DCartesian = 1;
  f2D_FileType_Parametric = 2;

{ TPlotThread2D }

constructor TPlotThread2D.Create;
begin
  inherited;
  PlotGraph := TPlotGraph2D.Create;
  FView2DZoomObj := TView2DZoomObj.Create;
  FSetting2DObj := TSetting2DObj.Create;
end;

destructor TPlotThread2D.Destroy;
begin
  PlotGraph.free;
  FView2DZoomObj.free;
  FSetting2DObj.free;
  inherited;
end;

procedure TPlotThread2D.DoAssignFieldObjects;
begin
  if FSetting2DObj.DidFieldsChange and
     (SettingBmpAttribsObj.BackgroundColor <> FSetting2DObj.BackColor) then
    begin
      //special fix up for backward file compatibility
      SettingBmpAttribsObj.BackgroundColor := FSetting2DObj.BackColor;
      SettingBmpAttribsObj.DidBackGroundColorChange := true;
    end; {if}
  if FView2DZoomObj.DidFieldsChange then
    begin
      TPlotGraph2D(PlotGraph).View2DZoomObj.Assign(FView2DZoomObj);
      FView2DZoomObj.DidFieldsChange := false;
    end; {if}
  if FSetting2DObj.DidFieldsChange then
    begin
      TPlotGraph2D(PlotGraph).Setting2DObj.Assign(FSetting2DObj);
      FSetting2DObj.DidFieldsChange := false;
    end; {if}
  inherited;
end;

function TPlotThread2D.GetPlotTypeString: string;
begin
  Result := '2D Cartesian';
end;

function TPlotThread2D.InitalizeFromFile(TheColorDepth: TPixelFormat;
  const TheFileName : string; ReadGraphFile: TReadGraphFile): boolean;

const
  ErCalcSpecValuesString = 'An error occured while calculating 2D function special values.';

{-------------------------------}

procedure ReadVer1Fields;
var
  I : integer;
  SaveFuncCount : integer;
  TmpFunc2DObj : TFunc2DObj;
begin {procedure ReadVer1Fields;}
  with ReadGraphFile do
    begin
      ReadSettingBmpAttribs(ReadGraphFile);
      if View2DZoomObj.ReadFromDisk(ReadGraphFile) then
        FileMsgViewCenterChanged;
      Setting2DObj.ReadFromDisk(ReadGraphFile);
      {begin Read objects owned by FuncObjList : TList;}
      SaveFuncCount := ReadSmallint;
      for I := 1 to SaveFuncCount do
        begin
          TmpFunc2DObj := TFunc2DObj.Create;
          try
            TmpFunc2DObj.ReadFromDisk(ReadGraphFile);
            FuncList.Add(TmpFunc2DObj);
          except
            TmpFunc2DObj.Free;
            raise;
          end; {try..except}
        end; {for}
      {end Read objects owned by FuncObjList : TList;}
      ReadAncestorDataFromDisk(ReadGraphFile);
    end; {with}
end; {procedure ReadVer1Fields;}

{-------------------------------}

procedure ReadVer2Fields;
var
  I : integer;
  SaveFuncCount : integer;
  TmpFunc2DObj : TFunc2DObj;
begin {procedure ReadVer2Fields;}
  with ReadGraphFile do
    begin
      ReadSettingBmpAttribs(ReadGraphFile);
      if View2DZoomObj.ReadFromDisk(ReadGraphFile) then
        FileMsgViewCenterChanged;
      Setting2DObj.ReadFromDisk(ReadGraphFile);
      {begin Read objects owned by FuncObjList : TList;}
      SaveFuncCount := ReadLongint;
      for I := 1 to SaveFuncCount do
        begin
          TmpFunc2DObj := TFunc2DObj.Create;
          try
            TmpFunc2DObj.ReadFromDisk(ReadGraphFile);
            FuncList.Add(TmpFunc2DObj);
          except
            TmpFunc2DObj.Free;
            raise;
          end; {try..except}
        end; {for}
      {end Read objects owned by FuncObjList : TList;}
      ReadAncestorDataFromDisk(ReadGraphFile);
    end; {with}
end; {procedure ReadVer2Fields;}

{-------------------------------}

procedure ReadVer3Fields;
var
  I : integer;
  SaveFuncCount : integer;
  CurrFuncType : longint;
  TmpBaseFuncObj : TBaseFuncObj;
  DidShowErCalcSpecValues : boolean;
begin {procedure ReadVer3Fields;}
  DidShowErCalcSpecValues := false;
  with ReadGraphFile do
    begin
      ReadSettingBmpAttribs(ReadGraphFile);
      if View2DZoomObj.ReadFromDisk(ReadGraphFile) then
        FileMsgViewCenterChanged;
      Setting2DObj.ReadFromDisk(ReadGraphFile);
      {begin Read objects owned by FuncObjList : TList;}
      SaveFuncCount := ReadLongint;
      for I := 1 to SaveFuncCount do
        begin
          CurrFuncType := ReadLongint;
          TmpBaseFuncObj := nil;
          case CurrFuncType of
            f2D_FileType_Parametric: TmpBaseFuncObj := TFuncParametricObj.Create;
            f2D_FileType_2DCartesian: TmpBaseFuncObj := TFunc2DObj.Create;
          end; {case}
          try
            if TmpBaseFuncObj is TFuncParametricObj then
              (TmpBaseFuncObj as TFuncParametricObj).ReadFromDisk(ReadGraphFile)
            else
              if not (TmpBaseFuncObj as TFunc2DObj).ReadFromDisk(ReadGraphFile) then
                if not DidShowErCalcSpecValues then
                  begin
                    DidShowErCalcSpecValues := true;
                    MessageDlgCenter(ErCalcSpecValuesString, mtError, [mbOK], 0, nil);
                  end; {if}
            FuncList.Add(TmpBaseFuncObj);
          except
            TmpBaseFuncObj.Free;
            raise;
          end; {try..except}
        end; {for}
      {end Read objects owned by FuncObjList : TList;}
      ReadAncestorDataFromDisk(ReadGraphFile);
    end; {with}
end; {procedure ReadVer3Fields;}

{-------------------------------}

procedure ReadVer4Fields;
var
  I : integer;
  SaveFuncCount : integer;
  CurrFuncType : longint;
  TmpBaseFuncObj : TBaseFuncObj;
  DidShowErCalcSpecValues : boolean;
begin {procedure ReadVer4Fields;}
  DidShowErCalcSpecValues := false;
  with ReadGraphFile do
    begin
      ReadSettingBmpAttribs(ReadGraphFile);
      ReadSettingGrBGPict(ReadGraphFile);      
      if View2DZoomObj.ReadFromDisk(ReadGraphFile) then
        FileMsgViewCenterChanged;
      Setting2DObj.ReadFromDisk(ReadGraphFile);
      {begin Read objects owned by FuncObjList : TList;}
      SaveFuncCount := ReadLongint;
      for I := 1 to SaveFuncCount do
        begin
          CurrFuncType := ReadLongint;
          TmpBaseFuncObj := nil;
          case CurrFuncType of
            f2D_FileType_Parametric: TmpBaseFuncObj := TFuncParametricObj.Create;
            f2D_FileType_2DCartesian: TmpBaseFuncObj := TFunc2DObj.Create;
          end; {case}
          try
            if TmpBaseFuncObj is TFuncParametricObj then
              (TmpBaseFuncObj as TFuncParametricObj).ReadFromDisk(ReadGraphFile)
            else
              if not (TmpBaseFuncObj as TFunc2DObj).ReadFromDisk(ReadGraphFile) then
                if not DidShowErCalcSpecValues then
                  begin
                    DidShowErCalcSpecValues := true;
                    MessageDlgCenter(ErCalcSpecValuesString, mtError, [mbOK], 0, nil);
                  end; {if}
            FuncList.Add(TmpBaseFuncObj);
          except
            TmpBaseFuncObj.Free;
            raise;
          end; {try..except}
        end; {for}
      {end Read objects owned by FuncObjList : TList;}
      ReadAncestorDataFromDisk(ReadGraphFile);
    end; {with}
end; {procedure ReadVer4Fields;}

{-------------------------------}

begin {function TPlotThread2D.InitalizeFromFile}
  Result := true;
  try
    if ReadGraphFile.FileVersion = 1 then
      ReadVer1Fields
    else
      if ReadGraphFile.FileVersion = 2 then
        ReadVer2Fields
      else
        if ReadGraphFile.FileVersion = 3 then
          ReadVer3Fields
        else
          ReadVer4Fields;
    ReadGraphFile.CloseDataFile;
  except
    Result := false;
    ReadGraphFile.CloseDataFile;
  end; {try..except}
  if Result then
    begin
      SettingBmpAttribsObj.BackgroundColor := Setting2DObj.BackColor;
      SettingBmpAttribsObj.BmpColorDepth := TheColorDepth;
      FFileName := TheFileName;
    end; {if}
end; {function TPlotThread2D.InitalizeFromFile}

function TPlotThread2D.IsColorAlreadyInUse(TheColor: TColor): boolean;
var
  I : integer;
begin
  Result := (TheColor = FSetting2DObj.BackColor);
  if not Result then
    begin
      for I := 0 to (FuncList.Count-1) do
        begin
          Assert(Assigned(FuncList[I]), 'nil in FuncList in function function function TPlotThread2D.IsColorAlreadyInUse');
          if (TObject(FuncList[I]) is TFunc2DObj) then
            if (TFunc2DObj(FuncList[I]).FuncColor = TheColor) then
              begin
                Result := true;
                exit;
              end; {if}
          if (TObject(FuncList[I]) is TFuncParametricObj) then
            if (TFuncParametricObj(FuncList[I]).FuncColor = TheColor) then
              begin
                Result := true;
                exit;
              end; {if}
        end; {for}
    end; {if}
end;

function TPlotThread2D.IsFreeFuncListObjSlot: boolean;
begin
  Result := FuncList.Count < Max2DFuncCount;
end;

function TPlotThread2D.WriteFileToDisk(const FString: string): boolean;
var
  I : integer;
  SaveFuncCount : integer;
  WriteGraphFile : TWriteGraphFile;
  TmpBaseFuncObj : TBaseFuncObj;
begin {function TPlotThread2D.WriteFileToDisk}
  Result := false;
  WriteGraphFile := TWriteGraphFile.Create;
  if not Assigned(WriteGraphFile) then
    Exit;
  try
    with WriteGraphFile do
      if RewriteSaveFile(FString, grType2D, DecimalSeparator) then
        begin
          SaveSettingBmpAttribs(WriteGraphFile);
          SaveSettingGrBGPict(WriteGraphFile);
          View2DZoomObj.SaveToDisk(WriteGraphFile);
          Setting2DObj.SaveToDisk(WriteGraphFile);
          {begin Save objects owned by FuncObjList : TList;}
          SaveFuncCount := FuncList.Count;
          WriteLongint(SaveFuncCount);
          for I := 1 to SaveFuncCount do
            begin
              TmpBaseFuncObj := FuncList[I-1];
              if TmpBaseFuncObj is TFunc2DObj then
                begin
                  WriteLongint(f2D_FileType_2DCartesian);
                  (TmpBaseFuncObj as TFunc2DObj).SaveToDisk(WriteGraphFile);
                end
              else
                if TmpBaseFuncObj is TFuncParametricObj then
                  begin
                    WriteLongint(f2D_FileType_Parametric);
                    (TmpBaseFuncObj as TFuncParametricObj).SaveToDisk(WriteGraphFile);
                  end; {else}
            end; {for}
          {end Save objects owned by FuncObjList : TList;}
          SaveAncestorDataToDisk(WriteGraphFile);
          Result := CloseValidFile;
        end; {if}
  except
    Result := False;
  end; {try..except}
  WriteGraphFile.Free;
end; {function TPlotThread2D.WriteFileToDisk}

end.
