unit PtTh3d;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  SysUtils, Classes, Graphics, GrGlobal, PtThList, Vw3dObj, Set3dObj, FileSave,
  funcObj, func3d, PtGr3d, Fg3dLine;

const
  Max3DFuncCount = PtGr3d.Max3DFuncCount;

type
  TPlotThread3D = class(TPlotThreadFuncList)
  private
    { Private declarations }
    FView3DZoomObj : TView3DZoomObj;
    FSetting3DObj : TSetting3DObj;
  protected
    { Protected declarations }
    function IsColorAlreadyInUse(TheColor : TColor) : boolean; override;
    procedure DoAssignFieldObjects; override;
    function WriteFileToDisk(const FString : string) : boolean; override;
    function GetPlotTypeString : string; override;
  public
    { Public declarations }
    property View3DZoomObj : TView3DZoomObj read FView3DZoomObj;
    property Setting3DObj : TSetting3DObj read FSetting3DObj;
    constructor Create;
    destructor Destroy; override;
    function IsFreeFuncListObjSlot : boolean; override;
    function InitalizeFromFile(TheColorDepth : TPixelFormat;
                               const TheFileName : string;
                               ReadGraphFile : TReadGraphFile) : boolean; override;
  end; {TPlotThread3D}

implementation

{ TPlotThread3D }

constructor TPlotThread3D.Create;
begin
  inherited;
  PlotGraph := TPlotGraph3D.Create;
  FView3DZoomObj := TView3DZoomObj.Create;
  FSetting3DObj := TSetting3DObj.Create;
end;

destructor TPlotThread3D.Destroy;
begin
  PlotGraph.free;
  FView3DZoomObj.free;
  FSetting3DObj.free;
  inherited;
end;

procedure TPlotThread3D.DoAssignFieldObjects;
begin
  if FSetting3DObj.DidFieldsChange and
     (SettingBmpAttribsObj.BackgroundColor <> FSetting3DObj.BackgroundColor) then
    begin
      //special fix up for backward file compatibility
      SettingBmpAttribsObj.BackgroundColor := FSetting3DObj.BackgroundColor;
      SettingBmpAttribsObj.DidBackGroundColorChange := true;
    end; {if}
  if FView3DZoomObj.DidFieldsChange then
    begin
      TPlotGraph3D(PlotGraph).View3DZoomObj.Assign(FView3DZoomObj);
      FView3DZoomObj.DidFieldsChange := false;
    end; {if}
  if FSetting3DObj.DidFieldsChange then
    begin
      TPlotGraph3D(PlotGraph).Setting3DObj.Assign(FSetting3DObj);
      FSetting3DObj.DidFieldsChange := false;
    end; {if}
  inherited;
end;

function TPlotThread3D.GetPlotTypeString: string;
begin
  Result := '3D Cartesian';
end;

function TPlotThread3D.InitalizeFromFile(TheColorDepth: TPixelFormat;
  const TheFileName : string; ReadGraphFile: TReadGraphFile): boolean;

{-------------------------------}

procedure ReadVer1Fields;
type
  TVer1_Func3DTypeEnum = (Ver1_fteRotFunc, Ver1_fte3DFunc);
var
  Ver1_FuncType : TVer1_Func3DTypeEnum;
  I : integer;
  SaveFuncCount : integer;
  TmpFunc : TBaseFunc3DObj;
begin {procedure ReadVer1Fields;}
  with ReadGraphFile do
    begin
      ReadSettingBmpAttribs(ReadGraphFile);
      View3DZoomObj.ReadFromDisk(ReadGraphFile);
      Setting3DObj.ReadFromDisk(ReadGraphFile);
      {begin Read objects owned by FuncObjList : TList;}
      SaveFuncCount := ReadSmallint;
      for I := 1 to SaveFuncCount do
        begin
          ReadRecord(Ver1_FuncType, sizeof(Ver1_FuncType));
          TmpFunc := nil;
          case Ver1_FuncType of
            Ver1_fteRotFunc: TmpFunc := TFuncRotateObj.Create;
            Ver1_fte3DFunc: TmpFunc := TFunc3DObj.Create;
          end; {case}
          try
            TmpFunc.ReadFromDisk(ReadGraphFile);
            FuncList.Add(TmpFunc);
          except
            TmpFunc.Free;
            raise;
          end; {try..except}
        end; {for}
      {end Read objects owned by FuncObjList : TList;}
      ReadAncestorDataFromDisk(ReadGraphFile);
    end; {with}
end; {procedure ReadVer1Fields;}

{-------------------------------}

procedure ReadVer2_3Fields;
var
  I : integer;
  SaveFuncCount : integer;
  FuncType : longint;
  TmpFunc : TBaseFunc3DObj;
begin {procedure ReadVer2_3Fields;}
  with ReadGraphFile do
    begin
      ReadSettingBmpAttribs(ReadGraphFile);
      View3DZoomObj.ReadFromDisk(ReadGraphFile);
      Setting3DObj.ReadFromDisk(ReadGraphFile);
      {begin Read objects owned by FuncObjList : TList;}
      SaveFuncCount := ReadLongint;
      for I := 1 to SaveFuncCount do
        begin
          FuncType := ReadLongint;
          TmpFunc := nil;
          case FuncType of
            f3D_FileType_Rotate: TmpFunc := TFuncRotateObj.Create;
            f3D_FileType_3D: TmpFunc := TFunc3DObj.Create;
          end; {case}
          try
            TmpFunc.ReadFromDisk(ReadGraphFile);
            FuncList.Add(TmpFunc);
          except
            TmpFunc.Free;
            raise;
          end; {try..except}
        end; {for}
      {end Read objects owned by FuncObjList : TList;}
      ReadAncestorDataFromDisk(ReadGraphFile);
    end; {with}
end; {procedure ReadVer2_3Fields;}

{-------------------------------}

procedure ReadVer4Fields;
var
  I : integer;
  SaveFuncCount : integer;
  FuncType : longint;
  TmpFunc : TBaseFunc3DObj;
begin {procedure ReadVer4Fields;}
  with ReadGraphFile do
    begin
      ReadSettingBmpAttribs(ReadGraphFile);
      ReadSettingGrBGPict(ReadGraphFile);
      View3DZoomObj.ReadFromDisk(ReadGraphFile);
      Setting3DObj.ReadFromDisk(ReadGraphFile);
      {begin Read objects owned by FuncObjList : TList;}
      SaveFuncCount := ReadLongint;
      for I := 1 to SaveFuncCount do
        begin
          FuncType := ReadLongint;
          TmpFunc := nil;
          case FuncType of
            f3D_FileType_Rotate: TmpFunc := TFuncRotateObj.Create;
            f3D_FileType_3D: TmpFunc := TFunc3DObj.Create;
          end; {case}
          try
            TmpFunc.ReadFromDisk(ReadGraphFile);
            FuncList.Add(TmpFunc);
          except
            TmpFunc.Free;
            raise;
          end; {try..except}
        end; {for}
      {end Read objects owned by FuncObjList : TList;}
      ReadAncestorDataFromDisk(ReadGraphFile);
    end; {with}
end; {procedure ReadVer4Fields;}

{-------------------------------}

begin {function TPlotThread3D.InitalizeFromFile}
  Result := true;
  try
    if ReadGraphFile.FileVersion = 1 then
      ReadVer1Fields
    else
      if ReadGraphFile.FileVersion <= 3 then
        ReadVer2_3Fields
      else
        ReadVer4Fields;
    ReadGraphFile.CloseDataFile;
  except
    Result := false;
    ReadGraphFile.CloseDataFile;
  end; {try..except}
  if Result then
    begin
      SettingBmpAttribsObj.BackgroundColor := Setting3DObj.BackgroundColor;
      SettingBmpAttribsObj.BmpColorDepth := TheColorDepth;
      FFileName := TheFileName;
    end;
end;

function TPlotThread3D.IsColorAlreadyInUse(TheColor: TColor): boolean;
var
  I : integer;
begin
  Result := (TheColor = FSetting3DObj.BackGroundColor);
  if not Result then
    begin
      for I := 0 to (FuncList.Count-1) do
        begin
          Assert(Assigned(FuncList[I]), 'nil in FuncList in function TPlotThread3D.IsColorAlreadyInUse');
          if (TObject(FuncList[I]) is TFuncRotateObj) then
            if TFuncRotateObj(FuncList[I]).FuncColor = TheColor then
              begin
                Result := true;
                exit;
              end; {if}
          if (TObject(FuncList[I]) is TFunc3DObj) then
            if TFunc3DObj(FuncList[I]).FuncColor = TheColor then
              begin
                Result := true;
                exit;
              end; {if}
        end; {if}
    end; {if}
end;

function TPlotThread3D.IsFreeFuncListObjSlot: boolean;
begin
  Result := FuncList.Count < Max3DFuncCount;
end;

function TPlotThread3D.WriteFileToDisk(const FString: string): boolean;
var
  I : integer;
  SaveFuncCount : integer;
  WriteGraphFile : TWriteGraphFile;
begin
  Result := false;
  WriteGraphFile := TWriteGraphFile.Create;
  if not Assigned(WriteGraphFile) then
    Exit;
  try
    with WriteGraphFile do
      if RewriteSaveFile(FString, grType3D, DecimalSeparator) then
        begin
          SaveSettingBmpAttribs(WriteGraphFile);
          SaveSettingGrBGPict(WriteGraphFile);
          View3DZoomObj.SaveToDisk(WriteGraphFile);
          Setting3DObj.SaveToDisk(WriteGraphFile);
          {begin Save objects owned by FuncObjList : TList;}
          SaveFuncCount := FuncList.Count;
          WriteLongint(SaveFuncCount);
          for I := 1 to SaveFuncCount do
            (TObject(FuncList[I-1]) as TBaseFunc3DObj).SaveToDisk(WriteGraphFile);
          {end Save objects owned by FuncObjList : TList;}
          SaveAncestorDataToDisk(WriteGraphFile);
          Result := CloseValidFile;
        end; {if}
  except
    Result := false;
  end; {try..except}
  WriteGraphFile.Free;
end;

end.
