unit PtThGraf;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Sysutils, Windows, Classes, Graphics, SyncObjs, PtShared, GrGlobal, PtBuffer,
  PtThread, PtGraph, SetGraph, FileSave;

type
  TGraphDrawingThread = class(TThread)
  private
    { Private declarations }
    FThreadErrorString : string;
    CriticalSection : TCriticalSection;
    EvHandle : THandle;
    FAnimatedPlotting : boolean; //access only threw Get and Set methods for thread safety
    FAnimatedRestart : boolean; //access only threw Get and Set methods for thread safety
    FAnimationDelay : integer; //access only threw Get and Set methods for thread safety
    FOnDoneGraphPloting : TNotifyEvent;
    FOnAnimatedGraphPlotingStep : TNotifyEvent;
    procedure IsTerminated(var ShouldTerminate : boolean);
    procedure DoDoneGraphPlotingEvent;
    procedure DoAnimatedGraphPlotingStepEvent;
  protected
    { Protected declarations }
    procedure Execute; override;
  public
    { Public declarations }
    PlotGraph : TPlotGraph;
    OutputBmpBuffer : TSafeBmpBuffer;
    property ThreadErrorString : string read FThreadErrorString;
    property OnDoneGraphPloting : TNotifyEvent read FOnDoneGraphPloting write FOnDoneGraphPloting;
    property OnAnimatedGraphPlotingStep : TNotifyEvent read FOnAnimatedGraphPlotingStep write FOnAnimatedGraphPlotingStep;
    constructor Create;
    destructor Destroy; override;
    procedure SignalAndTerminate;
    procedure SetAnimatedPlotting(TheAnimatedPlotting : boolean);
    function GetAnimatedPlotting : boolean;
    procedure SetAnimatedRestart(TheAnimatedRestart : boolean);
    function GetAnimatedRestart : boolean;
    procedure SetAnimationDelay(TheAnimationDelay : integer);
    function GetAnimationDelay : integer;
  end; {TGraphDrawingThread}

  TPlotThreadGraph = class(TPlotThread)
  private
    { Private declarations }
    FAnimatedPlotting : boolean;
    FAnimatedRestart : boolean;
    FAnimationDelay : integer;
    FSettingGrBGPictObj : TSettingGrBGPictObj;
    GraphThread : TGraphDrawingThread;
    procedure HandleDoneGraphPlotingEvent(Sender: TObject);
    procedure HandleAnimatedGraphPlotingStepEvent(Sender: TObject);
    procedure SetAnimatedPlotting(TheAnimatedPlotting : boolean);
    procedure SetAnimatedRestart(TheAnimatedRestart : boolean);
    procedure SetAnimationDelay(TheAnimationDelay : integer);
  protected
    { Protected declarations }
    PlotGraph : TPlotGraph; //must be created and destroyed in descendant class
    function GetIsAThreadRunning: boolean; override;
    function GetIsGraphThreadRunning : boolean; override;
    procedure SaveSettingGrBGPict(WriteGraphFile : TWriteGraphFile);
    procedure ReadSettingGrBGPict(ReadGraphFile : TReadGraphFile);
    procedure DoAssignFieldObjects; override; //this method is not thread safe
  public
    { Public declarations }
    property AnimatedPlotting : boolean read FAnimatedPlotting write SetAnimatedPlotting;
    property AnimatedRestart : boolean read FAnimatedRestart write SetAnimatedRestart;
    property AnimationDelay : integer read FAnimationDelay write SetAnimationDelay; //in milliseconds
    property SettingGrBGPictObj : TSettingGrBGPictObj read FSettingGrBGPictObj;
    constructor Create;
    destructor Destroy; override;
    procedure InitalizeNew(TheColorDepth : TPixelFormat;
                           TheDefaultBmpWidth, TheDefaultBmpHeight : integer); override;
    procedure ReStartGraphThread; override;
    procedure TerminateGraphThread; override;
    procedure WaitForTermGraphThread; override;
  end; {TPlotThreadGraph}

implementation

{---------------------------------------------------------------------}

{ TPlotThreadGraph }

constructor TPlotThreadGraph.Create;
begin {constructor TPlotThreadGraph.Create}
  inherited Create;
  FAnimatedPlotting := false;
  FAnimatedRestart := false;
  FAnimationDelay := 500;
  GraphThread := nil;
  FSettingGrBGPictObj := TSettingGrBGPictObj.Create;
end; {constructor TPlotThreadGraph.Create}

{---------------------------------------------------------------------}

destructor TPlotThreadGraph.Destroy;
var
  LocalTmpGraphThread : TGraphDrawingThread;
begin {destructor TPlotThreadGraph.Destroy;}
  if Assigned(GraphThread) then
    begin
      LocalTmpGraphThread := GraphThread;
      GraphThread := nil;
      LocalTmpGraphThread.SignalAndTerminate;
      LocalTmpGraphThread.WaitFor;
      LocalTmpGraphThread.free;
    end; {if}
  FSettingGrBGPictObj.free;
  inherited;
end; {destructor TPlotThreadGraph.Destroy;}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.DoAssignFieldObjects;
begin {procedure TPlotThreadGraph.DoAssignFieldObjects;}
  inherited;
  if FSettingGrBGPictObj.DidTileModeChange then
    PlotGraph.BackgroundTileMode := FSettingGrBGPictObj.BackgroundTileMode;
  if SettingBmpAttribsObj.DidBackGroundColorChange then
    PlotGraph.BackgroundColor := SettingBmpAttribsObj.BackgroundColor;
  if SettingBmpAttribsObj.DidWidthHeightChange then
    PlotGraph.SetResizeBitmap(SettingBmpAttribsObj.BmpWidth,
                              SettingBmpAttribsObj.BmpHeight);
  if SettingBmpAttribsObj.DidColorDepthChange or FSettingGrBGPictObj.DidPictureChange then
    begin
      if Assigned(FSettingGrBGPictObj.BackgroundBitmap) then
        PlotGraph.SetBackgroundBitmap(FSettingGrBGPictObj.BackgroundBitmap,
                                      SettingBmpAttribsObj.BmpColorDepth)
      else
        if Assigned(FSettingGrBGPictObj.BackgroundJPEG) then
          PlotGraph.SetBackgroundJPEG(FSettingGrBGPictObj.BackgroundJPEG,
                                      SettingBmpAttribsObj.BmpColorDepth)
        else
          PlotGraph.SetBackgroundNoPicture(SettingBmpAttribsObj.BmpColorDepth);
    end; {if}
  SettingBmpAttribsObj.SetAllDidChange(false);
  FSettingGrBGPictObj.SetAllDidChange(false);
end; {procedure TPlotThreadGraph.DoAssignFieldObjects;}

{---------------------------------------------------------------------}

function TPlotThreadGraph.GetIsAThreadRunning: boolean;
begin {function TPlotThreadGraph.GetIsAThreadRunning}
  Result := inherited GetIsAThreadRunning;
  Result := Result or IsGraphThreadRunning;
end; {function TPlotThreadGraph.GetIsAThreadRunning}

{---------------------------------------------------------------------}

function TPlotThreadGraph.GetIsGraphThreadRunning : boolean;
begin {function TPlotThreadGraph.GetIsGraphThreadRunning}
  Result := Assigned(GraphThread);
end; {function TPlotThreadGraph.GetIsGraphThreadRunning}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.HandleDoneGraphPlotingEvent(Sender: TObject);
var
  ThreadErrorString : string;
begin {procedure TPlotThreadGraph.HandleDoneGraphPlotingEvent}
  if Assigned(GraphThread) then
  begin
    ThreadErrorString := GraphThread.ThreadErrorString;
    UniqueString(ThreadErrorString);
    GraphThread.FreeOnTerminate := true;
    GraphThread.Terminate;
    GraphThread := nil; //thread is done and will destroy itself so loose the refrence
    if ThreadErrorString = '' then
      RunAgainAfterTermLabelThread //thread finished successfully
    else
      DoPlotErrorEvent(ThreadErrorString); //thread finished with an error
    DoRunningStartEndEvent;
  end; {if}
end; {procedure TPlotThreadGraph.HandleDoneGraphPlotingEvent}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.HandleAnimatedGraphPlotingStepEvent(Sender: TObject);
begin {procedure TPlotThreadGraph.HandleAnimatedGraphPlotingEvent}
  RunAgainAfterTermLabelThread;
end; {procedure TPlotThreadGraph.HandleAnimatedGraphPlotingEvent}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.InitalizeNew(TheColorDepth: TPixelFormat;
  TheDefaultBmpWidth, TheDefaultBmpHeight: integer);
begin {procedure TPlotThreadGraph.InitalizeNew}
  inherited;
  PlotGraph.InitalizeNew(TheColorDepth, TheDefaultBmpWidth, TheDefaultBmpHeight);
end; {procedure TPlotThreadGraph.InitalizeNew}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.ReadSettingGrBGPict(
  ReadGraphFile: TReadGraphFile);
begin {procedure TPlotThreadGraph.ReadSettingGrBGPict}
  FSettingGrBGPictObj.ReadFromDisk(ReadGraphFile);
end; {procedure TPlotThreadGraph.ReadSettingGrBGPict}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.ReStartGraphThread;
var
  LocalTmpGraphThread : TGraphDrawingThread;
  ThreadErrorString : string;
begin {procedure TPlotThreadGraph.ReStartGraphThread;}
  if Assigned(GraphThread) then
    begin
      LocalTmpGraphThread := GraphThread;
      GraphThread := nil;
      LocalTmpGraphThread.SignalAndTerminate;
      LocalTmpGraphThread.WaitFor;
      ThreadErrorString := LocalTmpGraphThread.ThreadErrorString;
      UniqueString(ThreadErrorString);
      LocalTmpGraphThread.free;
      if ThreadErrorString <> '' then
        DoPlotErrorEvent(ThreadErrorString);
    end; {if}
  GraphThread := TGraphDrawingThread.Create;
  GraphThread.SetAnimatedPlotting(FAnimatedPlotting);
  GraphThread.SetAnimationDelay(FAnimationDelay);
  GraphThread.OnDoneGraphPloting := HandleDoneGraphPlotingEvent;
  GraphThread.OnAnimatedGraphPlotingStep := HandleAnimatedGraphPlotingStepEvent;
  GraphThread.PlotGraph := PlotGraph;
  GraphThread.OutputBmpBuffer := GraphOutLabelInBitmapBuffer;
  GraphThread.Resume;
  DoRunningStartEndEvent;
end; {procedure TPlotThreadGraph.ReStartGraphThread;}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.SaveSettingGrBGPict(
  WriteGraphFile: TWriteGraphFile);
begin {procedure TPlotThreadGraph.SaveSettingGrBGPict}
  FSettingGrBGPictObj.SaveToDisk(WriteGraphFile);
end; {procedure TPlotThreadGraph.SaveSettingGrBGPict}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.SetAnimatedPlotting(
  TheAnimatedPlotting: boolean);
begin {procedure TPlotThreadGraph.SetAnimatedPlotting}
  FAnimatedPlotting := TheAnimatedPlotting;
  if Assigned(GraphThread) then
    GraphThread.SetAnimatedPlotting(FAnimatedPlotting);
end; {procedure TPlotThreadGraph.SetAnimatedPlotting}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.SetAnimatedRestart(
  TheAnimatedRestart: boolean);
begin {procedure TPlotThreadGraph.SetAnimatedRestart}
  FAnimatedRestart := TheAnimatedRestart;
  if Assigned(GraphThread) then
    GraphThread.SetAnimatedRestart(FAnimatedRestart);
end; {procedure TPlotThreadGraph.SetAnimatedRestart}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.SetAnimationDelay(TheAnimationDelay: integer);
begin {procedure TPlotThreadGraph.SetAnimationDelay}
  FAnimationDelay := TheAnimationDelay;
  if Assigned(GraphThread) then
    GraphThread.SetAnimationDelay(FAnimationDelay);
end; {procedure TPlotThreadGraph.SetAnimationDelay}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.TerminateGraphThread;
begin {procedure TPlotThreadGraph.TerminateGraphThread;}
  if Assigned(GraphThread) then
    GraphThread.SignalAndTerminate;
end; {procedure TPlotThreadGraph.TerminateGraphThread;}

{---------------------------------------------------------------------}

procedure TPlotThreadGraph.WaitForTermGraphThread;
var
  LocalTmpGraphThread : TGraphDrawingThread;
  ThreadErrorString : string;
begin {procedure TPlotThreadGraph.WaitForTermGraphThread;}
  if Assigned(GraphThread) then
    begin
      LocalTmpGraphThread := GraphThread;
      GraphThread := nil;
      LocalTmpGraphThread.SignalAndTerminate;
      LocalTmpGraphThread.WaitFor;
      ThreadErrorString := LocalTmpGraphThread.ThreadErrorString;
      UniqueString(ThreadErrorString);
      LocalTmpGraphThread.free;
      if ThreadErrorString <> '' then
        DoPlotErrorEvent(ThreadErrorString);
    end; {if}
  DoRunningStartEndEvent;
end; {procedure TPlotThreadGraph.WaitForTermGraphThread;}

{---------------------------------------------------------------------}

{ TGraphDrawingThread }

constructor TGraphDrawingThread.Create;
begin {constructor TGraphDrawingThread.Create;}
  inherited Create(True);
  FreeOnTerminate := false;
  FThreadErrorString := '';
  CriticalSection := TCriticalSection.Create;
  EvHandle := CreateEvent(nil, false, false, nil);
  FAnimatedPlotting := false;
  FAnimatedRestart := false;
  FAnimationDelay := 500; //in miliseconds
  PlotGraph := nil;
  OutputBmpBuffer := nil;
  FOnDoneGraphPloting := nil;
  FOnAnimatedGraphPlotingStep := nil;
end; {constructor TGraphDrawingThread.Create;}

{---------------------------------------------------------------------}

destructor TGraphDrawingThread.Destroy;
begin {destructor TGraphDrawingThread.Destroy;}
  CloseHandle(EvHandle);
  CriticalSection.free;
  FThreadErrorString := '';
  inherited;
end; {destructor TGraphDrawingThread.Destroy;}

{---------------------------------------------------------------------}

procedure TGraphDrawingThread.DoDoneGraphPlotingEvent;
begin {procedure TGraphDrawingThread.DoDoneGraphPlotingEvent;}
  try
    if Assigned(FOnDoneGraphPloting) then
      FOnDoneGraphPloting(self);
  except  //stop all exceptions. an exception is very unlikely here.
  end;
end; {procedure TGraphDrawingThread.DoDoneGraphPlotingEvent;}

{---------------------------------------------------------------------}

procedure TGraphDrawingThread.DoAnimatedGraphPlotingStepEvent;
begin {procedure TGraphDrawingThread.DoAnimatedGraphPlotingStepEvent;}
  if Assigned(FOnAnimatedGraphPlotingStep) then
    FOnAnimatedGraphPlotingStep(self);
end; {procedure TGraphDrawingThread.DoAnimatedGraphPlotingStepEvent;}

{---------------------------------------------------------------------}

procedure TGraphDrawingThread.Execute;

{---------------------}

procedure RunAnimatedPlotting;
var
  DoneAnimatedPlotting : boolean;
begin {procedure RunAnimatedPlotting;}
  DoneAnimatedPlotting := false;
  if Terminated then Exit;
  PlotGraph.GraphBitmap.Canvas.Lock;
  try
    PlotGraph.RestartAnimatedGraphPaint;
    if not Terminated then
      OutputBmpBuffer.SetBufferBitmap(PlotGraph.GraphBitmap);
  finally
    PlotGraph.GraphBitmap.Canvas.Unlock;
  end;
  if Terminated then Exit;
  Synchronize(DoAnimatedGraphPlotingStepEvent);
  if Terminated then Exit;
  WaitForSingleObject(EvHandle, GetAnimationDelay);
  if Terminated then Exit;
  while (not Terminated) and (DoneAnimatedPlotting = false) do
    begin
      if Terminated then Exit;
      PlotGraph.GraphBitmap.Canvas.Lock;
      try
        DoneAnimatedPlotting := PlotGraph.StepAnimatedGraphPaint;
        if not Terminated then
          OutputBmpBuffer.SetBufferBitmap(PlotGraph.GraphBitmap);
      finally
        PlotGraph.GraphBitmap.Canvas.Unlock;
      end;
      if Terminated then Exit;
      if (DoneAnimatedPlotting = false) then //don't do last Sync, will be done later.
        Synchronize(DoAnimatedGraphPlotingStepEvent);
      if Terminated then Exit;
      WaitForSingleObject(EvHandle, GetAnimationDelay);
      if Terminated then Exit;
    end; {while}
end; {procedure RunAnimatedPlotting;}

{---------------------}

procedure RunAnimatedRestart;
var
  DoneAnimatedPlotting : boolean;
begin {procedure RunAnimatedRestart;}
  while (not Terminated) do
  begin
    DoneAnimatedPlotting := false;
    if Terminated then Exit;
    PlotGraph.GraphBitmap.Canvas.Lock;
    try
      PlotGraph.RestartAnimatedGraphPaint;
      if not Terminated then
        OutputBmpBuffer.SetBufferBitmap(PlotGraph.GraphBitmap);
    finally
      PlotGraph.GraphBitmap.Canvas.Unlock;
    end;
    if Terminated then Exit;
    Synchronize(DoAnimatedGraphPlotingStepEvent);
    if Terminated then Exit;
    WaitForSingleObject(EvHandle, GetAnimationDelay);
    if Terminated then Exit;
    while (not Terminated) and (DoneAnimatedPlotting = false) do
      begin
        if Terminated then Exit;
        PlotGraph.GraphBitmap.Canvas.Lock;
        try
          DoneAnimatedPlotting := PlotGraph.StepAnimatedGraphPaint;
          if not Terminated then
            OutputBmpBuffer.SetBufferBitmap(PlotGraph.GraphBitmap);
        finally
          PlotGraph.GraphBitmap.Canvas.Unlock;
        end;
        if Terminated then Exit;
        Synchronize(DoAnimatedGraphPlotingStepEvent);
        if Terminated then Exit;
        WaitForSingleObject(EvHandle, GetAnimationDelay);
        if Terminated then Exit;
      end; {while}
  end; {while}
end; {procedure RunAnimatedRestart;}

{---------------------}

begin {procedure TGraphDrawingThread.Execute;}
  FreeOnTerminate := false;
  FThreadErrorString := '';
  try
    if not Terminated then
      begin
        //begin ploting graph
        PlotGraph.OnIsTerminated := IsTerminated;
        if GetAnimatedPlotting then
          begin
            if GetAnimatedRestart then
              RunAnimatedRestart
            else
              RunAnimatedPlotting;
          end
        else
          begin
            if not Terminated then
              begin
                PlotGraph.GraphBitmap.Canvas.Lock;
                try
                  FThreadErrorString := PlotGraph.FullGraphPaint;
                  UniqueString(FThreadErrorString);
                  if not Terminated then
                    OutputBmpBuffer.SetBufferBitmap(PlotGraph.GraphBitmap);
                finally
                  PlotGraph.GraphBitmap.Canvas.UnLock;
                end;
              end; {if}
          end; {else}
      end; {if}
  except
    //floating point errors
    on EOverFlow do FThreadErrorString := ErrorStr_OverFlow;
    on EUnderFlow do FThreadErrorString := ErrorStr_UnderFlow;
    on EZeroDivide do FThreadErrorString := ErrorStr_ZeroDivide;
    on EInvalidOp do FThreadErrorString := ErrorStr_InvalidOp;
    on EMathError do FThreadErrorString := ErrorStr_MathError;
    //integer errors
    on EDivByZero do FThreadErrorString := ErrorStr_DivByZero;
    on EIntOverflow do FThreadErrorString := ErrorStr_IntOverflow;
    on EIntError do FThreadErrorString := ErrorStr_IntError;
    //other errors
    on EOutOfMemory do FThreadErrorString := ErrorStr_OutOfMemory;
    on EInvalidPointer do FThreadErrorString := ErrorStr_InvalidPointer;
    on EConvertError do FThreadErrorString := ErrorStr_ConvertError;
    else FThreadErrorString := ErrorStr_UnknownError;
  end; {try..except}
  UniqueString(FThreadErrorString);
  if (not Terminated) then
    Synchronize(DoDoneGraphPlotingEvent);
  //clear parameters because thread is about to terminate incase Execute runs again.
  PlotGraph.OnIsTerminated := nil;
  PlotGraph := nil;
  OutputBmpBuffer := nil;
  FOnDoneGraphPloting := nil;
  FOnAnimatedGraphPlotingStep := nil;
end; {procedure TGraphDrawingThread.Execute;}

{---------------------------------------------------------------------}

function TGraphDrawingThread.GetAnimatedPlotting: boolean;
begin {function TGraphDrawingThread.GetAnimatedPlotting}
  CriticalSection.Acquire;
  try
    Result := FAnimatedPlotting;
  finally
    CriticalSection.Release;
  end; {try..finally}
end; {function TGraphDrawingThread.GetAnimatedPlotting}

{---------------------------------------------------------------------}

function TGraphDrawingThread.GetAnimatedRestart: boolean;
begin {function TGraphDrawingThread.GetAnimatedRestart}
  CriticalSection.Acquire;
  try
    Result := FAnimatedRestart;
  finally
    CriticalSection.Release;
  end; {try..finally}
end; {function TGraphDrawingThread.GetAnimatedRestart}

{---------------------------------------------------------------------}

function TGraphDrawingThread.GetAnimationDelay: integer;
begin {function TGraphDrawingThread.GetAnimationDelay}
  CriticalSection.Acquire;
  try
    Result := FAnimationDelay;
  finally
    CriticalSection.Release;
  end; {try..finally}
end; {function TGraphDrawingThread.GetAnimationDelay}

{---------------------------------------------------------------------}

procedure TGraphDrawingThread.IsTerminated(var ShouldTerminate: boolean);
begin {procedure TGraphDrawingThread.IsTerminated}
  ShouldTerminate := Terminated;
end; {procedure TGraphDrawingThread.IsTerminated}

{---------------------------------------------------------------------}

procedure TGraphDrawingThread.SetAnimatedPlotting(
  TheAnimatedPlotting: boolean);
begin {procedure TGraphDrawingThread.SetAnimatedPlotting}
  CriticalSection.Acquire;
  try
    FAnimatedPlotting := TheAnimatedPlotting;
  finally
    CriticalSection.Release;
  end; {try..finally}
end; {procedure TGraphDrawingThread.SetAnimatedPlotting}

{---------------------------------------------------------------------}

procedure TGraphDrawingThread.SetAnimatedRestart(
  TheAnimatedRestart: boolean);
begin {procedure TGraphDrawingThread.SetAnimatedRestart}
  CriticalSection.Acquire;
  try
    FAnimatedRestart := TheAnimatedRestart;
  finally
    CriticalSection.Release;
  end; {try..finally}
end; {procedure TGraphDrawingThread.SetAnimatedRestart}

{---------------------------------------------------------------------}

procedure TGraphDrawingThread.SetAnimationDelay(
  TheAnimationDelay: integer);
begin {procedure TGraphDrawingThread.SetAnimationDelay}
  CriticalSection.Acquire;
  try
    FAnimationDelay := TheAnimationDelay;
  finally
    CriticalSection.Release;
  end; {try..finally}
end; {procedure TGraphDrawingThread.SetAnimationDelay}

{---------------------------------------------------------------------}

procedure TGraphDrawingThread.SignalAndTerminate;
begin {procedure TGraphDrawingThread.SignalAndTerminate;}
  Terminate;
  SetEvent(EvHandle);
end; {procedure TGraphDrawingThread.SignalAndTerminate;}

{---------------------------------------------------------------------}

end.
