unit PtThread;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Windows, sysutils, classes, graphics, GrGlobal, PtLabel, FileSave, PtShared,
  PtBuffer, GrDraw, DrawObjs, dialogs, CentrDlg, SetGraph;

type
  TLabelDrawingThread = class(TThread)
  private
    { Private declarations }
    FThreadErrorString : string;
    FOnDoneLabelPloting : TNotifyEvent;
    procedure IsTerminated(var ShouldTerminate : boolean); //Event handler
    procedure DoDoneLabelPlotingEvent; //executes in main thread
  protected
    { Protected declarations }
    procedure Execute; override;
  public
    { Public declarations }
    PlotLabelsBuffer : TPlotLabelsBuffer;
    InputBmpBuffer : TSafeBmpBuffer;
    property ThreadErrorString : string read FThreadErrorString;
    property OnDoneLabelPloting : TNotifyEvent read FOnDoneLabelPloting write FOnDoneLabelPloting;
    constructor Create;
    destructor Destroy; override;
  end; {TLabelDrawingThread}

  TPlotThread = class
  private
    { Private declarations }
    PrevRunningStartEnd : char;
    FOnRunningStart : TNotifyEvent;
    FOnRunningEnd : TNotifyEvent;
    FPlotLabelsBuffer : TPlotLabelsBuffer;
    FGraphOutLabelInBitmapBuffer : TSafeBmpBuffer;
    FSettingBmpAttribsObj : TSettingBmpAttribsObj;
    FOnPlotError : TPlotErrorEvent;
    FLabelList : TList;
    FOnDonePloting : TDonePlotingEvent;
    LabelThread : TLabelDrawingThread;
    ShouldRunAgainAfterTermLabelThread : boolean;
    function GetFileNameNoPath : string;
    function GetCaptionTitleStringWithBrackets : string;
    function GetIsLabelThreadRunning : boolean;
    procedure DoDonePlotingEvent();
    procedure HandleDoneLabelPlotingEvent(Sender: TObject);
  protected
    { Protected declarations }
    FFileName : string;
    FHasChangedSinceSave : boolean;
    property GraphOutLabelInBitmapBuffer : TSafeBmpBuffer read FGraphOutLabelInBitmapBuffer;
    procedure DoRunningStartEndEvent();
    procedure DoPlotErrorEvent(ThreadErrorString : string); //method is not thread safe
    procedure DoAssignFieldObjects; virtual; //method is not thread safe
    function GetIsAThreadRunning : boolean; virtual;
    function GetIsGraphThreadRunning : boolean; virtual; abstract;
    procedure SaveSettingBmpAttribs(WriteGraphFile : TWriteGraphFile);
    procedure ReadSettingBmpAttribs(ReadGraphFile : TReadGraphFile);
    procedure SaveAncestorDataToDisk(WriteGraphFile : TWriteGraphFile);
    procedure ReadAncestorDataFromDisk(ReadGraphFile : TReadGraphFile);
    function WriteFileToDisk(const FString : string) : boolean; virtual; abstract;
    function GetPlotTypeString : string; virtual; abstract;
  public
    { Public declarations }
    property OnRunningStart : TNotifyEvent read FOnRunningStart write FOnRunningStart;
    property OnRunningEnd : TNotifyEvent read FOnRunningEnd write FOnRunningEnd;
    property IsGraphThreadRunning : boolean read GetIsGraphThreadRunning;
    property IsLabelThreadRunning : boolean read GetIsLabelThreadRunning;
    property IsAThreadRunning : boolean read GetIsAThreadRunning;
    property SettingBmpAttribsObj : TSettingBmpAttribsObj read FSettingBmpAttribsObj;
    property FileName : string read FFileName;
    property FileNameNoPath : string read GetFileNameNoPath;
    property CaptionTitleStringWithBrackets : string read GetCaptionTitleStringWithBrackets;
    property PlotTypeString : string read GetPlotTypeString;
    property OnPlotError : TPlotErrorEvent read FOnPlotError write FOnPlotError;
    property OnDonePloting : TDonePlotingEvent read FOnDonePloting write FOnDonePloting;
    property LabelList : TList read FLabelList;
    property HasChangedSinceSave : boolean read FHasChangedSinceSave;
    procedure SetHasChangedSinceSaveTrue;
    constructor Create;
    destructor Destroy; override;
    //graph thread control methods
    procedure ReStartGraphThread; virtual; abstract;
    procedure TerminateGraphThread; virtual; abstract;
    procedure WaitForTermGraphThread; virtual; abstract;
    //label thread control methods
    procedure ReStartLabelThread;
    procedure RunAgainAfterTermLabelThread;
    procedure TerminateLabelThread;
    procedure WaitForTermLabelThread;
    //all thread control methods
    procedure TerminateAllThreads;
    procedure WaitForTermAllThreads;
    procedure InitalizeNew(TheColorDepth : TPixelFormat;
                           TheDefaultBmpWidth, TheDefaultBmpHeight : integer); virtual;
    function InitalizeFromFile(TheColorDepth : TPixelFormat;
                               const TheFileName : string;
                               ReadGraphFile : TReadGraphFile) : boolean; virtual; abstract;
    function SaveFile : boolean;
    function SaveFileAs(const TheFileName : string) : boolean;
    procedure ClearLabelList;
    procedure AssignFieldObjects; //thread(s) must be stopped before calling this
    procedure AssignLabelObjects(HiddenLabelIndex : integer); //thread(s) must be stopped before calling this
    procedure AssignAllObjectsAfterInitFromFile; //thread(s) must be stopped before calling this
   end; {TPlotThread}

implementation

{ TPlotThread }

procedure TPlotThread.AssignAllObjectsAfterInitFromFile;
begin
  DoAssignFieldObjects;
  FPlotLabelsBuffer.AssignLabelList(FLabelList, lbHiddenLabelNone);
end;

procedure TPlotThread.AssignFieldObjects;
begin
  SetHasChangedSinceSaveTrue;
  DoAssignFieldObjects;
end;

procedure TPlotThread.AssignLabelObjects(HiddenLabelIndex : integer);
begin
  SetHasChangedSinceSaveTrue;
  FPlotLabelsBuffer.AssignLabelList(FLabelList, HiddenLabelIndex);
end;

procedure TPlotThread.ClearLabelList;
var
  I : integer;
begin
  SetHasChangedSinceSaveTrue;
  WaitForTermLabelThread;
  if FLabelList.Count >= 1 then
    for I := 0 to FLabelList.Count -1 do
      if Assigned(FLabelList[I]) then
        TObject(FLabelList[I]).Free;
  FLabelList.Clear;
  FPlotLabelsBuffer.ClearLabelList;
end;

constructor TPlotThread.Create;
begin
  inherited;
  ShouldRunAgainAfterTermLabelThread := false;
  PrevRunningStartEnd := 'N';
  FOnRunningStart := nil;
  FOnRunningEnd := nil;
  FOnPlotError := nil;
  LabelThread := nil;
  FOnDonePloting := nil;
  FSettingBmpAttribsObj := TSettingBmpAttribsObj.Create;
  FLabelList := TList.Create;
  FGraphOutLabelInBitmapBuffer := TSafeBmpBuffer.Create;
  FPlotLabelsBuffer := TPlotLabelsBuffer.Create;
  FFileName := '';
  FHasChangedSinceSave := false;
end;

destructor TPlotThread.Destroy;
var
  I : integer;
  LocalTmpLabelThread : TLabelDrawingThread;
begin
  ShouldRunAgainAfterTermLabelThread := false;
  if Assigned(LabelThread) then
    begin
      LocalTmpLabelThread := LabelThread;
      LabelThread := nil;
      LocalTmpLabelThread.Terminate;
      LocalTmpLabelThread.WaitFor;
      LocalTmpLabelThread.free;
    end; {if}
  FSettingBmpAttribsObj.free;
  if FLabelList.Count >= 1 then
    for I := 0 to FLabelList.Count -1 do
      if Assigned(FLabelList[I]) then
        TObject(FLabelList[I]).Free;
  FLabelList.free;
  FGraphOutLabelInBitmapBuffer.free;
  FPlotLabelsBuffer.free;
  inherited;
end;

procedure TPlotThread.DoAssignFieldObjects;
begin
  with FSettingBmpAttribsObj do
    begin
      if DidBackGroundColorChange then
        FPlotLabelsBuffer.SetBackgroundColor(BackgroundColor);
      if DidColorDepthChange then
        begin
          FGraphOutLabelInBitmapBuffer.SetColorDepth(BmpColorDepth);
          FPlotLabelsBuffer.SetColorDepth(BmpColorDepth);
        end; {if}
      if DidWidthHeightChange then
        begin
          FGraphOutLabelInBitmapBuffer.SetSize(BmpWidth, BmpHeight);
          FPlotLabelsBuffer.SetSize(BmpWidth, BmpHeight);
        end; {if}
    end; {with}
end;

procedure TPlotThread.DoDonePlotingEvent();
begin
  try
    //fire event
    if Assigned(FOnDonePloting) then
      FOnDonePloting(FPlotLabelsBuffer.Bitmap);
  except  //stop all exceptions. an exception is very unlikely here.
  end;
end;

procedure TPlotThread.DoPlotErrorEvent(ThreadErrorString : string);
begin
  try
    UniqueString(ThreadErrorString);
    //fire event
    if Assigned(FOnPlotError) then
      FOnPlotError(Self, ThreadErrorString);
  except //stop all exceptions. an exception is very unlikely here.
  end;
end;

procedure TPlotThread.DoRunningStartEndEvent;
begin
  try
    //fire event
    if IsAThreadRunning then
    begin
      if Assigned(FOnRunningStart) and (PrevRunningStartEnd <> 'S') then
      begin
        PrevRunningStartEnd := 'S'; //used to stop redundant events
        FOnRunningStart(Self);
      end;
    end
    else
    begin
      if Assigned(FOnRunningEnd) and (PrevRunningStartEnd <> 'E') then
      begin
        PrevRunningStartEnd := 'E'; //used to stop redundant events
        FOnRunningEnd(Self);
      end;
    end;
  except  //stop all exceptions. an exception is very unlikely here.
    PrevRunningStartEnd := 'N'; //reset on error so event will fire next time.
  end;
end;

procedure TPlotThread.HandleDoneLabelPlotingEvent(Sender: TObject);
var
  ThreadErrorString : string;
begin
  if Assigned(LabelThread) then
  begin
    ThreadErrorString := LabelThread.ThreadErrorString;
    UniqueString(ThreadErrorString);
    LabelThread.FreeOnTerminate := true;
    LabelThread.Terminate;
    LabelThread := nil; //thread is done and will destroy itself so loose the refrence
    if ThreadErrorString = '' then
      DoDonePlotingEvent() //thread finished successfully
    else
      DoPlotErrorEvent(ThreadErrorString); //thread finished with an error
    //restart thread if it has been previously requested
    if ShouldRunAgainAfterTermLabelThread then
       ReStartLabelThread; //this sets ShouldRunAgainAfterTermLabelThread = false
    DoRunningStartEndEvent;
  end; {if}
end;

function TPlotThread.GetCaptionTitleStringWithBrackets: string;
begin
  if FileName <> '' then
    Result := PlotTypeString + ' - [' + ExtractFileName(FileName) + ']'
  else
    Result := PlotTypeString;
end;

function TPlotThread.GetIsAThreadRunning: boolean;
begin
  Result := IsLabelThreadRunning;
end;

function TPlotThread.GetFileNameNoPath: string;
begin
  if FileName <> '' then
    Result := ExtractFileName(FileName)
  else
    Result := '';
end;

procedure TPlotThread.InitalizeNew(TheColorDepth: TPixelFormat;
  TheDefaultBmpWidth, TheDefaultBmpHeight: integer);
begin
  FSettingBmpAttribsObj.BmpColorDepth := TheColorDepth;
  FSettingBmpAttribsObj.BmpWidth := TheDefaultBmpWidth;
  FSettingBmpAttribsObj.BmpHeight := TheDefaultBmpHeight;
  FPlotLabelsBuffer.InitalizeNew(TheColorDepth, TheDefaultBmpWidth, TheDefaultBmpHeight);
  FGraphOutLabelInBitmapBuffer.InitalizeNew(TheColorDepth, TheDefaultBmpWidth, TheDefaultBmpHeight);
end;

procedure TPlotThread.ReadAncestorDataFromDisk(
  ReadGraphFile: TReadGraphFile);

{----------------------------------}

procedure ShowOutOfMemDlg;
begin
  try
    MessageDlgErrorOutOfMemoryNoWait;
  except
  end; {try..except}
end;

{----------------------------------}

procedure ReadVer1Fields;
type
  TVer1_DrawingTool = (Ver1_dtOff, Ver1_dtLine, Ver1_dtFloodFill,
       Ver1_dtRectangle, Ver1_dtEllipse, Ver1_dtRoundRect, Ver1_dtText);
var
  I, TmpLabelCount : integer;
  Ver1_LabelType : TVer1_DrawingTool;
  TmpLabel : TDrawBaseClass;
begin {procedure ReadVer1Fields;}
  {begin read labels}
  TmpLabelCount := ReadGraphFile.ReadSmallint;
  for I := 1 to TmpLabelCount do
    begin
      ReadGraphFile.ReadRecord(Ver1_LabelType, sizeof(Ver1_LabelType));
      TmpLabel := nil;
      case Ver1_LabelType of
        Ver1_dtLine: TmpLabel := TDrawLine.Create;
        Ver1_dtRectangle:
            TmpLabel := TDrawInclosedShape.Create(dtRectangle);
        Ver1_dtEllipse:
            TmpLabel := TDrawInclosedShape.Create(dtEllipse);
        Ver1_dtRoundRect:
            TmpLabel := TDrawInclosedShape.Create(dtRoundRect);
        Ver1_dtText: TmpLabel := TDrawText.Create;
      end; {case}
      if TmpLabel = nil then
        ReadGraphFile.RaiseExceptionLoading;
      try
        TmpLabel.ReadFromDisk(ReadGraphFile, FPlotLabelsBuffer.Bitmap,
                              FPlotLabelsBuffer.BitmapAttribRec);
      except
        TmpLabel.Free;
        raise;
      end; {try..except}
      try
        LabelList.Add(TmpLabel);
      except
        on EOutOfMemory do
          begin
            TmpLabel.Free;
            ShowOutOfMemDlg;
            raise;
          end; {on EOutOfMemory do}
      end; {try..except}
    end; {for}
  {end read labels}
  {begin read flood fills}
  TmpLabelCount := ReadGraphFile.ReadSmallint;
  for I := 1 to TmpLabelCount do
    begin
      TmpLabel := TDrawFloodFill.Create;
      if TmpLabel = nil then
        ReadGraphFile.RaiseExceptionLoading;
      try
        TmpLabel.ReadFromDisk(ReadGraphFile, FPlotLabelsBuffer.Bitmap,
                              FPlotLabelsBuffer.BitmapAttribRec);
      except
        TmpLabel.Free;
        raise;
      end; {try..except}
      try
        LabelList.Add(TmpLabel);
      except
        on EOutOfMemory do
          begin
            TmpLabel.Free;
            ShowOutOfMemDlg;
            raise;
          end; {on EOutOfMemory do}
      end; {try..except}
    end; {for}
  {end read flood fills}
end; {procedure ReadVer1Fields;}

{----------------------------------}

procedure ReadVer2_3Fields;
var
  I, TmpLabelCount : integer;
  LabelType : longint;
  TmpLabel : TDrawBaseClass;
begin {procedure ReadVer2_3Fields;}
  TmpLabelCount := ReadGraphFile.ReadLongint;
  for I := 1 to TmpLabelCount do
    begin
      LabelType := ReadGraphFile.ReadLongint;
      TmpLabel := nil;
      case LabelType of
        dl_FileType_Line: TmpLabel := TDrawLine.Create;
        dl_FileType_FloodFill: TmpLabel := TDrawFloodFill.Create;
        dl_FileType_Rectangle:
            TmpLabel := TDrawInclosedShape.Create(dtRectangle);
        dl_FileType_Ellipse:
            TmpLabel := TDrawInclosedShape.Create(dtEllipse);
        dl_FileType_RoundRect:
            TmpLabel := TDrawInclosedShape.Create(dtRoundRect);
        dl_FileType_Text: TmpLabel := TDrawText.Create;
        dl_FileType_FreeHand: TmpLabel := TDrawFreeHand.Create;
      end; {case}
      if TmpLabel = nil then
        ReadGraphFile.RaiseExceptionLoading;
      try
        TmpLabel.ReadFromDisk(ReadGraphFile, FPlotLabelsBuffer.Bitmap,
                              FPlotLabelsBuffer.BitmapAttribRec);
      except
        TmpLabel.Free;
        raise;
      end; {try..except}
      try
        LabelList.Add(TmpLabel);
      except
        on EOutOfMemory do
          begin
            TmpLabel.Free;
            ShowOutOfMemDlg;
            raise;
          end; {on EOutOfMemory do}
      end; {try..except}
    end; {for}
end; {procedure ReadVer2_3Fields;}

{----------------------------------}

begin {procedure TPlotThread.ReadAncestorDataFromDisk}
  Assert((FSettingBmpAttribsObj.BmpWidth > 0) and
         (FSettingBmpAttribsObj.BmpHeight > 0),
         'Error FPlotLabelsBuffer is not fully initalized');
  with FSettingBmpAttribsObj do
    begin
      if DidBackGroundColorChange then
        FPlotLabelsBuffer.SetBackgroundColor(BackgroundColor);
      if DidColorDepthChange then
        FPlotLabelsBuffer.SetColorDepth(BmpColorDepth);
      if DidWidthHeightChange then
        FPlotLabelsBuffer.SetSize(BmpWidth, BmpHeight);
    end; {with}
  Assert((FPlotLabelsBuffer.BitmapAttribRec.BmpWidth > 0) and
         (FPlotLabelsBuffer.BitmapAttribRec.BmpHeight > 0),
         'Error FPlotLabelsBuffer is not fully initalized');
  if ReadGraphFile.FileVersion = 1 then
    ReadVer1Fields
  else {if FileVersion >= 2 then}
    ReadVer2_3Fields;
  FPlotLabelsBuffer.AssignLabelList(FLabelList, lbHiddenLabelNone);
end; {procedure TPlotThread.ReadAncestorDataFromDisk}

procedure TPlotThread.ReadSettingBmpAttribs(ReadGraphFile: TReadGraphFile);
begin
  FSettingBmpAttribsObj.ReadFromDisk(ReadGraphFile);
end;

procedure TPlotThread.ReStartLabelThread;
var
  LocalTmpLabelThread : TLabelDrawingThread;
  ThreadErrorString : string;
begin
  ShouldRunAgainAfterTermLabelThread := false;
  if Assigned(LabelThread) then
    begin
      LocalTmpLabelThread := LabelThread;
      LabelThread := nil;
      LocalTmpLabelThread.Terminate;
      LocalTmpLabelThread.WaitFor;
      ThreadErrorString := LocalTmpLabelThread.ThreadErrorString;
      UniqueString(ThreadErrorString);
      LocalTmpLabelThread.free;
      if ThreadErrorString <> '' then
        DoPlotErrorEvent(ThreadErrorString);
    end; {if}
  LabelThread := TLabelDrawingThread.Create;
  LabelThread.PlotLabelsBuffer := FPlotLabelsBuffer;
  LabelThread.InputBmpBuffer := FGraphOutLabelInBitmapBuffer;
  LabelThread.OnDoneLabelPloting := HandleDoneLabelPlotingEvent;
  LabelThread.Resume;
  DoRunningStartEndEvent;
end;

procedure TPlotThread.RunAgainAfterTermLabelThread;
begin
  if IsLabelThreadRunning then
    ShouldRunAgainAfterTermLabelThread := true
  else
    ReStartLabelThread;
end;

procedure TPlotThread.SaveAncestorDataToDisk(
  WriteGraphFile: TWriteGraphFile);
var
  I, TmpLabelCount : integer;
begin
  TmpLabelCount := LabelList.Count;
  WriteGraphFile.WriteLongint(TmpLabelCount);
  for I := 0 to (TmpLabelCount - 1) do
    TDrawBaseClass(LabelList[I]).SaveToDisk(WriteGraphFile);
end;

function TPlotThread.SaveFile : boolean;
begin
  Result := WriteFileToDisk(FFileName);
  if Result then
    FHasChangedSinceSave := false;
end;

function TPlotThread.SaveFileAs(const TheFileName: string) : boolean;
begin
  Result := WriteFileToDisk(TheFileName);
  if Result then
    begin
      FHasChangedSinceSave := false;
      FFileName := TheFileName;
    end; {if}
end;

procedure TPlotThread.SaveSettingBmpAttribs(
  WriteGraphFile: TWriteGraphFile);
begin
  FSettingBmpAttribsObj.SaveToDisk(WriteGraphFile);
end;

procedure TPlotThread.SetHasChangedSinceSaveTrue;
begin
  FHasChangedSinceSave := true;
end;

procedure TPlotThread.TerminateAllThreads;
begin
  ShouldRunAgainAfterTermLabelThread := false;
  TerminateGraphThread;
  TerminateLabelThread;
end;

procedure TPlotThread.WaitForTermAllThreads;
begin
  ShouldRunAgainAfterTermLabelThread := false;
  TerminateGraphThread;
  TerminateLabelThread;
  WaitForTermGraphThread;
  WaitForTermLabelThread;
end;

procedure TPlotThread.TerminateLabelThread;
begin
  ShouldRunAgainAfterTermLabelThread := false;
  if Assigned(LabelThread) then
    LabelThread.Terminate;
end;

procedure TPlotThread.WaitForTermLabelThread;
var
  LocalTmpLabelThread : TLabelDrawingThread;
  ThreadErrorString : string;
begin
  ShouldRunAgainAfterTermLabelThread := false;
  if Assigned(LabelThread) then
    begin
      LocalTmpLabelThread := LabelThread;
      LabelThread := nil;
      LocalTmpLabelThread.Terminate;
      LocalTmpLabelThread.WaitFor;
      ThreadErrorString := LocalTmpLabelThread.ThreadErrorString;
      UniqueString(ThreadErrorString);
      LocalTmpLabelThread.free;
      if ThreadErrorString <> '' then
        DoPlotErrorEvent(ThreadErrorString);
    end; {if}
  DoRunningStartEndEvent;
end;

function TPlotThread.GetIsLabelThreadRunning: boolean;
begin
  Result := Assigned(LabelThread);
end;

{ TLabelDrawingThread }

constructor TLabelDrawingThread.Create;
begin
  inherited Create(True);
  FreeOnTerminate := false;
  FThreadErrorString := '';
  FOnDoneLabelPloting := nil;
  PlotLabelsBuffer := nil;
  InputBmpBuffer := nil;
end;

destructor TLabelDrawingThread.Destroy;
begin
  FThreadErrorString := '';
  inherited;
end;

procedure TLabelDrawingThread.DoDoneLabelPlotingEvent;
//executes in main thread
begin
  if Assigned(FOnDoneLabelPloting) then
    FOnDoneLabelPloting(PlotLabelsBuffer.Bitmap); //fire event
end;

procedure TLabelDrawingThread.Execute;
begin
  FreeOnTerminate := false;
  FThreadErrorString := '';
  try
    if not Terminated then
      begin
        PlotLabelsBuffer.Bitmap.Canvas.Lock;
        try
          InputBmpBuffer.GetBufferBitmap(PlotLabelsBuffer.Bitmap);
          if not Terminated then
            PlotLabelsBuffer.PaintDrawLables(IsTerminated);
        finally
          PlotLabelsBuffer.Bitmap.Canvas.Unlock;
        end;
      end; {if}
  except
    //floating point errors
    on EOverFlow do FThreadErrorString := ErrorStr_OverFlow;
    on EUnderFlow do FThreadErrorString := ErrorStr_UnderFlow;
    on EZeroDivide do FThreadErrorString := ErrorStr_ZeroDivide;
    on EInvalidOp do FThreadErrorString := ErrorStr_InvalidOp;
    on EMathError do FThreadErrorString := ErrorStr_MathError;
    //integer errors
    on EDivByZero do FThreadErrorString := ErrorStr_DivByZero;
    on EIntOverflow do FThreadErrorString := ErrorStr_IntOverflow;
    on EIntError do FThreadErrorString := ErrorStr_IntError;
    //other errors
    on EOutOfMemory do FThreadErrorString := ErrorStr_OutOfMemory;
    on EInvalidPointer do FThreadErrorString := ErrorStr_InvalidPointer;
    on EConvertError do FThreadErrorString := ErrorStr_ConvertError;
    else FThreadErrorString := ErrorStr_UnknownError;
  end; {try..except}
  UniqueString(FThreadErrorString);
  if (not Terminated) then
    Synchronize(DoDoneLabelPlotingEvent);
  //clear parameters because thread is about to terminate incase Execute runs again.
  FOnDoneLabelPloting := nil;
  PlotLabelsBuffer := nil;
  InputBmpBuffer := nil;
end;

procedure TLabelDrawingThread.IsTerminated(var ShouldTerminate: boolean);
begin
  ShouldTerminate := Terminated;
end;

end.
