unit SetGraph;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Classes, Graphics, Jpeg, GrGlobal, FileSave;

type
  TSettingBmpAttribsObj = class
  private
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
    DidWidthHeightChange : boolean;
    DidColorDepthChange : boolean;
    DidBackGroundColorChange : boolean;
    BmpWidth : integer;
    BmpHeight : integer;
    BmpColorDepth : TPixelFormat; //field not saved to disk
    BackgroundColor : TColor; //currently being set by graph setting objects. Change in v5
    constructor Create;
    procedure SetAllDidChange(DidChange : boolean);
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile);
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile);
  end; {TSettingBmpAttribsObj}

  TSettingGrBGPictObj = class
  private
    { Private declarations }
    FBackgroundBitmap: TBitmap;
    FBackgroundJPEG: TJPEGImage;
    procedure SetBackgroundBitmap(TheNewBitmap : TBitmap);
    procedure SetBackgroundJPEG(TheNewJPEGImage : TJPEGImage);
  protected
    { Protected declarations }
  public
    { Public declarations }
    DidTileModeChange : boolean;
    DidPictureChange : boolean;
    BackgroundTileMode : TGraphBackgroundTileMode;
    property BackgroundBitmap : TBitmap read FBackgroundBitmap write SetBackgroundBitmap;
    property BackgroundJPEG : TJPEGImage read FBackgroundJPEG write SetBackgroundJPEG;
    constructor Create;
    destructor Destroy; override;
    procedure ClearBackgroundPicture;
    procedure SetAllDidChange(DidChange : boolean);
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile);
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile);
  end; {TSettingGrBGPictObj}

implementation

const
  FileType_BGPicture_None = 0;
  FileType_BGPicture_Bitmap = 1;
  FileType_BGPicture_JPEG = 2;

{---------------------------------------------------------------------}

{ TSettingBmpAttribsObj }

constructor TSettingBmpAttribsObj.Create;
begin {constructor TSettingBmpAttribsObj.Create}
  inherited Create;
  DidWidthHeightChange := false;
  DidColorDepthChange := false;
  DidBackGroundColorChange := false;
  BmpWidth := -1;
  BmpHeight := -1;
  BmpColorDepth := pf24bit;
  BackgroundColor := clWhite;
end; {constructor TSettingBmpAttribsObj.Create}

{---------------------------------------------------------------------}

procedure TSettingBmpAttribsObj.ReadFromDisk(
  ReadGraphFile: TReadGraphFile);
begin {procedure TSettingBmpAttribsObj.ReadFromDisk}
  SetAllDidChange(true);
  with ReadGraphFile do
    begin
      if FileVersion = 1 then
        begin
          BmpHeight := ReadSmallint;
          BmpWidth := ReadSmallint;
        end
      else {if FileVersion >= 2 then}
        begin
          BmpHeight := ReadLongint;
          BmpWidth := ReadLongint;
          if FileVersion >= 4 then
            ReadRecord(BackgroundColor, sizeof(BackgroundColor));
        end; {else}
    end; {with}
end; {procedure TSettingBmpAttribsObj.ReadFromDisk}

{---------------------------------------------------------------------}

procedure TSettingBmpAttribsObj.SaveToDisk(
  WriteGraphFile: TWriteGraphFile);
begin {procedure TSettingBmpAttribsObj.SaveToDisk}
  with WriteGraphFile do
    begin
      WriteLongint(BmpHeight);
      WriteLongint(BmpWidth);
      WriteRecord(BackgroundColor, sizeof(BackgroundColor));
    end; {with}
end; {procedure TSettingBmpAttribsObj.SaveToDisk}

{---------------------------------------------------------------------}

procedure TSettingBmpAttribsObj.SetAllDidChange(DidChange: boolean);
begin {procedure TSettingBmpAttribsObj.SetAllDidChange}
  DidWidthHeightChange := DidChange;
  DidColorDepthChange := DidChange;
  DidBackGroundColorChange := DidChange;
end; {procedure TSettingBmpAttribsObj.SetAllDidChange}

{---------------------------------------------------------------------}

{ TSettingGrBGPictObj }

procedure TSettingGrBGPictObj.ClearBackgroundPicture;
begin {procedure TSettingGrBGPictObj.ClearBackgroundPicture;}
  if Assigned(FBackgroundBitmap) then
    FBackgroundBitmap.free;
  FBackgroundBitmap := nil;
  if Assigned(FBackgroundJPEG) then
    FBackgroundJPEG.free;
  FBackgroundJPEG := nil;
end; {procedure TSettingGrBGPictObj.ClearBackgroundPicture;}

{---------------------------------------------------------------------}

constructor TSettingGrBGPictObj.Create;
begin {constructor TSettingGrBGPictObj.Create;}
  inherited;
  DidTileModeChange := false;
  DidPictureChange := false;
  BackgroundTileMode := btCenter;
  FBackgroundBitmap := nil;
  FBackgroundJPEG := nil;
end; {constructor TSettingGrBGPictObj.Create;}

{---------------------------------------------------------------------}

destructor TSettingGrBGPictObj.Destroy;
begin {destructor TSettingGrBGPictObj.Destroy;}
  if Assigned(FBackgroundBitmap) then
    FBackgroundBitmap.free;
  if Assigned(FBackgroundJPEG) then
    FBackgroundJPEG.free;
  inherited;
end; {destructor TSettingGrBGPictObj.Destroy;}

{---------------------------------------------------------------------}

procedure TSettingGrBGPictObj.ReadFromDisk(ReadGraphFile: TReadGraphFile);
var
  TmpLongint : Longint;
begin {procedure TSettingGrBGPictObj.ReadFromDisk}
  SetAllDidChange(true);
  ClearBackgroundPicture;
  with ReadGraphFile do
    begin
      if FileVersion >= 4 then
        begin
          ReadRecord(BackgroundTileMode, sizeof(BackgroundTileMode));
          TmpLongint := ReadLongint;
          //if TmpLongint = FileType_BGPicture_None then do nothing
          if TmpLongint = FileType_BGPicture_Bitmap then
            FBackgroundBitmap := ReadCreateBitmap
          else
            if TmpLongint = FileType_BGPicture_JPEG then
              FBackgroundJPEG :=  ReadCreateJPEGImage;
        end; {if}
    end; {with}
end; {procedure TSettingGrBGPictObj.ReadFromDisk}

{---------------------------------------------------------------------}

procedure TSettingGrBGPictObj.SaveToDisk(WriteGraphFile: TWriteGraphFile);
begin {procedure TSettingGrBGPictObj.SaveToDisk}
  with WriteGraphFile do
    begin
      WriteRecord(BackgroundTileMode, sizeof(BackgroundTileMode));
      if Assigned(FBackgroundBitmap) then
        begin
          WriteLongint(FileType_BGPicture_Bitmap);
          WriteBitmap(FBackgroundBitmap);
        end
      else
        if Assigned(FBackgroundJPEG) then
          begin
            WriteLongint(FileType_BGPicture_JPEG);
            WriteJPEGImage(FBackgroundJPEG);
          end
        else
          WriteLongint(FileType_BGPicture_None);
    end; {with}
end; {procedure TSettingGrBGPictObj.SaveToDisk}

{---------------------------------------------------------------------}

procedure TSettingGrBGPictObj.SetAllDidChange(DidChange: boolean);
begin {procedure TSettingGrBGPictObj.SetAllDidChange}
  DidTileModeChange := DidChange;
  DidPictureChange := DidChange;
end; {procedure TSettingGrBGPictObj.SetAllDidChange}

{---------------------------------------------------------------------}

procedure TSettingGrBGPictObj.SetBackgroundBitmap(TheNewBitmap: TBitmap);
begin {procedure TSettingGrBGPictObj.SetBackgroundBitmap}
  if FBackgroundBitmap <> TheNewBitmap then
    begin
      if Assigned(FBackgroundBitmap) then
        FBackgroundBitmap.free;
      FBackgroundBitmap := TheNewBitmap;
    end; {if}
end; {procedure TSettingGrBGPictObj.SetBackgroundBitmap}

{---------------------------------------------------------------------}

procedure TSettingGrBGPictObj.SetBackgroundJPEG(
  TheNewJPEGImage: TJPEGImage);
begin {procedure TSettingGrBGPictObj.SetBackgroundJPEG}
  if FBackgroundJPEG <> TheNewJPEGImage then
    begin
      if Assigned(FBackgroundJPEG) then
        FBackgroundJPEG.free;
      FBackgroundJPEG := TheNewJPEGImage;
    end; {if}
end; {procedure TSettingGrBGPictObj.SetBackgroundJPEG}

{---------------------------------------------------------------------}

end.
