unit SetPoObj;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses Graphics, FloatStr, MathCalc, MathFunc, FileSave, DefFont;

const
  MaxPolarAxisCount = 360;
  SizeDashStr = 'Dash Length';
  SizeCircleStr = 'Circle Width';
  SizeNoTicksStr = 'No Tick Marks';

type
  TTickMarkType = (Dashes,Circles,NoTicks);
  PPolarAxisAngles = ^TPolarAxisAngles;
  TPolarAxisAngles = packed array[1..MaxPolarAxisCount] of TMathFloat;

  TSettingPolarObj = class
  public
    DidFieldsChange : boolean;
    BackColor : TColor;
    AxisCount : integer;
    AxisAngles : TPolarAxisAngles;
    AxisColor : TColor;
    AxisWidth : integer;
    AxisStyle : TPenStyle;
    TickMarkIncStr : string;
    TickMarkIncrements : TMathValue;
    TickMarkType : TTickMarkType;
    TickMarkColor : TColor;
    DashLength : integer;
    CircleWidth : integer;
    CircleStyle : TPenStyle;
    LabelTickMarks : boolean;
    TMLabelFormat : TFloatNumFormat;
    ShowTMLabelComa : boolean;
    ShowTMLabelPlus : boolean;
    LabelDecimalPlaces : integer;
    Restrict : boolean;
    RestrictAmtStr : string;
    RestrictAmtVal : TMathValue;
    {TFont values}
    LabelFont : TFont;
    constructor Create;
    destructor Destroy; override;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile);
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile);
    procedure Assign(Source : TSettingPolarObj);
  end; {TSettingPolarObj = class}

implementation

type
  TDiskV3_SettingPolarRec = packed record
    BackColor : TColor;
    AxisCount : integer;
    AxisAngles : TPolarAxisAngles;
    AxisColor : TColor;
    AxisWidth : integer;
    AxisStyle : TPenStyle;
    TickMarkIncrements : TPackedMathValue;
    TickMarkType : TTickMarkType;
    TickMarkColor : TColor;
    DashLength : integer;
    CircleWidth : integer;
    CircleStyle : TPenStyle;
    LabelTickMarks : boolean;
    TMLabelFormat : TFloatNumFormat;
    ShowTMLabelComa : boolean;
    ShowTMLabelPlus : boolean;
    LabelDecimalPlaces : integer;
    Restrict : boolean;
    RestrictAmtVal : TPackedMathValue;
  end; {TDiskV3_SettingPolarRec = packed record}

  TDiskV2_SettingPolarRec = packed record
    BackColor : TColor;
    AxisCount : integer;
    AxisAngles : TPolarAxisAngles;
    AxisColor : TColor;
    AxisWidth : integer;
    AxisStyle : TPenStyle;
    TickMarkIncrements : TMathFloat;
    TickMarkType : TTickMarkType;
    TickMarkColor : TColor;
    DashLength : integer;
    CircleWidth : integer;
    CircleStyle : TPenStyle;
    LabelTickMarks : boolean;
    TMLabelFormat : TFloatNumFormat;
    ShowTMLabelComa : boolean;
    ShowTMLabelPlus : boolean;
    LabelDecimalPlaces : integer;
    Restrict : boolean;
    RestrictAmtVal : TMathFloat;
  end; {TDiskV2_SettingPolarRec = packed record}

  TDiskV1_SettingPolarRec = packed record
    BackColor : TColor;
    AxisCount : smallint;
    AxisAngles : TPolarAxisAngles;
    AxisColor : TColor;
    AxisWidth : smallint;
    AxisStyle : TPenStyle;
    TickMarkIncStr : string[50];
    TickMarkIncrements : TMathFloat;
    TickMarkType : TTickMarkType;
    TickMarkColor : TColor;
    DashLength : smallint;
    CircleWidth : smallint;
    CircleStyle : TPenStyle;
    LabelTickMarks : boolean;
    LabelDecimalPlaces : smallint;
    Restrict : boolean;
    RestrictAmtStr : string[50];
    RestrictAmtVal : TMathFloat;
  end; {TDiskV1_SettingPolarRec = packed record}

{---------------------------------------------------------------------}

{ TSettingPolarObj }

constructor TSettingPolarObj.Create;
var
  I : integer;
  BeginAngle : single;
begin {constructor TSettingPolarObj.Create;}
  inherited Create;
  DidFieldsChange := false;
  BackColor := clWhite;
  for I := 1 to MaxPolarAxisCount do
    AxisAngles[I] := 0;
  AxisCount := 8;
  BeginAngle := 0;
  for I := 1 to AxisCount do
    begin
      AxisAngles[I] := BeginAngle;
      BeginAngle := BeginAngle + 45;
    end; {for}
  AxisColor := clBlack;
  AxisWidth := 1;
  AxisStyle := psSolid;
  TickMarkIncStr := '1';
  TickMarkIncrements := LongintToMathValue(1);
  TickMarkType := Dashes;
  TickMarkColor := clblack;
  CircleWidth := 1;
  DashLength := 5;
  CircleStyle := psSolid;
  LabelTickMarks := true;
  TMLabelFormat := lfMaxDecimal;
  ShowTMLabelComa := true;
  ShowTMLabelPlus := false;
  LabelDecimalPlaces := GetFloatCordDecPlacesDefault(lfMaxDecimal);
  Restrict := false;
  RestrictAmtStr := '5';
  RestrictAmtVal := LongintToMathValue(5);
  {TFont values}
  LabelFont := CreateANewTFontClass;
end; {constructor TSettingPolarObj.Create;}

{---------------------------------------------------------------------}

destructor TSettingPolarObj.Destroy;
begin {destructor TSettingPolarObj.Destroy;}
  LabelFont.Free;
  inherited Destroy;
end; {destructor TSettingPolarObj.Destroy;}

{---------------------------------------------------------------------}

procedure TSettingPolarObj.SaveToDisk(WriteGraphFile: TWriteGraphFile);
var
  DiskRec : TDiskV3_SettingPolarRec;
begin {procedure TSettingPolarObj.SaveToDisk}
  DiskRec.BackColor := BackColor;
  DiskRec.AxisCount := AxisCount;
  DiskRec.AxisAngles := AxisAngles;
  DiskRec.AxisColor := AxisColor;
  DiskRec.AxisWidth := AxisWidth;
  DiskRec.AxisStyle := AxisStyle;
  DiskRec.TickMarkIncrements := PackMathValue(TickMarkIncrements);
  DiskRec.TickMarkType := TickMarkType;
  DiskRec.TickMarkColor := TickMarkColor;
  DiskRec.DashLength := DashLength;
  DiskRec.CircleWidth := CircleWidth;
  DiskRec.CircleStyle := CircleStyle;
  DiskRec.LabelTickMarks := LabelTickMarks;
  DiskRec.TMLabelFormat := TMLabelFormat;
  DiskRec.ShowTMLabelComa := ShowTMLabelComa;
  DiskRec.ShowTMLabelPlus := ShowTMLabelPlus;
  DiskRec.LabelDecimalPlaces := LabelDecimalPlaces;
  DiskRec.Restrict := Restrict;
  DiskRec.RestrictAmtVal := PackMathValue(RestrictAmtVal);
  with WriteGraphFile do
    begin
      WriteRecord(DiskRec, sizeof(DiskRec));
      WriteMathString(TickMarkIncStr);
      WriteMathString(RestrictAmtStr);
      {TFont values}
      WriteFont(LabelFont);
    end; {with}
end; {procedure TSettingPolarObj.SaveToDisk}

{---------------------------------------------------------------------}

procedure TSettingPolarObj.ReadFromDisk(ReadGraphFile: TReadGraphFile);

{-------------------------------------}

procedure ReadVer1Fields;
var
  DiskRec : TDiskV1_SettingPolarRec;
begin {procedure ReadVer1Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  BackColor := DiskRec.BackColor;
  AxisCount := DiskRec.AxisCount;
  AxisAngles := DiskRec.AxisAngles;
  AxisColor := DiskRec.AxisColor;
  AxisWidth := DiskRec.AxisWidth;
  AxisStyle := DiskRec.AxisStyle;
  TickMarkIncStr := DiskRec.TickMarkIncStr;
  TickMarkIncrements := FloatToMathValue(DiskRec.TickMarkIncrements);
  TickMarkType := DiskRec.TickMarkType;
  TickMarkColor := DiskRec.TickMarkColor;
  DashLength := DiskRec.DashLength;
  CircleWidth := DiskRec.CircleWidth;
  CircleStyle := DiskRec.CircleStyle;
  LabelTickMarks := DiskRec.LabelTickMarks;
  LabelDecimalPlaces := DiskRec.LabelDecimalPlaces;
  Restrict := DiskRec.Restrict;
  RestrictAmtStr := DiskRec.RestrictAmtStr;
  RestrictAmtVal := FloatToMathValue(DiskRec.RestrictAmtVal);
  {TFont values}
  ReadGraphFile.Read16BitFont(LabelFont);
  {fix and calc math strings}
  with ReadGraphFile do
    begin
      FixCalcMathConstString(TickMarkIncStr, TickMarkIncrements);
      FixCalcMathConstString(RestrictAmtStr, RestrictAmtVal);
    end; {with}
end; {procedure ReadVer1Fields;}

{-------------------------------------}

procedure ReadVer2Fields;
var
  DiskRec : TDiskV2_SettingPolarRec;
begin {procedure ReadVer2Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  BackColor := DiskRec.BackColor;
  AxisCount := DiskRec.AxisCount;
  AxisAngles := DiskRec.AxisAngles;
  AxisColor := DiskRec.AxisColor;
  AxisWidth := DiskRec.AxisWidth;
  AxisStyle := DiskRec.AxisStyle;
  TickMarkIncrements := FloatToMathValue(DiskRec.TickMarkIncrements);
  TickMarkType := DiskRec.TickMarkType;
  TickMarkColor := DiskRec.TickMarkColor;
  DashLength := DiskRec.DashLength;
  CircleWidth := DiskRec.CircleWidth;
  CircleStyle := DiskRec.CircleStyle;
  LabelTickMarks := DiskRec.LabelTickMarks;
  TMLabelFormat := DiskRec.TMLabelFormat;
  ShowTMLabelComa := DiskRec.ShowTMLabelComa;
  ShowTMLabelPlus := DiskRec.ShowTMLabelPlus;
  LabelDecimalPlaces := DiskRec.LabelDecimalPlaces;
  Restrict := DiskRec.Restrict;
  RestrictAmtVal := FloatToMathValue(DiskRec.RestrictAmtVal);
  with ReadGraphFile do
    begin
      ReadMathConstString(TickMarkIncStr, TickMarkIncrements, False);
      ReadMathConstString(RestrictAmtStr, RestrictAmtVal, False);
      {TFont values}
      Read32BitFont(LabelFont);
    end; {with}
end; {procedure ReadVer2Fields;}

{-------------------------------------}

procedure ReadVer3Fields;
var
  DiskRec : TDiskV3_SettingPolarRec;
begin {procedure ReadVer3Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  BackColor := DiskRec.BackColor;
  AxisCount := DiskRec.AxisCount;
  AxisAngles := DiskRec.AxisAngles;
  AxisColor := DiskRec.AxisColor;
  AxisWidth := DiskRec.AxisWidth;
  AxisStyle := DiskRec.AxisStyle;
  TickMarkIncrements := UnPackMathValue(DiskRec.TickMarkIncrements);
  TickMarkType := DiskRec.TickMarkType;
  TickMarkColor := DiskRec.TickMarkColor;
  DashLength := DiskRec.DashLength;
  CircleWidth := DiskRec.CircleWidth;
  CircleStyle := DiskRec.CircleStyle;
  LabelTickMarks := DiskRec.LabelTickMarks;
  TMLabelFormat := DiskRec.TMLabelFormat;
  ShowTMLabelComa := DiskRec.ShowTMLabelComa;
  ShowTMLabelPlus := DiskRec.ShowTMLabelPlus;
  LabelDecimalPlaces := DiskRec.LabelDecimalPlaces;
  Restrict := DiskRec.Restrict;
  RestrictAmtVal := UnPackMathValue(DiskRec.RestrictAmtVal);
  with ReadGraphFile do
    begin
      ReadMathConstString(TickMarkIncStr, TickMarkIncrements, False);
      ReadMathConstString(RestrictAmtStr, RestrictAmtVal, False);
      {TFont values}
      Read32BitFont(LabelFont);
    end; {with}
end; {procedure ReadVer3Fields;}

{-------------------------------------}

begin {procedure TSettingPolarObj.ReadFromDisk}
  DidFieldsChange := true;
  if ReadGraphFile.FileVersion = 1 then
    ReadVer1Fields
  else
    if ReadGraphFile.FileVersion = 2 then
      ReadVer2Fields
    else {if ReadGraphFile.FileVersion >= 3 then}
      ReadVer3Fields;
end; {procedure TSettingPolarObj.ReadFromDisk}

{---------------------------------------------------------------------}

procedure TSettingPolarObj.Assign(Source: TSettingPolarObj);
begin {procedure TSettingPolarObj.Assign}
  BackColor := Source.BackColor;
  AxisCount := Source.AxisCount;
  AxisAngles := Source.AxisAngles;
  AxisColor := Source.AxisColor;
  AxisWidth := Source.AxisWidth;
  AxisStyle := Source.AxisStyle;
  TickMarkIncStr := Source.TickMarkIncStr;
  TickMarkIncrements := Source.TickMarkIncrements;
  TickMarkType := Source.TickMarkType;
  TickMarkColor := Source.TickMarkColor;
  DashLength := Source.DashLength;
  CircleWidth := Source.CircleWidth;
  CircleStyle := Source.CircleStyle;
  LabelTickMarks := Source.LabelTickMarks;
  TMLabelFormat := Source.TMLabelFormat;
  ShowTMLabelComa := Source.ShowTMLabelComa;
  ShowTMLabelPlus := Source.ShowTMLabelPlus;
  LabelDecimalPlaces := Source.LabelDecimalPlaces;
  Restrict := Source.Restrict;
  RestrictAmtStr := Source.RestrictAmtStr;
  RestrictAmtVal := Source.RestrictAmtVal;
  {TFont values}
  LabelFont.Assign(Source.LabelFont);
  {Make Memory Unique}
  UniqueString(TickMarkIncStr);
  UniqueString(RestrictAmtStr);
end; {procedure TSettingPolarObj.Assign}

{---------------------------------------------------------------------}

end.
