unit custedit;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Windows, Controls, Classes, Graphics, Sysutils, Messages, GrDraw;

type
  TStartEditingEvent = procedure(Sender: TObject; EditTopLeft : TPoint;
                             EditWidth, EditHeight : integer) of object;

  TCustomTextEdit = class(TCustomControl)
    private
      FOnStartEditing : TStartEditingEvent;
      FOnDoneEditing : TNotifyEvent;
      FTextEditVisible : boolean;
      FTextString : string;
      CaretVisible : boolean;
      CaretPosX : integer;
      CaretPosAfterChar : integer;
      procedure ResizeToFitText;
      procedure MoveCaretAfterChar(TheCaretPos : integer);
    public
      property TextString : string read FTextString;
      property TextEditVisible : boolean read FTextEditVisible;
      property OnStartEditing : TStartEditingEvent read FOnStartEditing
                                                   write FOnStartEditing;
      property OnDoneEditing : TNotifyEvent read FOnDoneEditing
                                            write FOnDoneEditing;
      constructor Create(AOwner : TComponent); override;
      destructor Destroy; override;
      procedure ShowTextEdit(TheBmpTopLeft, TheScrollTopLeft : TPoint;
                             TheTextFont : TFont; const TheTextStr : string);
      procedure HideTextEdit;
      procedure Paint; override;
      procedure WndProc(var Message: TMessage); override;
      procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
      procedure CustomKeyPress(Sender: TObject; var Key: Char);
      procedure CustomKeyDown(Sender: TObject; var Key: Word;
        Shift: TShiftState);
      procedure CustomMouseDown(Sender: TObject; Button: TMouseButton;
                                Shift: TShiftState; X, Y: Integer);
      procedure CustomExit(Sender: TObject);
  end; {TCustomTextEdit}

implementation

const
  MinCustEditWidth = 8;
  TextMargin = 2;
  MaxEditHeight = 10000;
  MaxEditWidth = 10000;
  MaxEditTextLength = MaxTextLabelLength;

{ TCustomTextEdit }

constructor TCustomTextEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOnStartEditing := nil;
  FOnDoneEditing := nil;
  FTextEditVisible := false;
  FTextString := '';
  TabStop := false;
  Enabled := false;
  Visible := false;
  Top := 1;
  Left := 1;
  Height := 0;
  Width := 0;
  Cursor := crIBeam;
  CaretVisible := false;
  CaretPosX := TextMargin;
  CaretPosAfterChar := 0;
  OnKeyPress := CustomKeyPress;
  OnKeyDown := CustomKeyDown;
  OnMouseDown := CustomMouseDown;
  OnExit := CustomExit;
end;

procedure TCustomTextEdit.ResizeToFitText;
var
  NewWidth, NewHeight : integer;
  HeightChanged : boolean;
begin
  NewHeight := Canvas.TextHeight('W'+FTextString) + (TextMargin*2);
  if FTextString = '' then
    NewWidth := MinCustEditWidth
  else
    begin
      NewWidth := Canvas.TextWidth(FTextString);
      if NewWidth < MinCustEditWidth then
        NewWidth := MinCustEditWidth;
    end; {else}
  NewWidth := NewWidth + (TextMargin*2);
  if NewHeight > MaxEditHeight then
    NewHeight := MaxEditHeight;
  if NewWidth > MaxEditWidth then
    NewWidth := MaxEditWidth;
  HeightChanged := Height <> NewHeight;
  Width := NewWidth;
  Height := NewHeight;
  {if height changed and the caret is visible then resize caret}
  if HeightChanged and CaretVisible then
    begin
      DestroyCaret;
      CreateCaret(self.Handle, 0, 0, abs(Height - (TextMargin*2)));
      SetCaretPos(CaretPosX, TextMargin);
      ShowCaret(self.Handle);
    end; {if}
  RePaint;
end;

procedure TCustomTextEdit.MoveCaretAfterChar(TheCaretPos: integer);
var
  NewCaretPosAfterChar, NewCaretPosX : integer;
begin
  if TheCaretPos <= 0 then
    begin
      NewCaretPosAfterChar := 0;
      NewCaretPosX := TextMargin;
    end
  else
    if TheCaretPos >= Length(FTextString) then
      begin
        NewCaretPosAfterChar := Length(FTextString);
        NewCaretPosX := Canvas.TextWidth(FTextString) + TextMargin;
      end
    else
      begin
        NewCaretPosAfterChar := TheCaretPos;
        NewCaretPosX :=
          Canvas.TextWidth(Copy(FTextString, 1, TheCaretPos)) + TextMargin;
      end; {else}
  if NewCaretPosX > ((Width - 1) - TextMargin) then
    NewCaretPosX := ((Width - 1) - TextMargin);
  CaretPosX := NewCaretPosX;
  CaretPosAfterChar := NewCaretPosAfterChar;
  if CaretVisible then
    SetCaretPos(NewCaretPosX, TextMargin);
end;

procedure TCustomTextEdit.HideTextEdit;
begin
  if FTextEditVisible then
    begin
      FTextEditVisible := false;
      Enabled := false;
      Visible := false;
      if Assigned(FOnDoneEditing) then
        FOnDoneEditing(Self);
      Top := 1;
      Left := 1;
      Height := 0;
      Width := 0;
      CaretPosX := TextMargin;
      CaretPosAfterChar := 0;
    end; {if}
end;

procedure TCustomTextEdit.Paint;
begin
  if Canvas.Font.Color <> clWhite then
    begin
      Canvas.Pen.Color := clBlack;
      Canvas.Brush.Color := clwhite;
    end
  else
    begin
      Canvas.Pen.Color := clWhite;
      Canvas.Brush.Color := clBlack;
    end; {else}
  Canvas.Pen.Style := psSolid;
  Canvas.Pen.Width := 1;
  Canvas.Brush.Style := bsSolid;
  Canvas.Rectangle(0, 0, Width, Height);
  Canvas.Brush.Color := clwhite;
  Canvas.Brush.Style := bsClear;
  Canvas.TextOut(TextMargin, TextMargin, FTextString);
end;

procedure TCustomTextEdit.ShowTextEdit(TheBmpTopLeft, TheScrollTopLeft : TPoint;
  TheTextFont: TFont; const TheTextStr: string);
begin
  if FTextEditVisible then
    HideTextEdit;
  FTextEditVisible := true;
  Top := TheScrollTopLeft.Y - TextMargin;
  Left := TheScrollTopLeft.X - TextMargin;
  Canvas.Font.Assign(TheTextFont);
  FTextString := TheTextStr;
  ResizeToFitText;
  if Assigned(FOnStartEditing) then
    begin
      with TheBmpTopLeft do
        begin
          X := X - TextMargin;
          Y := Y - TextMargin;
        end; {with}
      FOnStartEditing(Self, TheBmpTopLeft, Width, Height);
    end; {if}
  Enabled := true;
  Visible := true;
  SetFocus;
end;

procedure TCustomTextEdit.WndProc(var Message: TMessage);
begin
  if Message.Msg = WM_KILLFOCUS then
    begin
      DestroyCaret;
      CaretVisible := false;
    end; {if}
  if Message.Msg = WM_SETFOCUS then
    begin
      CreateCaret(self.Handle, 0, 0, abs(Height - (TextMargin*2)));
      SetCaretPos(CaretPosX, TextMargin);
      ShowCaret(self.Handle);
      CaretVisible := true;
    end; {if}
  inherited WndProc(Message);
end;

procedure TCustomTextEdit.CustomKeyPress(Sender: TObject; var Key: Char);
var
  TmpStr : string;
begin
  if not(Key in [#0, #9, #10, #127]) then //nul, <tab>, line feed, <delete>
    case Key of
      #13, #27: //carrage return, <escape>
        begin
          HideTextEdit;
        end;
      #8: //<backspace>
        begin
          if CaretPosAfterChar > 0 then
            begin
              Delete(FTextString, CaretPosAfterChar, 1);
              ResizeToFitText;
              MoveCaretAfterChar(CaretPosAfterChar - 1);
            end; {if}
        end;
      else
        begin
          if Length(FTextString) < MaxEditTextLength then
            begin
              TmpStr := Key;
              Insert(TmpStr, FTextString, CaretPosAfterChar + 1);
              ResizeToFitText;
              MoveCaretAfterChar(CaretPosAfterChar + 1);
            end; {if}
        end; {else}
    end; {case}
end;

procedure TCustomTextEdit.CustomKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_LEFT:
      begin
        if CaretPosAfterChar > 0 then
          MoveCaretAfterChar(CaretPosAfterChar - 1);
      end;
    VK_RIGHT:
      begin
        if CaretPosAfterChar < Length(FTextString) then
          MoveCaretAfterChar(CaretPosAfterChar + 1);
      end;
    VK_DELETE:
      begin
        if CaretPosAfterChar < Length(FTextString) then
          begin
            Delete(FTextString, (CaretPosAfterChar + 1), 1);
            ResizeToFitText;
            MoveCaretAfterChar(CaretPosAfterChar);
          end; {if}
      end;
    VK_HOME:
      begin
        if CaretPosAfterChar > 0 then
          MoveCaretAfterChar(0);
      end;
    VK_END:
      begin
        if CaretPosAfterChar < Length(FTextString) then
          MoveCaretAfterChar(Length(FTextString));
      end;
  end; {case}
end;

procedure TCustomTextEdit.CustomMouseDown(Sender: TObject;
            Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  I, CurrDist, NearestDist, NearestCaretPos : integer;
begin
  if Button = mbLeft then
    begin
      NearestDist := abs(TextMargin - X);
      NearestCaretPos := 0;
      for I := 1 to Length(FTextString) do
        begin
          CurrDist := abs((Canvas.TextWidth(Copy(FTextString,1,I))+TextMargin)-X);
          if CurrDist <= NearestDist then
            begin
              NearestDist := CurrDist;
              NearestCaretPos := I;
            end
          else
            break;
        end; {for}
      if NearestCaretPos <> CaretPosAfterChar then
        MoveCaretAfterChar(NearestCaretPos);
      Windows.SetFocus(self.handle);
      SetFocus; //do this just incase.
    end; {if}
end;

procedure TCustomTextEdit.CustomExit(Sender: TObject);
begin
  HideTextEdit;
end;

procedure TCustomTextEdit.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := Message.Result or DLGC_WANTARROWS;
  { Don't need this code for this version; Save for future use.
  Message.Result := Message.Result or DLGC_WANTALLKEYS;
  Message.Result := Message.Result or DLGC_WANTTAB;
  Message.Result := Message.Result or DLGC_WANTCHARS;}
end;

destructor TCustomTextEdit.Destroy;
begin
  if CaretVisible then
    DestroyCaret;
  inherited Destroy;
end;

end.
