unit editmath;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, MathCalc, MathFunc, CentrDlg;

type
  TValueCompareType = (vct_Equal, vct_NotEqual, vct_Greater, vct_Less,
                       vct_GreaterOrEqual, vct_LessOrEqual);

type
  TEditMathBounds = set of (AllowEmpty, CheckLoBound, LoBoundCanEqual,
                                        CheckHiBound, HiBoundCanEqual);
  TEditMathConst = class(TEdit)
  private
    { Private declarations }
    FBoundsCheck : TEditMathBounds;
    FLastErrorCode : TMathErrorCode;
    FLastErrorPosition : integer;
    FLastMathValue : TMathValue;
    FLastWasEmpty : boolean;
    FLastWasValid : boolean;
    FCtrlErrorName : string;
    FLoBoundStr : string;
    FLoBoundVal : TMathFloat;
    FHiBoundStr : string;
    FHiBoundVal : TMathFloat;
  protected
    { Protected declarations }
  public
    { Public declarations }
    property LastErrorCode : TMathErrorCode read FLastErrorCode;
    property LastErrorPosition : integer read FLastErrorPosition;
    property LastValue : TMathFloat read FLastMathValue.FloatValue;
    property LastMathValue : TMathValue read FLastMathValue;
    property LastWasEmpty : boolean read FLastWasEmpty;
    property LastWasValid : boolean read FLastWasValid;
    constructor Create(AOwner: TComponent); override;
    function GetUserErrorStr : string;
    procedure ShowUserErrorMsg;
    function ValidateMathStr : boolean;
    function ValidateMathStrNoErrMsg : boolean;
    function GetOwnerForm : TForm;
  published
    { Published declarations }
    property emBoundsCheck : TEditMathBounds read FBoundsCheck
      write FBoundsCheck default [];
    property emCtrlErrorName : string read FCtrlErrorName
      write FCtrlErrorName;
    property emLoBoundStr : string read FLoBoundStr
      write FLoBoundStr;
    property emLoBoundVal : TMathFloat read FLoBoundVal
      write FLoBoundVal;
    property emHiBoundStr : string read FHiBoundStr
      write FHiBoundStr;
    property emHiBoundVal : TMathFloat read FHiBoundVal
      write FHiBoundVal;
  end;

procedure ShowFuncErrorMsg(const TheMathErrorStr, TheCtrlErrorName : string;
                           TheCenterOnForm : TForm);
function Validate2EditMathCtrls(TheLoValCtrl : TEditMathConst;
                                TheValueCompareType : TValueCompareType;
                                TheHiValCtrl : TEditMathConst;
                                TheCenterOnForm : TForm) : boolean;

procedure Register;

implementation

{---------------------------------------------------------------------}

procedure ShowFuncErrorMsg(const TheMathErrorStr, TheCtrlErrorName : string;
                           TheCenterOnForm : TForm);
begin {procedure ShowFuncErrorMsg}
  MessageDlgCenter('Error in: ' + TheCtrlErrorName+': ' + TheMathErrorStr,
                    mtError, [mbOK], 0, TheCenterOnForm);
end; {procedure ShowFuncErrorMsg}

{---------------------------------------------------------------------}

function Validate2EditMathCtrls(TheLoValCtrl : TEditMathConst;
                                TheValueCompareType : TValueCompareType;
                                TheHiValCtrl : TEditMathConst;
                                TheCenterOnForm : TForm) : boolean;
begin {function Validate2EditMathCtrls}
  Result := false;
  if (not TheLoValCtrl.LastWasValid)or(not TheHiValCtrl.LastWasValid) then
    exit;
  if (TheLoValCtrl.LastWasEmpty)or(TheHiValCtrl.LastWasEmpty) then
    begin
      Result := true;
      exit;
    end; {if}
  case TheValueCompareType of
    vct_Equal :
      begin
        if TheLoValCtrl.LastValue = TheHiValCtrl.LastValue then
          Result := true
        else
          MessageDlgCenter('Error: '+TheLoValCtrl.emCtrlErrorName+' must equal '+
          TheHiValCtrl.emCtrlErrorName , mtError, [mbOK], 0, TheCenterOnForm);
      end;
    vct_NotEqual :
      begin
        if TheLoValCtrl.LastValue <> TheHiValCtrl.LastValue then
          Result := true
        else
          MessageDlgCenter('Error: '+TheLoValCtrl.emCtrlErrorName+' can not equal '+
          TheHiValCtrl.emCtrlErrorName , mtError, [mbOK], 0, TheCenterOnForm);
      end;
    vct_Greater :
      begin
        if TheLoValCtrl.LastValue > TheHiValCtrl.LastValue then
          Result := true
        else
          MessageDlgCenter('Error: '+TheLoValCtrl.emCtrlErrorName+' must be > '+
          TheHiValCtrl.emCtrlErrorName , mtError, [mbOK], 0, TheCenterOnForm);
      end;
    vct_Less :
      begin
        if TheLoValCtrl.LastValue < TheHiValCtrl.LastValue then
          Result := true
        else
          MessageDlgCenter('Error: '+TheLoValCtrl.emCtrlErrorName+' must be < '+
          TheHiValCtrl.emCtrlErrorName , mtError, [mbOK], 0, TheCenterOnForm);
      end;
    vct_GreaterOrEqual :
      begin
        if TheLoValCtrl.LastValue >= TheHiValCtrl.LastValue then
          Result := true
        else
          MessageDlgCenter('Error: '+TheLoValCtrl.emCtrlErrorName+' must be >= '+
          TheHiValCtrl.emCtrlErrorName , mtError, [mbOK], 0, TheCenterOnForm);
      end;
    vct_LessOrEqual :
      begin
        if TheLoValCtrl.LastValue <= TheHiValCtrl.LastValue then
          Result := true
        else
          MessageDlgCenter('Error: '+TheLoValCtrl.emCtrlErrorName+' must be <= '+
          TheHiValCtrl.emCtrlErrorName , mtError, [mbOK], 0, TheCenterOnForm);
      end;
    else Result := false;
  end; {case}
end; {function Validate2EditMathCtrls}

{---------------------------------------------------------------------}

procedure Register;
begin
  RegisterComponents('MathGV', [TEditMathConst]);
end;

{---------------------------------------------------------------------}

constructor TEditMathConst.Create(AOwner: TComponent);
begin {constructor TEditMathConst.Create}
  inherited Create(AOwner);
  FLastErrorCode := meNulExp;
  FLastErrorPosition := 0;
  InitMathValue(FLastMathValue);
  FLastWasValid := false;
  FLastWasEmpty := false;
end; {constructor TEditMathConst.Create}

{---------------------------------------------------------------------}

function TEditMathConst.GetOwnerForm: TForm;
begin {function TEditMathConst.GetOwnerForm}
  if assigned(Owner) then
    begin
      if Owner is TForm then
        Result := (Owner as TForm)
      else
        Result := nil;
    end
  else
    Result := nil;
end; {function TEditMathConst.GetOwnerForm}

{---------------------------------------------------------------------}

function TEditMathConst.GetUserErrorStr : string;
var
  TmpStr : string;
begin {function TEditMathConst.GetUserErrorStr}
  Result := '';
  if (not FLastWasValid) or (FLastErrorCode <> meNoError) then
    begin
      if FLastErrorCode <> meNoError then
        Result := 'Error in: ' + FCtrlErrorName + ': ' +
                  GetMLErrorStrFunc(FLastErrorCode, FLastErrorPosition)
      else
        begin
          TmpStr := '';
          if CheckLoBound in FBoundsCheck then
            begin
              if LoBoundCanEqual in FBoundsCheck then
                TmpStr := '>= '
              else
                TmpStr := '> ';
              if FLoBoundStr <> '' then
                TmpStr := TmpStr + FLoBoundStr
              else
                TmpStr := TmpStr + FloatToStr(FLoBoundVal);
            end; {if}
          if CheckHiBound in FBoundsCheck then
            begin
              if TmpStr <> '' then
                TmpStr := TmpStr + ' and ';
              if HiBoundCanEqual in FBoundsCheck then
                TmpStr := TmpStr + '<= '
              else
                TmpStr := TmpStr + '< ';
              if FHiBoundStr <> '' then
                TmpStr := TmpStr + FHiBoundStr
              else
                TmpStr := TmpStr + FloatToStr(FHiBoundVal);
            end; {if}
          if TmpStr <> '' then
            Result := 'Error: ' + FCtrlErrorName + ' must be ' + TmpStr;
        end; {else}
    end; {if}
end; {function TEditMathConst.GetUserErrorStr}

{---------------------------------------------------------------------}

procedure TEditMathConst.ShowUserErrorMsg;
begin {procedure TEditMathConst.ShowUserErrorMsg;}
  MessageDlgCenter(GetUserErrorStr, mtError, [mbOK], 0, GetOwnerForm);
end; {procedure TEditMathConst.ShowUserErrorMsg;}

{---------------------------------------------------------------------}

function TEditMathConst.ValidateMathStr : boolean;
var
  TmpParentWinControl : TWinControl;
  TmpPageControl : TPageControl;
begin {function TEditMathConst.ValidateMathStr}
  Result := ValidateMathStrNoErrMsg;
  if not Result then
    begin
      ShowUserErrorMsg;
      //set focus even if it's a tabed dialog
      TmpParentWinControl := Self.Parent;
      while Assigned(TmpParentWinControl) do
        begin
          if TmpParentWinControl is TTabSheet then
            begin
              TmpPageControl := (TmpParentWinControl as TTabSheet).PageControl;
              if Assigned(TmpPageControl) then
                if TmpPageControl.Visible then
                  TmpPageControl.ActivePageIndex := (TmpParentWinControl as TTabSheet).PageIndex;
              break;
            end;
          TmpParentWinControl := TmpParentWinControl.parent;
        end; {while}
      if Self.Visible then
        Self.SetFocus;
    end; {if}
end; {function TEditMathConst.ValidateMathStr}

{---------------------------------------------------------------------}

function TEditMathConst.ValidateMathStrNoErrMsg : boolean;
var
  MathConstantObj : TMathFunction;
  DidPassBoundsCheck : boolean;
begin {function TEditMathConst.ValidateMathStrNoErrMsg}
  FLastWasValid := false;
  if (AllowEmpty in FBoundsCheck) and (self.text = '') then
    begin
      FLastErrorCode := meNoError;
      FLastErrorPosition := 0;
      InitMathValue(FLastMathValue);
      FLastWasEmpty := true;
      FLastWasValid := true;
    end
  else
    begin
      FLastWasEmpty := false;
      MathConstantObj := TMathFunction.Create;
      try
        with MathConstantObj do
          begin
            FuncString := self.text;
            FLastWasValid := VerifyMathFunc;
            FLastErrorCode := ErrorCode;
            FLastErrorPosition := ErrorPosition;
            FLastMathValue := MathResultValue;
          end; {with}
      except
        FLastWasValid := false;
        FLastErrorCode := meUnknownException;
        FLastErrorPosition := 0;
        InitMathValue(FLastMathValue);
      end; {try..except}
      MathConstantObj.Free;
      if FLastWasValid then
        begin
          DidPassBoundsCheck := true;
          if CheckLoBound in FBoundsCheck then
            begin
              if (LoBoundCanEqual in FBoundsCheck) then
                begin
                  if LastValue < FLoBoundVal then
                    DidPassBoundsCheck := false;
                end
              else
                begin
                  if LastValue <= FLoBoundVal then
                    DidPassBoundsCheck := false;
                end; {else}
            end; {if}
          if DidPassBoundsCheck and (CheckHiBound in FBoundsCheck) then
            begin
              if (HiBoundCanEqual in FBoundsCheck) then
                begin
                  if LastValue > FHiBoundVal then
                    DidPassBoundsCheck := false;
                end
              else
                begin
                  if LastValue >= FHiBoundVal then
                    DidPassBoundsCheck := false;
                end; {else}
            end; {if}
          FLastWasValid := DidPassBoundsCheck;
        end; {if}
    end; {else}
  Result := FLastWasValid;
end; {function TEditMathConst.ValidateMathStrNoErrMsg}

{---------------------------------------------------------------------}

end.
