unit func2d;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses Classes, Graphics, MathCalc, MathFunc, FuncObj, FileSave, GapSort;

const
  MaxPointItemsCount = 1000;
  MaxSeriesItemsCount = 50;
  MaxPointsPerSeries = 5000;
  MaxGridEditorRows = (MaxPointItemsCount + MaxSeriesItemsCount + 32 + 1);

type
  TFromSideBehave = (sbAutomatic, sbExtraAccuracy, sbNoExtraAccuracy,
                     sbDrawPosInfinity, sbDrawNegInfinity, sbNotContinuous);
  TAtPointBehave = (pbAutoDetect, pbUndefined);
  TItemType = (itSingleValue, itSeries, itFuncSeries);

  TSpecialPointRec = record
    PointValue : TMathValue;
    ListIndexPriority : integer;
    LeftSideBehave : TFromSideBehave;
    AtPointBehave : TAtPointBehave;
    RightSideBehave : TFromSideBehave;
  end; {TSpecialPointRec}

  TSpecialPointArray = array of TSpecialPointRec;

  TFunc2DSpecialItem = class
  private
    { Private declarations }
  public
    { Public declarations }
    FirstStr, SecondStr, EndStr : string;
    FirstVal, SecondVal, EndVal : TMathValue;
    FunctionStr : string;
    LeftSideBehave : TFromSideBehave;
    AtPointBehave : TAtPointBehave;
    RightSideBehave : TFromSideBehave;
    SeriesPointCount : integer;
    ItemType : TItemType;
    constructor Create;
    destructor Destroy; override;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile);
    procedure ReadFromDisk(ReadGraphFile : TReadGraphFile);
    procedure Assign(Source : TFunc2DSpecialItem);
  end; {TFunc2DSpecialItem}

  TFunc2DSpecialValues = class
  private
    { Private declarations }
    FItemsList : TList;
    FSpecialPointArray : TSpecialPointArray;
  public
    { Public declarations }
    property ItemsList : TList read FItemsList;
    property SpecialPointArray : TSpecialPointArray read FSpecialPointArray;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure GapSortSwap(Item1, Item2: integer);
    procedure GapSortCompare(Item1, Item2: integer; var CompResult : integer);
    function CalcPointsArray(SpecialValueFunc : TMathFunction) : boolean;
    procedure SaveToDisk(WriteGraphFile : TWriteGraphFile);
    function ReadFromDisk(ReadGraphFile : TReadGraphFile) : boolean;
    procedure AssignMoveSpPointsArray(Source : TFunc2DSpecialValues);
  end; {TFunc2DSpecialValues}

  TMath2DFunc = class(TMathGraphFunc)
  private
    { Private declarations }
  public
    { Public declarations }
    constructor Create;
  end; {TMath2DFunc = class(TMathGraphFunc)}

  TFunc2DObj = class(TBaseFuncObj)
    private
    { Private declarations }
      FFunc2DSpecialValues : TFunc2DSpecialValues;
    protected
    { Protected declarations }
      function GetListFuncStr: string; override;
      function CreateNewSeqObj : TObject; override;
      procedure AssignSeqObj(Source : TObject); override;
    public
    { Public declarations }
      FuncCalc : TMath2DFunc;
      LoBoundStr,
      HiBoundStr : string;
      LoBlank,
      HiBlank : boolean;
      LoBoundVal,
      HiBoundVal : TMathValue;
      AsymFixUp : boolean;
      ExtraUndefAccuracy : boolean;
      FuncColor : TColor;
      LineWidth : integer;
      property Func2DSpecialValues : TFunc2DSpecialValues read FFunc2DSpecialValues;
      constructor Create;
      destructor Destroy; override;
      procedure SaveToDisk(WriteGraphFile : TWriteGraphFile);
      function ReadFromDisk(ReadGraphFile : TReadGraphFile) : boolean;
      procedure AssignAndMoveSpPoints(Source : TFunc2DObj);
  end; {TFunc2DObj = class(TBaseFuncObj)}

function VerifySeriesPoints(const FirstVal, SecondVal, EndVal : TMathValue) : boolean;
function CalcSeriesPointCount(const FirstVal, SecondVal, EndVal : TMathValue;
                              var PointCount : integer) : boolean;

implementation

type
  TDiskV3_Func2DSpItemRec = packed record
    FirstVal, SecondVal, EndVal : TPackedMathValue;
    LeftSideBehave : TFromSideBehave;
    AtPointBehave : TAtPointBehave;
    RightSideBehave : TFromSideBehave;
    SeriesPointCount : integer;
    ItemType : TItemType;
  end; {TDiskV3_Func2DSpItemRec = packed record}

  TDiskV1_Func2DRec = packed record
    JunkBytes : packed array[1..4] of byte;
    LoBoundStr,
    HiBoundStr : string[50];
    LoBlank,
    HiBlank : boolean;
    LoBoundVal,
    HiBoundVal : Extended;
    AsymFixUp : boolean;
    FuncColor : TColor;
    LineWidth : smallint;
  end; {TDiskV1_Func2DRec = record}

  TDiskV2_Func2DRec = packed record
    LoBlank,
    HiBlank : boolean;
    LoBoundVal,
    HiBoundVal : TMathFloat;
    AsymFixUp : boolean;
    ExtraUndefAccuracy : boolean;
    FuncColor : TColor;
    LineWidth : integer;
  end; {TDiskV2_Func2DRec = packed record}

  TDiskV3_Func2DRec = packed record
    LoBlank,
    HiBlank : boolean;
    LoBoundVal,
    HiBoundVal : TPackedMathValue;
    AsymFixUp : boolean;
    ExtraUndefAccuracy : boolean;
    FuncColor : TColor;
    LineWidth : integer;
  end; {TDiskV3_Func2DRec = packed record}

{-----------------------------------------------------------------------}

function VerifySeriesPoints(const FirstVal, SecondVal, EndVal : TMathValue) : boolean;
begin {function VerifySeriesPoints}
  Assert((FirstVal.ValueType <> vtEmpty) and (SecondVal.ValueType <> vtEmpty) and
         (EndVal.ValueType <> vtEmpty), 'Empty series value');
  Result := ((FirstVal.FloatValue < SecondVal.FloatValue) and (SecondVal.FloatValue < EndVal.FloatValue))or
            ((FirstVal.FloatValue > SecondVal.FloatValue) and (SecondVal.FloatValue > EndVal.FloatValue));
end; {function VerifySeriesPoints}

{-----------------------------------------------------------------------}

function CalcSeriesPointCount(const FirstVal, SecondVal, EndVal : TMathValue;
                              var PointCount : integer) : boolean;
var
  I : integer;
  Diff, Curr : TMathFloat;
begin {function CalcSeriesPointCount}
  Assert((FirstVal.ValueType <> vtEmpty) and (SecondVal.ValueType <> vtEmpty) and
         (EndVal.ValueType <> vtEmpty), 'Empty series value');
  Result := false;
  PointCount := 0;
  if ((FirstVal.FloatValue < SecondVal.FloatValue) and (SecondVal.FloatValue < EndVal.FloatValue)) then
    begin
      PointCount := 3;
      Diff := abs(SecondVal.FloatValue - FirstVal.FloatValue);
      I := 1;
      Curr := SecondVal.FloatValue + (Diff * I);
      while Curr < EndVal.FloatValue do
        begin
          inc(PointCount);
          if PointCount > MaxPointsPerSeries then
            Exit; //point count to large
          inc(I);
          Curr := SecondVal.FloatValue + (Diff * I);
        end; {while}
      Result := true;
    end
  else
    if ((FirstVal.FloatValue > SecondVal.FloatValue) and (SecondVal.FloatValue > EndVal.FloatValue)) then
      begin
        PointCount := 3;
        Diff := abs(FirstVal.FloatValue - SecondVal.FloatValue);
        I := 1;
        Curr := SecondVal.FloatValue - (Diff * I);
        while Curr > EndVal.FloatValue do
          begin
            inc(PointCount);
            if PointCount > MaxPointsPerSeries then
              Exit; //point count to large
            inc(I);
            Curr := SecondVal.FloatValue - (Diff * I);
          end; {while}
        Result := true;
      end;
end; {function CalcSeriesPointCount}

{-----------------------------------------------------------------------}

constructor TMath2DFunc.Create;
begin {constructor TMath2DFunc.Create;}
  inherited Create;
  VarCount := 1;
  VarStrings[0] := 'X';
end; {constructor TMath2DFunc.Create;}

{-----------------------------------------------------------------------}

function TFunc2DObj.GetListFuncStr: string;
begin {function TFunc2DObj.GetListFuncStr}
  GetListFuncStr := 'Y=f(x)=' + FuncCalc.FuncString;
end; {function TFunc2DObj.GetListFuncStr}

{-----------------------------------------------------------------------}

constructor TFunc2DObj.Create;
begin {constructor TFunc2DObj.Create;}
  inherited Create;
  FuncCalc := TMath2DFunc.Create;
  FFunc2DSpecialValues := TFunc2DSpecialValues.Create;
  LoBoundStr := '';
  HiBoundStr := '';
  LoBlank := true;
  HiBlank := true;
  InitMathValue(LoBoundVal); //old LoBoundVal := -1;
  InitMathValue(HiBoundVal); //old HiBoundVal := 1;
  AsymFixUp := false;
  ExtraUndefAccuracy := false;
  FuncColor := clBlue;
  LineWidth := 1;
end; {constructor TFunc2DObj.Create;}

{-----------------------------------------------------------------------}

destructor TFunc2DObj.Destroy;
begin {destructor TFunc2DObj.Destroy;}
  FuncCalc.Free;
  FFunc2DSpecialValues.Free;
  inherited Destroy;
end; {destructor TFunc2DObj.Destroy;}

{-----------------------------------------------------------------------}

procedure TFunc2DObj.SaveToDisk(WriteGraphFile : TWriteGraphFile);
var
  DiskRec : TDiskV3_Func2DRec;
begin {procedure TFunc2DObj.SaveToDisk}
  DiskRec.LoBlank := LoBlank;
  DiskRec.HiBlank := HiBlank;
  DiskRec.LoBoundVal := PackMathValue(LoBoundVal);
  DiskRec.HiBoundVal := PackMathValue(HiBoundVal);
  DiskRec.AsymFixUp := AsymFixUp;
  DiskRec.ExtraUndefAccuracy := ExtraUndefAccuracy;
  DiskRec.FuncColor := FuncColor;
  DiskRec.LineWidth := LineWidth;
  with WriteGraphFile do
    begin
      WriteRecord(DiskRec, sizeof(DiskRec));
      WriteMathString(LoBoundStr);
      WriteMathString(HiBoundStr);
    end; {with}
  Func2DSpecialValues.SaveToDisk(WriteGraphFile);
  FuncCalc.SaveToDisk(WriteGraphFile);
end; {procedure TFunc2DObj.SaveToDisk}

{-----------------------------------------------------------------------}

function TFunc2DObj.ReadFromDisk(ReadGraphFile : TReadGraphFile) : boolean;
{Returns false if an error occured while calculating 2D function special values.}

{-----------------------------}

procedure ReadVer1Fields;
var
  DiskRec : TDiskV1_Func2DRec;
begin {procedure ReadVer1Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  LoBoundStr := DiskRec.LoBoundStr;
  HiBoundStr := DiskRec.HiBoundStr;
  LoBlank := DiskRec.LoBlank;
  HiBlank := DiskRec.HiBlank;
  LoBoundVal := FloatToMathValue(DiskRec.LoBoundVal);
  HiBoundVal := FloatToMathValue(DiskRec.HiBoundVal);
  AsymFixUp := DiskRec.AsymFixUp;
  ExtraUndefAccuracy := false;
  FuncColor := DiskRec.FuncColor;
  LineWidth := DiskRec.LineWidth;
  FuncCalc.ReadFromDisk(ReadGraphFile);
  {fix and calc math strings}
  with ReadGraphFile do
    begin
      if not LoBlank then
        FixCalcMathConstString(LoBoundStr, LoBoundVal);
      if not HiBlank then
        FixCalcMathConstString(HiBoundStr, HiBoundVal);
    end; {with}
end; {procedure ReadVer1Fields;}

{-----------------------------}

procedure ReadVer2Fields;
var
  DiskRec : TDiskV2_Func2DRec;
begin {procedure ReadVer2Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  LoBlank := DiskRec.LoBlank;
  HiBlank := DiskRec.HiBlank;
  LoBoundVal := FloatToMathValue(DiskRec.LoBoundVal);
  HiBoundVal := FloatToMathValue(DiskRec.HiBoundVal);
  AsymFixUp := DiskRec.AsymFixUp;
  ExtraUndefAccuracy := DiskRec.ExtraUndefAccuracy;
  FuncColor := DiskRec.FuncColor;
  LineWidth := DiskRec.LineWidth;
  with ReadGraphFile do
    begin
      ReadMathConstString(LoBoundStr, LoBoundVal, LoBlank);
      ReadMathConstString(HiBoundStr, HiBoundVal, HiBlank);
    end; {with}
  Result := Func2DSpecialValues.ReadFromDisk(ReadGraphFile);
  FuncCalc.ReadFromDisk(ReadGraphFile);
end; {procedure ReadVer2Fields;}

{-----------------------------}

procedure ReadVer3Fields;
var
  DiskRec : TDiskV3_Func2DRec;
begin {procedure ReadVer3Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  LoBlank := DiskRec.LoBlank;
  HiBlank := DiskRec.HiBlank;
  LoBoundVal := UnPackMathValue(DiskRec.LoBoundVal);
  HiBoundVal := UnPackMathValue(DiskRec.HiBoundVal);
  AsymFixUp := DiskRec.AsymFixUp;
  ExtraUndefAccuracy := DiskRec.ExtraUndefAccuracy;
  FuncColor := DiskRec.FuncColor;
  LineWidth := DiskRec.LineWidth;
  with ReadGraphFile do
    begin
      ReadMathConstString(LoBoundStr, LoBoundVal, LoBlank);
      ReadMathConstString(HiBoundStr, HiBoundVal, HiBlank);
    end; {with}
  Result := Func2DSpecialValues.ReadFromDisk(ReadGraphFile);
  FuncCalc.ReadFromDisk(ReadGraphFile);
end; {procedure ReadVer3Fields;}

{-----------------------------}

begin {procedure TFunc2DObj.ReadFromDisk}
  DidFieldsChange := true;
  Result := true;
  if ReadGraphFile.FileVersion = 1 then
    ReadVer1Fields
  else
    if ReadGraphFile.FileVersion = 2 then
      ReadVer2Fields
    else {if ReadGraphFile.FileVersion >= 3 then}
      ReadVer3Fields;
end; {procedure TFunc2DObj.ReadFromDisk}

{-----------------------------------------------------------------------}

{ TFunc2DSpecialValues }

procedure TFunc2DSpecialValues.AssignMoveSpPointsArray(
  Source: TFunc2DSpecialValues);
begin {procedure TFunc2DSpecialValues.AssignMoveSpPointsArray}
  Clear;
  FSpecialPointArray := Source.FSpecialPointArray;
  Source.FSpecialPointArray := nil;
end; {procedure TFunc2DSpecialValues.AssignMoveSpPointsArray}

{-----------------------------------------------------------------------}

function TFunc2DSpecialValues.CalcPointsArray(SpecialValueFunc : TMathFunction) : boolean;
var
  ItemsListNdx : integer;
  EstimatedPointCount, PointCount : integer;
  TmpObj : TFunc2DSpecialItem;
  CurrPointRec : TSpecialPointRec;

{------------------------------}

procedure DoResizeArray(NewSize : integer);
begin {procedure DoResizeArray}
  if NewSize <= 0 then
    FSpecialPointArray := nil
  else
    if NewSize <> Length(FSpecialPointArray) then
      SetLength(FSpecialPointArray, NewSize);
end; {procedure DoResizeArray}

{------------------------------}

procedure AddCurrPointToArray;
begin {procedure AddCurrPointToArray}
  with TmpObj do
    begin
      CurrPointRec.LeftSideBehave := LeftSideBehave;
      CurrPointRec.AtPointBehave := AtPointBehave;
      CurrPointRec.RightSideBehave := RightSideBehave;
    end; {with}
  inc(PointCount);
  if PointCount >= Length(FSpecialPointArray) then
    DoResizeArray(PointCount + 64);
  if FSpecialPointArray = nil then
    Exit;
  FSpecialPointArray[PointCount - 1] := CurrPointRec;
end; {procedure AddCurrPointToArray}

{------------------------------}

procedure CalcSingleValue;
begin {procedure CalcSingleValue;}
  CurrPointRec.PointValue := TmpObj.FirstVal;
  AddCurrPointToArray;
end; {procedure CalcSingleValue;}

{------------------------------}

procedure CalcSeriesValues;
var
  Steps, Points : integer;
  Diff, Curr : TMathValue;

{--------------}

procedure DoAddCurrValue;
begin {procedure DoAddCurrValue}
  CurrPointRec.PointValue := Curr;
  if TmpObj.ItemType = itFuncSeries then
    begin
      SpecialValueFunc.MathVarValues[0] := Curr;
      if SpecialValueFunc.EvaluateMathFunc then
        CurrPointRec.PointValue := SpecialValueFunc.MathResultValue;
    end; {if}
  AddCurrPointToArray;
end; {procedure DoAddCurrValue}

{--------------}

begin {procedure CalcSeriesValues;}
  Assert((TmpObj.FirstVal.ValueType <> vtEmpty) and (TmpObj.SecondVal.ValueType <> vtEmpty) and
         (TmpObj.EndVal.ValueType <> vtEmpty), 'Empty series value');
  with TmpObj do
    begin
      //parse and verify the function
      if ItemType = itFuncSeries then
        begin
          SpecialValueFunc.FuncString := TmpObj.FunctionStr;
          if not SpecialValueFunc.VerifyMathFunc then
            begin
              Result := false;
              Exit;
            end; {if}
        end; {if}
      //add first, second and end values
      Curr := FirstVal;
      DoAddCurrValue;
      Curr := SecondVal;
      DoAddCurrValue;
      if ((FirstVal.FloatValue < SecondVal.FloatValue) and (SecondVal.FloatValue < EndVal.FloatValue)) then
        begin
          Points := 3;
          //Diff := abs(SecondVal - FirstVal);
          Diff := AbsFuncMathValue(SubMathValue(SecondVal, FirstVal));
          Steps := 1;
          //Curr := SecondVal.FloatValue + (Diff * Steps);
          Curr := AddMathValue(SecondVal, MulMathValue(Diff, LongintToMathValue(Steps)));
          while Curr.FloatValue < EndVal.FloatValue do
            begin
              inc(Points);
              if Points > (MaxPointsPerSeries + 32) then
                Break; //point count to large
              DoAddCurrValue;
              inc(Steps);
              //Curr := SecondVal.FloatValue + (Diff * Steps);
              Curr := AddMathValue(SecondVal, MulMathValue(Diff, LongintToMathValue(Steps)));
            end; {while}
        end
      else
        if ((FirstVal.FloatValue > SecondVal.FloatValue) and (SecondVal.FloatValue > EndVal.FloatValue)) then
          begin
            Points := 3;
            //Diff := abs(FirstVal.FloatValue - SecondVal.FloatValue);
            Diff := AbsFuncMathValue(SubMathValue(FirstVal, SecondVal));
            Steps := 1;
            //Curr := SecondVal.FloatValue - (Diff * Steps);
            Curr := SubMathValue(SecondVal, MulMathValue(Diff, LongintToMathValue(Steps)));
            while Curr.FloatValue > EndVal.FloatValue do
              begin
                inc(Points);
                if Points > (MaxPointsPerSeries + 32) then
                  Break; //point count to large
                DoAddCurrValue;
                inc(Steps);
                //Curr := SecondVal.FloatValue - (Diff * Steps);
                Curr := SubMathValue(SecondVal, MulMathValue(Diff, LongintToMathValue(Steps)));
              end; {while}
          end;
      Curr := EndVal;
      DoAddCurrValue;
    end; {with}
end; {procedure CalcSeriesValues;}

{------------------------------}

procedure DoSortRecords;
begin {procedure DoSortRecords;}
  GapSortZeroBasedArray(Length(FSpecialPointArray), GapSortCompare, GapSortSwap);
end; {procedure DoSortRecords;}

{------------------------------}

procedure DeleteDupeRecords;
var
  I : integer;
  DupeCount : integer;
  HoleNdx, FirstDupeNdx : integer;
begin {procedure DeleteDupeRecords;}
  DupeCount := 0;
  FirstDupeNdx := 0;
  //Identify dupes
  for I := 1 to (Length(FSpecialPointArray) - 1) do
    begin
      if FSpecialPointArray[I-1].PointValue.FloatValue = FSpecialPointArray[I].PointValue.FloatValue then
        begin
          FSpecialPointArray[I].ListIndexPriority := -1;
          inc(DupeCount);
          if DupeCount = 1 then
            FirstDupeNdx := I;
        end; {if}
    end; {for}
  if DupeCount = 0 then
    Exit; //found no dupes
  //close empty dupes holes
  HoleNdx := FirstDupeNdx;
  for I := (FirstDupeNdx + 1) to (Length(FSpecialPointArray) - 1) do
    begin
      if FSpecialPointArray[I].ListIndexPriority <> (-1) then
        begin
          FSpecialPointArray[HoleNdx] := FSpecialPointArray[I];
          inc(HoleNdx);
        end; {if}
    end; {for}
  //Shrink array if necessary
  DoResizeArray(Length(FSpecialPointArray) - DupeCount);
end; {procedure DeleteDupeRecords;}

{------------------------------}

begin {function TFunc2DSpecialValues.CalcPointsArray}
  try
    Result := true;
    PointCount := 0;
    //count points
    EstimatedPointCount := 0;
    for ItemsListNdx := 0 to (FItemsList.Count - 1) do
      EstimatedPointCount := EstimatedPointCount + TFunc2DSpecialItem(FItemsList.Items[ItemsListNdx]).SeriesPointCount;
    //set array size
    DoResizeArray(EstimatedPointCount);
    //add points to array
    for ItemsListNdx := 0 to (FItemsList.Count - 1) do
      begin
        CurrPointRec.ListIndexPriority := ItemsListNdx;
        TmpObj := TFunc2DSpecialItem(FItemsList.Items[ItemsListNdx]);
        if TmpObj.ItemType = itSingleValue then
          CalcSingleValue
        else {itSeries, itFuncSeries}
          CalcSeriesValues;
        if not Result then
          begin
            FSpecialPointArray := nil;
            Exit;
          end; {if}
      end; {for}
    //Shrink array if necessary
    DoResizeArray(PointCount);
    if Length(FSpecialPointArray) >= 2 then
      begin
        //Sort array
        DoSortRecords;
        //Delete Dupes
        DeleteDupeRecords;
      end; {if}
  except
    Result := false;
    FSpecialPointArray := nil;
  end; {try..except}
end; {function TFunc2DSpecialValues.CalcPointsArray}

{-----------------------------------------------------------------------}

procedure TFunc2DSpecialValues.Clear;
var
  I : integer;
begin {procedure TFunc2DSpecialValues.Clear;}
  for I := 0 to (FItemsList.Count - 1) do
    TFunc2DSpecialItem(FItemsList.Items[I]).Free;
  FItemsList.Clear;
  FSpecialPointArray := nil;
end; {procedure TFunc2DSpecialValues.Clear;}

{-----------------------------------------------------------------------}

constructor TFunc2DSpecialValues.Create;
begin {constructor TFunc2DSpecialValues.Create;}
  inherited;
  FItemsList := TList.Create;
  FSpecialPointArray := nil;
end; {constructor TFunc2DSpecialValues.Create;}

{-----------------------------------------------------------------------}

destructor TFunc2DSpecialValues.Destroy;
begin {destructor TFunc2DSpecialValues.Destroy;}
  Clear;
  FItemsList.Free;
  FSpecialPointArray := nil;
  inherited;
end; {destructor TFunc2DSpecialValues.Destroy;}

{-----------------------------------------------------------------------}

procedure TFunc2DSpecialValues.GapSortCompare(Item1, Item2: integer;
  var CompResult: integer);
begin {procedure TFunc2DSpecialValues.GapSortCompare}
  if FSpecialPointArray[Item1].PointValue.FloatValue < FSpecialPointArray[Item2].PointValue.FloatValue then
    CompResult := -1
  else
    if FSpecialPointArray[Item1].PointValue.FloatValue > FSpecialPointArray[Item2].PointValue.FloatValue then
      CompResult := 1
    else //if FSpecialPointArray[Item1].PointValue.FloatValue = FSpecialPointArray[Item2].PointValue.FloatValue then
      begin
        if FSpecialPointArray[Item1].ListIndexPriority < FSpecialPointArray[Item2].ListIndexPriority then
          CompResult := -1
        else
          if FSpecialPointArray[Item1].ListIndexPriority > FSpecialPointArray[Item2].ListIndexPriority then
            CompResult := 1
          else //if FSpecialPointArray[Item1].ListIndexPriority = FSpecialPointArray[Item2].ListIndexPriority then
            CompResult := 0;
      end; {else}
end; {procedure TFunc2DSpecialValues.GapSortCompare}

{-----------------------------------------------------------------------}

procedure TFunc2DSpecialValues.GapSortSwap(Item1, Item2: integer);
var
  SwapRec : TSpecialPointRec;
begin {procedure TFunc2DSpecialValues.GapSortSwap}
  Assert((Item1 <> Item2), 'Sort swap error.');
  SwapRec := FSpecialPointArray[Item1];
  FSpecialPointArray[Item1] := FSpecialPointArray[Item2];
  FSpecialPointArray[Item2] := SwapRec;
end; {procedure TFunc2DSpecialValues.GapSortSwap}

{-----------------------------------------------------------------------}

function TFunc2DSpecialValues.ReadFromDisk(ReadGraphFile: TReadGraphFile) : boolean;
{Returns false if an error occured while calculating 2D function special values.}
var
  SpecialPointsReadSuccessfuly : boolean;

{----------------------------}

procedure ReadVer2Fields;
var
  TmpFunc2DSpecialItem : TFunc2DSpecialItem;
  TheString : string;
  TheMathValue : TMathValue;
  ItemCount, I : integer;
begin {procedure ReadVer2Fields;}
  ItemCount := ReadGraphFile.ReadLongint;
  if ItemCount >= 16 then
    ItemsList.Capacity := ItemCount;
  for I := 1 to ItemCount do
    begin
      ReadGraphFile.ReadMathConstString(TheString, TheMathValue, false);
      TmpFunc2DSpecialItem := TFunc2DSpecialItem.Create;
      try
        TmpFunc2DSpecialItem.FirstStr := TheString;
        TmpFunc2DSpecialItem.FirstVal := TheMathValue;
        ItemsList.Add(TmpFunc2DSpecialItem);
      except
        SpecialPointsReadSuccessfuly := false;
        TmpFunc2DSpecialItem.free;
        raise;
      end; {try..except}
    end; {for}
  ItemCount := ReadGraphFile.ReadLongint;
  for I := 1 to ItemCount do
    ReadGraphFile.SkipBytes(sizeof(Extended));
end; {procedure ReadVer2Fields;}

{----------------------------}

procedure ReadVer3Fields;
var
  TmpFunc2DSpecialItem : TFunc2DSpecialItem;
  ItemCount, I : integer;
begin {procedure ReadVer3Fields;}
  ItemCount := ReadGraphFile.ReadLongint;
  if ItemCount >= 16 then
    ItemsList.Capacity := ItemCount;
  for I := 1 to ItemCount do
    begin
      TmpFunc2DSpecialItem := TFunc2DSpecialItem.Create;
      try
        TmpFunc2DSpecialItem.ReadFromDisk(ReadGraphFile);
        with TmpFunc2DSpecialItem do
          begin
            if ItemType in [itSeries, itFuncSeries] then
              begin
                if VerifySeriesPoints(FirstVal, SecondVal, EndVal) then
                  begin
                    if CalcSeriesPointCount(FirstVal, SecondVal, EndVal, SeriesPointCount) then
                      ItemsList.Add(TmpFunc2DSpecialItem)
                    else
                      begin
                        SpecialPointsReadSuccessfuly := false;
                        TmpFunc2DSpecialItem.free;
                      end; {else}
                  end
                else
                  begin
                    SpecialPointsReadSuccessfuly := false;
                    TmpFunc2DSpecialItem.free;
                  end; {else}
              end
            else
              ItemsList.Add(TmpFunc2DSpecialItem);
          end; {with}
      except
        SpecialPointsReadSuccessfuly := false;
        TmpFunc2DSpecialItem.free;
        raise;
      end; {try..except}
    end; {for}
end; {procedure ReadVer3Fields;}

{----------------------------}

procedure InitCalculatePointArray;
var
  SpecialValueFunc : TMathFunction;
begin {procedure InitCalculatePointArray;}
  SpecialValueFunc := TMathFunction.Create;
  if Assigned(SpecialValueFunc) then
    begin
      try
        SpecialValueFunc.VarCount := 1;
        SpecialValueFunc.VarStrings[0] := 'X';
        if not CalcPointsArray(SpecialValueFunc) then
          SpecialPointsReadSuccessfuly := false;
      finally
        SpecialValueFunc.free;
      end; {try..finally}
    end
  else
    SpecialPointsReadSuccessfuly := false;
end; {procedure InitCalculatePointArray;}

{----------------------------}

begin {function TFunc2DSpecialValues.ReadFromDisk}
  SpecialPointsReadSuccessfuly := true;
  if ReadGraphFile.FileVersion = 1 then
    Clear //no special values in version 1
  else
    if ReadGraphFile.FileVersion = 2 then
      ReadVer2Fields
    else {if ReadGraphFile.FileVersion >= 3 then}
      ReadVer3Fields;
  InitCalculatePointArray;
  Result := SpecialPointsReadSuccessfuly;
end; {function TFunc2DSpecialValues.ReadFromDisk}

{-----------------------------------------------------------------------}

procedure TFunc2DSpecialValues.SaveToDisk(WriteGraphFile: TWriteGraphFile);
var
  TmpFunc2DSpecialItem : TFunc2DSpecialItem;
  ItemCount, I : integer;
begin {procedure TFunc2DSpecialValues.SaveToDisk}
  ItemCount := ItemsList.Count;
  WriteGraphFile.WriteLongint(ItemCount);
  for I := 0 to (ItemCount - 1) do
    begin
      TmpFunc2DSpecialItem := ItemsList.Items[I];
      TmpFunc2DSpecialItem.SaveToDisk(WriteGraphFile);
    end; {for}
end; {procedure TFunc2DSpecialValues.SaveToDisk}

{-----------------------------------------------------------------------}

{ TFunc2DSpecialItem }

procedure TFunc2DSpecialItem.Assign(Source: TFunc2DSpecialItem);
begin {procedure TFunc2DSpecialItem.Assign}
  FirstStr := Source.FirstStr;
  SecondStr := Source.SecondStr;
  EndStr := Source.EndStr;
  FirstVal := Source.FirstVal;
  SecondVal := Source.SecondVal;
  EndVal := Source.EndVal;
  FunctionStr := Source.FunctionStr;
  LeftSideBehave := Source.LeftSideBehave;
  AtPointBehave := Source.AtPointBehave;
  RightSideBehave := Source.RightSideBehave;
  SeriesPointCount := Source.SeriesPointCount;
  ItemType := Source.ItemType;
  {Make Memory Unique}
  UniqueString(FirstStr);
  UniqueString(SecondStr);
  UniqueString(EndStr);
  UniqueString(FunctionStr);
end; {procedure TFunc2DSpecialItem.Assign}

{-----------------------------------------------------------------------}

constructor TFunc2DSpecialItem.Create;
begin {constructor TFunc2DSpecialItem.Create;}
  inherited;
  FirstStr := '';
  SecondStr := '';
  EndStr := '';
  InitMathValue(FirstVal);
  InitMathValue(SecondVal);
  InitMathValue(EndVal);
  FunctionStr := '';
  LeftSideBehave := sbAutomatic;
  AtPointBehave := pbAutoDetect;
  RightSideBehave := sbAutomatic;
  SeriesPointCount := 1;
  ItemType := itSingleValue;
end; {constructor TFunc2DSpecialItem.Create;}

{-----------------------------------------------------------------------}

destructor TFunc2DSpecialItem.Destroy;
begin {destructor TFunc2DSpecialItem.Destroy;}
  FirstStr := '';
  SecondStr := '';
  EndStr := '';
  FunctionStr := '';
  inherited;
end; {destructor TFunc2DSpecialItem.Destroy;}

{-----------------------------------------------------------------------}

procedure TFunc2DSpecialItem.ReadFromDisk(ReadGraphFile: TReadGraphFile);
var
  Func2DSpItemRec : TDiskV3_Func2DSpItemRec;
begin {procedure TFunc2DSpecialItem.ReadFromDisk}
  with ReadGraphFile do
    begin
      //read simple types
      ReadRecord(Func2DSpItemRec, sizeof(Func2DSpItemRec));
      FirstVal := UnPackMathValue(Func2DSpItemRec.FirstVal);
      SecondVal := UnPackMathValue(Func2DSpItemRec.SecondVal);
      EndVal := UnPackMathValue(Func2DSpItemRec.EndVal);
      LeftSideBehave := Func2DSpItemRec.LeftSideBehave;
      AtPointBehave := Func2DSpItemRec.AtPointBehave;
      RightSideBehave := Func2DSpItemRec.RightSideBehave;
      SeriesPointCount := Func2DSpItemRec.SeriesPointCount;
      ItemType := Func2DSpItemRec.ItemType;
      //read First string
      FirstStr := ReadMathString;
      if FirstStr = '' then
        InitMathValue(FirstVal)
      else
        FixCalcMathConstString(FirstStr, FirstVal);
      //read Second string
      SecondStr := ReadMathString;
      if SecondStr = '' then
        InitMathValue(SecondVal)
      else
        FixCalcMathConstString(SecondStr, SecondVal);
      //read End string
      EndStr := ReadMathString;
      if EndStr = '' then
        InitMathValue(EndVal)
      else
        FixCalcMathConstString(EndStr, EndVal);
      //read function string
      FunctionStr := ReadMathString;
    end; {with}
end; {procedure TFunc2DSpecialItem.ReadFromDisk}

{-----------------------------------------------------------------------}

procedure TFunc2DSpecialItem.SaveToDisk(WriteGraphFile: TWriteGraphFile);
var
  Func2DSpItemRec : TDiskV3_Func2DSpItemRec;
begin {procedure TFunc2DSpecialItem.SaveToDisk}
  with WriteGraphFile do
    begin
      //write simple types
      Func2DSpItemRec.FirstVal := PackMathValue(FirstVal);
      Func2DSpItemRec.SecondVal := PackMathValue(SecondVal);
      Func2DSpItemRec.EndVal := PackMathValue(EndVal);
      Func2DSpItemRec.LeftSideBehave := LeftSideBehave;
      Func2DSpItemRec.AtPointBehave := AtPointBehave;
      Func2DSpItemRec.RightSideBehave := RightSideBehave;
      Func2DSpItemRec.SeriesPointCount := SeriesPointCount;
      Func2DSpItemRec.ItemType := ItemType;
      WriteRecord(Func2DSpItemRec, sizeof(Func2DSpItemRec));
      //write string
      WriteMathString(FirstStr);
      WriteMathString(SecondStr);
      WriteMathString(EndStr);
      WriteMathString(FunctionStr);
    end; {with}
end; {procedure TFunc2DSpecialItem.SaveToDisk}

{-----------------------------------------------------------------------}

procedure TFunc2DObj.AssignAndMoveSpPoints(Source: TFunc2DObj);
begin {procedure TFunc2DObj.AssignAndMoveSpPoints}
  FFunc2DSpecialValues.AssignMoveSpPointsArray(Source.FFunc2DSpecialValues);
  FuncCalc.Assign(Source.FuncCalc);
  LoBoundStr := Source.LoBoundStr;
  HiBoundStr := Source.HiBoundStr;
  LoBlank := Source.LoBlank;
  HiBlank := Source.HiBlank;
  LoBoundVal := Source.LoBoundVal;
  HiBoundVal := Source.HiBoundVal;
  AsymFixUp := Source.AsymFixUp;
  ExtraUndefAccuracy := Source.ExtraUndefAccuracy;
  FuncColor := Source.FuncColor;
  LineWidth := Source.LineWidth;
  {make memory strings unique}
  UniqueString(LoBoundStr);
  UniqueString(HiBoundStr);
end; {procedure TFunc2DObj.AssignAndMoveSpPoints}

{-----------------------------------------------------------------------}

procedure TFunc2DObj.AssignSeqObj(Source: TObject);
begin {procedure TFunc2DObj.AssignSeqObj}
  AssignAndMoveSpPoints(Source as TFunc2DObj);
end; {procedure TFunc2DObj.AssignSeqObj}

{-----------------------------------------------------------------------}

function TFunc2DObj.CreateNewSeqObj: TObject;
begin {function TFunc2DObj.CreateNewSeqObj}
  Result := TFunc2DObj.Create;
end; {function TFunc2DObj.CreateNewSeqObj}

{-----------------------------------------------------------------------}

end.
