unit funcpara;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses Graphics, MathCalc, MathFunc, FuncObj, FileSave;

const
  MaxParametricDomain = 10000*pi;
  MinParametricDomain = -10000*pi;
  MaxParametricDomainStr = '10,000*pi';
  MinParametricDomainStr = '-10,000*pi';

type
  TMathParametricFunc = class(TMathGraphFunc)
  private
    { Private declarations }
  public
    { Public declarations }
    constructor Create;
  end; {TMathParametricFunc}

  TFuncParametricObj = class(TBaseFuncObj)
    private
    { Private declarations }
    protected
    { Protected declarations }
      function GetListFuncStr: string; override;
      function CreateNewSeqObj : TObject; override;
      procedure AssignSeqObj(Source : TObject); override;
    public
    { Public declarations }
      XFuncCalc,
      YFuncCalc : TMathParametricFunc;
      LoBoundStr,
      HiBoundStr : string;
      LoBoundVal,
      HiBoundVal : TMathValue;
      Samples : integer;
      FuncColor : TColor;
      LineWidth : integer;
      constructor Create;
      destructor Destroy; override;
      procedure SaveToDisk(WriteGraphFile : TWriteGraphFile);
      procedure ReadFromDisk(ReadGraphFile : TReadGraphFile);
      procedure Assign(Source : TFuncParametricObj);
  end; {TFuncParametricObj}

implementation

type
  TDiskV3_ParametricFuncRec = packed record
    LoBoundVal,
    HiBoundVal : TPackedMathValue;
    Samples : integer;
    FuncColor : TColor;
    LineWidth : integer;
  end; {TDiskV3_ParametricFuncRec = packed record}

{----------------------------------------------------------------------}

constructor TMathParametricFunc.Create;
begin {constructor TMathParametricFunc.Create;}
  inherited Create;
  VarCount := 1;
  VarStrings[0] := 'T';
end; {constructor TMathParametricFunc.Create;}

{----------------------------------------------------------------------}

function TFuncParametricObj.GetListFuncStr: string;
begin {function TFuncParametricObj.GetListFuncStr}
  GetListFuncStr := 'X=' + XFuncCalc.FuncString + '; Y=' + YFuncCalc.FuncString;
end; {function TFuncParametricObj.GetListFuncStr}

{----------------------------------------------------------------------}

constructor TFuncParametricObj.Create;
begin {constructor TFuncParametricObj.Create;}
  inherited Create;
  XFuncCalc := TMathParametricFunc.Create;
  YFuncCalc := TMathParametricFunc.Create;
  LoBoundStr := '0';
  HiBoundStr := '2pi';
  LoBoundVal := LongintToMathValue(0);
  HiBoundVal := FloatToMathValue(2*pi);
  Samples := 10000;
  FuncColor := clBlue;
  LineWidth := 1;
end; {constructor TFuncParametricObj.Create;}

{----------------------------------------------------------------------}

destructor TFuncParametricObj.Destroy;
begin {destructor TFuncParametricObj.Destroy;}
  XFuncCalc.Free;
  YFuncCalc.Free;
  inherited Destroy;
end; {destructor TFuncParametricObj.Destroy;}

{----------------------------------------------------------------------}

procedure TFuncParametricObj.SaveToDisk(WriteGraphFile : TWriteGraphFile);
var
  DiskRec : TDiskV3_ParametricFuncRec;
begin {procedure TFuncParametricObj.SaveToDisk}
  DiskRec.LoBoundVal := PackMathValue(LoBoundVal);
  DiskRec.HiBoundVal := PackMathValue(HiBoundVal);
  DiskRec.Samples := Samples;
  DiskRec.FuncColor := FuncColor;
  DiskRec.LineWidth := LineWidth;
  with WriteGraphFile do
    begin
      WriteRecord(DiskRec, sizeof(DiskRec));
      WriteMathString(LoBoundStr);
      WriteMathString(HiBoundStr);
    end; {with}
  XFuncCalc.SaveToDisk(WriteGraphFile);
  YFuncCalc.SaveToDisk(WriteGraphFile);
end; {procedure TFuncParametricObj.SaveToDisk}

{----------------------------------------------------------------------}

procedure TFuncParametricObj.ReadFromDisk(ReadGraphFile : TReadGraphFile);

{---------------------------------}

procedure ReadVer3Fields;
var
  DiskRec : TDiskV3_ParametricFuncRec;
begin {procedure ReadVer3Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  LoBoundVal := UnPackMathValue(DiskRec.LoBoundVal);
  HiBoundVal := UnPackMathValue(DiskRec.HiBoundVal);
  Samples := DiskRec.Samples;
  FuncColor := DiskRec.FuncColor;
  LineWidth := DiskRec.LineWidth;
  with ReadGraphFile do
    begin
      ReadMathConstString(LoBoundStr, LoBoundVal, False);
      ReadMathConstString(HiBoundStr, HiBoundVal, False);
    end; {with}
  XFuncCalc.ReadFromDisk(ReadGraphFile);
  YFuncCalc.ReadFromDisk(ReadGraphFile);
end; {procedure ReadVer3Fields;}

{---------------------------------}

begin {procedure TFuncParametricObj.ReadFromDisk}
  DidFieldsChange := true;
  Assert((ReadGraphFile.FileVersion >= 3), 'File version number error in read parametric');
  ReadVer3Fields;
end; {procedure TFuncParametricObj.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TFuncParametricObj.Assign(Source: TFuncParametricObj);
begin {procedure TFuncParametricObj.Assign}
  XFuncCalc.Assign(Source.XFuncCalc);
  YFuncCalc.Assign(Source.YFuncCalc);
  LoBoundStr := Source.LoBoundStr;
  HiBoundStr := Source.HiBoundStr;
  LoBoundVal := Source.LoBoundVal;
  HiBoundVal := Source.HiBoundVal;
  Samples := Source.Samples;
  FuncColor := Source.FuncColor;
  LineWidth := Source.LineWidth;
  {Make Memory Unique}
  UniqueString(LoBoundStr);
  UniqueString(HiBoundStr);
end; {procedure TFuncParametricObj.Assign}

{----------------------------------------------------------------------}

procedure TFuncParametricObj.AssignSeqObj(Source: TObject);
begin {procedure TFuncParametricObj.AssignSeqObj}
  Assign(Source as TFuncParametricObj);
end; {procedure TFuncParametricObj.AssignSeqObj}

{----------------------------------------------------------------------}

function TFuncParametricObj.CreateNewSeqObj: TObject;
begin {function TFuncParametricObj.CreateNewSeqObj}
  Result := TFuncParametricObj.Create;
end; {function TFuncParametricObj.CreateNewSeqObj}

{----------------------------------------------------------------------}

end.
