unit funcpolr;

(*
Copyright 1999-2011 Greg VanMullem
This program is distributed under the terms of the GNU General Public License Version 3

This file is part of MathGV.

MathGV is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MathGV is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MathGV.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses Graphics, MathCalc, MathFunc, FuncObj, FileSave;

const
  MaxPolarDomain = 10000*pi;
  MinPolarDomain = -10000*pi;
  MaxPolarDomainStr = '10,000*pi';
  MinPolarDomainStr = '-10,000*pi';

type
  TMathPolarFunc = class(TMathGraphFunc)
  private
    { Private declarations }
  public
    { Public declarations }
    constructor Create;
  end; {TMathPolarFunc}

  TFuncPolarObj = class(TBaseFuncObj)
    private
    { Private declarations }
    protected
    { Protected declarations }
      function GetListFuncStr: string; override;
      function CreateNewSeqObj : TObject; override;
      procedure AssignSeqObj(Source : TObject); override;
    public
    { Public declarations }
      FuncCalc : TMathPolarFunc;
      LoBoundStr,
      HiBoundStr : string;
      LoBoundVal,
      HiBoundVal : TMathValue;
      Samples : integer;
      FuncColor : TColor;
      LineWidth : integer;
      constructor Create;
      destructor Destroy; override;
      procedure SaveToDisk(WriteGraphFile : TWriteGraphFile);
      procedure ReadFromDisk(ReadGraphFile : TReadGraphFile);
      procedure Assign(Source : TFuncPolarObj);
  end; {TFuncPolarObj}

implementation

type
  TDiskV1_PolarFuncRec = packed record
    JunkBytes : packed array[1..4] of byte;
    LoBoundStr,
    HiBoundStr : string[50];
    LoBoundVal,
    HiBoundVal : Extended;
    Samples : smallint;
    FuncColor : TColor;
    LineWidth : smallint;
  end; {TDiskV1_PolarFuncRec = packed record}

  TDiskV2_PolarFuncRec = packed record
    LoBoundVal,
    HiBoundVal : Extended;
    Samples : integer;
    FuncColor : TColor;
    LineWidth : integer;
  end; {TDiskV2_PolarFuncRec = packed record}

  TDiskV3_PolarFuncRec = packed record
    LoBoundVal,
    HiBoundVal : TPackedMathValue;
    Samples : integer;
    FuncColor : TColor;
    LineWidth : integer;
  end; {TDiskV3_PolarFuncRec = packed record}

{----------------------------------------------------------------------}

constructor TMathPolarFunc.Create;
begin {constructor TMathPolarFunc.Create;}
  inherited Create;
  VarCount := 2;
  VarStrings[0] := 'THETA';
  VarStrings[1] := 'X';
end; {constructor TMathPolarFunc.Create;}

{----------------------------------------------------------------------}

function TFuncPolarObj.GetListFuncStr: string;
begin {function TFuncPolarObj.GetListFuncStr}
  GetListFuncStr := 'R=f(Theta)=' + FuncCalc.FuncString;
end; {function TFuncPolarObj.GetListFuncStr}

{----------------------------------------------------------------------}

constructor TFuncPolarObj.Create;
begin {constructor TFuncPolarObj.Create;}
  inherited Create;
  FuncCalc := TMathPolarFunc.Create;
  LoBoundStr := '0';
  HiBoundStr := '2pi';
  LoBoundVal := LongintToMathValue(0);
  HiBoundVal := FloatToMathValue(2*pi);
  Samples := 10000;
  FuncColor := clBlue;
  LineWidth := 1;
end; {constructor TFuncPolarObj.Create;}

{----------------------------------------------------------------------}

destructor TFuncPolarObj.Destroy;
begin {destructor TFuncPolarObj.Destroy;}
  FuncCalc.Free;
  inherited Destroy;
end; {destructor TFuncPolarObj.Destroy;}

{----------------------------------------------------------------------}

procedure TFuncPolarObj.SaveToDisk(WriteGraphFile : TWriteGraphFile);
var
  DiskRec : TDiskV3_PolarFuncRec;
begin {procedure TFuncPolarObj.SaveToDisk}
  DiskRec.LoBoundVal := PackMathValue(LoBoundVal);
  DiskRec.HiBoundVal := PackMathValue(HiBoundVal);
  DiskRec.Samples := Samples;
  DiskRec.FuncColor := FuncColor;
  DiskRec.LineWidth := LineWidth;
  with WriteGraphFile do
    begin
      WriteRecord(DiskRec, sizeof(DiskRec));
      WriteMathString(LoBoundStr);
      WriteMathString(HiBoundStr);
    end; {with}
  FuncCalc.SaveToDisk(WriteGraphFile);
end; {procedure TFuncPolarObj.SaveToDisk}

{----------------------------------------------------------------------}

procedure TFuncPolarObj.ReadFromDisk(ReadGraphFile : TReadGraphFile);

{---------------------------------}

procedure ReadVer1Fields;
var
  DiskRec : TDiskV1_PolarFuncRec;
begin {procedure ReadVer1Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  LoBoundStr := DiskRec.LoBoundStr;
  HiBoundStr := DiskRec.HiBoundStr;
  LoBoundVal := FloatToMathValue(DiskRec.LoBoundVal);
  HiBoundVal := FloatToMathValue(DiskRec.HiBoundVal);
  Samples := DiskRec.Samples;
  FuncColor := DiskRec.FuncColor;
  LineWidth := DiskRec.LineWidth;
  FuncCalc.ReadFromDisk(ReadGraphFile);
  {fix and calc math strings}
  with ReadGraphFile do
    begin
      FixCalcMathConstString(LoBoundStr, LoBoundVal);
      FixCalcMathConstString(HiBoundStr, HiBoundVal);
    end; {with}
end; {procedure ReadVer1Fields;}

{---------------------------------}

procedure ReadVer2Fields;
var
  DiskRec : TDiskV2_PolarFuncRec;
begin {procedure ReadVer2Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  LoBoundVal := FloatToMathValue(DiskRec.LoBoundVal);
  HiBoundVal := FloatToMathValue(DiskRec.HiBoundVal);
  Samples := DiskRec.Samples;
  FuncColor := DiskRec.FuncColor;
  LineWidth := DiskRec.LineWidth;
  with ReadGraphFile do
    begin
      ReadMathConstString(LoBoundStr, LoBoundVal, False);
      ReadMathConstString(HiBoundStr, HiBoundVal, False);
    end; {with}
  FuncCalc.ReadFromDisk(ReadGraphFile);
end; {procedure ReadVer2Fields;}

{---------------------------------}

procedure ReadVer3Fields;
var
  DiskRec : TDiskV3_PolarFuncRec;
begin {procedure ReadVer3Fields;}
  ReadGraphFile.ReadRecord(DiskRec, sizeof(DiskRec));
  LoBoundVal := UnPackMathValue(DiskRec.LoBoundVal);
  HiBoundVal := UnPackMathValue(DiskRec.HiBoundVal);
  Samples := DiskRec.Samples;
  FuncColor := DiskRec.FuncColor;
  LineWidth := DiskRec.LineWidth;
  with ReadGraphFile do
    begin
      ReadMathConstString(LoBoundStr, LoBoundVal, False);
      ReadMathConstString(HiBoundStr, HiBoundVal, False);
    end; {with}
  FuncCalc.ReadFromDisk(ReadGraphFile);
end; {procedure ReadVer3Fields;}

{---------------------------------}

begin {procedure TFuncPolarObj.ReadFromDisk}
  DidFieldsChange := true;
  if ReadGraphFile.FileVersion = 1 then
    ReadVer1Fields
  else
    if ReadGraphFile.FileVersion = 2 then
      ReadVer2Fields
    else {if ReadGraphFile.FileVersion >= 3 then}
      ReadVer3Fields;
end; {procedure TFuncPolarObj.ReadFromDisk}

{----------------------------------------------------------------------}

procedure TFuncPolarObj.Assign(Source: TFuncPolarObj);
begin {procedure TFuncPolarObj.Assign}
  FuncCalc.Assign(Source.FuncCalc);
  LoBoundStr := Source.LoBoundStr;
  HiBoundStr := Source.HiBoundStr;
  LoBoundVal := Source.LoBoundVal;
  HiBoundVal := Source.HiBoundVal;
  Samples := Source.Samples;
  FuncColor := Source.FuncColor;
  LineWidth := Source.LineWidth;
  {Make Memory Unique}
  UniqueString(LoBoundStr);
  UniqueString(HiBoundStr);
end; {procedure TFuncPolarObj.Assign}

{----------------------------------------------------------------------}

procedure TFuncPolarObj.AssignSeqObj(Source: TObject);
begin {procedure TFuncPolarObj.AssignSeqObj}
  Assign(Source as TFuncPolarObj);
end; {procedure TFuncPolarObj.AssignSeqObj}

{----------------------------------------------------------------------}

function TFuncPolarObj.CreateNewSeqObj: TObject;
begin {function TFuncPolarObj.CreateNewSeqObj}
  Result := TFuncPolarObj.Create;
end; {function TFuncPolarObj.CreateNewSeqObj}

{----------------------------------------------------------------------}

end.
